/*****************************************************************************
** tiff2bp.c
** =========
Convert tiff files (from Zeiss Phodis scanner) to BP.

14-Jul-1997 - Add option to 'bin' down
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "myLibs:picdefs.h"

#define VERSION "1.1"

#define MAX(x,y)  (((x)>(y)) ? (x) : (y))  
#define MIN(x,y)  (((x)<(y)) ? (x) : (y))

typedef unsigned char UBYTE;

typedef struct __tifTag_t__ {
    int    tagid;
    int    dataType;
    int    dataCount;
    int    dataOffset;
  } tifTag_t;

static
int  WriteBP(UBYTE *buffer, int dx, int dy);
static
int  ReadTiff(UBYTE **buffer, int *dx, int *dy);
static
void ConvertOD(UBYTE *buffer, const int dx, const int dy);
static
void BinPixels(UBYTE *buffer, int *dx, int *dy);
static
void TiffTagType(tifTag_t tag);
static
void DifferentEndian(UBYTE *inStr, int length, int endianSwitch);
static
void CopyArray( UBYTE *dst, int ddx, int ddy
	      , UBYTE *src, int sdx, int sdy
	      , int dx0, int dy0
	      );


// Tiff file parameters
// ====================
int bitsPerSample, compression, pmi, orientation, samplesPerPixel,
  planarConfig, tileWidth,tileLength, tileOffsets,tileByteCounts;
int tileByteCountsType;
int tilesInImage,tilesAcross,tilesDown;
int ncol, nrow;

/*****************************************************************************
** main
** ====
*/
int main(void)
{
  UBYTE *buffer;
  int    dx, dy;

  printf("tiff2bp - convert tiff format into BP format\n");
  printf("============================================\n");
  printf("Version %s compiled %s %s\n", VERSION, __DATE__, __TIME__);
  printf("JF Conway, LSBR/NIAMS/NIH - James_Conway@nih.gov\n");
  printf("\n");

  if(ReadTiff(&buffer, &dx, &dy) == 0)
    exit(0);

  ConvertOD(buffer, dx, dy); // optional
  BinPixels(buffer, &dx, &dy); // optional

  WriteBP(buffer, dx, dy);

  free(buffer);
}
/*********************************************************************
** WriteBP
** =======
*/
static
int WriteBP(UBYTE *buffer, int dx, int dy)
{
  char  fname[1024];
  const int FOR_UNIT = 10;
  int y, tenth, block;

  do  {
    printf("BP filename ? ");
    scanf("%s", fname);
    if(PicOpenOutFile(fname, FOR_UNIT, dx, dy, fname) != OK)
      printf("Error opening BP file for output\n");
    else
      break;
  } while(1);

  printf("- Writing BP:");
  tenth = dy / 10 + 1;	// make sure 10% or more, but not less
  for(y=0; y<dy; y+=tenth)  {
    printf(" %d%%", (y*100)/dy);
    fflush(stdout);
    block = MIN(tenth, dy-y);
    if(PicWriteAllLines(FOR_UNIT, dx, block, buffer) != OK)  {
      printf("Error writing BP file\n");
      PicCloseFile(FOR_UNIT);
      return(0);
    }
    buffer += block*dx;
  }    
  printf(" 100%%\n");

  PicCloseFile(FOR_UNIT);
  return(1);
}
/*********************************************************************
** ConvertOD
** =========
0 maps to 0, 255 maps to 255, OD is log(1/(255-pixel+1)).
First off, invert the input value (255-pixel), then add one to the transmission
value to get a scale 1-256, then we can take logs and avoid log of 0. And
conveniently, log(1) = 0 so one end of the output scale is nicely set. The
(negative) scale factor then returns us to 0-255 but inverts, and so we must
undo this inversion by subtracting from 255 because the OD scale runs backwards
to the transmission scale.
Also, rather than use log10 for each pixel, we create a 256-entry lookup table.
*/
static
void ConvertOD(UBYTE *buffer, const int dx, const int dy)
{
  char instr[1000];
  const double scale = 255.0 / log10(1.0/256.0); // rescale back to 0-255
  int i, x, y, tenth, nextPercent;
  UBYTE table[256];

  printf("Convert to OD values (y/n) ? ");
  scanf("%s", instr);
  if(toupper(instr[0]) != 'Y')  {
    printf("- No OD conversion done\n");
    return;
  }

// Set up table for mapping 8-bit tranmission value onto 8-bit OD value

  for(i=0; i<256; i++)
    table[i] = 255 - (UBYTE) (scale * log10(1.0 / (double) (255 - i + 1)));

// Convert transmissions into OD's

  printf("- OD conversion beginning:");
  tenth = dy / 10 + 1;	// make sure 10% or more, but not less
  nextPercent = 0;
  for(y=0; y<dy; y++)  {
    if(y >= nextPercent)  {
      printf(" %d%%", (y*100)/dy);
      fflush(stdout);
      nextPercent += tenth;
    }
    for(x=0; x<dx; x++, buffer++)
      *buffer = table[*buffer];
  }
  printf(" 100%%\n");
}
/*********************************************************************
** BinPixels
** =========
Simple averaging of 2x2, 3x3, 4x4, ... into singple 'bins. We do this in
software because the Zeiss Scanner software does not do it properly in the
Y-direction.
*/
static
void BinPixels(UBYTE *buffer, int *dx, int *dy)
{
  int binFactor, binSize, finalDx, finalDy, sum;
  int x, y, xx, yy, tenth, nextPercent;
  unsigned long offset;
  UBYTE *bufIn, *bufOut;

  do  {
    printf("Bin pixels by integer factor of 1,2,3,...> ");
    scanf("%d", &binFactor);
  } while(binFactor<1);

  if(binFactor == 1)  {
    printf("- No 'binning' done\n");
    return;
  }

// Bin down - only keep complete 'bin'ed' rows and columns

  finalDx = (*dx) / binFactor;
  finalDy = (*dy) / binFactor;
  binSize = binFactor * binFactor;

  printf("- Image size of (%d,%d) scaled down by %d to (%d,%d)\n"
			, *dx, *dy
			, binFactor
			, finalDx, finalDy
			);
  printf("- Binning beginning:");
  tenth = finalDy / 10 + 1;	// make sure 10% or more, but not less
  nextPercent = 0;
  bufOut = buffer;	// overwrite input array

  for(y=0; y<finalDy; y++)  {
    if(y >= nextPercent)  {
      printf(" %d%%", (y*100)/finalDy);
      fflush(stdout);
      nextPercent += tenth;
    }
    bufIn = buffer + y*binFactor*(*dx);		// start of row
    for(x=0; x<finalDx; x++, bufIn+=binFactor)  {
      sum = 0;
      offset = 0;
      for(yy=0; yy<binFactor; yy++, offset+=(*dx))
        for(xx=0; xx<binFactor; xx++)
          sum += *(bufIn+offset+xx);
      *bufOut++ = (sum + binSize/2) / binSize;  // round value with binSize/2
    }    // along binned rows
  }    // down binned columns
  printf(" 100%%\n");
  *dx = finalDx;
  *dy = finalDy;
}
/*********************************************************************
** ReadTiff
** ========
*/
static
int ReadTiff(UBYTE **buffer, int *dx, int *dy)
{
  FILE *fptr;

  do  {
    char  fname[1024];
    printf("Tiff filename ? ");
    scanf("%s", fname);
    if((fptr = fopen(fname, "r")) == NULL)
      perror("Error opening TIFF file for input");
    else
      break;
  } while(1);

  TIFF_ReadImageHeader(fptr, dx, dy);

// Allocate memory

  {
    const size_t bytes = sizeof(UBYTE) * ((*dx)*(*dy));
    *buffer = (UBYTE *) malloc(bytes);
    if(*buffer == NULL)  {
      printf("Can't allocate [%d] sized buffer \n", bytes);
      fclose(fptr);
      return(0);
    }
  }

  TIFF_ReadImageData(fptr, *buffer, *dx, *dy);
  fclose(fptr);
  return(1);
}
/****************************************************************************/
