/*****************************************************************************
** tiff2bp.c
** =========
Convert tiff files (from Zeiss Phodis scanner) to BP.

14-Jul-1997 - Add option to 'bin' down
*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "myLibs:picdefs.h"

#define VERSION "1.1"

#define MAX(x,y)  (((x)>(y)) ? (x) : (y))  
#define MIN(x,y)  (((x)<(y)) ? (x) : (y))

typedef unsigned char UBYTE;

typedef struct __tifTag_t__ {
    int    tagid;
    int    dataType;
    int    dataCount;
    int    dataOffset;
  } tifTag_t;

static
int  WriteBP(UBYTE *buffer, int dx, int dy);
static
int  ReadTiff(UBYTE **buffer, int *dx, int *dy);
static
void ConvertOD(UBYTE *buffer, const int dx, const int dy);
static
void BinPixels(UBYTE *buffer, int *dx, int *dy);
static
void TiffTagType(tifTag_t tag);
static
void DifferentEndian(UBYTE *inStr, int length, int endianSwitch);
static
void CopyArray( UBYTE *dst, int ddx, int ddy
	      , UBYTE *src, int sdx, int sdy
	      , int dx0, int dy0
	      );


// Tiff file parameters
// ====================
int bitsPerSample, compression, pmi, orientation, samplesPerPixel,
  planarConfig, tileWidth,tileLength, tileOffsets,tileByteCounts;
int tileByteCountsType;
int tilesInImage,tilesAcross,tilesDown;
int ncol, nrow;

/*****************************************************************************
** main
** ====
*/
int main(void)
{
  UBYTE *buffer;
  int    dx, dy;

  printf("tiff2bp - convert tiff format into BP format\n");
  printf("============================================\n");
  printf("Version %s compiled %s %s\n", VERSION, __DATE__, __TIME__);
  printf("JF Conway, LSBR/NIAMS/NIH - James_Conway@nih.gov\n");
  printf("\n");

  if(ReadTiff(&buffer, &dx, &dy) == 0)
    exit(0);

  ConvertOD(buffer, dx, dy); // optional
  BinPixels(buffer, &dx, &dy); // optional

  WriteBP(buffer, dx, dy);

  free(buffer);
}
/*********************************************************************
** WriteBP
** =======
*/
static
int WriteBP(UBYTE *buffer, int dx, int dy)
{
  char  fname[1024];
  const int FOR_UNIT = 10;
  int y, tenth, block;

  do  {
    printf("BP filename ? ");
    scanf("%s", fname);
    if(PicOpenOutFile(fname, FOR_UNIT, dx, dy, fname) != OK)
      printf("Error opening BP file for output\n");
    else
      break;
  } while(1);

  printf("- Writing BP:");
  tenth = dy / 10 + 1;	// make sure 10% or more, but not less
  for(y=0; y<dy; y+=tenth)  {
    printf(" %d%%", (y*100)/dy);
    fflush(stdout);
    block = MIN(tenth, dy-y);
    if(PicWriteAllLines(FOR_UNIT, dx, block, buffer) != OK)  {
      printf("Error writing BP file\n");
      PicCloseFile(FOR_UNIT);
      return(0);
    }
    buffer += block*dx;
  }    
  printf(" 100%%\n");

  PicCloseFile(FOR_UNIT);
  return(1);
}
/*********************************************************************
** ConvertOD
** =========
0 maps to 0, 255 maps to 255, OD is log(1/(255-pixel+1)).
First off, invert the input value (255-pixel), then add one to the transmission
value to get a scale 1-256, then we can take logs and avoid log of 0. And
conveniently, log(1) = 0 so one end of the output scale is nicely set. The
(negative) scale factor then returns us to 0-255 but inverts, and so we must
undo this inversion by subtracting from 255 because the OD scale runs backwards
to the transmission scale.
Also, rather than use log10 for each pixel, we create a 256-entry lookup table.
*/
static
void ConvertOD(UBYTE *buffer, const int dx, const int dy)
{
  char instr[1000];
  const double scale = 255.0 / log10(1.0/256.0); // rescale back to 0-255
  int i, x, y, tenth, nextPercent;
  UBYTE table[256];

  printf("Convert to OD values (y/n) ? ");
  scanf("%s", instr);
  if(toupper(instr[0]) != 'Y')  {
    printf("- No OD conversion done\n");
    return;
  }

// Set up table for mapping 8-bit tranmission value onto 8-bit OD value

  for(i=0; i<256; i++)
    table[i] = 255 - (UBYTE) (scale * log10(1.0 / (double) (255 - i + 1)));

// Convert transmissions into OD's

  printf("- OD conversion beginning:");
  tenth = dy / 10 + 1;	// make sure 10% or more, but not less
  nextPercent = 0;
  for(y=0; y<dy; y++)  {
    if(y >= nextPercent)  {
      printf(" %d%%", (y*100)/dy);
      fflush(stdout);
      nextPercent += tenth;
    }
    for(x=0; x<dx; x++, buffer++)
      *buffer = table[*buffer];
  }
  printf(" 100%%\n");
}
/*********************************************************************
** BinPixels
** =========
Simple averaging of 2x2, 3x3, 4x4, ... into singple 'bins. We do this in
software because the Zeiss Scanner software does not do it properly in the
Y-direction.
*/
static
void BinPixels(UBYTE *buffer, int *dx, int *dy)
{
  int binFactor, binSize, finalDx, finalDy, sum;
  int x, y, xx, yy, tenth, nextPercent;
  unsigned long offset;
  UBYTE *bufIn, *bufOut;

  do  {
    printf("Bin pixels by integer factor of 1,2,3,...> ");
    scanf("%d", &binFactor);
  } while(binFactor<1);

  if(binFactor == 1)  {
    printf("- No 'binning' done\n");
    return;
  }

// Bin down - only keep complete 'bin'ed' rows and columns

  finalDx = (*dx) / binFactor;
  finalDy = (*dy) / binFactor;
  binSize = binFactor * binFactor;

  printf("- Image size of (%d,%d) scaled down by %d to (%d,%d)\n"
			, *dx, *dy
			, binFactor
			, finalDx, finalDy
			);
  printf("- Binning beginning:");
  tenth = finalDy / 10 + 1;	// make sure 10% or more, but not less
  nextPercent = 0;
  bufOut = buffer;	// overwrite input array

  for(y=0; y<finalDy; y++)  {
    if(y >= nextPercent)  {
      printf(" %d%%", (y*100)/finalDy);
      fflush(stdout);
      nextPercent += tenth;
    }
    bufIn = buffer + y*binFactor*(*dx);		// start of row
    for(x=0; x<finalDx; x++, bufIn+=binFactor)  {
      sum = 0;
      offset = 0;
      for(yy=0; yy<binFactor; yy++, offset+=(*dx))
        for(xx=0; xx<binFactor; xx++)
          sum += *(bufIn+offset+xx);
      *bufOut++ = (sum + binSize/2) / binSize;  // round value with binSize/2
    }    // along binned rows
  }    // down binned columns
  printf(" 100%%\n");
  *dx = finalDx;
  *dy = finalDy;
}
/*********************************************************************
** ReadTiff
** ========
*/
static
int ReadTiff(UBYTE **buffer, int *dx, int *dy)
{
  FILE *fptr;
  char  fname[1024];
  int m, k, l, nimg, imageX, imageY, tenth, nextPercent;
  unsigned short numDirEntries;
  tifTag_t taglist[32];
  int firstIFD, endianSwitch, *tileTable;
  UBYTE file_id, *tileBuffer, *imageArray;

  do  {
    printf("Tiff filename ? ");
    scanf("%s", fname);
    if((fptr = fopen(fname, "r", "dna=.tif")) == NULL)
      perror("Error opening TIFF file for input");
    else
      break;
  } while(1);

// Read off the header

  fread(&file_id, sizeof(UBYTE), 1, fptr);  
  endianSwitch = FALSE;
  if(file_id==77)
    endianSwitch = TRUE;
  if(endianSwitch)
    printf("Endian mode = big (hi-byte first)\n");
  else
    printf("Endian mode = little (lo-byte first)\n");

  fread(&file_id, sizeof(UBYTE), 1, fptr);  
  fread(&file_id, sizeof(UBYTE), 1, fptr);  
  fread(&file_id, sizeof(UBYTE), 1, fptr);  
  fread(&firstIFD, sizeof(int), 1, fptr);  
  DifferentEndian((UBYTE *)&firstIFD, 4, endianSwitch);
  printf("FirstIFD = %d\n", firstIFD);

// Read TIFF ifd

  fseek(fptr, firstIFD, 0);
  fread(&numDirEntries, sizeof(short), 1, fptr);  
  DifferentEndian((UBYTE *) &numDirEntries, 2, endianSwitch);
  printf("Num DirENTRies = %hd\n", numDirEntries);
  if(numDirEntries > 32)  {	// overflow
    printf("Too many tags [%hd]\n", numDirEntries);
    fclose(fptr);
    return(0);
  }
  for(m=0; m<numDirEntries; m++)  {	// Read each tag
    short thisShort;
    fseek(fptr, (firstIFD+2+(m*12)), 0);
    fread(&thisShort, sizeof(short), 1, fptr);
    DifferentEndian((UBYTE *)&thisShort, 2, endianSwitch);
    taglist[m].tagid= thisShort;
    fread(&thisShort, sizeof(short), 1, fptr);
    DifferentEndian((UBYTE *)&thisShort, 2, endianSwitch);
    taglist[m].dataType= thisShort;
    fread(&(taglist[m].dataCount), sizeof(int), 1, fptr);
    DifferentEndian((UBYTE *)&(taglist[m].dataCount), 4, endianSwitch);
    fread(&(taglist[m].dataOffset), sizeof(int), 1, fptr);
    DifferentEndian((UBYTE *)&(taglist[m].dataOffset), 4, endianSwitch); 
//    printf("id=%d  type=%d  count=%d  off=%d\n",
//	     taglist[m].tagid,
//	     taglist[m].dataType,
//	     taglist[m].dataCount,
//	     taglist[m].dataOffset
//	     );
    TiffTagType(taglist[m]);
  }

// Let's check to see we've got all the info we need

  if(tileWidth == 0 || tileLength == 0 || ncol==0 || nrow==0)  {
    printf("Not enough info in header to continue...\n");
    fclose(fptr);
    return(0);
  }

  if(planarConfig != 1)  {
    printf("Not configured for this planar config [%d]\n",planarConfig);
    fclose(fptr);
    return(0);
  }

// Gather tile information

  tilesAcross = (ncol + (tileWidth-1)) / tileWidth;
  tilesDown   = (nrow + (tileLength-1)) / tileLength;

  tilesInImage = tilesAcross * tilesDown;
  printf("Tiles: size=(%d,%d)  #across=%d  #down=%d  total=%d  \n"
	, tileWidth, tileLength, tilesAcross, tilesDown, tilesInImage);

// Allocate memory
  nimg = 1;      /* Hard code this for now */
  imageArray = (UBYTE *)malloc(sizeof(UBYTE)* ((ncol*nrow)*nimg));
  if(imageArray==0)  {
    printf("Can't allocate [%d] sized buffer \n"
		, sizeof(UBYTE) * ((ncol*nrow)*nimg));
    fclose(fptr);
    return(0);
  }
  tileBuffer = (UBYTE *)malloc(sizeof(UBYTE) * tileWidth*tileLength);
  if(tileBuffer==0)  {
    printf("Can't allocate temp buffer \n");
    fclose(fptr);
    return(0);
  }
  tileTable = malloc(tilesInImage * sizeof(*tileTable));
  if(tileTable==0)  {
    printf("Can't allocate tile table\n");
    fclose(fptr);
    return(0);
  }

// Read tile offsets

  fseek(fptr, tileOffsets, 0);
  fread(tileTable, sizeof(int), tilesInImage, fptr);
  for(m=0; m<tilesInImage; m++)  	// Fix up endian
    DifferentEndian((UBYTE *) &tileTable[m], 4, endianSwitch);

// Read the image data

  printf("- Reading tiff data:");
  tenth = tilesInImage/10 + 1;	// make sure 10% or more, but not less
  nextPercent = 0;
  for(m=0; m<tilesInImage; m++)  {	// Look at all tiles
    if(m >= nextPercent)  {
      printf(" %d%%", (m*100)/tilesInImage);
      fflush(stdout);
      nextPercent += tenth;
    }
    fseek(fptr, tileTable[m], 0);	// Now read the tile
    if(fread(tileBuffer, (sizeof(UBYTE)*tileWidth), tileLength, fptr) != tileLength)  {
      perror("ReadTiff - fread error");
      free(tileBuffer);
      free(tileTable);
      fclose(fptr);
      return(0);
    }

// Now copy tile buffer into main array
    imageX= (m % tilesAcross) * tileWidth;
    imageY= (m / tilesAcross) * tileLength;
    CopyArray( imageArray, ncol, nrow
	     , tileBuffer, tileWidth, tileLength
	     , imageX, imageY
	     );
  }
  printf(" 100%%\n");
  free(tileBuffer);
  free(tileTable);
  fclose(fptr);
  *dx = ncol;
  *dy = nrow;
  *buffer = imageArray;
  return(1);
}
/*****************************************************************************
** CopyArray
** =========
Copy all of src (size sdx,sdy) into dst (size ddx,ddy) starting at
dst(dx0,dy0). Make sure we copy only the areas that overlap.
*/
static
void CopyArray( UBYTE *dst, int ddx, int ddy
	      , UBYTE *src, int sdx, int sdy
	      , int dx0, int dy0
	      )
{
  UBYTE *d, *s;
  const int dx1 = MIN(ddx, dx0+sdx);	// clip to dst boundaries
  const int dy1 = MIN(ddy, dy0+sdy);
  const int dinc = ddx - (dx1-dx0);
  const int sinc = sdx - (dx1-dx0);	// this may be zero
  int x, y;

  d = dst + dy0*ddx + dx0;		// start byte of dst(dx0,dy0)
  s = src;				// start byte of src(0,0)
  for(y=dy0; y<dy1; y++, d+=dinc, s+=sinc)	// inc d&s to next line
    for(x=dx0; x<dx1; x++)
      *d++ = *s++; 
}
/*****************************************************************************
** TiffTagType
** ===========
*/
static
void TiffTagType(tifTag_t tag)
{
  int i,j;

  switch(tag.tagid)  {
    case 256:	// width
      if(tag.dataType==3)
	ncol = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      else
        ncol = (int)tag.dataOffset;
      printf("Image width = %d\n", ncol);
      break;
    case 257:	// height
      if(tag.dataType==3)
	nrow = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      else
        nrow = (int)tag.dataOffset;
      printf("Image height = %d\n",nrow);
      break;
    case 258:	// bitspersample
      bitsPerSample = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      printf("Bits-per-sample = %d\n", bitsPerSample);
      break;
    case 259:	// compression
      compression = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      printf("Compression = ");
      switch(compression)  {
	  case 1 :
	    printf("uncompressed\n");
	    break;
	  case 2 :
	    printf("CCITT 1D\n");
	    break;
	  case 3 :
	    printf("CCITT grp3\n");
	    break;
	  case 4 :
	    printf("CCITT grp4\n");
	    break;
	  case 5 :
	    printf("lzw compression\n");
	    break;
	  case 6 :
	    printf("jpeg\n");	  
	    break;
	  case 32771 :
	    printf("uncompressed222\n");	  
	    break;
	  case 32773 :
	    printf("packbits\n");	  
	    break;
	  default :
	    printf("unknown\n");
	    break;
	}
      break;
    case 262:	// photometricinterpretation
      printf("Photometric Interpretation: ");
      pmi = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      switch(pmi) {
	  case 0 :
	    printf("Black is zero.\n");
	    break;
	  case 1 :
	    printf("White is Zero.\n");
	    break;
	  case 2 :
	    printf("rgb.\n");
	    break;
	  case 3 :
	    printf("rgb pallette.\n");
	    break;
	  case 4 :
	    printf("transparancey mask.\n");	  
	    break;
	  case 5 :
	    printf("cmyk.\n");	  
	    break;
	  case 6 :
	    printf("ycbcr.\n");	  
	    break;
	  case 8 :
	    printf("cielab.\n");	  
	    break;
	  default :
	    printf("unknown.\n");
	    break;
	}
      break;
    case 274:	// orientation
      orientation = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      printf("Orientation = %d\n", orientation);
      break;
    case 277:	// sampleperspixel
      samplesPerPixel = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      printf("Samples per pixel = %d\n", samplesPerPixel);
      break;
    case 284:	// planar configuration
      planarConfig = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      printf("Planar configuration = %d\n", planarConfig);
      break;
    case 305:	// software
      {
      char inChar;
      inChar = (char)( (tag.dataOffset & 0xFF000000) >> 24 );
      printf("Software = (%d) '%c'\n", (int) inChar, inChar);
      }
      break;
    case 322:	// tileWidth
      if(tag.dataType==3)
	tileWidth = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      else
        tileWidth = tag.dataOffset;
      printf("Tile width = %d\n", tileWidth);
      break;
    case 323:	// tileLength
      if(tag.dataType==3)
	tileLength = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      else
        tileLength = tag.dataOffset;
      printf("Tile length = %d\n", tileLength);
      break;
    case 324:	// tileOffsets
      if(tag.dataType==3)
	tileOffsets = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      else
        tileOffsets = tag.dataOffset;
      printf("Tile offsets = %d\n", tileOffsets);
      break;
    case 325:	// tileByteCounts
      if(tag.dataType==3)
	tileByteCounts = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      else
        tileByteCounts = tag.dataOffset;
      printf("Tile byte-counts = %d\n", tileByteCounts);
      tileByteCountsType = tag.dataType;
      break;
    case 32997:	// Terminating tag?
      printf("Terminating tag (32997)??? \n");
//    i = (short)((tag.dataOffset & 0xFFFF0000) >> 16);
      break;
    default:
      printf("Don't know tag [%d]\n",tag.tagid);
      break;
  }
}
/*****************************************************************************
** DifferentEndian
** ===============
*/
static
void DifferentEndian(UBYTE *inStr, int length, int endianSwitch)
{
  int i, lm1;
  UBYTE thisStr[8];

  if(endianSwitch)  {
    for(i=0; i<length; i++)
      thisStr[i] = inStr[i];
    lm1 = length - 1;
    for(i=0; i<length; i++)
      inStr[i] = thisStr[lm1-i];
  }
}
/****************************************************************************/
