/*****************************************************************************
** PFTChanges.c
** ============

Version 1.2 James Conway,  10-Mar-94
Version 1.1 James Conway,   8-Mar-94
Version 1.0 James Conway,   7-Mar-94

Quick and nasty to compare two EMPFTREF.DAT output files:
	* The particles may be in any order in either file
	* Some may be present in one and missing in the other (no comparison)
	* Output is delta theta/phi/omega/x/y/cc1/cc2/other

Mods V1.2
=========
Changed subtraction from start-final to final-start.

Mods V1.1
=========
Add second average row, this time of absolute differences.

*****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>	/* for fabs */

typedef struct {
  int    id;
  double theta;
  double phi;
  double omega;
  double x;
  double y;
  double cc1;
  double cc2;
  double cc3;
  int    matchedFlag;
} PARTICLEDATA;

/* Macros */
#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

#define MIN(m,n) ( (m<n) ? m : n )

/* Constants */

#define MAX_ROWS (10000)		/* we can process 10000 rows	*/

/* Global variables */

char	*rowDataList[MAX_ROWS];		/* save rows here		*/
int	rowDataIndex;			/* count of rows		*/
double	newParameter[MAX_ROWS];		/* new column			*/
char    *notMatchedList[MAX_ROWS];	/* appear in file 2 but not 1	*/
int     notMatchedIndex, averageCount;
PARTICLEDATA particle1Data[MAX_ROWS], average, absAverage;

/* Function prototypes */

int  main(void);
void ReadAllRows(void);
void Process(void);
int  AddToList(char *list[], int *index, char *line);
int  ReadNextRow(char *buffer, int bufSize, FILE *fptr);
int  Convert(char *buffer, PARTICLEDATA *pdata);
int  GetDifferences(PARTICLEDATA *theParticle);
void PrintUnmatched1(FILE *fptr);
void PrintUnmatched2(FILE *fptr);
void PrintParticle(FILE *fptr, PARTICLEDATA *pptr);
void PrintAverages(FILE *fptr);
void AccumulateAverages(PARTICLEDATA *pptr);
void ZeroAverages(void);

/*****************************************************************************
** main
** ====
*/
int main(void)
{
  printf("PFTCHANGES - EMPFTREF.DAT comparison program (v1.2)\n");
  printf("===================================================\n");
  printf("James Conway, 7-Feb-94\n");
  printf("Mods %s\n", __DATE__);
  printf(\
"Calculates differences in orientations between particles before and after\n\
PFTREF. The particles are in EMPFTREF.DAT-format files, and may be in any\n\
order. Particles may be missing from either file.\n\
	\n");

  rowDataIndex = 0;
  notMatchedIndex = 0;
  ReadAllRows();
  Process();
}
/*****************************************************************************
** ReadAllRows
** ===========
Read all non-empty rows as strings into the rowDataList structure.
*/
void ReadAllRows(void)
{
  FILE *ifptr;
  char buffer[1000];

  do {
    printf("Input filename 1 (before EMPFTREF) ? ");
    scanf("%s", buffer);
    ifptr = fopen(buffer, "r");
  }  while(ifptr==NULL);

  printf(">> Reading all rows...\n");
  while(ReadNextRow(buffer, 999, ifptr))  {
    if(rowDataIndex >= MAX_ROWS)  {
      printf("WARNING - reached max input rows: %d. Processing these rows only.\n"
							, rowDataIndex);
      break;
    }
    if(!AddToList(rowDataList, &rowDataIndex, buffer))  {
      printf("ERROR - out of memory on row %d. Processing these rows only.\n"
							, rowDataIndex+1);
      break;
    }
    if(!Convert(buffer, &particle1Data[rowDataIndex-1]))  {
      rowDataIndex -= 1;
      printf("ERROR - unable to process row %d - will skip it\n", rowDataIndex);
      continue;
    }
    particle1Data[rowDataIndex-1].matchedFlag = FALSE;
  }
  printf(">> %d rows read\n", rowDataIndex);
  fclose(ifptr);
}
/*****************************************************************************
** Process
** =======
Read a line from the second input file, and compare with all rows from first 
to see if it is there. If not, add to list of notMatched particles to write out
at end. If it does match, then calculate differences.
*/
void Process(void)
{
  char *cptr;
  double c1, c2, c3;
  int row;
  FILE *ifptr, *ofptr;
  char buffer[1000];
  PARTICLEDATA particle2;

  do {
    printf("Input filename 2 (after EMPFTREF) ? ");
    scanf("%s", buffer);
    ifptr = fopen(buffer, "r");
  }  while(ifptr==NULL);
  do {
    printf("Output filename ? ");
    scanf("%s", buffer);
    ofptr = fopen(buffer, "w");
  }  while(ofptr==NULL);

  printf(">> Processing rows...");

  ZeroAverages();
  row = 0;
  while(ReadNextRow(buffer, 999, ifptr))  {
     row += 1;
    if(!Convert(buffer, &particle2))  {
      printf("ERROR - unable to process row %d - will skip it\n", row);
      continue;
    }
    if(!GetDifferences(&particle2))  {
      AddToList(notMatchedList, &notMatchedIndex, buffer);
      continue;
    }
    PrintParticle(ofptr, &particle2);
    AccumulateAverages(&particle2);
  }
  fclose(ifptr);
  PrintAverages(ofptr);
  PrintUnmatched1(ofptr);	/* particles in file 1 not matched in file 2 */
  PrintUnmatched2(ofptr);	/* particles in file 2 not matched in file 1 */

  fclose(ofptr);
}
/*****************************************************************************
** AddToList
** =========
*/
int AddToList(char *list[], int *index, char *line)
{
  if((list[*index] = malloc(strlen(line)+1)) == NULL)  {
    printf("ERROR - out of memory for storing line data\n");
    return(FALSE);
  }
  strcpy(list[*index], line);
  *index += 1;

  return(TRUE);
}
/*****************************************************************************
** ReadNextRow
** ===========
*/
int ReadNextRow(char *buffer, int bufSize, FILE *fptr)
{
  do  {
    if(fgets(buffer, bufSize, fptr) == NULL)	/* end of file */
      return(FALSE);
    while(strlen(buffer)>0)
      if(buffer[strlen(buffer)-1]=='\n' || buffer[strlen(buffer)-1]=='\r')
        buffer[strlen(buffer)-1] = '\0';
      else
	break;
  }  while(strlen(buffer) == 0);    /* if non-empty, finish, else continue */
  return(TRUE);
}
/*****************************************************************************
** Convert
** =======
*/
int Convert(char *buffer, PARTICLEDATA *pdata)
{
  int count;

  count = sscanf(buffer, "%d,%lf,%lf,%lf,%lf,%lf,%*lf, ,%lf,%lf,%lf"
                   , &(pdata->id)
                   , &(pdata->theta), &(pdata->phi), &(pdata->omega)
                   , &(pdata->x), &(pdata->y)
                   , &(pdata->cc1), &(pdata->cc2), &(pdata->cc3)
		   );
  switch(count)  {
      case 8:	/* extra column not present - calculate it */
      case 9:	/* extra column present - recalculate accurate cc3 anyway */
        pdata->cc3 = pdata->cc1 * pdata->cc2 * 100.0;
	break;
      case 6:
        pdata->cc1 = 0;
      case 7:
        pdata->cc2 = 0;
        pdata->cc3 = 0;
	break;
      default:
        printf("ERROR: row does not have correct format: %d items read\n"
								, count);
	return(FALSE);
        break;
    }
  return(TRUE);
}
/*****************************************************************************
** GetDifferences
** ==============
Search for a matching id in particle1Data, then put difference in theParticle.
*/
int GetDifferences(PARTICLEDATA *theParticle)
{
  int i;

  for(i=0; i<rowDataIndex; i++)
    if(theParticle->id == particle1Data[i].id)  {   /* a match */
      theParticle->theta -= particle1Data[i].theta;
      theParticle->phi   -= particle1Data[i].phi;
      theParticle->omega -= particle1Data[i].omega;
      theParticle->x     -= particle1Data[i].x;
      theParticle->y     -= particle1Data[i].y;
      theParticle->cc1   -= particle1Data[i].cc1;
      theParticle->cc2   -= particle1Data[i].cc2;
      theParticle->cc3   -= particle1Data[i].cc3;
      if(particle1Data[i].matchedFlag)  {
        printf("Warning - particle %d of input file 1 has at least two matches\n");
	printf("          in the other input file\n", particle1Data[i].id);
      }
      else
        particle1Data[i].matchedFlag = TRUE;
      return(TRUE);
    }
  return(FALSE);
}
/*****************************************************************************
** PrintUnmatched1
** ===============
*/
void PrintUnmatched1(FILE *fptr)
{
  int i, noneYetFlag=TRUE;

  for(i=0; i<rowDataIndex; i++)
    if(!particle1Data[i].matchedFlag)  {
      if(noneYetFlag)  {
	noneYetFlag = FALSE;
        fprintf(fptr, "\nParticles in file 1 not matched in file 2\n");
      }
     PrintParticle(fptr, &particle1Data[i]);
    }
}
/*****************************************************************************
** PrintUnmatched2
** ===============
*/
void PrintUnmatched2(FILE *fptr)
{
  int i;

  if(notMatchedIndex>0)
    fprintf(fptr, "\nParticles in file 2 not matched in file 1\n");
  for(i=0; i<notMatchedIndex; i++)
      fprintf(fptr, "%s\n", notMatchedList[i]);
}
/*****************************************************************************
** PrintParticle
** =============
*/
void PrintParticle(FILE *fptr, PARTICLEDATA *pptr)
{
 fprintf(fptr,"%4d,%8.3lf,%8.3lf,%8.3lf,%8.3lf,%8.3lf,%9.3lf,%7.3lf,%7.3lf\n"
		, pptr->id
		, pptr->theta, pptr->phi, pptr->omega
		, pptr->x, pptr->y
		, pptr->cc1, pptr->cc2, pptr->cc3
		);
}
/*****************************************************************************
** PrintAverages
** =============
*/
void PrintAverages(FILE *fptr)
{
  if(averageCount != 0)  {
    average.theta /= (double) averageCount;
    average.phi   /= (double) averageCount;
    average.omega /= (double) averageCount;
    average.x     /= (double) averageCount;
    average.y     /= (double) averageCount;
    average.cc1   /= (double) averageCount;
    average.cc2   /= (double) averageCount;
    average.cc3   /= (double) averageCount;
    fprintf(fptr, "\nAverage differences over %d particles\n", averageCount);
    PrintParticle(fptr, &average);
    printf("\nAverage differences over %d particles\n", averageCount);
    PrintParticle(stdout, &average);
    absAverage.theta /= (double) averageCount;
    absAverage.phi   /= (double) averageCount;
    absAverage.omega /= (double) averageCount;
    absAverage.x     /= (double) averageCount;
    absAverage.y     /= (double) averageCount;
    absAverage.cc1   /= (double) averageCount;
    absAverage.cc2   /= (double) averageCount;
    absAverage.cc3   /= (double) averageCount;
    fprintf(fptr, "Average absolute differences over %d particles\n"
							, averageCount);
    PrintParticle(fptr, &absAverage);
    printf("Average absolute differences over %d particles\n"
							, averageCount);
    PrintParticle(stdout, &absAverage);
  }
}
/*****************************************************************************
** AccumulateAverages
** ==================
*/
void AccumulateAverages(PARTICLEDATA *pptr)
{
  average.theta += pptr->theta;
  average.phi   += pptr->phi;
  average.omega += pptr->omega;
  average.x     += pptr->x;
  average.y     += pptr->y;
  average.cc1   += pptr->cc1;
  average.cc2   += pptr->cc2;
  average.cc3   += pptr->cc3;
  absAverage.theta += fabs(pptr->theta);
  absAverage.phi   += fabs(pptr->phi);
  absAverage.omega += fabs(pptr->omega);
  absAverage.x     += fabs(pptr->x);
  absAverage.y     += fabs(pptr->y);
  absAverage.cc1   += fabs(pptr->cc1);
  absAverage.cc2   += fabs(pptr->cc2);
  absAverage.cc3   += fabs(pptr->cc3);
  averageCount  += 1;
}
/*****************************************************************************
** ZeroAverages
** ============
*/
void ZeroAverages(void)
{
  average.id    = 0;
  average.theta = 0.0;
  average.phi   = 0.0;
  average.omega = 0.0;
  average.x     = 0.0;
  average.y     = 0.0;
  average.cc1   = 0.0;
  average.cc2   = 0.0;
  average.cc3   = 0.0;
  average.id    = 0;
  absAverage.theta = 0.0;
  absAverage.phi   = 0.0;
  absAverage.omega = 0.0;
  absAverage.x     = 0.0;
  absAverage.y     = 0.0;
  absAverage.cc1   = 0.0;
  absAverage.cc2   = 0.0;
  absAverage.cc3   = 0.0;
  averageCount  = 0;
}
/****************************************************************************/
