/* ------------------------------------------------------------------
   Lists.c 
   Description:  Contains all the functions used in list management
   	Functions to add a member to a list, display the list, find
	the number in a given range, and get the current node are 
	included
   Created By: Brant French
   Last Modified:  5-11-98
  ------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include "math.h"
#include "lists.h"
#include "../libR/common.h"

INFO info;  /* A global info structure */

/* -----------------------------------------------------------------
  Add_Particle:  
  Description:  Adds a Particle to the list.  It always adds to the
  	end of the list.  Also sets all values to those passed in
	by the function.
  ------------------------------------------------------------------ */
P_INFO *Add_Particle(P_INFO *head,int particle_number,
		     float theta,float phi,float omega,float x,float y,
		     float mag,float cc1,float cc2,float cc3,
		     int state, int ncol, int nrow)
{
  int rank;
  P_INFO *temp; /* temp structure used in holding the current location
		   in the list. */

  rank = 1;
  if (head==NULL) 
    {
      head = (P_INFO *)malloc(sizeof(P_INFO));
      temp = head;
    }
  else 
    {
      temp = head;
      rank++;
      while (temp->next!=NULL) 
	{/* Find next position */
	  temp=temp->next;
	  rank++;
	}/* Find next position */
      temp->next = (P_INFO *)malloc(sizeof(P_INFO));
      temp = temp->next;
    }
  temp->particle_number=particle_number;
  temp->theta=theta; temp->phi=phi; temp->omega=omega; 
  temp->x=x; temp->y=y; temp->mag=mag;
  temp->origX = x;
  temp->origY = y;
  temp->deltaX = x - (float)(((float)ncol / 2.0)-0.5);
  temp->deltaY = y - (float)(((float)nrow / 2.0)-0.5);
  deltaMax = MAX(temp->deltaX, deltaMax);
  deltaMax = MAX(temp->deltaY, deltaMax);
  deltaMin = MIN(temp->deltaX, deltaMin);
  deltaMin = MIN(temp->deltaY, deltaMin);
  deltaXmin= MIN(temp->deltaX, deltaXmin);
  deltaYmin= MIN(temp->deltaY, deltaYmin);
  deltaXmax= MAX(temp->deltaX, deltaXmax);
  deltaYmax= MAX(temp->deltaY, deltaYmax);
  temp->distance = sqrt((temp->deltaX*temp->deltaX)+
			(temp->deltaY*temp->deltaY));
  distanceMin = MIN(distanceMin,temp->distance);
  distanceMax = MAX(distanceMax,temp->distance);
  temp->rank  = rank;
  temp->cc1=cc1; temp->cc2=cc2; temp->cc3=cc3;
  temp->toggle = state;  /* Set particle to on/off */
  if(particle_number < 0)
    temp->toggle &= INFILEBAD;
  temp->projection = NULL;
  temp->next=NULL;
  return(head);
}

/* -----------------------------------------------------------------
  Display_Patricle:
  Description:  Displays all the links in the list.  Takes the list
	as input and runs through until Null is hit and then stops.
  ------------------------------------------------------------------ */
void Display_Particle_List(INFO head)
{
  P_INFO *temp=head.info;

  printf ("%s\n",head.filename);
  printf ("%f %f %f %f\n%f %f %f\n",head.a,head.b,head.c,head.d,head.e,
				head.f,head.g);
  while (temp!=NULL) {
    printf ("%d %f %f %f %f %f %f %f %f %f\n",temp->particle_number,
			temp->theta,temp->phi,temp->omega,temp->x,temp->y,
			temp->mag,temp->cc1,temp->cc2,temp->cc3);
    temp=temp->next;
  }
}

/* ----------------------------------------------------------------------
  find_max_height:
  Description:  Finds the maximum height value in this given range of
	particles.
  ----------------------------------------------------------------------- */
float find_max_height(P_INFO *head,float start,float end,float step,
		      int paramNum, int *badCount) 
{
  P_INFO *temp = head;
  float maxheight=-99999.9;
  float x;
  float height;
 
  for (x=start;x<end;x+=step) {
    height = find_number_in_range(head,x,x+step,paramNum, badCount);
    if ((height+(*badCount)) > maxheight)
      maxheight = (height+(*badCount));
  }
  if(maxheight <= 0.0)
    maxheight = 1.0;
  return (maxheight);
}

/* ----------------------------------------------------------------------
  find_number_in_range:
  Description:  Finds the number of particles in the range specified
	in the parameter list.  
  ----------------------------------------------------------------------- */
int find_number_in_range(P_INFO *head,float start,float end,int paramNum,
			 int *badCount)
/* This function finds the number of particles cc1 values in the range 
   start to end in O(n) time.  It then returns that number. */
{/* FIND_NUMBER_IN_RANGE */
  P_INFO *temp = head;
  int count;
  
  count = *badCount = 0;
  while (temp!=NULL) {
    if(paramNum==THETAT)
      {/* theta */
	if ((temp->theta<end) && (temp->theta>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* theta */
    else if(paramNum==PHIT)
      {/* phi */
	if ((temp->phi<end) && (temp->phi>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* phi */
    else if(paramNum==OMEGAT)
      {/* omega */
	if ((temp->omega<end) && (temp->omega>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* omega */
    else if(paramNum==XT)
      {/* x */
	if ((temp->x<end) && (temp->x>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* x */
    else if(paramNum==DELTAXT)
      {/* x */
	if ((temp->deltaX<end) && (temp->deltaX>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* x */
    else if(paramNum==ABSDELTAXT)
      {/* x */
	if ((fabs(temp->deltaX)<end) && 
	    (fabs(temp->deltaX)>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* x */
    else if(paramNum==YT)
      {/* y */
	if ((temp->y<end) && (temp->y>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* y */
    else if(paramNum==DELTAYT)
      {/* x */
	if ((temp->deltaY<end) && (temp->deltaY>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* x */
    else if(paramNum==ABSDELTAYT)
      {/* x */
	if ((fabs(temp->deltaY)<end) && 
	    (fabs(temp->deltaY)>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* x */
    else if(paramNum==DISTANCET)
      {/* x */
	if ((temp->distance<end) && 
	    (temp->distance>=start))
	  if (!temp->toggle)
	    count++;
	  else (*badCount)++;
      }/* x */
    else if(paramNum==MAGT)
      {/* mag */
	if ((temp->mag<end) && (temp->mag>=start))
	  if (temp->toggle & HISTMAGBAD)
	    (*badCount)++;
	  else  count++;
      }/* mag */
    else if(paramNum==PFTCC)
      {/* pftcc */
	if ((temp->cc1<end) && (temp->cc1>=start))
	  if (temp->toggle & HISTPFTBAD)
	    (*badCount)++;
	  else  count++;
      }/* pftcc */
    else if(paramNum==PRJCC)
      {/* prjcc */
	if ((temp->cc2<end) && (temp->cc2>=start))
	  if (temp->toggle & HISTPRJBAD)
	    (*badCount)++;
	  else  count++;
      }/* prjcc */
    else if(paramNum==CMPCC)
      {/* cmpcc */
	if ((temp->cc3<end) && (temp->cc3>=start))
	  if (temp->toggle & HISTCMPBAD)
	    (*badCount)++;
	  else  count++;
      }/* cmpcc */

    temp=temp->next;
  }

  return (count);
}/* FIND_NUMBER_IN_RANGE */



void find_minmax(P_INFO *head)
{
  P_INFO *temp = head;

  thetamin = head->theta;
  thetamax = head->theta;
  phimin = head->phi;
  phi_max = head->phi;
  omegamin = head->omega;
  omegamax = head->omega;
  xmin = head->x;
  xmax = head->x;
  ymin = head->y;
  ymax = head->y;
  magmin = head->mag;
  magmax = head->mag;
  cc1min = head->cc1;
  cc1max = head->cc1;
  cc2min = head->cc2;
  cc2max = head->cc2;
  cc3min = head->cc3;
  cc3max = head->cc3;
  while (temp!=NULL) {
    if (!temp->toggle) {
      if (temp->theta < thetamin) thetamin = temp->theta;
      if (temp->theta > thetamax) thetamax = temp->theta;
      if (temp->phi < phimin) phimin = temp->phi;
      if (temp->phi > phi_max) phi_max = temp->phi;
      if (temp->omega < omegamin) omegamin = temp->omega;
      if (temp->omega > omegamax) omegamax = temp->omega;
      if (temp->x < xmin) xmin = temp->x;
      if (temp->x > xmax) xmax = temp->x;
      if (temp->y < ymin) ymin = temp->y;
      if (temp->y > ymax) ymax = temp->y;
      if (temp->mag < magmin) magmin = temp->mag;
      if (temp->mag > magmax) magmax = temp->mag;
      if (temp->cc1 < cc1min) cc1min = temp->cc1;
      if (temp->cc1 > cc1max) cc1max = temp->cc1;
      if (temp->cc2 < cc2min) cc2min = temp->cc2;
      if (temp->cc2 > cc2max) cc2max = temp->cc2;
      if (temp->cc3 < cc3min) cc3min = temp->cc3;
      if (temp->cc3 > cc3max) cc3max = temp->cc3;
    }
    
    temp = temp->next;
  }
/*  printf ("thetamin[%f] thetamax[%f]\n",thetamin,thetamax);
  printf ("phimin[%f] phi_max[%f]\n",phimin,phi_max);
  printf ("omegamin[%f] omegamax[%f]\n",omegamin,omegamax);
  printf ("xmin[%f] xmax[%f]\n",xmin,xmax);
  printf ("ymin[%f] ymax[%f]\n",ymin,ymax);
  printf ("magmin[%f] magmax[%f]\n",magmin,magmax);
  printf ("cc1min[%f] cc1max[%f]\n",cc1min,cc1max);
  printf ("cc2min[%f] cc2max[%f]\n",cc2min,cc2max);
  printf ("cc3min[%f] cc3max[%f]\n",cc3min,cc3max); */
}

/* ---------------------------------------------------------------- 
  Get_Particle:
  Description:  Gets an individual particle from the file specified
	by fIn and places data into the list specified by head.
  ----------------------------------------------------------------- */
void Get_Particle_Info(INFO *head, FILE *fIn, int ncol, int nrow)
{
  float theta,phi,omega,x,y,mag,f7;  /* These variables are used to store */
  float cc1,cc2,cc3;		     /* the individual parameters that */
  char junk;			     /* can then be passed to the      */
  int particle_number;		     /* Add_Particle function          */
  int quit = 0,ijunk,tiltFormat,tot;
  char inStr [512];

  fscanf(fIn,"%s",head->filename); 
  tiltFormat= FALSE;
  if(EQU(head->filename,"TILT_FORMAT"))
    tiltFormat= TRUE;

  if(!tiltFormat)
    {/* Get second line */
#ifdef BEFORE_WE_ADDED_UNITS
      fscanf(fIn,"%f%c %f%c %f%c %f%c %f%c %f%c %f",&theta,&junk,&phi,&junk,
	     &omega,&junk,&x,&junk,&y,&junk,
	     &mag,&junk,&f7);
#else
#ifdef OLDWAY_BRANT
      fscanf(fIn,"%f%c %d%c %f%c %f%c %f%c %f%c %f%c %f",
	     &theta,&junk,
	     &ijunk,&junk,
	     &phi,&junk,
	     &omega,&junk,&x,&junk,&y,&junk,
	     &mag,&junk,&f7);
#else
      get_str(fIn,'\n',inStr);
      get_str(fIn,'\n',inStr);
      sscanf(inStr,"%f%c %d%c %f%c %f%c %f%c %f%c %f%c %f",
	     &theta,&junk,
	     &ijunk,&junk,
	     &phi,&junk,
	     &omega,&junk,&x,&junk,&y,&junk,
	     &mag,&junk,&f7);
#endif /* ifdef vms */
#endif
    }/* Get second line */

  x= y= 1.0;
  mag = 1.0;
  cc1=cc2=cc3= 1.0;

  head->a = theta; head->b = phi; head->c = omega; head->d = x;
  head->e = y; head->f = mag; head->g = f7;

/* OLDWAY Brantt
  if (fscanf(fIn,"%d%c",&particle_number,&junk)!=2) quit = 1;
printf("particle number [%d]\n",particle_number);
*/
  tot = 0;
  while (!quit) {
    if(tiltFormat)
      {/* special */
	fscanf(fIn,"%d%c %f%c %f%c %f",
	       &particle_number,&junk,&theta,&junk,&phi,&junk,&omega);
      }/* special */
    else
      {/* Normal */
#ifdef OLDWAYBRANT
	fscanf(fIn,"%f%c %f%c %f%c %f%c %f%c %f%c",&theta,&junk,&phi,&junk,
	       &omega,&junk,
	       &x,&junk,&y,&junk,&mag,&junk);
	fscanf(fIn,"%f%c %f%c %f",&cc1,&junk,&cc2,&junk,&cc3);
#endif
	get_str(fIn,'\n',inStr);
	if(EQU(inStr,""))
	  particle_number = -99;
	else
	  sscanf(inStr,
		 "%d%c %f%c %f%c %f%c %f%c %f%c %f%c %f%c %f%c %f",
		 &particle_number, &junk,
		 &theta,&junk,&phi,&junk,
		 &omega,&junk,
		 &x,&junk,&y,&junk,&mag,&junk,
		 &cc1,&junk,&cc2,&junk,&cc3
		 );
      }/* Normal */

    if(particle_number > 0 && particle_number < 999777)
      head->info = Add_Particle(head->info,particle_number,theta,phi,omega,
				x,y,mag,
				cc1,cc2,cc3,FALSE, ncol, nrow);
    if(feof(fIn))
      quit = TRUE;
/* OLDWAY Brant 
    if (fscanf(fIn,"%d%c",&particle_number,&junk)!=2) quit = 1;
printf("222  particle number [%d]\n",particle_number);
*/
  }
}/* GET_PARTICLE_INFO */





/********************************************************************
 * add_missing_particles() - We've got to add in blank records for 
 * missing particles so the progtram has something to use when 
 * displaying them.
*/
void add_missing_particles(INFO *head, int nimg, int ncol, int nrow)
{/* ADD_MISSING_PARTICLES */
  int i,foundIt;
  P_INFO *thisInfo;

  for(i=0;i<nimg;i++)
    {/* Find them all */

      foundIt = FALSE;
      thisInfo = head->info;
      while(thisInfo != NULL)
	{/* Do all of them */
	  if(thisInfo->particle_number==(i+1))
	    {/* Found it */
	      foundIt = TRUE;
	      break;
	    }/* Found it */
	  thisInfo = thisInfo->next;
	}/* Do all of them */

      if(!foundIt)
	{/* The add it */
	  head->info = Add_Particle(head->info,
				    (i+1),
				    0.0,0.0,0.0,
				    0.0,0.0,0.0,
				    0.0,0.0,0.0,
				    FALSE,ncol,nrow);

	}/* The add it */

    }/* Find them all */

}/* ADD_MISSING_PARTICLES */





/* ---------------------------------------------------------------- 
  Free_Particle_Info:
  Description:  Free the memory.
  ----------------------------------------------------------------- */
void Free_Particle_Info(INFO *head)
{
  P_INFO *thisInfo,*this2;

  thisInfo = head->info;
  while(thisInfo != NULL)
    {/* Free all */
      if(thisInfo->projection)
	free(thisInfo->projection);
      thisInfo->projection = NULL;
      this2 = thisInfo->next;
      free(thisInfo);
      thisInfo = this2;
    }/* Free all */
  head->info = NULL;

}/* FREE_PARTICLE_INFO */



/*****************************************************************
 * SortParticles()-
*/
void sortParticles(P_INFO *head, int maxParts, int key, int incSort)
{/* SORTPARTICLES */
  int totParts,i,j,k;
  P_INFO *temp; /* temp structure used in holding the current location
		   in the list. */
  int *fids;
  int *ids;
  float *ts;
  float *ps;
  float *os;
  float *xs;
  float *ys;
  float *mags;
  float *cc1s;
  float *cc2s;
  float *cc3s;
  float *deltaXs;
  float *deltaYs;
  float *dists;

  int sorted;

  fids = (int *)malloc(sizeof(int)*maxParts);
  ids = (int *)malloc(sizeof(int)*maxParts);
  ts = (float *)malloc(sizeof(float)*maxParts);
  ps = (float *)malloc(sizeof(float)*maxParts);
  os = (float *)malloc(sizeof(float)*maxParts);
  xs = (float *)malloc(sizeof(float)*maxParts);
  ys = (float *)malloc(sizeof(float)*maxParts);
  mags = (float *)malloc(sizeof(float)*maxParts);
  cc1s = (float *)malloc(sizeof(float)*maxParts);
  cc2s = (float *)malloc(sizeof(float)*maxParts);
  cc3s = (float *)malloc(sizeof(float)*maxParts);
  deltaXs = (float *)malloc(sizeof(float)*maxParts);
  deltaYs = (float *)malloc(sizeof(float)*maxParts);
  dists = (float *)malloc(sizeof(float)*maxParts);

  totParts = 0;
  if (head==NULL) 
    {
      /* Do nothing */
    }
  else 
    {/* Sort it */
      temp = head;
      while (temp!=NULL) 
	{/* Find next position */
	  fids[totParts] = 1;
	  ids[totParts]  = temp->particle_number;
	  ts[totParts]   = temp->theta;
	  ps[totParts]   = temp->phi;
	  os[totParts]   = temp->omega;
	  xs[totParts]   = temp->x;
	  ys[totParts]   = temp->y;
	  mags[totParts]   = temp->mag;
	  cc1s[totParts]   = temp->cc1;
	  cc2s[totParts]   = temp->cc2;
	  cc3s[totParts]   = temp->cc3;
	  deltaXs[totParts]= temp->deltaX;
	  deltaYs[totParts]= temp->deltaY;
	  dists[totParts]  = temp->distance;

	  temp=temp->next;
	  totParts++;
	}/* Find next position */

      /* Do the sort */
      param_srt_rwa(&key,&totParts,fids,ids,ts,ps,os,xs,ys,
		mags,cc1s,cc2s,cc3s,deltaXs,deltaYs,dists,&incSort);

      /* Stuff the sort info backinto the linked list */
      temp = head;
      while (temp!=NULL) 
	{/* Find next position */

	  for(i=0;i<totParts;i++)
	    if(ids[i]==temp->particle_number)
	      {/* Found it */
		temp->rank = i+1;
		break;
	      }/* Found it */

	  temp=temp->next;
	}/* Find next position */

    }/* Sort it */

  free(fids);
  free(ids);
  free(ts);
  free(ps);
  free(os);
  free(xs);
  free(ys);
  free(cc1s);
  free(cc2s);
  free(cc3s);
  free(deltaXs);
  free(deltaYs);
  free(dists);

}/* SORTPARTICLES */





/*int main(void) {
  FILE *fIn;
  INFO info;

  fIn = fopen("dummy.dat","r");
  if (!fIn) {
	  printf ("File can't be opened\n");
	  exit(0);
  }
  info.info = NULL;

  Get_Particle_Info(&info,fIn);
  Display_Particle_List(info);
  return 1;
} */




