      subroutine fftanaly(npix_row, npix_col, pixels, center,
     +             nrfft_omega, rfft_omega, nrfft_row, 
     +             nrfft_col, rfft, rfft_trns, nfc_2d_row, 
     +             nfc_2d_col, fc_2d )
C                                                               FFTANALY
C **********************************************************************
C *                                                               *
C * Subroutine: fftanaly                                          *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1997                                                  *
C * Used in: aamain -- The main routine                           *
C * Function: Fourier transform of a pixel frame and              *
C *           interpolation                                       *
C * Usage:                                                        *
C *   call fftanaly( npix_row, npix_col, pixels, center,          * 
C *                  nrfft_omega, rfft_omega, nrfft_row,          * 
C *                  nrfft_col, rfft, rfft_trns, nfc_2d_row,      *
C *                  nfc_2d_col, fc_2d )                          *
C * Input:                                                        *
C *   nrfft_row (Integer) -- The row number of array "rfft" used  *
C *                          in Fourier transform.                *
C *   nrfft_col (Integer) -- The column number of array "rfft"    *
C *                          used in Fourier transform.           *
C *   nrfft_omega (Integer) -- The limit on the size of array     *
C *     "rfft_omega", the working site for calling LAPACK         *
C *     routine.                                                  *
C *   nfc_2d_row (Integer) -- The row number of array "fc_2d".    *
C *   nfc_2d_col (Integer) -- The column number of array "fc_2d". *
C *   npix_row, npix_col (Integer) -- Dimension of array PIXELS.  *
C *   pixels (Real) -- The pixel frame.                           *
C *   center (Real) -- The center of the pixel frame.             *
C *   rfft, rfft_trns (Real) -- Working array for 2D DFT.         *
C * Output:                                                       *
C *   fc_2d (Complex) -- The result of 2D DFT.                    *
C *                                                               *
C * Algorithm:                                                    *
C *   The pixel values are rearranged and stored in the larger    *
C *   array                                                       *
C *        RFFT( 1:NRFFT_ROW, 1:NRFFT_COL )                       *
C *   where both NRFFT_ROW and NRFFT_COL are even; the extra      *
C *   spaces in RFFT are filled with zeros.  The discrete complex *
C *   Fourier coefficients of the nonzero rows of RFFT are        *
C *   computed and then the coefficients of the columns are       *
C *   computed.  By changing the phases of the coefficients, the  *
C *   origin is shifted so that the origin of the pixel values is *
C *   moved to (CENTER(1),CENTER(2)). The complex coefficients    *
C *   are stored in the array CFFT so that                        *
C *   CFFT( 1:NCFFT_ROW, !:NCFFT_COL ) contains the coefficients  *
C *   arranged in the standard way that DFFT store values.  These *
C *   are rearranged so that interpolation is easy to carry out.  *
C *   The output array is complex:                                *
C *                                                               *
C *        FC_2d( -NFC_2D_ROW:NCF_2D_ROW, 0:NFC_2D_COL)           *
C *                                                               *
C *   The first index corresponds to k and the second to h which  *
C *   are the transformations subscripts for y and x,             *
C *   respectively.                                               *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 0.6, Jul.6, 1998                      *
C *       Move code for Fourier transform and interpolation from  *
C *       the main routine to this subroutine, localize some      *
C *       dynamically allocated arrays                            *
C *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *     - Hong Lin, Version 2.0, Oct. 4, 1999                     *
C *       Adding "accurate" interpolation method in which the     *
C *       weight of the values used in interpolation is           *
C *       determined by SINC function.                            *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                               *
C *****************************************************************
C   Input;   -----------------------------------------------------------
 
C  NPIX_ROW = NPIX_COL  INTEGER*4; dimensions of square array PIXELS
C          NPIX_ROW .eq. NPIX_COL

C  PIXELS( 1:NPIX_ROW, 1:NPIX_COL)  REAL*4;  square array containing
C          pixel values.  PIXELS(irow,icol) contains the value at
C          U = IROW - 1, V = ICOL - 1
 
C  CENTER( 1:2 )  REAL*4; pixel coordinates, (CENTER_U,CENTER_V), of the
C          projection of the center of the particle.  E.g.
C             if CENTER_U .eq. 10.2, CENTER_V .eq. 11.75,
C          then, because the indices of array PIXELS begin at 1,
C          the projection of the center is in the interior of the pixel
C          stored at PIXELS( 11, 12 ) [Input]

C WORK ARRAYS:  --------------------------------------------------------
 
C  NRFFT_OMEGA  INTEGER*4;  array dimension of RFFT_OMEGA; this must be
C          .ge. NRFFT_COL+15 where NRFFT_COL is the column dimension of
C          RFFT.

C  RFFT_OMEGA( 1:NRFFT_OMEGA )  REAL*4;  array of used in the DFFT 
C          analyses of rows of RFFT.  The values in RFFT_OMEGA must
C          be assigned before the first call of routine VRFFTF:
C                CALL VRFFTI( NRFFT_COL, RFFT_OMEGA )
C          The values in RFFT_OMEGA are initialized once; they are
C          used each time VRFFTF is called; the contents of RFFT_OMEGA
C          must not be changed between calls of VRFFTF.
 
C  NRFFT_ROW = NRFFT_COL  INTEGER*4;  array dimensions for the (square)
C          work arrays RFFT and RFFT_TRNS,  NRFFT_ROW ,eq, NRFFT_COL and
C          NRFFT_ROW must be even

C  RFFT( 1:NRFFT_ROW, 1:NRFFT_COL )  REAL*4;  square work array.
C          RFFT is filled with pixel values and zero-fill when RFFT
C          is larger than PIXELS. DFFT analyses of its rows are
C          carried out by VRFFTF.

C  RFFT_TRNS( 1:NRFFT_COL, 1:NRFFT_ROW )  REAL*4;  square work array
C          used during the execution of VRFFTF.

C  NFC_2D_ROW = NFC-2D_COL  INTEGER*4; limits on the indices of the
C          output array FC_2D.
 
C OUTPUT:  -------------------------------------------------------------
 
C  FC_2D( 0:NFC_2D_ROW, -NFC_2D_COL:NFC_2D_COL )  COMPLEX*4;  array with
C          NFC_2D_ROW .eq. NFC-2D_COL.  It will contain the Fourier
C          coefficients, FC_2D( H, K ) the pixels values as stored in
C          the array RFFT.  The FFT with respect to U yields
C          coefficients with indices H, the FFT with respect to V yields
C          coefficients with indices K.  Conjugate symmetry gives the
C          coefficient for all H,K:
C               -NFC_2D_ROW <= H, K <= NFC_2D_ROW = HFC_2D_COL
C          The values  FC_2D( H, K ), H = -1, -2, ..., -NFC_2D_ROW
C          are not stored.  Such a value is equal to the complex
C          conjugate of FC_2D( H,-K) which is stored.
 
C  ---------------------------------------------------------------------
C Outline of execution:

C The input pixel values are rearranged, as explained below, in the
C array RFFT which is larger than the pixel array.  Zeroes are put into
C RFFT to fill it.  A vectorized real FFT is applied to each of the rows
C of RFFT.  RFFT is transposed into RFFT_TRNS and the same FFT is
C applied to RFFT_TRNS.  The resulting NRFFT_ROW*NRFFT_COL real values
C are combined to form (RFFT_ROW+1)*NRFFT_COL/2 complex Fourier
C coefficients and stored in
C         FC_2D(0:NRFFT_ROW/2,-NRFFT_COL/2:NRFFT_COL/2),
C with FC_2D(H,K) corresponding to the U,V axes of the original pixels.
C Rows of FC_2D are multpiled by EXP(-2pi i DU H / U_PERIOD ) and
C columns are multiplied by EXP(-2pi i DV K / V_PERIOD ) to obtain the
C Fourier coefficients of the pixels whose origin has been translated
C by (DU,DV)
C  ---------------------------------------------------------------------
      implicit none
      include '../include/allprogs.inc'
C
C Arguments
      integer*4 npix_row, npix_col, nrfft_row, nrfft_col, nrfft_omega,
     +        nfc_2d_row, nfc_2d_col
      complex fc_2d( 0:nfc_2d_row, -nfc_2d_col:nfc_2d_col )
      real*4  pixels(npix_row, npix_col), rfft(nrfft_row, nrfft_col),
     +       rfft_trns(nrfft_col, nrfft_row), rfft_omega(nrfft_omega),
     +        center(2)
C
C Local
      complex cmplx_recip, c_exp, c_exp_1
      integer*4 icol, irow, icol_a, icol_a_1, irow_a_1, irow_b_1_rfft,
     +        icol_c, icol_c_1_rfft, irow_c_1_rfft, irow_d_1,
     +        icol_rfft, h, k, num_cols_a, num_rows_a, num_rows_b,
     +        num_cols_c, num_rows_c, num_rows_d, nfc_2d_end,
     +        nrfft_d2,
     +        nrfft_trns_row, nrfft_trns_col
      real*4    center_u, center_v, cc, ss, sc, cs, du, dv,
     +        thetau, thetav, pixelmin, pixelmax
      character*9 subname
      data        subname / 'FFTANALY' /

C
C External
C     BLAS routines; typically these are optimized code from vendor.
C     See modern books on numerical linear algbra or the syntax given
C     in E. Anderson, et al., LAPACK User's Guide, 2nd Edit., Siam
C     Publications, 1996.
C
      external cscal, ccopy, scopy
C
C     Syntax:
C       CALL CSCAL( N, CON, C, INC )
C         Multiply a vector C by a constant CON
C            N    INTEGER*4: number of entries of C to be scaled
C            CON  COMPLEX*4: constant
C            C    COMPLEX*4: vector of length at least (N-1)*INC+1
C            INC  INTEGER*4: stride of C
C       The procedure is equivalent to
C            DO JC = 1, ( (N-1)*INC + 1 ), INC 
C              C(JC) = CON * C(JC)
C            ENDDO
C
C       With `x' is `C' when x and Y are complex; `x' is `S' when X and
C       Y are REAL*4
C       CALL xCOPY( N, X, INCX, Y, INCY )
C         Copy vector X to vector Y
C             N    INTEGER*4: number of entries of X to be copied
C             X    COMPLEX*4 or REAL*4: vector to be copied to Y;
C                     X must have length at least N*INCX
C             INCX INTEGER*4: stride of X
C             Y    COMPLEX*4 or REAL*4; vector copied from X;
C                     Y must have length at least N*INCY
C             INCY INTEGER*4: stride of Y
C       The procedure is equivalent to
C            JY = 1
C            DO JX = 1, ( (N-1)*INCX + 1 ), INCX
C              Y(JY) = X(JX)
C              JY = JY + INCY
C            ENDDO
C
C     REAL DFFT analysis routines: Sweet, R.A., et al., VFFTPK, A
C     Vectorized Pacakage of Fortran Subprograms for the Fast Fourier
C     Transform of Multiple Real Sequences, 1990; obtainable by e-mail,
C     ftp, or www from netlib@ornl.gov
C
      external vrffti, vrfftf
C**********************************************************************
C
      nrfft_trns_row = nrfft_col    
      nrfft_trns_col = nrfft_row
C
C     The PIXELS array is square.
C
C     The array, PIXELS( 1:NPIX_ROW 1:NPIX_COL) is     --------- cols
C     partitioned into 4 arrays A, B, C, D:            | C | B |
C         A has nonnegative u and v coordinates        ----------- v
C         B has negative u, nonnegative v              | D | A |    
C         C has negative u and v                       ----|----   
C         D has nonnegative u and negative v           |   |   |
C                                                      |   | 
C                                                    rows  u 
C 
C     Transfer and rearrange PIXELS to the larger work array RFFT
C     and fill the rest of RFFT with zeros.
C
C     After rearrangement, the pixel values are      --------------- v
C     stored in RFFT as indicated to the right:      | A | 0 | D |
C                                                    -------------
C                                                    | 0 | 0 | 0 |
C                                                    -------------
C                                                    | B | 0 | C |
C                                                    -------------
C                                                    |
C                                                    u
C
C     See NOTE #1 at end of this subroutine.
C
      center_u = center(1)
      center_v = center(2)
C
C     The U-coordinate begins with 0 at row 1 of PIXELS and is oriented
C     down PIXELS rows.  The V-coordinate begins with 0 at column 1
C     of PIXELS and is oriented across the columns of PIXELS.
C              V-coordinate  0  1  2  3  4  5  ...
C              PIXELS(U,:)   1  2  3  4  5  6  ...
C     If CENTER_V .eq. V = 3.2, then IFIX( CENTER_V ) + 1 is 4
C     which is the column number of the PIXELS containing the V-origin;
C     similarly for CENTER_U
C
      irow_a_1   = ifix( center_u ) + 1
      icol_a_1   = ifix( center_v ) + 1
      num_rows_a = npix_row + 1 - irow_a_1
      num_cols_a = npix_col + 1 - icol_a_1
C 
C     The first row number of A in PIXELS is the same as the first row
C     number of D in PIXELS.
C
      irow_d_1   = irow_a_1
C 
C     The first column number of A in PIXELS is the same as the first
C     column number of B in PIXELS
C
C     The number of rows of D is the same as the number of rows of A
C
      num_rows_d = num_rows_a
C
C     The number of columns of B and A are the same.
C 
C     Find the number of rows and columns of C --- these are each
C     one less than the first row and column number of A in PIXELS
C
      num_rows_c    = irow_a_1 - 1
      num_cols_c    = icol_a_1 - 1
C
C     Find the first row and column numbers of C in array RFFT.
C
      irow_c_1_rfft = nrfft_row + 1 - num_rows_c
      icol_c_1_rfft = nrfft_col + 1 - num_cols_c
C
C     The first column number of C in RFFT is the same as the first
C     column number of D in RFFT.
C     The number of the first row of C in RFFT is the same as the
C     number of the first row of B in RFFT
C 
C     The number of rows of B is the same as the number of rows of C
C
      num_rows_b    = num_rows_c
C
C     The first row of B in RFFT is the same as the first row of C
C
      irow_b_1_rfft = irow_c_1_rfft
C 
C     Zero RFFT
C
      do 1010 icol = 1, nrfft_col
        do 1000 irow = 1, nrfft_row
          rfft( irow, icol ) = 0.
 1000   continue
 1010 continue
      
C
C     Use the BLAS routine SCOPY to copy real vectors from one location
C     to another.
C     Syntax:
C       SCOPY(vector_length,input_vector,stride,output_vector,stride)
C
      icol_rfft = 0
      do 1020 icol_a = icol_a_1, npix_col
        icol_rfft = icol_rfft + 1
C
C       Copy a column of A in PIXELS to a column of A in RFFT.
C
        call scopy( num_rows_a, pixels(irow_a_1,icol_a), 1,
     +                          rfft(1,icol_rfft),  1 )
C
C       Copy a column of B in PIXELS to a column of B in RFFT.
C
        call scopy( num_rows_b, pixels(1,icol_a), 1,
     +                          rfft(irow_b_1_rfft,icol_rfft), 1 )
 1020 continue
C
      icol_rfft = icol_c_1_rfft - 1
      do 1030 icol_c = 1, num_cols_c
        icol_rfft = icol_rfft + 1
C
C       Copy a column of C in PIXELS to a column of C in RFFT.
C
        call scopy( num_rows_c, pixels(1,icol_c), 1,
     +                          rfft(irow_c_1_rfft,icol_rfft), 1 )
C
C       Copy a column of D in PIXELS to a column of D in RFFT.
C
        call scopy( num_rows_d, pixels(irow_d_1,icol_c), 1,
     +                          rfft(1,icol_rfft), 1 )
 1030 continue
C 
C     Carry out real FFT analysis (foward transformation) of each row
C     of  RFFT.
C
C     Transformation parameters, RFFT_OMEGA, were computed the first
C     time this routine (FFTANALY) is called.  The values in RFFT_OMEGA
C     must not change between successive calls of VRFFTF unless the
C     size of RFFT changes.  The input is the array RFFT of rearranged
C     pixel values and a work array RFFT_TRNS.  The transformation is
C     with respect to the column index, V, which becomes the index K of
C     the Fourier coefficients.  The Fourier coefficents are returned in
C     RFFT whose column coordinate is now K and whose row coordinate
C     remains U.
C 
      if( npix_row .lt. nrfft_row ) then
C
C       Array RFFT has more rows than PIXELS; array RFFT contains
C       rows of zeros --- do not transform the rows of zeros.
C                                                            CALL VRFFTF
C                                                            ***********
C       Find Fourier coefficients of rows containing A and D.
C
        call vrfftf(num_rows_a, nrfft_col,
     +              rfft(1,1), rfft_trns(1,1),
     +              nrfft_col, rfft_omega)
C
C       Find Fourier coefficients of rows containing B and C.
C
        call vrfftf(num_rows_c, nrfft_col,
     +              rfft(irow_c_1_rfft,1), rfft_trns(irow_c_1_rfft,1),
     +              nrfft_col, rfft_omega)
      else
C
C       Array RFFT has same number of rows as PIXELS.  Find Fourier
C       coefficients of each row of RFFT
C
        call vrfftf(nrfft_row, nrfft_col,
     +              rfft(1,1), rfft_trns(1,1),
     +              nrfft_col, rfft_omega)
      endif
C 
C     See NOTE #2 at end of this subroutine.
C 
C     Transpose RFFT and store the transpose in RFFT_TRNS.  The column
C     index of RFFT_TRNS is U and its row index is K.
C
      do 1050 irow = 1, nrfft_row
        do 1040 icol = 1, nrfft_col
          rfft_trns(icol,irow) = rfft(irow,icol)
 1040   continue
 1050 continue
C 
C     Carry out a real FFT analysis (foward transformation) of each row
C     of RFFT_TRNS, i.e., with respect to U and the index of the Fourier
C     coefficients is H.  The Fourier coefficients are returned in
C     RFFT_TRNS whose column coordinate is now H and row index is K.
C
C                                                            CALL VRFFTF
C                                                            ***********
      call vrfftf(nrfft_trns_row, nrfft_trns_col,
     +            rfft_trns(1,1), rfft(1,1),
     +            nrfft_col, rfft_omega)
C 
C     Zero the complex output array                  ---------- k
C     FC_2D( 0:NFC_2d_ROW, -NFC_2D_COL:NFC_2D_COL )  |   |   |
C                                                    ----|----
C                                                        |
C                                                        h
C 
      do 1070 icol = -nfc_2d_col, nfc_2d_col
        do 1060 irow = 0, nfc_2d_row
          fc_2d( irow, icol ) = ( 0., 0. )
 1060   continue
 1070 continue

C 
C     The complex array FC_2D( 0:NFC_2D_ROW, -NFC_2D_COL:NFC_2D_COL )
C     will contain the complex valued Fourier coefficients of the
C     real values which were originally stored in RFFT (obtained
C     by rearranging the pixel values and zero fill).  The current
C     real values in the array RFFT_TRNS(1:NRFFT_COL, 1:NRFFT_ROW)
C     have to be combined to get the complex valued Fourier
C     coefficients.  The Fourier coefficients are periodic with periods
C     COL_PERIOD = 2*NFC_2D_COL and ROW_PERIOD = 2*NFC_2D_ROW
C
C     See NOTE #3 at end of this subroutine.
C
C     Set the value of the complex Fourier coefficient at the origin.
C
      fc_2d( 0, 0 ) = cmplx( rfft_trns( 1, 1), 0.)
C
      nrfft_d2 = nrfft_trns_col / 2
C
CXXXXXif( 2*nrfft_d2 .eq. nrfft_trns_col ) then
CXXXXXIn this version, only even NRFFT_TRNS_COL is allowed.
C       There is an even number of rows and colums in RFFT.
C       NFC_2D_END is equal to NRFFT_D2
C
        nfc_2d_end = nrfft_d2
C
C       Set complex value at end of H-axis.
C
        fc_2d(nrfft_d2, 0) = cmplx( rfft_trns(1,nrfft_trns_col), 0.)
C 
C       Set complex values at four corners; use periodicity.
C
        fc_2d(0, nrfft_d2) = cmplx(rfft_trns(nrfft_trns_row,1), 0.)
        fc_2d(0,-nrfft_d2) = fc_2d(0, nrfft_d2)
        fc_2d(nrfft_d2, nrfft_d2)
     +    = cmplx(rfft_trns(nrfft_trns_row,nrfft_trns_col), 0.)
        fc_2d(nrfft_d2,-nrfft_d2) = fc_2d(nrfft_d2, nrfft_d2)

C       Set rest of values along along H = 0 and H = NRFFT_D2
C
        do 1080 k = 1, nrfft_d2 - 1
          cc                    = rfft_trns( 2*k,   1 )
          sc                    = rfft_trns( 2*k+1, 1 )
          fc_2d( 0,  k )        = cmplx( cc, sc )
C
C         FFT of real function is conjugate symmetric.
C
          fc_2d( 0, -k )        = cmplx( cc,-sc )

          cc                    = rfft_trns( 2*k,   nrfft_row )
          sc                    = rfft_trns( 2*k+1, nrfft_row )
          fc_2d( nrfft_d2, k ) = cmplx( cc, sc )
C
C         FFT of real function is conjugate symmetric.
C
          fc_2d( nrfft_d2,-k ) = cmplx( cc,-sc )
 1080   continue

C       Set rest of values along along K = 0 and K =   NRFFT_D2
C       and use periodicity to set values along  K = - NRFFT_D2
C
        do 1090 h = 1, (nrfft_d2 - 1)
          cc                    = rfft_trns( 1, 2*h   )
          cs                    = rfft_trns( 1, 2*h+1 )
          fc_2d( h, 0 )         = cmplx( cc, cs )
          cc                    = rfft_trns( nrfft_col, 2*h   )
          cs                    = rfft_trns( nrfft_col, 2*h+1 )
          fc_2d( h, nrfft_d2 ) = cmplx( cc, cs )
C
C         Impose periodic conditions.
C
          fc_2d( h,-nrfft_d2 ) = cmplx( cc, cs )
 1090   continue

C       Fill FC_2D for  H =  1, 2, ... ,  NRFFT_D2 - 1
C                       K =  1, 2, ... ,  NRFFT_D2 - 1
C                       K = -1,-2, ... , -NRFFT_D2 + 1
 
        do 1110 k = 1, (nrfft_d2 - 1)
          do 1100 h = 1, (nrfft_d2 - 1)
            cc            = rfft_trns( 2*k,   2*h   )
            ss            = rfft_trns( 2*k+1, 2*h+1 )
            sc            = rfft_trns( 2*k+1, 2*h   )
            cs            = rfft_trns( 2*k,   2*h+1 )
            fc_2d( h, k ) = cmplx( cc - ss,  sc + cs )
            fc_2d( h,-k ) = cmplx( cc + ss, -sc + cs )
 1100     continue
 1110   continue

Cji*************************************************************
Cji  exp( 2 pi i [ H * DU / ROW_PERIOD + K * DV / COL_PERIOD ] )
Cji  = complex(cos(2 pi [ H * DU / ROW_PERIOD + K * DV / COL_PERIOD ]),
Cji            sin(2 pi [ H * DU / ROW_PERIOD + K * DV / COL_PERIOD ]))
Cji  center correction by ( DU, DV )

C     Modify the Fourier coefficient so as to translate the origin
C     of the pixels by ( DU, DV ).  The coefficient FC_2D(K,H)
C     is multiplied by
C          exp( 2 pi i [ H * DU / ROW_PERIOD + K * DV / COL_PERIOD ] )
C        = exp( i H THETAU ) * exp( i K THETAV )

      dv      = center_v - float( ifix( center_v ) )
      thetav  = twopi * dv / float( 2*nfc_2d_end )
      c_exp_1 = cmplx( cos( thetav ), sin( thetav ) )

C     Use the BLAS routine CSCAL to multiply the K-th column of complex
C     FC_2D by C_EXP**K for K = 1, ..., NFC_2D_END
C     and by CONJG( C_EXP(K) ) for K = -1, ..., -NFC_2D_END. 
C     Syntax:
C        CSCAL(vectorlength, complex constant, complex vector, stride)

      c_exp   = c_exp_1
      do 1150 k = 1, nfc_2d_end
        call cscal(nfc_2d_end,        c_exp  , fc_2d( 1,  k), 1 )
        call cscal(nfc_2d_end, conjg( c_exp ), fc_2d( 1, -k), 1 )
        c_exp = c_exp * c_exp_1
 1150 continue

C     Impose periodic boundary conditions.
      do 1160 h = 1, nfc_2d_end
        fc_2d( h, nfc_2d_end ) = 0.5 * (  fc_2d( h,   nfc_2d_end )
     +                                  + fc_2d( h, - nfc_2d_end ) )
        fc_2d( h,-nfc_2d_end ) = fc_2d( h, nfc_2d_end )
 1160 continue

      du      = center_u - float( ifix( center_u ) )
      thetau  = twopi * du / float( 2*nfc_2d_end )
      c_exp_1 = cmplx( cos( thetau ), sin( thetau ) )

C     Use the BLAS routine CSCAL to multiply the H-th row of FC_2D
C     by C_EXP_1**H

      c_exp   = c_exp_1
      do 1170 h =  1, nfc_2d_end
        call cscal( (2*nfc_2d_end+1), c_exp, fc_2d(h,-nfc_2d_end),
     +              (nfc_2d_end+1))
        c_exp = c_exp * c_exp_1
 1170 continue

C     Impose periodic boundary conditions; use conjugate symmetry
      do 1180 k = 1, nfc_2d_end
        fc_2d( nfc_2d_end,  k ) = 0.5 * ( fc_2d( nfc_2d_end, k )
     +                              + conjg( fc_2d( nfc_2d_end, -k ) ) )
        fc_2d( nfc_2d_end, -k ) = conjg( fc_2d( nfc_2d_end, k ) )
 1180 continue

CXXXXX The following code was used when NRFFT_TRNS_COL was allowed to
CXXXXX be odd.  The code is retained in case someone finds it
CXXXXX advantageious to use an odd value of NRFFT_TRNS_COL.
CXXXXX      else
CXXXXXC       There is an odd number of rows and columns in RFFT.
CXXXXX        nrfft_m1_d2 = ( nrfft_trns_col - 1 ) / 2
CXXXXX        nfc_2d_end  = nrfft_m1_d2
CXXXXX 
CXXXXXC       Set complex values along rest of H = 0 and rest of K = 0
CXXXXXC       axes.
CXXXXX        do 1120 hk = 1, nrfft_m1_d2
CXXXXX          cc                    = rfft_trns( 1, 2*hk   )
CXXXXX          cs                    = rfft_trns( 1, 2*hk+1 )
CXXXXX          fc_2d( hk, 0 )        = cmplx( cc, cs)
CXXXXX
CXXXXX          cc                    = rfft_trns( 2*hk,   1 )
CXXXXX          sc                    = rfft_trns( 2*hk+1, 1 )
CXXXXX          fc_2d( 0,  hk )        = cmplx( cc, sc )
CXXXXX          fc_2d( 0, -hk )        = cmplx( cc,-sc )
CXXXXX
CXXXXX 1120   continue
CXXXXX
CXXXXXC       Fill FC_2D for  H =  1, 2, ... ,  NRFFT_M1_D2
CXXXXXC                       K =  1, 2, ... ,  NRFFT_M1_D2
CXXXXXC                       K = -1,-2, ... , -NRFFT_M1_D2
CXXXXX 
CXXXXX        do 1140 k = 1, nrfft_m1_d2
CXXXXX          do 1130 h = 1, nrfft_m1_d2
CXXXXX            cc            = rfft_trns( 2*k,   2*h   )
CXXXXX            ss            = rfft_trns( 2*k+1, 2*h+1 )
CXXXXX            sc            = rfft_trns( 2*k+1, 2*h   )
CXXXXX            cs            = rfft_trns( 2*k,   2*h+1 )
CXXXXX            fc_2d( h, k ) = cmplx( cc - ss,  sc + cs )
CXXXXX            fc_2d( h,-k ) = cmplx( cc + ss, -sc + cs )
CXXXXX 1130     continue
CXXXXX 1140   continue
CXXXXX      endif

C >>> NOTE #1 **********************************************************
C     The array, PIXELS( 1:NPIX_ROW 1:NPIX_COL) is partitioned 
C     into 4 arrays A, B, C, D:                          --------- cols
C         A has nonnegative u and v coordinates          | C | B |
C         B has nonnegative u, negative v                ----------- v
C         C has negative u and y                         | D | A |
C         D has negative u and nonnegative v             ----|----
C     Near the end of FFTANALY, a slight adjustment      |   |   |
C     made to the Fourier coefficients to change the     |   | 
C     origin of the pixels to (center_u,center_v)      rows  u 

C     With   ICOL_A_1 = first pixel column of A
C            IROW_A_1 = first pixel row of A
C     the subarrays are:
C        A = PIXELS( IROW_A_1 : NPIX_ROW,   ICOL_A_1 : NPIX_COL  )
C        B = PIXELS(        1 : IROW_A_1-1, ICOL_A_1 : NPIX_COL  )
C        C = PIXELS(        1 : IROW_A_1-1,        1 : ICOL_A_1-1 )
C        D = PIXELS( IROW_A_1 : NPIX_ROW,          1 : ICOL_A_1-1 )
 
C     The pixel values are moved to the array        -------------
C     RFFT in which an FFT is carried out with       | A | 0 | D |
C     respect to v.  The blocks A,B,C,D are          -------------
C     arranged so that the origin A is closest to    | 0 | 0 | 0 |
C     RFFT(1,1).  Array RFFT is larger than PIXELS;  -------------
C     the rest of RFFT is filled with zeros.  The    | B | 0 | C |
C     FFT imposes  periodicity on the values         -------------
C     with respect to v.  The original periodicity is changed from

C          -------------------------
C       ...| D | A | D | A | D | A | ...
C          -------------------------
C       ...| C | B | C | B | C | B | ...
C          -------------------------
C     to
C          -------------------------------------
C       ...| A | 0 | D | A | 0 | D | A | 0 | D | ...
C          -------------------------------------
C       ...| 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | 0 | ...
C          -------------------------------------
C       ...| B | 0 | C | B | 0 | C | B | 0 | C | ...
C          -------------------------------------
 
C     The pixel containing the center, (center_u,center_v), is put into
C     RFFT(1,1), the upper left corner of RFFT or, equivalently, A

C     The coordinate system used for the pixels has the origin at
C     the upper left corner of the array PIXELS.  Thus, PIXELS(1,1)
C     corresponds to u = 0, v = 0.  Hence when finding the indices
C     in PIXELS for the center, we add 1 to ifix(center).
C     Because of the way the array PIXELS is filled by reading the
C     data input file, the u-axis points downward ( row 3 corresponds
C     to u = 2 ) and the v-axis point to the left ( col 3 corresponds
C     to v = 2 ).  The center is usually close to the center of
C     the array, so if PIXELS has 100 rows and 100 column, then
C     the center is near row 50 and column 50.
C >>> END NOTE #1 ******************************************************

C >>> NOTE #2 **********************************************************
C     VRFFTF overwrites the each row of RFFT with the real DFFT of the
C     row.  Row J of the output array, RFFT( J, 1:NRFFT_COL), contains
C     the real and the imaginary parts of the the independent Fourier
C     coefficients, F(J,K), of the J-th row of the initial values
C     stored in RFFT( J, 1:NRFFT_COL).  The input values stored in RFFT
C     are real.

C     If  N  is even, then there are N/2 = NRFFT_COL/2 independent
C     complex values, two of which are real:
C       REAL(F(0)) REAL(F(1)) IMAG(F(1)), ..., IMAG(F(N/2-1)), REAL(N/2)
C     This gives the complete set of Fourier coefficients because
C     F(-K) = ( REAL(F(-K), IMAG(F(-K) ) = ( REAL(F(K), -IMAG(F(K) )

C     If  N  is odd, then there are (N+1)/2 = (NRFFT_COL+1)/2
C     independent complex values, one of which is real:
C       REAL(F(0)) REAL(F(1)) IMAG(F(1)), ..., REAL(F(N-1)/2),
C                                                  IMAG( F((N-1)/2)
 
C     The output array contains
C
C          RFFT(J,1)     = REAL( F(J,0) ),
C                        = SQRT(1/N)*SUM(L=0,N-1)[ X(J,L) ],
C
C          RFFT(J,2*K)   = REAL( F(J,K) )
C                        = SQRT(1/N)*SUM(L=0,N-1)[X(J,L)*COS(2L*K*PI/N)]
C
C                   for K = 1, 2, ... , (N+1)/2 - 1 when N is odd
C               and for K = 1, 2, ... , N/2 - 1     when N is even 
C
C          RFFT(J,2*K+1) = IMAG( F(J,K) )
C                        =-SQRT(1/N)*SUM(L=0,N-1)[X(J,L)*SIN(2L*K*PI/N)]
C
C                   for K = 1, 2, ... , (N+1)/2 - 1 when N is odd
C               and for K = 1, 2, ... , N/2 - 1     when N is even 
C
C          In addition, when N is even:
C
C          RFFT(J,N) = REAL( F(J,N/2) ).
C                    = SQRT(1/N)*SUM(L=0,N-1)[ (-1)**L*X(J,L) ].
C
C       FFC(J,0) = A(J,0) = RFFT(J,1) is real.
C     When N is even,
C       FFC(J,N/2) = A(j,N/2) = RFFT(J,N) is real
C     Otherwise, FFC is complex:
C    
C           FFC(J,K) = A(J,K) + i B(J,K) = RFFT(J,2K) + i RFFT(J,2K+1)
C                for K = 1, 2, ..., (N+1)/2 - 1 when N is odd
C                for K = 1, 2, ..., N/2 - 1     when N is even
C >>> END NOTE #2 ******************************************************

C >>> NOTE #3 **********************************************************
 
C     After the VRFFTF has computed the Fourier coefficients of the
C     rows of RFFT, RFFT is transposed and stored in RFFT_TRNS
 
C     VRFFTF is called again, this time with RFFT_TRNS as the input
C     array and it computes the Fourier coefficients of the rows
C     of RFFT_TRNS (fomerly the columns of RFFT).  After VRFFTF
C     finishes, the entries stored in RFFT_TRNS can be combined to form
C     the real and imaginary parts of the complex valued double
C     transform of the pixel values as originally stored in RFFT.

C     Let FFT denote an array containing the Fourier coefficients and
C     let N = NRFFT_TRNS_ROW = NRFFT_TRNS_COL.  Let H denote the
C     subscript associated with transformation with respect to U and
C     let K noete the subscript associated with transformation with
C     respect to V.  In the array RFFT_TRNS, the H are column indices
C     and the K are row indices.  In FFT we use H as row indices and
C     K as column indices.  Because of conjugate symmetry,
C            FFT(-H, K) = CONJG( FFT( H,-K) ) and
C            FFT(-H,-K) = CONJG( FFT( H, K) ),
C     We store values in FFT
C       for even N:  H = 0, 1, ..., N/2
C                    K = -N/2+1,   ..., -1, 0, 1, ..., N/2
C       for odd  N:  H = 0, 1, ..., (N-1)/2
C                    K = -(N-1)/2, ..., -1, 0, 1, ..., (N-1)/2

C Interior of domain: For positive integers
C    H,K = 1,2, ..., N/2 - 1, N even [or (N-1)/2 for odd N]:

C FFT(  H,  K) =   (  RFFT_TRNS(2K,    2H) - RFFT_TRNS(2K + 1, 2H + 1) )
C              + i (  RFFT_TRNS(2K + 1,2H) + RFFT_TRNS(2K,     2H + 1) )

C FFT(  H, -K) =   (  RFFT_TRNS(2K,    2H) + RFFT_TRNS(2K + 1, 2H + 1) )
C              + i ( -RFFT_TRNS(2K + 1,2H) + RFFT_TRNS(2K,     2H + 1) )

C FFT(  0,  0) =      RFFT_TRNS( 1, 1)
C FFT(  H,  0) =      RFFT_TRNS( 1, 2H) + i (  RFFT_TRNS( 1, 2H+1) )
C and when N is even, 
C FFT(N/2,  0) =      RFFT_TRNS( 1, N)

C FFT(  0,  K) =      RFFT_TRNS(2K, 0) + i (  RFFT_TRNS( 2K + 1, 1 ) )
C FFT(  0, -K) =      RFFT_TRNS(2K, 0) + i ( -RFFT_TRNS( 2K + 1, 1 ) )
C and when N is even, 
C FFT(N/2,N/2) =      RFFT_TRNS( N, N)
 
C When needed, a values for negative H is obtained by conjugate
C symmetry:  FFT( -H, K ) = Real( FFT(H,K) ) - Imag( FFT(H,K) )
 
C The values FFT are stored in the array FC_2D which is larger than
C FFT.  Ther rest of FC_2D is filled with zeros.  An edge of the array
C FC_2D is a little larger than SQRT(3)*[edge of FFT].  Then, for any
C orientation of the pixel plane, there will be Fourier coefficients
C throughout a plane inside the Fourier image of the output density
C grid.
C >>> END NOTE #3 ******************************************************
      return
      end
