C *****************************************************************
C *                                                               *
C * Subroutine: set_indices                                       *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1998                                                  *
C * Used in: intlz_arrays -- The subroutine initializing arrays   *
C *            recording the partitioning information of data     *
C *            structures when running in multiple nodes          *
C * Function: Partitioning indices among multiple nodes so that   *
C *           the length of indices segments are evenly           *
C *           distributed                                         *
C * Usage:                                                        *
C *   call set_indices(numnodes, min_index, max_index,            *
C *                    index_array, index_slab)                   *
C * Input:                                                        *
C *   min_index (Integer) -- The minimum index                    *
C *   max_index (Integer) -- The maximum index                    *
C *   numnodes (Integer) -- The number of nodes                   *
C * Output:                                                       *
C *   index_array (Integer) -- The array describing the first and *
C *                            last index for each node           *
C *   index_slab (Integer) -- The maximum length of index segment *
C *                                                               *
C * Algorithm:                                                    *
C *   Initialize indices: INDEX_ARRAY( 1:2, 0:NUMNODES - 1 )      *
C *   Node NODE processes entries in an array from                *
C *   INDEX_ARRAY(1,NODE) to INDEX_ARRAY(1,NODE).  The number of  *
C *   entries processed by each node is approximately the same as *
C *   any other node. If NUM_NODES .eq. 5, MIN_INDEX .eq. 1 and   *
C *   MAX_INDEX .eq. 3, then For MIN_INDEX .eq. 1, MAX_INDEX .eq. *
C *   5, then the items processed by each node are                *
C *                                                               *
C *              NODE number 0  1  2  3  4  5  6  7  8  9 10      *
C *   NUM_NODES = 1          5                                    *
C *   NUM_NODES = 2          3  2                                 *
C *   NUM_NODES = 3          2  2  1                              *
C *   NUM_NODES = 4          2  1  1  1                           *
C *   NUM_NODES = 5 or more  1  1  1  1  1                        *
C *                                                               *
C *   If MIN_INDEX .eq. -8, MAX_INDEX .eq. 8                      *
C *                                                               *
C *              NODE number 0  1  2  3  4  5  6  7  8  9 10      *
C *   NUM_NODES = 1         17                                    *
C *   NUM_NODES = 2          9  8                                 *
C *   NUM_NODES = 3          6  6  5                              *
C *   NUM_NODES = 4          5  4  4  4                           *
C *   NUM_NODES = 5          4  4  3  3  3                        *
C *   NUM_NODES = 6          3  3  3  3  3  2                     *
C *   NUM_NODES = 7          3  3  3  2  2  2  2                  *
C *   etc.                                                        *
C *                                                               *
Cji change by ji to assign less to node 0.   10/02/00             *
Cji such as the following :
C *              NODE number 0  1  2  3  4  5  6  7  8  9 10      *
C *   NUM_NODES = 1         17                                    *
C *   NUM_NODES = 2          8  9                                 *
C *   NUM_NODES = 3          5  6  6                              *
C *   NUM_NODES = 4          4  5  4  4                           *
C *   NUM_NODES = 5          3  4  4  3  3                        *
C *   NUM_NODES = 6          2  3  3  3  3  3                     *
C *   NUM_NODES = 7          2  3  3  3  2  2  2                  *
C *   etc.                                                        *

C * Modification history:                                         *
C *     - Hong Lin, Version 2.0, Oct. 7, 1999                     *
C *       Correct some errors.                                    *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *****************************************************************
C
      subroutine set_indices(numnodes, min_index, max_index,
     +                       index_array, index_slab)

C                                                            SET_INDICES
C                                                            ***********
      implicit none
      include '../include/allprogs.inc'
 
C Arguments
C
      integer numnodes, min_index, max_index, index_slab
 
C     NUMNODES .eq. NUM_NODES in COMMON in ALLPROGS.inc
 
      integer index_array( 1:2, 0:num_nodes-1 )
 
C Local
C
      integer delta_index, node, quotient, remainder, total_num

      character*12 subname
      data         subname / 'SET_INDICES' /
C ----------------------------------------------------------------------
C     Fill INDEX_ARRAY

      if( num_nodes .eq. 1 ) then
        index_array(1,0) = min_index
        index_array(2,0) = max_index
      else
        total_num = ( - min_index + max_index + 1 )
        quotient  = total_num / num_nodes
        remainder = total_num - quotient * num_nodes
C 
        if( quotient .eq. 0 ) then
C
C         There are more nodes than MAX_INDEX.  Assign one index to each
C         node; for excess nodes, make difference between first and
C         last negative
C
            index_array(1,0) =  0
            index_array(2,0) = -1

	  do 1000 node = 1, max_index-min_index+1
	    index_array(1,node) = node+min_index-1
	    index_array(2,node) = node+min_index-1
 1000     continue
C
          if (total_num .EQ. num_nodes-1) goto 1011
          do 1010 node = max_index-min_index+2, num_nodes - 1
            index_array(1,node) =  0
            index_array(2,node) = -1
 1010     continue
 1011     continue
 
        else

C         There are at least as many nodes as there indices.
C         Fill INDEX_ARRAY(2,*) with the increment of index from one
C         node to the next.
C
          delta_index = quotient - 1
          do 1020 node = 0, num_nodes - 1
            index_array(2,node) = delta_index
 1020     continue
 
          if( remainder .gt. 0 ) then
C           Distribute the remaining indices to the first few nodes.
            do 1030 node = 1, remainder
              index_array(2,node) = index_array(2,node) + 1
 1030       continue
          endif
 
C         Fill INDEX_ARRAY(1,*) and reset INDEX(2,*)
C
          index_array(1,0) = min_index
          index_array(2,0) = min_index + index_array(2,0)
          do 1040 node = 1, num_nodes - 1
            index_array(1,node) = index_array(2,node - 1) + 1
            index_array(2,node) = index_array(1,node)
     +                          + index_array(2,node)
 1040     continue
        endif
      endif
 
C     Find the maximum slab width
C
      index_slab = 0
      do 1050 node = 0, num_nodes - 1
        index_slab = max( index_slab, ( - index_array(1,node)
     +                                  + index_array(2,node) + 1 ) )
 1050 continue
 
      return
      end
