      PROGRAM P3DR
C     Parallel 3D Cartesian Reconstruction program
C                                                     MAIN PROGRAM: P3DR
C **********************************************************************
C *                                                                    *
C * Main routine: P3DR in file aamain.f                                *
C * Author: Robert E. Lynch (rel@cs.purdue.edu) and                    *
C *         Hong Lin (linh@cs.purdue.edu)                              *
C *         Yongchang Ji (yji@cs.ucf.edu)                              *
C * Date:   1997 -- 2002                                               *
C * Function: Main program for cartesian reconstruction                *
C * Algorithm: P3DR uses a set of pixel frames of projections of       *
C *   an object to reconstruct the 3D density map of the object.       *
C *   Firstly, a 2D Fourier transform is carried out on each           *
C *   pixel frame; then the 2D DFT is used in estimating 3D DFT        *
C *   of the object; finally, a reverse Fourier transform is           *
C *   carried out on the 3D DFT to get the density values of the       *
C *   object.                                                          *
C *      P3DR reads in orientations and pixel frames from two          *
C *      files with extension .dat and .pif and carries out            *
C *      reconstruction based on these data.                           *
C * Modification history:                                              *
C *   - Robert E. Lynch, Version 0.0, 1997                           *
C *   - Hong Lin, Version 0.1, Apr.20, 1998                          *
C *   - Hong Lin, Version 0.6, Jul.6, 1998                           *
C *   - Hong Lin, Version 1.0, Aug.18, 1998                          *
C *   - Hong Lin, Version 2.0, Sep.29, 1999.                         *
C *   - Hong Lin, Version 2.9, June 17, 2000                           *
C *   - R.E. Lynch, Aug 2000                                           *
C *        Program name set to P3DR                                    *
C *        Moved print of time and memory-use to subroutine MEM_TIME   *
C *        from main routine --- THIS DOESN"T WORK !!!!!!              *
C *        The setting of the value of  diameter  was removed --- the  *
C *          number of angstroms per pixels is not known until the     *
C *          header of the orientation file is read.                   *
C *        Used new variable  interpolants_size  to store the size     *
C *          of the array  interpolants  which is passed to several    *
C *          routines.  Simplified the routine  intlz_interp           *
C *                                                                    *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *
C *                                                                    *
C **********************************************************************

C confine_3D -- When confine_3D is .true., the edge-length of 3D
C   DFT is set to the same as the edge-length of pixel frames. 
C 
      implicit none
C
      include 'mpif.h'
      include '../include/allprogs.inc'
      include '../include/info.inc'
      include '../include/com_01.inc'
      include '../include/mem_time.inc'

      DATA PROGNAME / 'P3DR' /

      logical confine_3D
      data confine_3D / .false. / 

C ----------------------------------------------------------------------
C Local variables, used only in the main routine
C
      integer first_iq, last_iq, max_iq_slab, max_node_eqns,
     +        first_z,last_z, first_hk,last_hk, max_hk_slab, num_hkl,
     +        den_min_x, den_max_x, den_min_y, den_max_y, 
     +        den_min_z, den_max_z, node, mem_size, 
     +        min_xyz, max_xyz, hkl_min, hkl_max, max_z_slab,
     +        first_rr_out, last_rr_out, max_rr_out_slab,
     +        first_u, last_u, max_u_slab,
     +        first_y, last_y, max_y_slab,
     +        interpolants_size
C
C     For FFTSYNTH
C
      integer nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +        nrffts_2_row, nrffts_2_col, nrffts_2_omega,
     +        max_axis
 
      logical iam_continues_cmpt_intrps, 
     +        iam_continues_solve_fftsynth_1,
     +        iam_continues_fftsynth_2
 
C -----------------------------------------------------------------
C     The next arrays variables have their lenghts defined dynamically
C     by calls to cmalloc; pointers are used for each of these.

      real f_vru(1), f_vru_tmp(1), pixels_array(1), F_XK_dZ(1),
     +  rho(1), interpolants(1), density_map(1), ctf_val(1),
     +  ctf2(1)
C
      complex f_hkl(1),f_hkl_ref(1)
C
      pointer ( pnt_f_vru,            f_vru            )
      pointer ( pnt_f_vru_tmp,        f_vru_tmp        )
      pointer ( pnt_pixels_array,     pixels_array     )
      pointer ( pnt_ctf_val     ,     ctf_val          )
      pointer ( pnt_ctf2     ,        ctf2             )
      pointer ( pnt_F_XK_dZ,          F_XK_dZ          )
      pointer ( pnt_rho,              rho              )
      pointer ( pnt_interpolants,     interpolants     )
      pointer ( pnt_density_map,      density_map      )
      pointer ( pnt_f_hkl,            f_hkl            )
      pointer ( pnt_f_hkl_ref,        f_hkl_ref        )

      character*18 subname
      data subname / 'P3DR main routine' /

      REAL    MAP_MIN, MAP_MAX, Weight
C
C     cmalloc assigns lenths of memory for variables
      integer cmalloc, i, j, k, total_views
      external cfree
      real    enlargefact, imgprjbig, imgprjstd

C *********************************************************************
C *                      Initialization                               *
C *********************************************************************

C test_time -- When test_time is .true., each node participating in the
C   computation reports the time used in each phase.
C test_memory -- When test_memory is .true., each node participating in
C   the computation reports the memory space allocated for global arrays
C   and in each main subroutine, and the peak memory space allocated 
C   in the computation.

      test_time   = .true.
      test_memory = .true.
      call usage
C                                                         INITIALIZE MPI
C                                                         **************
      call MPI_INIT ( mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +    'Error in MPI_INIT')
        call MPI_COMM_RANK ( MPI_COMM_WORLD, iam, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +    'Error in MPI_COMM_RANK')
       call MPI_COMM_SIZE ( MPI_COMM_WORLD, num_nodes, mpi_err )
      if (mpi_err.ne.0) call error_stop(progname, subname, 
     +    'Error in MPI_COMM_SIZE')
      if (num_nodes .gt. max_numnodes) then
           write(stdoutput, 9001) num_nodes, max_numnodes
 9001      format(//' Requested number of nodes = NUM_NODES    = ',i10/
     +              ' Available number of nodes = MAX_NUMNODES = ',i10/)
          call error_stop(progname, subname,  
     +    'NUM_NODES requested exceeds MAX_NUMNODES')
      endif
C
      t1 = wtime()
C
      if( iam .eq. 0 ) then
        print *, ' Number of Nodes =', num_nodes
C                                                              *********
C       Read User's Control-Input
Cji     I have change the control_input to multi files.
C
        call info
      endif
C
      if( num_nodes .gt. 1 ) then
C                                                  CALL BCAST_PARAMETERS
C                                        *******************************
C       Broadcast User's parameters and PIF File Header to other nodes.
        call bcast_parameters 
      endif
C                                                      CALL INTLZ_PARAMS
C                                                      *****************
      call intlz_params(hkl_min, hkl_max, num_hkl, min_xyz, max_xyz, 
     +        nrffta_row, nrffta_col, nrffta_omega, nfc_2d_row, 
     +        nfc_2d_col, nrffts_1_row, nrffts_1_col, 
     +        nrffts_1_omega, nrffts_2_row, nrffts_2_col, 
     +        nrffts_2_omega, den_min_x, den_max_x, den_min_y, 
     +        den_max_y, den_min_z, den_max_z,
     +        confine_3D,
     +        max_axis )

Cji 01/06/01 added 2 print statements after call of intlz_params
        if( iam .eq. 0 ) then
          print *, ' zero_fill after  call of INTLZ_PARAMS =',zero_fill
          print *, ' nrffta_row after call of INTLZ_PARAMS =',nrffta_row
        endif
C                                                      CALL INTLZ_ARRAYS
C                                                      *****************
      call intlz_arrays( hkl_min, hkl_max, min_xyz, max_xyz,
     +       min_y, max_y, min_z, max_z, max_views, num_symm,
     +       first_hk, last_hk, max_hk_slab, first_iq, last_iq,
     +       max_iq_slab, first_rr_out, last_rr_out, max_rr_out_slab,
     +       first_u, last_u, max_u_slab, first_view, last_view,
     +       first_y, last_y, max_y_slab, first_z, last_z, max_z_slab,
     +       max_node_eqns, npix_row,
     +       iam_continues_cmpt_intrps,
     +       iam_continues_solve_fftsynth_1, iam_continues_fftsynth_2)
C
C ---------- End of Initializationl --- Synchronize all the nodes ------

      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_init = t2 - t1
      
C **********************************************************************
C *             2D DFT and estimating 3D DFT from 2D DFT               *
C **********************************************************************
C
      t1 = wtime()
C     Allocate memory for array interpolants.
 

      interpolants_size = 3 * ( last_iq - first_iq + 1 ) 
     +                     * (( hkl_max - hkl_min + 1 )**2)
      mem_size = interpolants_size
      pnt_interpolants = cmalloc( 4 * mem_size )
      mem_interpolants = dble(4 * mem_size) 

      if( pnt_interpolants .eq. 0 ) then
          print 9011, subname,  pnt_interpolants
 9011     format(/' Memory allocation failed in >> ',a,' <<'/
     +          '  PNT_INTERPOLANTS    =', i15)
          call error_stop (progname, subname,  
     +      ' Memory allocation pnt_interpolants failed')
      endif 

      mem_total = mem_total + mem_interpolants 
      mem_peak  = max(mem_peak, mem_total)

C       Zero INTERPOLANTS array
Cji
      do 1000 i = 1, interpolants_size
       interpolants( i ) = 0.0
 1000 continue
C
Cji  initialize the CTF mem for half fft_size matrix
C
      mem_size = (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
      pnt_ctf_val = cmalloc( 4 * mem_size )
      pnt_ctf2 = cmalloc( 4 * mem_size )
      mem_ctf_val = dble( 8 * mem_size)

      if( pnt_ctf_val .eq. 0 .or. pnt_ctf2 .eq. 0) then
        print 9010, subname,  pnt_ctf2
 9010   format(/' Memory allocation failed in >> ',a,' <<'/
     +          '  pnt_ctf_val =', i15)
        call error_stop (progname, subname,
     +      ' Memory allocation pnt_ctf failed')
      endif

      mem_total = mem_total + mem_ctf_val
      mem_peak  = max(mem_peak, mem_total)

C
Cji  loop for all the input parameter files to do cmpt_intrps.
      t_analy         = 0
      t_fftanaly      = 0
      total_views     = 0
      mem_cmpt_intrps = 0

      do 2000 i = 1, NFILES

Cji      initialize the parameters for one image file.

        if (iam .eq. 0) then
          pix_file_name = IMG_FNAMES(i)
          ort_file_name = PARAM_FNAMES(i)
C                                             ******PIF_OPEN************
          call PIF_OPEN (pix_file_name, pix_file_unit, 'r')
C
Cji because each files has different standard division
Cji we'll apply this ratio enlargefact to each pixel file
Cji we'll combine this ratio to the ctf factor below
C
          call read_1_pif(imgprjbig) 
          if (i.eq.1) then
            imgprjstd   = imgprjbig
            enlargefact = 1. 
          else 
            enlargefact = imgprjstd / imgprjbig
          endif
          print *, 'enlargefact=',enlargefact, imgprjbig, imgprjstd
        endif
C
        call mpi_bcast(enlargefact, 1, mpi_real,
     +                 0, mpi_comm_world, mpi_err)
        if( mpi_err .ne. 0 ) call error_stop(progname,subname,
     +     'Error in calling MPI_BCAST enlargefact')
C
        npix_row    =  NPIX(i)
        npix_col    =  NPIX(i)
        max_views   =  NIDS(i)
        angstrom_per_pix = PIXSIZS(i)

        mem_size =   view_perread * npix_row * npix_col
        pnt_pixels_array = cmalloc( 4 * mem_size )
        mem_pixels_array = dble(4 * mem_size)
C
        if( pnt_pixels_array .eq. 0) then
          print 9012, subname, pnt_pixels_array 
 9012     format(/' Memory allocation failed in >> ',a,' <<'/
     +          '  pnt_pixels_array =', i15)
          call error_stop (progname, subname,  
     +      ' Memory allocation pnt_pixels_array failed')
        endif
        mem_total = mem_total + mem_pixels_array
        mem_peak  = max(mem_peak, mem_total)
C
Cji  initialize the CTF value  for half fft_size matrix
        do k = 1, (nfc_2d_row +1) * (nfc_2d_col * 2 +1)
            ctf_val(k) = 0.0
            ctf2(k) = 1.0
        enddo

        IF (CTFMODE.GT.0 .and. CTFMODE.LE.4)  then
C
Cji we'll combine this enlarge ratio to ctf factor
          Weight = enlargefact
C                                                          CALL CTF_PARA
C                                                      *****************
          call ctf_para(ctf_val, ctf2, nfc_2d_row, nfc_2d_col, CTFMODE, 
     +         Filter, TEMPFAC, WIENER, WIENER2,
     +         PIXSIZS(i), units, VOLTS(i), DELF_MAJS(i),
     +         DELF_MINS(i), ANG_MAJS(i), AMPFACS(i), CS_COEFFS(i),
     +         Weight, res_max, res_min)
        endif
C
C       For each image carry out Fourier analysis and determine
C       the interpolated values of the Fourier coefficients
C                                                       CALL CMPT_INTRPS
C                                                       ****************
        mem_cmpt = 0

        call cmpt_intrps( pixels_array, hkl_min, hkl_max,
     +         max_axis, interpolants, interpolants_size, 
     +         t_analy, t_fftanaly, max_iq_slab, 
     +         mem_cmpt, total_views, ctf_val, ctf2)

        mem_cmpt_intrps = max(mem_cmpt_intrps, mem_cmpt)
        mem_peak  = max(mem_peak, mem_total+mem_cmpt)

C     Free memory for the array used in loading pixel frames.
        call cfree(pnt_pixels_array)
        mem_total = mem_total - mem_pixels_array
Cji
Cji    finish read the pif file, Close PIF file.
Cji
        if(  iam .eq. 0 ) then
          call pif_close (pix_file_unit)
          print *, 'total_view = ', total_views
        endif

 2000 continue

      call cfree(pnt_ctf_val)
      call cfree(pnt_ctf2)
      mem_total = mem_total - mem_ctf_val

Cji end of all the parameter files
C -- End of 2D DFT and estimating 3D DFT.  Synchronize all the nodes ---
      call mpi_barrier(mpi_comm_world, mpi_err)

      t2 = wtime()
      t_interp = t2 - t1 - t_fftanaly - t_analy
C
C **********************************************************************
C *        Combining estimates of 3D DFT along three directions        *
C *        arrange the data before 3D DFT                              * 
C **********************************************************************
C 
      t1 = wtime()
C     Allocate memory.
C
      if (iam_continues_solve_fftsynth_1) then
        mem_size = (num_hkl**2) * (-first_iq+last_iq+1)
        pnt_f_hkl = cmalloc(8 * mem_size)
        if (pnt_f_hkl.eq.0) call error_stop(progname, subname,
     +    'Error in allocating memory for F_HKL')
        mem_f_hkl = dble(8 * mem_size)
        mem_total = mem_total + mem_f_hkl
        mem_peak  = max(mem_peak, mem_total)
C
Cji the following is add by ji, convert interpolants to F_HKL
C
        call realtocomplx( hkl_min, hkl_max,
     +             first_iq, last_iq, interpolants, f_hkl )

C       Free memory used in estimating 3D DFT by interpolation.
C
        call cfree(pnt_interpolants)
        mem_total = mem_total - mem_interpolants
      endif

C ----- This is the end of combining 3D DFT along 3 directions -----
C ----- Before move onto the next step, synchronize all the nodes ------
C
      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_solvesys = t2 - t1
C
C **********************************************************************
C *                      2D Fourier synthesis                          *
C **********************************************************************
C
      t1 = wtime()
C
C     Allocate memory
C
      if (iam_continues_solve_fftsynth_1) then
        mem_size = ((-min_xyz+max_xyz+1)**2) *
     +               (-first_rr_out+last_rr_out+1)
        pnt_f_vru = cmalloc(4 * mem_size)
        if (pnt_f_vru.eq.0) call error_stop(progname, subname, 
     +    'Error in allocating memory for F_VRU')
        mem_f_vru = 4 * mem_size
        mem_total = mem_total + mem_f_vru
        mem_peak  = max(mem_peak, mem_total)
C                                                             FFTSYNTH_1
C                                                             **********
        call fftsynth_1_m_slab( hkl_min, hkl_max, first_iq,
     +           last_iq, f_hkl, nrffts_1_row, nrffts_1_col,
     +           nrffts_1_omega, min_x, max_x, min_y, max_y,
     +           min_z, max_z, min_xyz, max_xyz, f_vru,
     +           mem_fftsynth_1_m )

        call cfree(pnt_f_hkl)
        mem_peak = max(mem_peak, mem_total+mem_fftsynth_1_m)
        mem_total = mem_total - mem_f_hkl

      endif
C
C ------------ This is the end of 2D Fourier synthesis -----------------
C ------ Before move onto the next step, synchronize all the nodes -----
C
      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_fftsynth_1 = t2 - t1
C
C **********************************************************************
C *           Exchanging data for 1D Fourier synthesis                 *
C **********************************************************************
C
      t1 = wtime()
C
C     Allocate storage for subroutine EXCHANGE_2
C
      mem_size = (-min_xyz+max_xyz+1) * (-min_xyz+max_xyz+1) *
     +           (-frst_lst_U(1,iam)+frst_lst_U(2,iam)+1)
      pnt_F_VRU_TMP = cmalloc(4 * mem_size)
      if (pnt_F_VRU_TMP.eq.0) call error_stop(progname,
     +  'exchange_2_m', 'Error in allocating memory for F_VRU_TMP!')
      mem_F_VRU_TMP = dble(4 * mem_size)
      mem_total = mem_total + mem_F_VRU_TMP
      mem_peak  = max(mem_peak, mem_total)
C
      call exchange_2_slab( min_xyz, max_xyz, F_VRU,
     +                      nrffts_1_row, F_VRU_TMP )

      if (iam_continues_solve_fftsynth_1) then
        call cfree(pnt_F_VRU)
        mem_total = mem_total - mem_F_VRU
      endif

      mem_size = nrffts_1_row*(-frst_lst_u(1,iam)+frst_lst_u(2,iam)+1)
      mem_F_XK_dZ = dble(4 * (-min_x+max_x+1) * mem_size)
      pnt_F_XK_dZ = cmalloc( 4 * (-min_x+max_x+1) * mem_size )
      if(pnt_F_XK_dZ.eq.0 ) then
        print 9000, subname, pnt_F_XK_dZ 
 9000   format(/' Memory allocation failed in >> ',a,' <<'/
     +          '  PNT_F_XK_dZ =', i15)
        call error_stop (progname, subname,  
     +    ' Memory allocation failed')
      endif
      mem_total = mem_total + mem_F_XK_dZ 
      mem_peak  = max(mem_peak, mem_total)

      call rearrange_2_slab( min_xyz, max_xyz, nrffts_1_row,
     +       F_VRU_TMP, min_x,max_x, min_y,max_y, min_z,max_z,
     +       F_XK_dZ )
C
C     Free memory used in 2D Fourier synthesis.
C
      if (iam_continues_solve_fftsynth_1) then
        call cfree(pnt_F_VRU_TMP)
        mem_total = mem_total - mem_F_VRU_TMP
      endif
C
C ----- This is the end of exchanging data for 1D Fourier synthesis ----
C ------ Before move onto the next step, synchronize all the nodes -----
C
      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_exchange_2 = t2 - t1

C **********************************************************************
C *                        1D Fourier synthesis                        *
C **********************************************************************
C
      t1 = wtime()
C     Allocate storage for 1D Fourier synthesis.
C
      if (iam_continues_fftsynth_2) then
        mem_size = ((max_x-min_x+1)**2) * (-first_z+last_z+1)
        pnt_rho = cmalloc(4 * mem_size)
        if (pnt_rho.eq.0) call error_stop(progname, subname, 
     +    'Error in allocating memory for RHO')
        mem_rho = dble(4 * mem_size)
        mem_total  = mem_total + mem_rho 
        mem_peak   = max(mem_peak, mem_total)
C                                                        CALL FFTSYNTH_2
C                                                        ***************
        call fftsynth_2_m_slab( nrffts_1_row, min_x, max_x, min_y, 
     +           max_y, min_z, max_z, F_XK_dZ, nrffts_2_row, 
     +           nrffts_2_col, nrffts_2_omega, rho, 
     +           mem_fftsynth_2_m )
        mem_peak = max(mem_peak, mem_total+mem_fftsynth_2_m)
      endif
C
C     Free memory used in 2D Fourier synthesis.
C
      call cfree(pnt_F_XK_dZ)
      mem_total = mem_total - mem_F_XK_dZ
C
C -------------- This is the end of 1D Fourier synthesis --------------
C ----- Before move onto the next step, synchronize all the nodes -----
C
      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_fftsynth_2 = t2 - t1
C
C *********************************************************************
Cji   clear the density map and output density map
C *********************************************************************

      t1 = wtime()
        mem_size = (-den_min_x+den_max_x+1)*
     +             (-den_min_y+den_max_y+1)* max_z_slab
Cji   according to the set_indice, node 1 alway has the most largest slab,

      pnt_density_map = cmalloc(4*mem_size)
      mem_density_map = dble(4*mem_size)
      mem_total = mem_total + mem_density_map
      mem_peak  = max(mem_peak, mem_total)
      if (pnt_density_map.eq.0) call error_stop(progname,
     +    subname,  'Error in allocating density_map')
        
      call density_clear (rho, min_x, max_x, min_y, max_y,
     +          den_min_x, den_max_x, density_map, MAP_MIN, MAP_MAX)

      call cfree(pnt_rho)
      mem_total = mem_total - mem_rho
C
Cji   Write density map into MAP file.
C
      call output_density( density_map, den_min_x,
     +                     den_max_x, max_z_slab, MAP_MIN, MAP_MAX )
C
C     Free memory for storing the density values.
C
       call cfree(pnt_density_map)
       mem_total = mem_total - mem_density_map
C
C ------- This the end of writing density values into MAP file --------
C ----- Before reporting the timing the memory allocating mesage, -----
C ------------------- synchronize all the nodes. ----------------------

      call mpi_barrier(mpi_comm_world, mpi_err)
      t2 = wtime()
      t_write_output = t2 - t1
C
C *********************************************************************
C *       Reporting the timing and memory allocating messages         *
C *********************************************************************
C
      if (mem_total .ne. 0) print *,'mem_total=',mem_total,
     +  'At the end, mem_total should be 0!!!'
C
      if ( iam .eq. 0 ) then
C     Print timing message.
C
        t_seq = t_init + t_analy + t_write_output
        t_comm = t_exchange_2 + t_gather
        t_par = t_interp  + t_fftanaly + 
     +          t_solvesys + t_fftsynth_1 + t_fftsynth_2
        time = t_seq + t_comm + t_par
C
        if (test_time) 
     +    print 40, iam, t_init,
     +                   t_analy,
     +                   t_fftanaly,
     +                   t_interp,
     +                   t_solvesys,
     +                   t_fftsynth_1,
     +                   t_exchange_2,
     +                   t_fftsynth_2,
     +                   t_gather,
     +                   t_write_output,
     +                   time, t_seq, t_comm, t_par
  40      format (/ 'Node ', i3, 
     +          ' Time used in initialization                      =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in read images                         =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in 2D FFT analysis                     =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in 2D interpolation                    =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in solving linear systems              =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in 2-D Fourier synthesis               =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in exchanging data before fftsynth_2   =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in 1-D Fourier synthesis               =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in exchanging data before output       =',
     +        f15.4, ' ' / 8x,
     +          ' Time used in writing output                      =', 
     +        f15.4, ' ' / 8x, 
     +          ' Total time used                                  =',
     +        f15.4, ' ' / 8x,
     +          '     (Seq, Com, Par):'/   
     +          '    (', f15.4, ',', f15.4, ',', f15.4, ')')
C
C     Print memory allocating message.
C
         if (test_memory) print 60, iam, mem_peak, mem_pixels_array,
     +       mem_interpolants, mem_f_hkl, mem_f_hkl_ref, mem_f_vru,
     +       mem_F_XK_dZ, mem_rho, mem_density_map,mem_read_files,
     +       mem_cmpt_intrps, mem_solvesys_m, mem_fftsynth_1_m,
     +       mem_exchange_2_m, mem_fftsynth_2_m
   60    format(/ 'Node ', i3, ' Memory allocation in BYTES:', /
     +                  8x, ' Peak memory allocated   =', f18.1 /
     +                  8x, ' Memory for PIXELS_ARRAY =', f18.1 /
     +                  8x, ' Memory for INTERPOLANTS =', f18.1 /
     +                  8x, ' Memory for F_HKL        =', f18.1 /
     +                  8x, ' Memory for f_hkl_ref    =', f18.1 /
     +                  8x, ' Memory for F_VRU        =', f18.1 /
     +                  8x, ' Memory for F_XK_dZ      =', f18.1 /
     +                  8x, ' Memory for RHO          =', f18.1 /
     +                  8x, ' Memory for density_map  =', f18.1 /
     +                  8x, ' Memory used in READ_FILES   =', f18.1 /
     +                  8x, ' Memory used in CMPT_INTRPS  =', f18.1 /
     +                  8x, ' Memory used in SOLVESYS_M   =', f18.1 /
     +                  8x, ' Memory used in FFTSYNTH_!_M =', f18.1 /
     +                  8x, ' Memory used in EXCHANGE_2_M =', f18.1 /
     +                  8x, ' Memory used in FFTSYNTH_2_M =', f18.1 )
C
      endif
C
C     Finalize MPI
C
      print *, '**********finished nicely**********'
      call MPI_FINALIZE ( mpi_err )
C
      stop
      end
