C  *****************************************************************************
C  *                                                                           *
C  *                  *****   SUBROUTINE TO_ASYM_UNIT   *****                  *
C  *                                                                           *
C  *    Calls routines to check if orientations are in the asymmetric unit.    *
C  *    If not, these routines calculate the equivalent view within the        *
C  *    asymmetric unit.  If OMEGA < 0, it adds 360 deg.  If PHI > 360, it     *
C  *    subtracts 360 deg.                                                     *
C  *                                                                           *
C  *    Purdue University DEC 3000, Jan,Jun 1994, D. Belnap                    *
C  *    Added 222, 522, 622, & 722 symmetry, Dec 1996, Jan 1997, D. Belnap     *
C  *    MODIFIED 22-JUN-1998 BY TSB (Updated dihedral symmetries).             *
C  *    Modified 29-JUN-1999 BY D. Belnap (Updated dihedral symmetries).       *
C  *****************************************************************************
C
      SUBROUTINE TO_ASYM_UNIT (SYMMETRY, THETA, PHI, OMEGA)
C
      IMPLICIT NONE
C
      INTEGER*4  SYMMETRY
      REAL*4     THETA, PHI, OMEGA
C -----------------------------------------------------------------------------
C
      IF (SYMMETRY.GE.2 .AND. SYMMETRY.LE.31) THEN       ! n-fold symmetries
         CALL TO_NFOLD (THETA, PHI, OMEGA, SYMMETRY)
       ELSEIF (SYMMETRY.EQ.532) THEN
         CALL RETURN_TO_UNIT_TRIANGLE (THETA, PHI, OMEGA, .TRUE.)
       ELSEIF (SYMMETRY.EQ.222) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 2)
       ELSEIF (SYMMETRY.EQ.32) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 3)
       ELSEIF (SYMMETRY.EQ.422) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 4)
       ELSEIF (SYMMETRY.EQ.52) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 5)
       ELSEIF (SYMMETRY.EQ.622) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 6)
       ELSEIF (SYMMETRY.EQ.72) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 7)
       ELSEIF (SYMMETRY.EQ.822) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 8)
       ELSEIF (SYMMETRY.EQ.92) THEN
         CALL TO_N22 (THETA, PHI, OMEGA, 9)
      ENDIF
C
      RETURN
      END
C
C  *****************************************************************************
C  *                                                                           *
C  *                   *****     SUBROUTINE TO_N22     *****                   *
C  *                                                                           *
C  *    Checks if orientation is within the specified n2 or n22 asymmetric     *
C  *    unit.  If not the equivalent position in the asym. unit is             *
C  *    calculated.                                                            *
C  *                                                                           *
C  *                                                                           *
C  *    Geometric convention:                                                  *
C  *    --------------------                                                   *
C  *    In a right-handed X-Y-Z coordinate system, PHI is a rotation about     *
C  *    the z-axis and THETA is a rotation about the "new" y-axis (after       *
C  *    the PHI rotation).  OMEGA is a rotation about the normal (deter-       *
C  *    mined by the THETA, PHI orientation).                                  *
C  *                                                                           *
C  *    In this routine, the n-fold symmetry component is about the z-axis.    *
C  *    The value of n is input as Z_SYM.  There is one 2-fold symmetry        *
C  *    axis about the x axis.  This is assumed in the routine.  (For n22      *
C  *    symmetries there also is a 2-fold symmetry axis on the y axis.)        *
C  *    The minimum and maximum angles of PHI (in the asymmetric unit) are     *
C  *    calculated from Z_SYM as PHI_MIN and PHI_MAX.  PHI_MIN and PHI_MAX     *
C  *    differ only in sign, with the midpoint being PHI = 0.0.  For all       *
C  *    n2 and n22 asymmetric units, the minimum and maximum THETA angles are  *
C  *    0 and 90, respectively.                                                *
C  *                                                                           *
C  *    For example, if you have 822 symmetry then the 8-fold symmetry         *
C  *    axis is defined along the Z axis.  Two two-fold axes are along         *
C  *    the X and Y axes.  The 822 asymmetric unit is, in THETA, 0 to 90       *
C  *    degrees, and in PHI, -22.5 to +22.5 degrees.  Therefore,               *
C  *            Z_SYM   = 8             PHI_DIMENSION = 45.0                   *
C  *            PHI_MIN = -22.5         PHI_MAX = 22.5                         *
C  *                                                                           *
C  *    See Belnap et al. (1997) J. Struct. Biol. 120:44-51 for a              *
C  *    definition of THETA, PHI, and OMEGA angles.                            *
C  *                                                                           *
C  *    PURDUE UNIVERSITY DEC 3000, 4 JANUARY 1994, D. Belnap                  *
C  *    Added PHI_DIMENSION, 31 Jan 1997, D. Belnap                            *
C  *****************************************************************************
C
      SUBROUTINE TO_N22 (THETA, PHI, OMEGA, Z_SYM)
C
      IMPLICIT NONE
C
      INTEGER*4  Z_SYM
      REAL*4     THETA, PHI, OMEGA, PHI_DIMENSION, PHI_MIN, PHI_MAX
C -----------------------------------------------------------------------------
C
C     ************************************************
C     *  Calculate PHI dimension of asymmetric unit  *
C     ************************************************
      PHI_DIMENSION = 360.0 / FLOAT(Z_SYM)
C     ***********************************
C     *  Calculate PHI_MIN and PHI_MAX  *
C     ***********************************
      PHI_MIN  =  0.0  -  ( PHI_DIMENSION/2.0 )
      PHI_MAX  =  0.0  +  ( PHI_DIMENSION/2.0 )
C     ********************************************************************
C     *  If THETA at pole, PHI = 0.0 and OMEGA ranges between PHI_MIN &  *
C     *  PHI_MAX, add or subtract PHI_DIMENSION if OMEGA outside range   *
C     ********************************************************************
      IF (THETA.EQ.0.0 .OR. THETA.EQ.180.0) THEN
        PHI = 0.0
        IF (THETA.EQ.180.0) THEN                      ! Calc. equivalent view
          THETA = 0.0
          OMEGA = -OMEGA
        ENDIF
        DO WHILE (OMEGA.LT.PHI_MIN)
          OMEGA = OMEGA + PHI_DIMENSION
        ENDDO
        DO WHILE (OMEGA.GT.PHI_MAX)
          OMEGA = OMEGA - PHI_DIMENSION
        ENDDO
        RETURN
      ENDIF
C     *******************************************************************
C     *  Determine if orientation is in asymmetric unit, exit if it is  *
C     *******************************************************************
      IF (THETA.LE.90.0 .AND. PHI.GE.PHI_MIN
     1    .AND. PHI.LE.PHI_MAX)  RETURN
C     ******************************************************
C     *  Return stray PHI orientations to asymmetric unit  *
C     *  by adding or subtracting PHI_DIMENSION            *
C     ******************************************************
      DO WHILE (PHI.LT.PHI_MIN)
        PHI = PHI + PHI_DIMENSION
      ENDDO
      DO WHILE (PHI.GT.PHI_MAX)
        PHI = PHI - PHI_DIMENSION
      ENDDO
C     *********************************************************
C     *  Return stray THETA orientations to asymmetric unit,  *
C     *  adjust PHI and OMEGA                                 *
C     *********************************************************
      IF (THETA.GT.90.0) THEN
        THETA = 180.0 - THETA
        PHI   = -PHI
        OMEGA = OMEGA + 180.0
        IF (OMEGA.GE.360.0)  OMEGA = OMEGA - 360.0
      ENDIF
C
      END

C  *****************************************************************************
C  *                          SUBROUTINE TO_NFOLD                              *
C  *                                                                           *
C  *    In this routine, the n-fold symmetry component is about the z-axis.    *
C  *    Routine tests if orientation is within the n-fold asymmetric unit,     *
C  *    if not the equivalent position in the asymmetric unit is calculated.   *
C  *                                                                           *
C  *    Geometric convention:  see description for TO_N22                      *
C  *                                                                           *
C  *    The value of n is input as Z_SYM.  There are no other symmetry axes.   *
C  *    The minimum and maximum angles of PHI (in the asymmetric unit) are     *
C  *    calculated from Z_SYM as PHI_MIN and PHI_MAX.  PHI_MIN and PHI_MAX     *
C  *    differ only in sign, with the midpoint being PHI = 0.0.  For all       *
C  *    n-fold asymmetric units, the minimum and maximum THETA angles are 0    *
C  *    and 180, respectively.                                                 *
C  *                                                                           *
C  *    For example, if you have 8-fold symmetry then the asymmetric unit      *
C  *    is, in THETA, 0 to 180 deg., and in PHI, -22.5 to +22.5 deg.           *
C  *    Therefore,                                                             *
C  *            Z_SYM   = 8             PHI_DIMENSION = 45.0                   *
C  *            PHI_MIN = -22.5         PHI_MAX = 22.5                         *
C  *                                                                           *
C  *    31 Jan 1997, D. Belnap, modified from TO_N22                           *
C  *                                                                           *
C  *    MODIFIED 19-OCT-2000 BY TSB/MZ (Check for OMEGA > 360.0).              *
C  *****************************************************************************
C
      SUBROUTINE TO_NFOLD (THETA, PHI, OMEGA, Z_SYM)
C
      IMPLICIT NONE
C
      INTEGER*4  Z_SYM
      REAL*4     THETA, PHI, OMEGA, PHI_DIMENSION, PHI_MIN, PHI_MAX
C -----------------------------------------------------------------------------
C
C     ************************************************
C     *  Calculate PHI dimension of asymmetric unit  *
C     ************************************************
      PHI_DIMENSION = 360.0 / FLOAT(Z_SYM)
C     ***********************************
C     *  Calculate PHI_MIN and PHI_MAX  *
C     ***********************************
      PHI_MIN  =  0.0  -  ( PHI_DIMENSION/2.0 )
      PHI_MAX  =  0.0  +  ( PHI_DIMENSION/2.0 )
C     ********************************************************************
C     *  If THETA at pole, PHI = 0.0 and OMEGA ranges between PHI_MIN &  *
C     *  PHI_MAX, add or subtract PHI_DIMENSION if OMEGA outside range   *
C     ********************************************************************
      IF (THETA.EQ.0.0 .OR. THETA.EQ.180.0) THEN
        PHI = 0.0
        DO WHILE (OMEGA.LT.PHI_MIN)
          OMEGA = OMEGA + PHI_DIMENSION
        ENDDO
        DO WHILE (OMEGA.GT.PHI_MAX)
          OMEGA = OMEGA - PHI_DIMENSION
        ENDDO
        RETURN
      ENDIF
      IF(OMEGA.GE.360.0) OMEGA = AMOD(OMEGA,360.0)
C     *******************************************************************
C     *  Determine if orientation is in asymmetric unit, exit if it is  *
C     *******************************************************************
      IF (PHI.GE.PHI_MIN .AND. PHI.LE.PHI_MAX)  RETURN
C     **************************************************
C     *  Return stray orientations to asymmetric unit  *
C     *  by adding or subtracting PHI_DIMENSION        *
C     **************************************************
      DO WHILE (PHI.LT.PHI_MIN)
        PHI = PHI + PHI_DIMENSION
      ENDDO
      DO WHILE (PHI.GT.PHI_MAX)
        PHI = PHI - PHI_DIMENSION
      ENDDO
C
      END
C
C  *****************************************************************************
C  *              subroutine  Return_to_Unit_Triangle
C  *     The subroutine checks the input orientation for
C  *     validity that is lying in the unique triangle
C  *     between the five folds at 90,+-31.7 and the
C  *     three fold at 69.01,0 (this region is called good).
C  *     sdf 4-dec-1985
C  *     mrc conversion sdf - 17-jan-87
C  *     revision for larger area of return
C  *     SDF 28-JAN-87
C  *     last modified 1-feb-87 sdf
C  *     completely re-written - previous version was wrong
C  *     sdf 12-jan88
C  *****************************************************************************
        subroutine Return_to_Unit_Triangle(THETA, PHI, OMEGA, L_DEGREE)
C
        implicit none
C
        real*4 THETA, PHI, OMEGA, DCOS(3), X(3)
        real*4 DEG_PER_RAD/57.29577951/
        integer IMAT
        logical L_DEGREE, GOOD
C
        real*4 THETAR, PHIR, OMEGAR
        logical L_GEOMETRY_INIT 
        real*4 ICOS_ROT(3,3,60)         ! icos rotation matrixes
        
        real*4 THREE_FOLD(3)    ! direction cosines for three fold
        real*4 FIVE_FOLD(3)     ! direction cosines for five fold
        real*4 ANG_35,ANG_55    ! angular distances btw 3fold and 5 folds
        real*4 AREA_STD         ! area of asymmetric unit in spherical measure

        common/icos_geometry/ L_GEOMETRY_INIT,
     1       ICOS_ROT, THREE_FOLD, FIVE_FOLD, ANG_35, ANG_55, AREA_STD
C
        if(.not.L_DEGREE) then
            THETA = THETA*DEG_PER_RAD
            PHI = PHI*DEG_PER_RAD
            OMEGA = OMEGA*DEG_PER_RAD
        endif
        call polar_to_cosines(THETA, PHI, DCOS)
C
        if( good(DCOS) ) return
C
        do IMAT = 1, 60
            call equivalent_view(THETA, PHI,OMEGA,
     1                       ICOS_ROT(1,1,IMAT),
     1                       THETAR, PHIR, OMEGAR)
            call polar_to_cosines(THETAR, PHIR, X)
            if(good(X)) then
                if(L_DEGREE) then
                    THETA = THETAR
                    PHI = PHIR
                    OMEGA=OMEGAR
                    return
                else
                    THETA = THETAR/DEG_PER_RAD
                    PHI = PHIR/DEG_PER_RAD
                    OMEGA = OMEGAR/DEG_PER_RAD
                    return
                endif
            endif
        enddo
C

        write(6,'('' RETURN_TO_UNIT_TRIANGLE '')')
        write(6,'('' no equivalent found '')')
        write(6,'('' input values '')')
        write(6,'('' theta '',g)') THETA
        write(6,'('' phi '',g)') PHI
        write(6,'('' omega '',g)') OMEGA
        write(6,'('' Last matrix tried '',i2)') IMAT
        return
        end

C
C  *****************************************************************************
        subroutine POLAR_TO_COSINES(THETA, PHI, X)
C
C               Converts theta, phi angles in degrees to direction
C               cosines
C
C               a tolerance is set to try to alleviate round off
C               problems and the total length of the direction cosine
C               vector is set to 1.0
C
C               sdf - 10-jan-88
C  *****************************************************************************
C

        real*4 THETA,PHI,X(3)
        real*4 TOLER/1.0e-8/

        X(3) = cosd(THETA)
        if(abs(x(3)-1.0).lt.TOLER) then
            X(2) = sind(PHI)
            X(1) = sind(PHI)
        else
            X(2) = sind(THETA)*sind(PHI)
            X(1) = sind(THETA)*cosd(PHI)
        endif
        do I = 1,3
           if(abs(X(I)).LT.TOLER) X(I)=0.0
        enddo

        call norm(X,X)

        return

        end

c
C  *****************************************************************************
        logical function good(X)
c
c               This logical function is true if a view is within the
c       assymetric icosahedral unit defined by the (90,+/- 31.7..) five folds
c       and the (69.0..,0.0) three-fold. The function does a rough check
c       first on the direction cosines of the view and then calculates
c       the spherical measure of the largest triangle formed by the test
c       point and the bounding symmetry axes. If this is larger than
c       that of the assymmetric unit, the view is outside the unit.
c
c               sdf 10-jan-88
C  *****************************************************************************
        implicit none
        real*4  TOLER/1.0e-6/
        real*4  X(3)
        real*4  A3X, A5PX, A5X, D3X, D5X, D5PX, AREA_X
        integer I
c
        logical L_GEOMETRY_INIT 
        real*4 ICOS_ROT(3,3,60)         ! icos rotation matrixes
        
        real*4 THREE_FOLD(3)    ! direction cosines for three fold
        real*4 FIVE_FOLD(3)     ! direction cosines for five fold
        real*4 ANG_35,ANG_55    ! angular distances btw 3fold and 5 folds
        real*4 AREA_STD         ! area of asymmetric unit in spherical measure
        common/icos_geometry/ L_GEOMETRY_INIT,
     1       ICOS_ROT, THREE_FOLD, FIVE_FOLD, ANG_35, ANG_55, AREA_STD
c
        real*4 sur,excess,A,B,C

        sur(A,B,C) = (A+B+C)/2.0
        excess(A,B,C) =
     1 4.0*atan(sqrt(abs(tan(sur(A,B,C)/2.)
     1 *tan((sur(A,B,C)-A)/2.0)
     1 *tan((sur(A,B,C)-B)/2.0)
     1 *tan((sur(A,B,C)-C)/2.0) ) )  )
c
        GOOD = .false.

        if(.not.L_GEOMETRY_INIT) call geometry_init

        if(X(3).lt.-TOLER .or.
     1        X(3).gt.THREE_FOLD(3)+TOLER.or.
     2        abs(X(2)).gt.FIVE_FOLD(2)+TOLER) then
        return
        endif

        D3X = 0.0
        D5X = 0.0
        D5PX = 0.0

        do I = 1,3
           D3X=(X(I)-THREE_FOLD(I))**2+D3X
           D5X=(X(I)-FIVE_FOLD(I))**2+D5X
           if(I.eq.2) then
              D5PX=(X(I)+FIVE_FOLD(I))**2+D5PX
           else
              D5PX=(X(I)-FIVE_FOLD(I))**2+D5PX
           endif
        enddo

        D3X = sqrt(D3X)
        D5X = sqrt(D5X)
        D5PX = sqrt(D5PX)

        A5X=2.0*asin(D5X/2.)
        A5PX=2.0*asin(D5PX/2.)
        A3X=2.0*asin(D3X/2.)
        
        AREA_X =   excess(A5X,A3X,ANG_35)+excess(A5PX,A3X,ANG_35)+
     1          excess(A5X,A5PX,ANG_55) -
     1       min(excess(A5PX,A3X,ANG_35),excess(A5X,A5PX,ANG_55),
     1               excess(A5X,A3X,ANG_35) )

        GOOD = AREA_X.lt.(AREA_STD+TOLER)

        return
        end

c
      subroutine equivalent_view(THETA,PHI,OMEGA,ROT,THETAR,PHIR,OMEGAR)
c
c               should find theta,phi and omega for an equivalent view
c       which is generated by rotation with rot.
c
c       new formulation which avoids the cummulative errors in the many
C       multiplications of the previous version
C           formulation suggested by Peter Metcalf
C       
c               sdf 29-oct-89
c--------------------------------------------------------------------------
        implicit none

        real*4 ROT(3,3)
c-------------------    rotation matrix to generate equivalents
        real*4 VIEW(3,3)
c------------------       equivalent rotation matrix to crowther angles
        real*4 PROT(3,3)
C----------------     product matrix


        real*4 THETA, PHI, OMEGA
        real*4 THETAR, PHIR, OMEGAR

c.....................................................................

        call crowther_to_matrix(THETA, PHI, OMEGA, VIEW)
        call matmul(PROT,VIEW,ROT)
        call matrix_to_crowther(THETAR, PHIR, OMEGAR, PROT)
        return
        end

c
        subroutine norm(NX,X)
c
c               Normalizes a vector X and returns the normalized value
c
c               sdf 8-jan-88
c-----------------------------------------------------------------------
        implicit none
        real NX(3),X(3),NDOT
        integer J
        real dot
c.......................................................................
        NDOT = sqrt(abs(dot(X,X)))
        if(NDOT.gt.1.e-20) then
            do J=1,3
                NX(J)=X(J)/NDOT
            enddo
        else
            print *,' norm - zero length vector ',X
            do J=1,3
                NX(J)=0.0
            enddo
        endif
        return
        end

c
        function dot(x,y)
c
c               Returns the dot product of two 3 dim vectors
c       
c
c               sdf 7-jan-88
c----------------------------------------------------------------------
        implicit none
        real DOT,X(3),Y(3)
c........................................................................
        DOT=X(1)*Y(1)+X(2)*Y(2)+X(3)*Y(3)
        return
        end

c
        subroutine GEOMETRY_INIT
c
c
c               Initializes the arrays and geometric constants needed by
c                       the return to unit triangle routines
c               The key l_geometry_init is used to force its call upon
c                       the first reference
c
c                       sdf 10-jan-88
c
c----------------------------------------------------------------------
        implicit none

        real*8 TAU, AREA_35, AREA_55, DIST_35, DIST_55
        real*8 THETA_THREE, PHI_THREE, THETA_FIVE, PHI_FIVE
        real*8 ARG

        integer I

c
        logical L_GEOMETRY_INIT 
        real*4 ICOS_ROT(3,3,60)         ! icos rotation matrixes
        
        real*4 THREE_FOLD(3)    ! direction cosines for three fold
        real*4 FIVE_FOLD(3)     ! direction cosines for five fold
        real*4 ANG_35,ANG_55    ! angular distances btw 3fold and 5 folds
        real*4 AREA_STD         ! area of asymmetric unit in spherical measure

        common/icos_geometry/ L_GEOMETRY_INIT,
     1       ICOS_ROT, THREE_FOLD, FIVE_FOLD, ANG_35, ANG_55, AREA_STD

c
        call GENROT(ICOS_ROT)
        TAU=(1.+dsqrt(5.0d0))/2.0
        ARG=(1.+TAU*TAU/3.d0-1./(3.d0*TAU*TAU))/(2.d0*TAU/dsqrt(3.0d0))
        THETA_THREE = 90. - dacosd(ARG)
        PHI_THREE = 0.0
        THREE_FOLD(3) = cosd(THETA_THREE)
        THREE_FOLD(2) = 0.0
        THREE_FOLD(1) = sind(THETA_THREE)

        THETA_FIVE = 90.0
        PHI_FIVE = datand(1.0/TAU)


        FIVE_FOLD(1) = cosd(PHI_FIVE)
        FIVE_FOLD(2) = sind(PHI_FIVE)
        FIVE_FOLD(3) = 0.0      


        DIST_35 = 0.0

        do I = 1,3
           DIST_35=(THREE_FOLD(I)-FIVE_FOLD(I))**2+DIST_35
        enddo
        DIST_35 = dsqrt(DIST_35)
        DIST_55 = FIVE_FOLD(2)*2.d0

        ANG_55=2.0*dasin(DIST_55/2.)
        ANG_35=2.0*dasin(DIST_35/2.)
        
        AREA_STD  = 4*3.141592653589793/60.
C                       one sixtieth of the sphere surface

        L_GEOMETRY_INIT = .true.

        return
        end

!!! -----  at 20:56:56 on 25-FEB-90
        subroutine crowther_to_matrix(THETA,PHI,OMEGA,ROT)
C
C           generates the rotation matrix corresponding to the crowther
C           angles THETA, PHI,OMEGA
C
C               sdf 29oct89
C
c--------------------------------------------
        implicit none
        real*4 ROT(3,3)
        real*4 THETA, PHI,OMEGA

        real*4 CT,ST
        real*4 CP,SP
        real*4 CO,SO

c--------------------------- Taking the crowther angles as eulerian angles

        ST=sind(THETA)
        CT=cosd(THETA)
        SP=sind(PHI)
        CP=cosd(PHI)
        SO=sind(OMEGA)
        CO=cosd(OMEGA)

        ROT(1,1)= SO*CT*CP-CO*SP
        ROT(1,2)=-SO*CT*SP-CO*CP
        ROT(1,3)=-SO*ST
        ROT(2,1)= CO*CT*CP+SO*SP
        ROT(2,2)=-CO*CT*SP+SO*CP        
        ROT(2,3)=-CO*ST
        ROT(3,1)= ST*CP
        ROT(3,2)=-ST*SP
        ROT(3,3)= CT

        end

c
                subroutine matrix_to_crowther(THETA,PHI,OMEGA,ROT)
C           returns the crowther angles corresponding to a rotation
C           matrix
C       
C       Peter's derivation is based on crowther matrix as:
c-
c       so*ct*cp-co*sp          -so*ct*sp-co*cp         -so*st
c       co*ct*cp+so*sp          -co*ct*sp+so*cp         -co*st
c       st*cp                   -st*sp                  ct 
C
C               sdf 29-oct-89
C.............................................................

        implicit none

        real*4 ROT(3,3)

        real*4 THETA, PHI, OMEGA

        real*4 OMEGA_MINUS_PHI, OMEGA_PLUS_PHI

        real*4 TOLER/1.0e-6/

c fix rotation matrix if not accurate by this method - bug fix

        if(abs(ROT(3,3))+TOLER .ge. 1.0) then
            ROT(3,3)=sign(1.0, ROT(3,3))
        endif                                             

c nb all minus signs in atan2 below necessary       
                
        THETA=acosd(ROT(3,3))

        if(sind(THETA).gt.TOLER) then
                OMEGA = atan2d(-ROT(1,3),-ROT(2,3))
                PHI = atan2d(-ROT(3,2),ROT(3,1))
        else  
                if(cosd(THETA).eq.1.0) then
                        OMEGA_MINUS_PHI=
     $                    atan2d(ROT(2,2),ROT(2,1))
                        OMEGA=0.0               ! arbitrary choice
                        PHI=OMEGA-OMEGA_MINUS_PHI     

                else                            ! cosd(theta)=-1
                        OMEGA_PLUS_PHI=
     $                     atan2d(ROT(2,2),-ROT(2,1))
                        OMEGA=0.0                 ! arbitrary choice
                        PHI=OMEGA_PLUS_PHI     
                endif
        endif          
        end


c
      subroutine matmul(a,b,c)
c
c       multiply matrix b by matrix c and return the product a
c
c               for double precision matrices
c
c               sdf 6-jan-1988
c--------------------------------------------------------
        implicit none
        real A(3,3),B(3,3),C(3,3),B_COPY(3,3),C_COPY(3,3)
        integer I,J,K
c........................................................................
        do I=1,3
            do J=1,3
                B_COPY(I,J)=B(I,J)
                C_COPY(I,J)=C(I,J)
            enddo
        enddo

        do I=1,3
            do J=1,3
                A(I,J)=0.0
                do K=1,3
                    A(I,J)=A(I,J)+B_COPY(I,K)*C_COPY(K,J)
                enddo
            enddo
        enddo
        return
        end

C
      SUBROUTINE   GENROT(ROT)                              

C     modified from G. Bricogne's program GENROT in the MRC package
C
C                 THIS SUBROUTINE GENERATES THE 60 ICOSAHEDRAL ROTATIONS
C                 AND THEIR MATRICES EXPRESSED IN THE ORTHNORMAL FRAME
C                 BASED ON THREE DYADS .
C
c               This is a double precision version
      IMPLICIT real*8 (A-H,O-Z)
C
      REAL*4   ROT(3,3,60)
      real*8    drot(3,3)
      REAL*8   P3(3,3)/9*0.0/,P5(3,3)/9*0.0/
      REAL*8   P3INV(3,3)/9*0.0/,P5INV(3,3)/9*0.0/,TMAT6(3,3)
      REAL*8   TMAT1(3,3),TMAT2(3,3),TMAT3(3,3),TMAT4(3,3),TMAT5(3,3)
      REAL*8   R2Y(3,3,2)/18*0.0/,R2X(3,3,2)/18*0.0/
      REAL*8   S3(3,3,3)/27*0.0/,S5(3,3,5)/45*0.0/
C
c       --------  sdf 6-jan-1988
c-------------------------------------------------------------


      PI = dacos(-1.0d0)
      TPIOV3=PI/1.5
      TPIOV5=PI/2.5
C
      ROOT5=SQRT(5.0)
C
      XX3=ROOT5/6.0
      COS3=SQRT(0.5+XX3)
      SIN3=SQRT(0.5-XX3)
C
      XX5=0.5/ROOT5
      COS5=SQRT(0.5+XX5)
      SIN5=SQRT(0.5-XX5)
C
C                  GENERATE Y-DYAD MATRICES .
      DO 10 K=1,3
   10 R2Y(K,K,1)=1.0
      R2Y(1,1,2)=-1.0
      R2Y(2,2,2)=+1.0
      R2Y(3,3,2)=-1.0
C
C                  GENERATE 3-FOLD MATRICES .
      THETA=-TPIOV3
      DO 20 K=1,3
      THETA=THETA+TPIOV3
      COSINE=COS(THETA)
      SINE=SIN(THETA)
      S3(1,1,K)=1.0
      S3(2,2,K)=COSINE
      S3(3,2,K)=SINE
      S3(2,3,K)=-SINE
   20 S3(3,3,K)=COSINE
C
C                  GENERATE 5-FOLD MATRICES .
      THETA=-TPIOV5
      DO 30 K=1,5
      THETA=THETA+TPIOV5
      COSINE=COS(THETA)
      SINE=SIN(THETA)
      S5(1,1,K)=COSINE
      S5(2,1,K)=SINE
      S5(1,2,K)=-SINE
      S5(2,2,K)=COSINE
   30 S5(3,3,K)=1.0
C
C                  GENERATE X-DYAD MATRICES .
      DO 40 K=1,3
   40 R2X(K,K,1)=1.0
      R2X(1,1,2)=1.0
      R2X(2,2,2)=-1.0
      R2X(3,3,2)=-1.0
C
C                  GENERATE  XZ-THREEFOLD TRANSITION MATRICES .
      P3(1,1)=COS3
      P3(3,1)=SIN3
      P3(2,2)=1.0
      P3(1,3)=-SIN3
      P3(3,3)=COS3
      P3INV(1,1)=COS3
      P3INV(3,1)=-SIN3
      P3INV(2,2)=1.0
      P3INV(1,3)=SIN3
      P3INV(3,3)=COS3
C
C                  GENERATE  XZ-FIVEFOLD TRANSITION MATRICES .
      P5(1,1)=COS5
      P5(3,1)=+SIN5
      P5(2,2)=1.0
      P5(1,3)=-SIN5
      P5(3,3)=COS5
      P5INV(1,1)=COS5
      P5INV(3,1)=-SIN5
      P5INV(2,2)=1.0
      P5INV(1,3)=+SIN5
      P5INV(3,3)=COS5
C
C
C                  GENERATE ICOSAHEDRAL ROTATION MATRICES .
      N=0
      DO 100 L2=1,2
      CALL   DMATMUL(tmat1,R2Y(1,1,L2),P5)
      DO 100 K5=1,5
      CALL   DMATMUL(tmat2,TMAT1,S5(1,1,K5))
      CALL   DMATMUL(tmat3,TMAT2,P5INV)
      DO 100 J2=1,2
      CALL   DMATMUL(tmat4,TMAT3,R2X(1,1,J2))
      CALL   DMATMUL(tmat5,TMAT4,P3)
      DO 100 I3=1,3
      N=N+1
      CALL   DMATMUL(tmat6,TMAT5,S3(1,1,I3))
      CALL   DMATMUL(drot(1,1),TMAT6,P3INV)
      do ii = 1,3
          do jj = 1,3
                rot(ii,jj,N)=drot(ii,jj)
          enddo
      enddo
  100 CONTINUE
C
      RETURN
      END


c
      subroutine DMATMUL(a,b,c)
c
c       multiply matrix b by matrix c and return the product a
c
c               for double precision matrices
c
c               sdf 6-jan-1988
c--------------------------------------------------------
        implicit none
        double precision A(3,3),B(3,3),C(3,3),B_COPY(3,3),C_COPY(3,3)
        integer I,J,K
c........................................................................
        do I=1,3
            do J=1,3
                B_COPY(I,J)=B(I,J)
                C_COPY(I,J)=C(I,J)
            enddo
        enddo

        do I=1,3
            do J=1,3
                A(I,J)=0.0d0
                do K=1,3
                    A(I,J)=A(I,J)+B_COPY(I,K)*C_COPY(K,J)
                enddo
            enddo
        enddo
        return
        end



