      subroutine setrotmat(angles_rad, rot_mat) 
C                                                              SETROTMAT
C **********************************************************************
C *                                                               *
C * Subroutine: setrotmat                                         *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1997                                                  *
C * Used in: cmpt_intrps -- The subroutine for Fourier transform  *
C *                         and interpolation                     *
C * Function: Form matrix which rotates pixel plane to standard   *
C *           coordinate system.                                  *
C * Usage:                                                        *
C *   call setrotmat( angles_rad, rot_mat)               *
C * Input:                                                        *
C *   angles_rad (Real) -- Rotation anlges THETA, PHI, OMEGA in   *
C *                        radians.                               *
C * Output:                                                       *
C *   rot_mat (Real) -- Rotation matrix which maps a point        *
C *     (U,V,W)^T in the space of pixel Fourier coefficients to   *
C *     the corresponding point (X,Y,Z)^T = ROT_MAT*(U,V,W)^T in  *
C *     the standard cartesian space of Fourier coefficients of   *
C *     the particle.                                             *
C *                                                               *
C * Algorithm:                                                    *
C *   Matrix which rotates a vector about the original y-axis:    *
C *                                                               *
C *               |   cos(theta)             sin(theta) |         *
C *       rot_1 = |    1            1          0        |         *
C *               |  -sin(theta)             cos(theta) |         *
C *                                                               *
C *   Matrix which rotates a vector about the original z-axis:    *
C *                                                               * 
C *               |  cos(phi)  -sin(phi)       0        |         *
C *       rot_2 = |  sin(phi)   cos(phi)       0        |         *
C *               |    0          0            1        |         *
C *                                                               * 
C *   The original x, y, z axes are now vectors x_21, y_21, z_21: *
C *                                                               * 
C *                        | 1 0 0 |   |                |         *
C *        rot_2 * rot_1 * | 0 1 0 | = | x_21 y_21 z_21 |         *
C *                        | 0 0 1 |   |                |         *
C *                                                               * 
C *   i.e, the original x-axis = ( 1, 0, 0)^T is now the vector   *
C *   x_21 where x_21 = rot_2 * rot_1 * ( 1, 0, 0 )               *
C *                                                               * 
C *   Rotation about the new z axis, which is the vector v_z, is  *
C *   obtained by                                                 *
C *      x_21 --> x_213 =  cos(omega) * x_21 + sin(omega) * y_21  *
C *      y_21 --> y_213 = -sin(omega) * x_21 + cos(omega) * y_21  *
C *      z_21 --> z_213                                           *
C *                                                               * 
C *   A vector expressed with respect to the axes x, y, z is      *
C *   transformated by the rotations is accomplished by           *
C *   multiplication of the vector by the matrix                  *
C *       rot = rot_2 * rot_1 * rot_3   where                     *
C *                                                               * 
C *               |  cos(omega)  -sin(omega)       0   |          *
C *       rot_3 = |  sin(omega)   cos(omega)       0   |          *
C *               |     0            0             1   |          *
C *                                                               * 
C *   Thus, ROT_MAT = ROT_1 * ROT_2 * ROT_3                       *
C *                                                               *
C * Modification history:                                         *
C *  - Hong Lin, Version 1.0, Aug.18, 1998                             *
C *  - Hong Lin, Version 2.0, Oct.11, 1999                             *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C **********************************************************************
C ----------------------------------------------------------------------
C Form matrix which rotates pixel plane to standard coordinate system
C Arguments  -----------------------------------------------------------
C  Input ---------------------------------------------------------------
C  ANGLES_RAD(1:3)  REAL*4: rotation anlges THETA, PHI, OMEGA in radians
C  Output --------------------------------------------------------------
C  ROT_MAT(1:3,1:3) REAL*4: rotation matrix which maps a point (U,V,W)^T
C        in the space of pixel Fourier coefficients to the corresponding
C        point (H,K,L)^T = ROT_MAT*(U,V,W)^T in the standard cartesian
C        space of Fourier coefficients of the particle.  In particular,
C        Fourier coefficients of the 2D array of pixels are at points
C        (U,V,0)^T in the plane W = 0.
C ----------------------------------------------------------------------
C
      implicit none
      include '../include/allprogs.inc'
C
C Argument
C
      real angles_rad(1:3), rot_mat(1:3,1:3)
C
C Local
C
      integer irow, icol, k
      real    swap
      real    rot_1(3,3), rot_2(3,3), rot_3(3,3), rot_2_1(3,3),
     +        theta_rad, phi_rad, omega_rad
C ----------------------------------------------------------------------
 
      do 1010 icol = 1, 3
        do 1000 irow = 1, 3
          rot_2_1(irow,icol) = 0.
          rot_mat(irow,icol) = 0.
 1000   continue
 1010 continue
C
      theta_rad  = angles_rad(1)
      phi_rad    = angles_rad(2)
      omega_rad  = angles_rad(3)
C
      rot_1(1,1) =   cos( theta_rad )
      rot_1(1,2) =   0.
      rot_1(1,3) =   sin( theta_rad )
      rot_1(2,1) =   0.
      rot_1(2,2) =   1.
      rot_1(2,3) =   0.
      rot_1(3,1) = - rot_1(1,3)
      rot_1(3,2) =   0.
      rot_1(3,3) =   rot_1(1,1)
 
      rot_2(1,1) =   cos( phi_rad )
      rot_2(1,2) = - sin( phi_rad )
      rot_2(1,3) =   0.
      rot_2(2,1) = - rot_2(1,2)
      rot_2(2,2) =   rot_2(1,1)
      rot_2(2,3) =   0.
      rot_2(3,1) =   0.
      rot_2(3,2) =   0.
      rot_2(3,3) =   1.
 
      rot_3(1,1) =   cos( omega_rad )
      rot_3(1,2) = - sin( omega_rad )
      rot_3(1,3) =   0.
      rot_3(2,1) = - rot_3(1,2)
      rot_3(2,2) =   rot_3(1,1)
      rot_3(2,3) =   0.
      rot_3(3,1) =   0.
      rot_3(3,2) =   0.
      rot_3(3,3) =   1.
C
C     Form the product  ROT_2_1 = ROT_2 * ROT_1
 
      do 1040 icol = 1, 3
	do 1020 irow = 1, 3
          do 1030 k = 1, 3
            rot_2_1(irow,icol) = rot_2_1(irow,icol)
     +                         + rot_2(irow,k)*rot_1(k,icol)
 1030     continue
 1020   continue
 1040 continue
 
C     Form the product ROT_MAT = ROT_2_1 * ROT_3
 
      do 1070 icol = 1, 3
	do 1050 irow = 1, 3
          do 1060 k = 1, 3
            rot_mat(irow,icol) = rot_mat(irow,icol)
     +                         + rot_2_1(irow,k)*rot_3(k,icol)
 1060     continue
 1050   continue
 1070 continue
C
C     The matrix  rot_mat  is Hermitian (its inverse is its transpose).
C     Store the inverse of  rot_mat  in  rot_mat_inv
C Transpose ROTMAT to get it inversion.
C
      do 1080 icol = 1, 3
        do 1090 irow = icol+1, 3
          swap = rot_mat(irow, icol)
          rot_mat(irow, icol) = rot_mat(icol, irow)
          rot_mat(icol, irow) = swap
 1090   enddo
 1080 enddo
      return
      end
