      real function sincpi(x)
C                                                                 SINCPI
C **********************************************************************
C *                                                               *
C * Subroutine: sincpi                                            *
C * Author: Robert E. Lynch (rel@cs.purdue.edu)                   *
C * Date:   1997                                                  *
C * Used in: matrixentries -- The subroutine estimating 3D DFT    *
C *            by interpolation from 2D DFT                       *
C * Function: Calculate function SINCPI                           *
C * Usage:                                                        *
C *   call sincpi(x)                                              *
C * Input:                                                        *
C *   x (real) -- The variable                                    *
C * Output:                                                       *
C *   sincpi(x) (Real) -- The function SINCPI on X                *
C *                                                               *
C * Algorithm:                                                    *
C *   Evaluate   sinc( pi * x ) = sin( pi * x ) / ( pi * x )      *
C *                                                               *
C *   After translation of  X   to interval  0 .LE. Y .LE. 2,     *
C *   sin( pi * y ) is evaluated, accurate to 0.25 * 10**( -8 ),  *
C *   by use of the cubic Taylor interpolant centered at the mesh *
C *   point nearest to pi*y. The first time that  SINCPI is       *
C *   called, Values of sin(pi*y) and cos(pi*y) are tabulated at  *
C *   200 equal spaced points between 0 and y = 2.                *
C *                                                               *
C *   A test with spacing  0.001, 0.01, 0.1 and random numbers    *
C *   uniformally distributed between in intervals                *
C *          [0, 0.01], [0, 10], and [0, 100]                     *
C *   indicates that that the maximum interpolation error is less *
C *   than 1.8e-07.                                               *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 2.0, Oct. 7, 1999                     *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                               *
C *****************************************************************
C ----------------------------------------------------------------------
C     SAVE needed so that the tabulated values are not destroyed between
C     successive calls of this routine
C
      implicit none
      save
C
C   Input/Output -------------------------------------------------------
C Argument
C
      real x
C
C Local
C
      integer n_max, iy, j_pi_x, j_pi_y
      parameter ( n_max = 200 )
      logical  first_time
      data first_time  / .true. /
      real     delta_pi_y, omega, omega_sqr, pi, pi_x, y, pi_y,
     +         rec_6, rec_del_pi_y, small_pi_x,
     +         sin_pi( 0:(n_max+1) ), cos_pi( 0:(n_max+1) )
      common /com_sincpi/ delta_pi_y, omega, omega_sqr, pi, pi_x, pi_y,
     +                    rec_6, rec_del_pi_y, small_pi_x, y,
     +                    sin_pi, cos_pi, iy, j_pi_x, j_pi_y, 
     +                    first_time
C ----------------------------------------------------------------------
C
      if( first_time ) then
C
C       Form tables of sin( j pi delta_y ) and cos( j pi delta_y ) for
C       DELTA_Y = 2/N_MAX and J = 0:1:(N_MAX+1)
C
        pi           = atan( 1. ) * 4.
        delta_pi_y   = 2. * pi / float( n_max )
        rec_del_pi_y = 1.0 / delta_pi_y
        small_pi_x   = pi * 0.005
        sin_pi( 0 ) = 0.
        cos_pi( 0 ) = 1.
        do 1000 iy = 1, (n_max + 1)
          pi_y = float( iy ) * delta_pi_y
          sin_pi(iy) = sin( pi_y )
          cos_pi(iy) = cos( pi_y )
 1000   continue
        rec_6 = 1.0 / 6.0
        first_time = .false.
      end if
      pi_x = pi * abs( x )
      if( pi_x .le. small_pi_x ) then
C
C       PI_X is close enough to zero so that the value of the quadratic
C       Taylor interpolant about X = 0 gives a value accurate to
C       0.25 * 10**(-8)
C
        sincpi = 1. - pi_x * pi_x * rec_6
      else
C
C       Find mesh point which is closest to PI_X
C
        j_pi_x = nint( pi_x * rec_del_pi_y )
C
C       Find the distance to nearest mesh point
C
        omega = pi_x - float( j_pi_x ) * delta_pi_y
C
C       Translate J_PI_X to interval 0 .le. y .le. 2
C
        j_pi_y = j_pi_x - ( j_pi_x / n_max ) * n_max
C
C       Use periodicity and cubic Taylor interpolant to evaluate sine
C
        omega_sqr = omega * omega
        sincpi = (  sin_pi( j_pi_y ) * (1. - 0.5 * omega_sqr)
     +            + cos_pi( j_pi_y ) * omega * (1. - omega_sqr * rec_6)
     +           ) / pi_x
      end if
C
      return
      end
