      subroutine fftsynth_1_m_slab(hkl_min, hkl_max, 
     +             first_iq, last_iq, f_hkl, nrffts_1_row, 
     +             nrffts_1_col, nrffts_1_omega, min_x, 
     +             max_x, min_y, max_y, min_z, max_z, 
     +             min_xyz, max_xyz, f_vru, mem_total )
C                                                      FFTSYNTH_1_M_SLAB
C *****************************************************************
C *                                                               *
C * Subroutine: fftsynth_1_m_slab                                 *
C * Author: Hong Lin (linh@cs.purdue.edu)                         *
C * Date:   Dec.10, 1998                                          *
C * Used in: aamain -- The main routine.                          *
C * Function: The control subroutine for 2-D Fourier synthesis,   *
C *           "piecewise constant" model.                         * 
C * Usage:                                                        *
C *   call fftsynth_1_m_slab(hkl_min, hkl_max, first_iq,          *
C *          last_iq, f_hkl, nrffts_1_row, nrffts_1_col,          *
C *          nrffts_1_omega, min_x, max_x, min_y, max_y, min_z,   *
C *          max_z, min_xyz, max_xyz, f_vru, mem_total)           *
C * Input:                                                        *
C *   F_HKL (Complex) -- The 3D Fourier coefficients estimated    *
C *                      from 2D DFT                              *
C *   hkl_min (Integer) -- The minimum index of HKL space.        *
C *   hkl_max (Integer) -- The maximum index of HKL space.        *
C *   first_iq (Integer) -- The starting index of IQ in this node *
C *   last_iq (Integer) -- The ending index of IQ in this node    *
C *   nrffts_1_row (Integer) -- The limits of row number of array *
C *                   "rffts_1_even", "rffts_1_odd" and the       *
C *                   column number of "rffts_1_trns".            *
C *   nrffts_1_col (Integer) -- The limits of row number of array *
C *                   "rffts_1_even", "rffts_1_odd" and the row   *
C *                   number of "rffts_1_trns".                   *
C *   nrffts_1_omega (Integer) -- The limits of size of array     *
C *                     "rffts_1_omega.                           *
C *   min_x (Integer) -- The minimum index in X axe of the real   *
C *                      space.                                   *
C *   min_y (Integer) -- The minimum index in Y axe of the real   *
C *                      space.                                   *
C *   min_z (Integer) -- The minimum index in Z axe of the real   *
C *                      space.                                   *
C *   max_x (Integer) -- The maximum index in X axe of the real   *
C *                      space.                                   *
C *   max_y (Integer) -- The maximum index in Y axe of the real   *
C *                      space.                                   *
C *   max_z (Integer) -- The maximum index in Z axe of the real   *
C *                      space.                                   *
C *   min_xyz (Integer) -- The minimum index in X, Y and Z axes   *
C *   max_xyz (Integer) -- The maximum index in X, Y and Z axes   *
C * Output:                                                       *
C *   F_VRU (Real) -- The result of 2-D Fourier synthesis         *
C *   mem_total (Double precision) -- The memory allocated in     *
C *     subroutine                                                *
C *                                                               *
C * Algorithm:                                                    *
C *   Allocating memory of subsidiary arrays "rffts_1_even",      *
C *   "rffts_1_odd", "rffts_1_trns" and "rffts_1_omega" and then  *
C *   call subroutine "fftsynth_1" to do 2-D Fourier synthesis.   *
C *   After returning from "fftsynth_1", free these subsidiary    *
C *   arrays.                                                     *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *     - Hong Lin, Version 2.0, Oct. 5, 1999                     *
C *       Changes in the list of parameters                       *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *
C *                                                               *
C *****************************************************************
      implicit none
      include '../include/allprogs.inc'
C
C Arguments
C
      integer hkl_min, hkl_max, first_iq, last_iq,
     +        nrffts_1_row, nrffts_1_col, nrffts_1_omega,
     +        min_x, max_x, min_y, max_y, min_z, max_z, 
     +        min_xyz, max_xyz
C
      complex f_hkl(hkl_min:hkl_max,hkl_min:hkl_max,
     +              first_iq:last_iq)
C
      real F_VRU( min_xyz:max_xyz,
     +            frst_lst_RR_out(1,iam):frst_lst_RR_out(2,iam),
     +            min_xyz:max_xyz )
      double precision mem_total
C
C ======================================================================
C
C Local
C
      real    rffts_1_even(1),
     +        rffts_1_odd (1),
     +        rffts_1_trns(1),
     +        rffts_1_omega(1),
     +        F_VU(1)
C
      pointer (pnt_rffts_1_even, rffts_1_even),
     +        (pnt_rffts_1_odd,  rffts_1_odd ),
     +        (pnt_rffts_1_trns, rffts_1_trns),
     +        (pnt_rffts_1_omega, rffts_1_omega),
     +        (pnt_F_VU,          F_VU         )
C
      integer mem_size, iiuu, iirr, iivv
      double precision mem_rffts_1_even, mem_rffts_1_odd, 
     +       mem_rffts_1_trns, mem_rffts_1_omega, mem_F_VU
      data   mem_rffts_1_even / 0.0 /, mem_rffts_1_odd / 0.0 /,
     +       mem_rffts_1_trns / 0.0 /, mem_rffts_1_omega / 0.0 /,
     +       mem_F_VU / 0.0 /
C
      integer cmalloc

C ======================================================================
C
      mem_size = nrffts_1_row * nrffts_1_col
      pnt_rffts_1_even = cmalloc(4 * mem_size)
      if (pnt_rffts_1_even.eq.0) call error_stop(progname,
     +  'fftsynth_1_m', 'Error in allocating memory for RFFTS_1_EVEN!')
      mem_rffts_1_even = dble(4 * mem_size)
C
      pnt_rffts_1_odd  = cmalloc(4 * mem_size)
      if (pnt_rffts_1_odd.eq.0) call error_stop(progname,
     +  'fftsynth_1_m', 'Error in allocating memory for RFFTS_1_ODD!')
      mem_rffts_1_odd  = dble(4 * mem_size)
C
      mem_size = nrffts_1_row * nrffts_1_col
      pnt_rffts_1_trns = cmalloc(4 * mem_size)
      if (pnt_rffts_1_trns.eq.0) call error_stop(progname,
     +  'fftsynth_1_m', 'Error in allocating memory for RFFTS_1_TRNS!')
      mem_rffts_1_trns = dble(4 * mem_size)
C
      pnt_rffts_1_omega = cmalloc(4 * nrffts_1_omega)
      if (pnt_rffts_1_omega.eq.0) call error_stop(progname,
     +  'fftsynth_1_m', 'Error in allocating memory for RFFTS_1_OMEGA!')
      mem_rffts_1_omega = dble(4 * nrffts_1_omega)
C
      mem_size = (-min_xyz+max_xyz+1)**2
      pnt_F_VU = cmalloc(4 * mem_size)
      if (pnt_F_VU.eq.0) call error_stop(progname,
     +  'fftsynth_1_m', 'Error in allocating memory for F_VU!')
      mem_F_VU = dble(4 * mem_size)
      mem_total = mem_rffts_1_even + mem_rffts_1_odd + mem_rffts_1_trns
     +          + mem_rffts_1_omega + mem_F_VU
C
      call fftsynth_1_slab( hkl_min, hkl_max, first_iq, last_iq, 
     +       f_hkl, nrffts_1_row, nrffts_1_col, rffts_1_even, 
     +       rffts_1_odd, rffts_1_trns, nrffts_1_omega, rffts_1_omega, 
     +       min_x, max_x, min_y, max_y, min_z, max_z, min_xyz, 
     +       max_xyz, F_VU, F_VRU )
C
      call cfree(pnt_rffts_1_even)
      call cfree(pnt_rffts_1_odd)
      call cfree(pnt_rffts_1_trns)
      call cfree(pnt_rffts_1_omega)
      call cfree(pnt_F_VU)
C
      return
      end
