      subroutine fftsynth_2_m_slab(rrnum, min_x, max_x, 
     +             min_y, max_y, min_z, max_z, F_XK_dZ, 
     +             nrffts_2_row, nrffts_2_col, nrffts_2_omega, 
     +             rho, mem_total )
C                                                      FFTSYNTH_2_M_SLAB
C **********************************************************************
C *                                                               *
C * Subroutine: fftsynth_2_m_slab                                 *
C * Author: Hong Lin (linh@cs.purdue.edu)                         *
C * Date:   Jul.6, 1998                                           *
C * Used in: aamain -- The main routine.                          *
C * Function: The driver of 1-Dimensional Fourier synthesis,      *
C *           "piecewise constant" model.                         *
C * Usage:                                                        *
C *   call fftsynth_2_m_slab( num_hkl, min_x, max_x, min_y,       *
C *          max_y, min_z, max_z, F_XK_dZ, nrffts_2_row,          *
C *          nrffts_2_col, nrffts_2_omega, rho, mem_total )       *
C * Input:                                                        *
C *   F_XK_dZ (Real) -- The output of 2-D Fourier synthesis       *
C *   min_x (Integer) -- The minimum index in X axe of the real   *
C *                      space.                                   *
C *   min_y (Integer) -- The minimum index in Y axe of the real   *
C *                      space.                                   *
C *   min_z (Integer) -- The minimum index in Z axe of the real   *
C *                      space.                                   *
C *   max_x (Integer) -- The maximum index in X axe of the real   *
C *                      space.                                   *
C *   max_y (Integer) -- The maximum index in Y axe of the real   *
C *                      space.                                   *
C *   max_z (Integer) -- The maximum index in Z axe of the real   *
C *                      space.                                   *
C *   num_hkl (Integer) -- The number of coordinates in H, K or L *
C *                        axe of the reciprocal space.           *
C *   nrffts_2_row, nrffts_2_col (Integer) -- The limits of size  *
C *                                           of array "rffts_2". *
C *   nrffts_2_omega (Integer) -- The limits of size of array     *
C *                               "rffts_2_omega".                *
C * Output:                                                       *
C *   rho (Real) -- The final calculated density.                 *
C *   mem_total (Double precision) -- The memeory allocated in    *
C *                                   this subroutine             *
C *                                                               *
C * Algorithm:                                                    *
C *   Allocating memory of subsidiary arrays "rffts_2",           *
C *   "rffts_2_trns" and "rffts_2_omega" and then call subroutine *
C *   "fftsynth_2" to do 1-D Fourier synthesis. After returning   *
C *   from "fftsynth_2", free these subsidiary arrays.            *
C *                                                               *
C * Modification history:                                         *
C *     - Hong Lin, Version 1.0, Aug.18, 1998                     *
C *     - Hong Lin, Version 2.0, Oct. 6, 1999                     *
C *       Data structure is changed.                              *
C *     - Hong Lin, Version 2.0, Dec.10, 1999                     *
C *       Remove extra data structure RHO_XZ.                     *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                               *
C *****************************************************************
      implicit none
      include '../include/allprogs.inc'
C
C Arguments
C
      integer rrnum, min_x,max_x, min_y,max_y, min_z,max_z,
     +        nrffts_2_row, nrffts_2_col, nrffts_2_omega
C
      real    F_XK_dZ( min_x:max_x, 1:rrnum,
     +                 frst_lst_u(1,IAM):frst_lst_u(2,IAM)),
     +        rho( min_x:max_x, min_y:max_y,
     +             frst_lst_z(1,IAM):frst_lst_z(2,IAM))
C
      double precision mem_total
C ======================================================================
C
C Local
C
      real    rffts_2(1), rffts_2_trns(1), rffts_2_omega(1)
C
      pointer (pnt_rffts_2, rffts_2),
     +        (pnt_rffts_2_trns, rffts_2_trns),
     +        (pnt_rffts_2_omega, rffts_2_omega)
C
      integer mem_size, xx, zz, kk
      double precision mem_rffts_2, mem_rffts_2_trns,
     +       mem_rffts_2_omega
      data   mem_rffts_2 / 0.0 /, mem_rffts_2_trns / 0.0 /,
     +       mem_rffts_2_omega / 0.0 /
C
      integer cmalloc

C =====================================================================
C
      mem_size = nrffts_2_row * nrffts_2_col
      pnt_rffts_2 = cmalloc(4 * mem_size)
      if (pnt_rffts_2.eq.0) call error_stop(progname,
     +  'fftsynth_2_m', 'Error in allocating memory for RFFTS_2!')
      mem_rffts_2 = dble(4 * mem_size)
C
      pnt_rffts_2_trns = cmalloc(4 * mem_size)
      if (pnt_rffts_2_trns.eq.0) call error_stop(progname,
     +  'fftsynth_2_m', 'Error in allocating memory for RFFTS_2_TRNS!')
      mem_rffts_2_trns = dble(4 * mem_size)
C
      pnt_rffts_2_omega = cmalloc(4 * nrffts_2_omega)
      if (pnt_rffts_2_omega.eq.0) call error_stop(progname,
     +  'fftsynth_2_m', 'Error in allocating memory for RFFTS_2_OMEGA!')
      mem_rffts_2_omega = dble(4 * nrffts_2_omega)
C     
      mem_total   = mem_rffts_2 + mem_rffts_2_trns + mem_rffts_2_omega
C
      call fftsynth_2_slab( rrnum, min_x, max_x, min_y, max_y, min_z, 
     +       max_z, F_XK_dZ, nrffts_2_row, nrffts_2_col, rffts_2, 
     +       rffts_2_trns, nrffts_2_omega, rffts_2_omega, rho )
C
      call cfree(pnt_rffts_2)
      call cfree(pnt_rffts_2_trns)
      call cfree(pnt_rffts_2_omega)
C
      return
      end
