      subroutine matrixentries_slab(nfc_2d_row, nfc_2d_col, 
     +             fc_2d, hkl_min, hkl_max, max_axis, HKL_DFT,
     +             max_count, datalen,
     +             rot_mat, ruv_mat_1, ruv_mat_2, ruv_sup_1, ruv_sup_2,
     +             num_annulus, ctf2, imgmagfactor, mapmagfactor)
C                                                     MATRIXENTRIES_SLAB
C **********************************************************************
C *                                                                    *
C * Subroutine: matrixentries_slab                                     *
C * Author: Hong Lin (linh@cs.purdue.edu)                              *
C * Date:   1997                                                       *
C * Used in: cmpt_intrps -- The subroutine for Fourier transform and   *
C *                         interpolation                              *
C * Function: Interpolation in "piecewise constant" method.            *
C * Usage:                                                             *
C *   call matrixentries_slab( nfc_2d_row, nfc_2d_col, fc_2d, hkl_min, *
C *          hkl_max, max_axis, HKL_DFT, rot_mat, ruv_mat_1, ruv_mat_2, *
C *          ruv_sup_1, ruv_sup_2 )                                    *
C * Input:                                                             *
C *   nfc_2d_row (Integer) -- Limits on the number of rows of the output     *
C *                           array FC_2D.                                   *
C *   nfc_2d_col (Integer) -- Limits on the number of columns of the output  *
C *                           array FC_2D.                                   *
C *   fc_2d (Complex) -- Array with NFC_2D_ROW .eq. NFC-2D_COL, containing   *
C *                      the Fourier coefficients.                           *
C *   hkl_min (Integer) -- The minimum index of HKL space.                   *
C *   hkl_max (Integer) -- The maximum index of HKL space.                   *
C *   max_axis (Integer) -- The maximum number of axis.                      *
C *   rot_mat (Real) -- The matrix for coordinate conversion between XYZ and *
C *                     PQR space.                                           *
C *   ruv_mat_1, ruv_mat_2, ruv_sup_1, ruv_sup_2 (Real) -- For axis IA, the  *
C *     matrix-vector product RUV_MAT(:,:,IA) * [IP,IQ]^T is equal to        *
C *     [R_PRIME,U_PRIME,V_PRIME]^T which gives the coordinates              *
C *     (R_PRIME,IP,IQ) of a point in the 3D standard space of Fourier       *
C *     coefficients and the corresponding point (U_PRIME,V_PRIME,-0.5) and  *
C *     (U_PRIME,V_PRIME,0.5) in the plane for Fouier coefficients of the    *
C *     pixel values.                                                        *
C *                                                                          *
C * Algorithm:                                                               *
C *   This routine computes the information necessary to combine to          *
C *   determine the 3D Fourier coefficients on a cubic mesh, FC(H,K,L),      *
C *   H_MIN:H_MAX, K_MIN:K_MAX, L_MIN:L_MAX.                                 *
C *                                                                          *
C *   The input includes the 2D Fourier coefficients, FC_2D(U,V),            *
C *   U_MIN:U_MAX, V_MIN:V_MAX, of the transform of the projection of the    *
C *   particle.  The plane of the projection has a general orientation with  *
C *   respect to the axes used for FC(H,K,L).  If the axis H is not parallel *
C *   to the plane, then for each pair, (K,L), the intersection of the line  *
C *   through (K,L) and parallel to the H-axis intersects the plane at       *
C *   (U',V') at (H',K,L).   Similarly, we can obtain 2 other intersections. *
C *   In this subroutine, "piecewise constant" method is used, which means   *
C *   that the width of the 2D DFT is 1 (instead of 0 in MATRIXENTRIES).     *
C *   Therefore, it is not necessary to obtain 3 intersections. Instead, we  *
C *   one direction to which the plane of 2D DFT is not parallel and         *
C *   calculate the grid points which are embraced by the 2D DFT slab. We    *
C *   consider the value at the nearest (U,V,W) point in 2D DFT slab as one  *
C *   of the estiamtes for the 3D point. Note that -0.5<=W<=0.5 and the      *
C *   value at (U,V,W) equals to FC_2D(U,V). Then, we average all estimates  *
C *   for the point in 3D HKL space from different pixel frames. As one node *
C *   only processes a portion of the pixel frames, we keep the number of    *
C *   estimates got in this node for each point and do a final average in    *
C *   subroutine SOLVE_BY_INTERP_SLAB.                                 *
C *                                                                    *
C * Modification history:                                              *
C *   - Hong Lin, Version 2.0, Oct. 7, 1999                            *
C *   - Yongchang Ji, Final Version, Aug.2000 -- Jan. 2002             *

C *                                                                    *
C **********************************************************************
C Arguments  -----------------------------------------------------------
C
C Input:   -----------------------------------------------------------
 
C  NFC_2D_ROW = NFC_2D_COL  INTEGER*4; dimensions of rectangular
C       array FC_2D.
 
C  FC_2D( 0:NFC_2D_ROW, -NFC_2D_COL:NFC_2D_COL)  COMPLEX*4; array of
C       complex Fourier coefficients of pixel values (after zero fill).
 
C  HKL_MIN, HKL_MAX  INTEGER*4;  lower and upper limits of the indices
C       H,K,L of the output Fouier coefficients.
 
C  MAX_AXIS   INTEGER*4;  number of different axes intersecting the
C       plane of Fourier coefficients of the pixel values (MAX_AXIS is
C       equal to 3).
 
C  RUV_MAT( 3, 2, 3 )  REAL*4;  for axis IA, the matrix-vector product
C       RUV_MAT(:,:,IA) * [IP,IQ]^T is equal to
C       [R_PRIME,U_PRIME,V_PRIME]^T which gives the coordinates
C       (R_PRIME,IP,IQ) of a point in the 3D standard space of
C       Fourier coefficients and the corresponding point
C       (U_PRIME,V_PRIME,0) in the plane for Fouier coefficients of
C       the pixel values.

C Output:  -----------------------------------------------------------
 
C INTERPOLANTS( 3, NUM_INTRP_EQNS, MAX_AXIS, HKL_MIN:HKL_MAX, 0:HKL_MAX)
C       REAL*4;  The values INTERPOLANTS(I,N,IA,IP,IQ) are:
C         I .eq. 1,2,3 for the Nth set of values R_PRIME, real( FC ),
C            and aimag( FC ), respectively, where R_PRIME is the
C            point of intersection of the line  (r,IP,IQ)  in the 3D
C            standard space of Fourier coefficients and a point in
C            the plane of Fourier coefficients of the pixel values
C            for the axis-system IA
C ----------------------------------------------------------------------
C
      implicit none
      include '../include/allprogs.inc'
C
C Arguments 
C
      integer nfc_2d_row, nfc_2d_col, hkl_min, hkl_max, max_axis,
     +        max_count, datalen
C
      complex fc_2d( 0:nfc_2d_row, - nfc_2d_col:nfc_2d_col )
C
      real    HKL_DFT( 6, max_count)
      real    ctf2(0:nfc_2d_row, -nfc_2d_col:nfc_2d_col)

      real    ruv_mat_1(3,2,3), ruv_mat_2(3,2,3), ruv_sup_1(3,3),
     +        ruv_sup_2(3,3), rot_mat(3,3), num_annulus, 
     +        imgmagfactor, mapmagfactor
C ----------------------------------------------------------------------
C
C Local
C
      complex c_rght_side, fc_ne, fc_nw, fc_sw, fc_se
C
      integer i,j
      integer ia, ip, iq, iu_nearest, iv_nearest,
     +        iu_minus, iu_plus, iv_minus, iv_plus, r_prime_up,
     +        r_prime_dn, case, ih, ik, il, ip_start, iq_start,
     +        ia_start, ip_end, iq_end, ia_end
C 
      real    real_ip, real_iq, real_ia, uv_max, r_prime_1, r_prime_2,
     +        r_prime_start, r_prime_end, u_prime, v_prime, del_u, 
     +        del_u_minus, del_u_plus, del_v, del_v_minus, del_v_plus, 
     +        radius, num_annulus_sqr, mag_annulus,
     +        sincpi_del_u_minus, sincpi_del_v_plus,
     +        sincpi_del_u_plus, sincpi_del_v_minus,
     +        w_tmp, real_ih, real_ik, real_il,
     +        ct_ne, ct_nw, ct_sw, ct_se, ct_all
C 
      logical accurate
C
C Note: accurate = .true. -- Using Fourier transform theory to
C                            calculate the interpolant;
C       accurate = .false. -- Using bilinear interpolation.
C
      data accurate / .false. /
C 
      character*18 subname
      data         subname /'MATRIXENTRIES_SLAB'/
C
      real sincpi
      external sincpi
C ----------------------------------------------------------------------
C
C     NFC_2D_ROW .eq. NFC_2D_COL .eq. - NFC_2D_COL
C
      mag_annulus      = num_annulus / mapmagfactor
      num_annulus_sqr  = num_annulus * num_annulus
      uv_max           = float( nfc_2d_col )
C 
C Decide which index among H, K and L will use as R_PRIME.
C
      if (abs(rot_mat(3,1)).ge.abs(rot_mat(3,2)) .and.
     +    abs(rot_mat(3,1)).ge.abs(rot_mat(3,3))) then
	case = 1
	ip_start = hkl_min
        ip_end   = hkl_max
	iq_start = 0
        iq_end   = hkl_max
	ia_start = hkl_min
        ia_end   = hkl_max
      elseif (abs(rot_mat(3,2)).ge.abs(rot_mat(3,1)) .and.
     +        abs(rot_mat(3,2)).ge.abs(rot_mat(3,3))) then
	case = 2
	ip_start = hkl_min
        ip_end   = hkl_max
	iq_start = hkl_min
        iq_end   = hkl_max
	ia_start = 0
        ia_end   = hkl_max
      elseif (abs(rot_mat(3,3)).ge.abs(rot_mat(3,1)) .and.
     +        abs(rot_mat(3,3)).ge.abs(rot_mat(3,2))) then
	case = 3
        ip_start = 0
        ip_end   = hkl_max
	iq_start = hkl_min
        iq_end   = hkl_max
	ia_start = hkl_min
        ia_end   = hkl_max
      else
	call error_stop(progname, subname, 
     +         'Error: cannot determine which case should be used!')
      endif
C
      ip_start = max( ip_start, int(-mag_annulus - 1.) )
      ip_end   = min( ip_end, int(mag_annulus + 1.) )
      iq_start = max( iq_start, int(-mag_annulus - 1.))
      iq_end   = min( iq_end, int(mag_annulus + 1.))
      ia_start = max( ia_start, int(-mag_annulus - 1.))
      ia_end   = min( ia_end, int(mag_annulus + 1.))
C
      do 1070 ip = ip_start,ip_end 
C       IP .eq. L, H, K for axis IA .eq. 1, 2, 3, respectively
Cji     apply the map mag factor to get the real original H,K,L 
C 
        real_ip = float( ip * mapmagfactor )
C
C       When the array  RUV_P_PRIME(1:3,1:3) is filled, its three
C       row will hold the contributions to R', U', V' with the first
C       column having the contribution for IA = 1 for axis = H, the
C       second for IA = 2 for axis K, and the third for IA = 3 for L.
C
        do 1060 iq = iq_start,iq_end 
C         IQ .eq. K, L, H for axis IA .eq. 1, 2, 3, respectively
 
          real_iq = float( iq * mapmagfactor)
C
C Calculate the start and end index of the R_PRIME's of those grid points
C which are in the slab.
C
	  r_prime_start = ruv_mat_1(1,1,case)*real_ip + 
     +                    ruv_mat_1(1,2,case)*real_iq + 
     +                    ruv_sup_1(1,case)
	  r_prime_end   = ruv_mat_2(1,1,case)*real_ip +
     +                    ruv_mat_2(1,2,case)*real_iq + 
     +                    ruv_sup_2(1,case)
C
	  r_prime_1     = min(r_prime_start,r_prime_end)
	  r_prime_2     = max(r_prime_start,r_prime_end)
C
Cji  apply back the map mag factor
C
          r_prime_1 = r_prime_1 / mapmagfactor
          r_prime_2 = r_prime_2 / mapmagfactor
C
	  if (r_prime_1.le.0.0) then
	    r_prime_dn  = int(r_prime_1)
	  else
	    r_prime_dn  = int(r_prime_1+1.0)
	  endif
	  if (r_prime_2.lt.0.0) then
	    r_prime_up  = int(r_prime_2-1.0)
	  else
	    r_prime_up  = int(r_prime_2)
	  endif
	  r_prime_dn    = max(ia_start, r_prime_dn)
	  r_prime_up    = min(ia_end,  r_prime_up)
C
	  if (r_prime_dn.gt.r_prime_up) goto 1060
C
C         Now each column of the array RUV_PQ_PRIME contains the values
C         of R', U', V' due to contributions from both IP and IQ; the
C         first column has values for IA .eq. 1 for H, second column for
C         IA .eq. 2 for K, and third column for IA .eq. 3 for L
 
          do 1050 ia = r_prime_dn, r_prime_up
 
C           The components RUV_PQ_PRIME( 1:3, ia ) are in units of the
C           the space of 3d Fourier coefficients whose synthesis gives
C           the output electron density.  The maximum grid point is
C           HKL_MAX.  The maximum grid point in the 2D Fourier
C           coefficients of the pixel values is NFC_2D_COL. The edge
C           lengths of the box and square containing the points are
C           equal.  The conversition factor is for the gird spacings is
C                  RAT_FRQ_3D_TO_2D = NFC_2D_COL/HKL_MAX
C
Cji     apply the map mag factor to get the real original H,K,L 
            real_ia = ia * mapmagfactor
C
            if (case.eq.1) then
              real_ih = real_ia
              real_ik = real_iq
              real_il = real_ip
              ih = ia
              ik = iq
              il = ip
            elseif (case.eq.2) then
              real_ih = real_ip
              real_ik = real_ia
              real_il = real_iq
              ih = ip
              ik = ia
              il = iq
            else
              real_ih = real_iq
              real_ik = real_ip
              real_il = real_ia
              ih = iq
              ik = ip
              il = ia
            endif
C
            u_prime = rot_mat(1,1)*real_ih + rot_mat(1,2)*real_ik +
     +                rot_mat(1,3)*real_il
            v_prime = rot_mat(2,1)*real_ih + rot_mat(2,2)*real_ik +
     +                rot_mat(2,3)*real_il
            w_tmp   = rot_mat(3,1)*real_ih + rot_mat(3,2)*real_ik +
     +                rot_mat(3,3)*real_il
Cji	      if (abs(w_tmp).gt.0.5001) then
            if (abs(w_tmp).gt.1.0) then
              print *, ' '
              print *, 'r_prime_start, r_prime_end,',
     +                 ' r_prime_dn, r_prime_up =',
     +                  r_prime_start, r_prime_end, 
     +                  r_prime_dn, r_prime_up
              print *, 'ih, ik, il, u_prime, v_prime, w_tmp =',
     +             real_ih, real_ik, real_il, u_prime, v_prime, w_tmp
              call error_stop(progname, subname,
     +               'Error: w value out of range!')
            endif

Cji u_prime, v_prime changed because the different size of particle image related to standard
C
            u_prime = u_prime / imgmagfactor
            v_prime = v_prime / imgmagfactor
C
 1040       if (uv_max.lt.abs(u_prime) .or. uv_max.lt.abs(v_prime) .or.
     +      (real_ih**2+real_ik**2+real_il**2).gt.num_annulus_sqr) then
C
C             Point (R_PRIME,V_PRIME,U_PRIME) is outside the mesh-box,
C             or outside the maximum annulum.
C 
            else
C             Accept U_PRIME, and V_PRIME
C 
              iu_nearest = nint( u_prime )
              iv_nearest = nint( v_prime )
C
C ----------------------------------------------------------------------
C Construct information for bilinear interpolation
 
C 2D case is tensor product of 1D case, which is indicated below

C    del_u positive              *         del_u negative            *
C                                |                      *            |
C               *                |                      |            |
C *             |                |    *                 |            |
C |-del_u_minus-|---del_u_plus---|    |---del_u_minus---|-del_u_plus-| 
C |----del_u----|                |    |                 |----del_u---|
C |-------------|----------------|    |-----------------|------------|
C |             |             
C iu_nearest   u_prime                              u_prime iu_nearest
C iu_minus                 iu_plus    iu_minus                 iu_plus
 
C Because iu_plus .eq. iu_minus + 1, the linear interpolant is given by
C   value(iu_minus) * del_u_plus + value(iu_plus) * del_u_minus
C ----------------------------------------------------------------------
C
              del_u         = u_prime - float( iu_nearest )
              if( del_u .ge. 0. ) then
                del_u_minus = del_u
                iu_minus    = iu_nearest
              else
                del_u_minus = 1. + del_u
                iu_minus    = iu_nearest - 1
              end if
              del_u_plus    = 1.0 - del_u_minus
              iu_plus       = 1   + iu_minus

              del_v         = v_prime - float( iv_nearest )
              if( del_v .ge. 0. ) then
                del_v_minus = del_v
                iv_minus    = iv_nearest
              else
                del_v_minus = 1. + del_v
                iv_minus    = iv_nearest - 1
              end if
              del_v_plus    = 1.0 - del_v_minus
              iv_plus       = 1   + iv_minus
 
C             The array of complex valued Fourier coefficients is
C             FC_2D( 0:IU_MAX, -IV_MAX:IV_MAX) where IU_MAX and IV_MAX
C             are even.  The case  IU_MAX = IV_MAX = 2 is illustrated
C             below where: `S' denotes a stored complex value and `C'
C             denotes a value not stored which is obtained from an `S'
C             value by using the conjugate symmetry:
C                   FC_2D( iu, iv ) = CONJG( FC_2D( -iu, -iv ) )
C                    iu = -2    C  C  C  C  C
C                         -1    C  C  C  C  C
C                          0    S  S  S  S  S
C                          1    S  S  S  S  S
C                          2    S  S  S  S  S
C                         iv = -2 -1  0  1  2
C 
C             Determine the indices and values of FC_2D to be used in
C             the interpolation.  With * denoting the point at which the
C             interpolant is evaluated, the values of FC_2D are used in
C             the interpolant as indicated; when necessary, a value is
C             obtained by use of conjugate symmetry.
C  
C              iu_minus     FC_NW               FC_NE
C              
C              u_prime  -------------- *
C                                      |
C              iu_plus      FC_SW      |        FC_SE
C                          iv_minus  v_prime    iv_plus

              if( ( iu_plus  .lt. - nfc_2d_row ) .or.
     +            ( iu_minus .gt.   nfc_2d_row ) .or.
     +            ( iv_plus  .lt. - nfc_2d_col ) .or.
     +            ( iv_minus .gt.   nfc_2d_col )      ) then
                                                              go to 1045
              endif
C
              if( iu_minus .lt. 0 ) then
C
C               We have a `C' point.  Use conjugates symmetry to obtain
C               the Fourier coefficients.
C               Store two values of the Fourier coefficients.
C
                fc_nw = conjg( fc_2d( - iu_minus, - iv_minus ) )
                fc_ne = conjg( fc_2d( - iu_minus, - iv_plus  ) )
                ct_nw = ctf2( - iu_minus, - iv_minus )
                ct_ne = ctf2( - iu_minus, - iv_plus  )
              else
C
C               The value of iu_minus is .ge. 0 and therefore so is the
C               value of iu_plus; store the values of the coefficients.
C
                fc_nw = fc_2d( iu_minus, iv_minus )
                fc_ne = fc_2d( iu_minus, iv_plus  )
                ct_nw = ctf2( iu_minus, iv_minus )
                ct_ne = ctf2( iu_minus, iv_plus  )
              endif
C
              if( iu_plus .lt. 0 ) then
C
C               Store two values of the Fourier coefficients; use
C               conjugate symmetry.
C
                fc_sw = conjg( fc_2d( - iu_plus,  - iv_minus ) )
                fc_se = conjg( fc_2d( - iu_plus,  - iv_plus  ) )
                ct_sw = ctf2( - iu_plus,  - iv_minus )
                ct_se = ctf2( - iu_plus,  - iv_plus  )
              else
C
C               The value of iu_plus is .ge. 0; the values of the
C               Fourier coefficients are stored.
C
                fc_sw = fc_2d( iu_plus,  iv_minus )
                fc_se = fc_2d( iu_plus,  iv_plus  )
                ct_sw = ctf2( iu_plus,  iv_minus )
                ct_se = ctf2( iu_plus,  iv_plus  )
              end if
C
	      if (accurate) then
		if (del_u_minus .lt. 0.00001) then
		  sincpi_del_u_minus = 1.0
		else
		  sincpi_del_u_minus = sin(pi*del_u_minus) /
     +                                 (pi*del_u_minus)
		endif
		if (del_u_plus .lt. 0.00001) then
                  sincpi_del_u_plus = 1.0
                else
                  sincpi_del_u_plus = sin(pi*del_u_plus) /
     +                                 (pi*del_u_plus)
		endif
		if (del_v_minus .lt. 0.00001) then
                  sincpi_del_v_minus = 1.0
                else
                  sincpi_del_v_minus = sin(pi*del_v_minus) /
     +                                 (pi*del_v_minus)
                endif
                if (del_v_plus .lt. 0.00001) then
                  sincpi_del_v_plus = 1.0
                else
                  sincpi_del_v_plus = sin(pi*del_v_plus) /
     +                                 (pi*del_v_plus)
                endif
C
Cji apply the scaled
		c_rght_side = (
     +            fc_nw*sincpi_del_u_minus*sincpi_del_v_minus +
     +            fc_ne*sincpi_del_u_minus*sincpi_del_v_plus +
     +            fc_sw*sincpi_del_u_plus*sincpi_del_v_minus +
     +            fc_se*sincpi_del_u_plus*sincpi_del_v_plus
Cji apply correction to the c_rght_side caused by image mag factor
     +            ) / (imgmagfactor**2)
Cji apply correction to the c_rght_side caused by map mag factor
     +            * (mapmagfactor**3)
                ct_all = 
     +            (
     +            ct_nw*sincpi_del_u_minus*sincpi_del_v_minus +
     +            ct_ne*sincpi_del_u_minus*sincpi_del_v_plus +
     +            ct_sw*sincpi_del_u_plus*sincpi_del_v_minus +
     +            ct_se*sincpi_del_u_plus*sincpi_del_v_plus
     +            ) * ( 1.0 - abs(w_tmp))
	      else
                c_rght_side =
     +            ((fc_nw*del_v_plus+fc_ne*del_v_minus)*del_u_plus
     +            +(fc_sw*del_v_plus+fc_se*del_v_minus)*del_u_minus)
Cji apply correction to the c_rght_side caused by image mag factor
     +            / (imgmagfactor**2)
Cji apply correction to the c_rght_side caused by map mag factor
     +            * (mapmagfactor**3)
                ct_all = 
     +            ((ct_nw*del_v_plus+ct_ne*del_v_minus)*del_u_plus
     +            +(ct_sw*del_v_plus+ct_se*del_v_minus)*del_u_minus)
     +            * ( 1.0 - abs(w_tmp))
	      endif
C
C             Store the current number of equations, and
C             the real and the imaginary parts of the interpolated
C             Fourier coefficient
C
              datalen = datalen + 1
              if ( datalen .GT. max_count) then
                 print *, max_count,'<', datalen
                 call error_stop (progname, subname,
     +                            'Error datalen execeed!!!')
              endif
Cji
              HKL_DFT(1, datalen) = ih
              HKL_DFT(2, datalen) = il
              HKL_DFT(3, datalen) = ik
              HKL_DFT(4, datalen) = real( c_rght_side )
              HKL_DFT(5, datalen) = aimag( c_rght_side )
              HKL_DFT(6, datalen) = ct_all
C
C           ENDIF of IF at 1040
            endif
 1045       continue
 1050       continue
 1060   continue
 1070 continue
C 
      return
      end
