
C *****************************************************************************
C *                                                                           *
C *                        *****     OOR.F         *****                      *
C *                                                                           *
C *    MODEL-BASED PROCEDURE TO DETERMINE ORIGINS, ORIENTATIONS, AND          *
C *     MAG SCALE FACTORS OF IMAGED PARTICLES.  IT CAN BE USED TO FIND OR     *
C *	 REFINE THESE PARAMETERS.                                             *
C *                                                                           *
C *    PROJECT CUBIC (NCOL=NROW=NSEC) 3D MAP (DEFAULT ASSUME 532 SYMMETRY)    *
C *     IN REGULARLY SPACED THETA,PHI ORIENTATIONS FOR A USER-DEFINED         *
C *     REGION OF 3D SPACE (DEFAULT IS ONE-HALF OF THE ASYMMETRIC UNIT).      *
C *     THESE CARTESIAN PROJECTIONS ARE THEN CONVERTED TO POLAR MAPS AND      *
C *     POLAR FFTS.                                                           *
C *                                                                           *
C *    LOGICAL UNITS:                                                         *
C *              1 = REFERENCE 3D MAP (READ IF MODE.GT.0)                     *
C *              2 = INPUT PARAMETERS DATA FILE(s)                            *
C *              3 = PRJ DATABASE FILE (DEFAULT = PFT.PRJS) (IF MODE.LT.0)    *
C *              4 = PFT DATABASE FILE (DEFAULT = PFT.PFTS) (IF MODE.LT.0)    *
C *              7 = INPUT RAW IMAGE DATA                                     *
C *              8 = STATS OUTPUT FILE (EMPFT.RES1)                           *
C *              9 = STATS OUTPUT FILE (EMPFT.RADS)                           *
C *             10 = STATS OUTPUT FILE (EMPFT.RES2)                           *
C *             11 = PARAMETER OUTPUT DATA FILES (EMPFT.NEW_DAT)              *
C *                                                                           *
C *    AUGUST 1993 PURDUE UNIVERSITY VAX 4000/90, T. BAKER.                   *
C *    See EMPFT.DOC for record of old modifications.                         *
C *                                                                           *
C *    MODIFIED 11-MAR-1998 BY TSB/RWA (Clean up PIF_* calls).                *
C *    MODIFIED  6-MAY-1998 BY TSB (Sliding window REFINE code).              *
C *    MODIFIED 26-MAY-1998 BY RWA/TSB (Bug in FILL_PARAMS).                  *
C *    MODIFIED  1-JUN-1998 BY TSB/SBW (Add CTF stuff).                       *
C *    MODIFIED 23-JUL-1999 BY TSB (Update SBW's CTF stuff).                  *
C *    MODIFIED 22-SEP-1999 BY TSB/RWA (Fix FFT_MAP_FILL calls).              *
C *    MODIFIED 30-NOV-1999 BY TSB (Get rid of FFT_IEND).                     *
C *    MODIFIED  2-MAY-2000 BY TSB (Implement new CTF_MULTIPLY routine).      *
C *    MODIFIED 10-NOV-2000 BY TSB (Add IMG low-pass filter).                 *
C *    MODIFIED Jan-2001 BY XZ/TSB (Debug multiple_def in GETPRJSPFTS.        *
C *    LAST MODIFIED Dec-2002 BY MBS (Read I1, I2 or I4 images)               *
C *****************************************************************************
C *****************************************************************************
C *  1. Refine OMEGA, XCEN, YCEN USING REAL PART OF PFTS WITHIN A 7*7 WIN.    *
C *     MODIFIED 23-JUL-2001 BY XING.                                         *
C *  2. ADD EVALUATION OF THE PFTS SF USING SPLINE FITTED CURVE.              *
C *     MODIFIED 23-JUL-2001 BY XING.                                         *
C *  3. Delete the noisy data of the raw images close to the CTF nodes.       *
C *     MODIFIED  23-JULY-2001 BY XING                                        *
C *  4. Add Option of Stepsize and Num for refining Center and Omega.         *
C *     MODIFIED  23-JULY-2001 BY XING                                        *
C *  5. Use Imaginary part of PFTS to determine Omega or Omega+180.0.         *
C *     MODIFIED  26-JULY-2001 BY XING                                        *
C *  6. Use The Addition of the Imaginary and Real parts of PFTS              *
C *     to determine Omega or Omega+180.0.                                    *
C *     MODIFIED  31-JULY-2001 BY XING                                        *
C *  7. Use CCOR of the Imaginary parts of PFTS to decide the Sign of Theta,  *
C *     and to determine Omega or Omega+180.0.                                *
C *     MODIFIED  01-Aug-2001 BY XING                                         *
C *****************************************************************************
C *****************************************************************************
C *  1. Use Refine Model.                                                     *
C *     MODIFIED 3-AUG-2001 BY XING.                                          *
C *  2. Delete All ABS(PHI) to make it compatible with minus Phi.             *
C *     MODIFIED 3-AUG-2001 BY XING.                                          *
C *  3. Add Code to handle multi-files.                                       *
C *     MODIFIED 3-AUG-2001 BY XING.                                          *
C *  4. Refine using Real and Imaginary part of PFTS.                         *
C *     MODIFIED 4-AUG-2001 BY XING.                                          *
C *  5. Fix a bug in the Global model (in GET_TPO_REAL_IMG).                  *
C *     MODIFIED 4-AUG-2001 BY XING.                                          *
C *  6. Handle the OMEGA+180.0 AND FLIP.                                      *
C *     MODIFIED 5-AUG-2001 BY XING.                                          *
C *****************************************************************************

C
	IMPLICIT NONE
C
	include 'PIFHEAD.INC'

	RECORD /PIFHEAD_T/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD

	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1 FOCUS_MIN, ANG_MAJ, CS_COEFF
C
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, NFILES, SORT_KEY, CTFMODE
	INTEGER*4 SYM, NCOL, NROW, NSEC, NROT, NANNULI, IDIM1, IDIM2
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, NPIX, JCUT, ANN_LO, ANN_HI
	INTEGER*4 NIDS_TOT, NIDS(999), NCOLBIN, NROWBIN, NSECBIN
	INTEGER*4 MAG_NUM, MAG_NORM, UNITS(999), FILE,Radius2
	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999), ANG_MAJ(999)
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    RADIUS, IMG_AVG, SIGCUT, PI, RADIAN, DEGREE
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT
C
C	ARRAYS AND POINTERS
C
	INTEGER*4 FIDS(1), IDS(1)
	REAL*4    TS(1), PS(1), OS(1), XS(1), YS(1), MAGS(1)
	REAL*4    PFT_CCS(1), PRJ_CCS(1), CMP_CCS(1)
	POINTER   (TS_P,TS), (PS_P,PS), (OS_P,OS), (MAGS_P,MAGS)
	POINTER   (FIDS_P,FIDS), (IDS_P,IDS), (XS_P,XS), (YS_P,YS)
	POINTER   (PFT_CCS_P,PFT_CCS), (PRJ_CCS_P,PRJ_CCS)
	POINTER   (CMP_CCS_P,CMP_CCS)
c$$$	REAL*4    IC,SC,IO,SO, BFACTOR
	REAL*4    IC, IO, BFACTOR
	INTEGER*4 SC, SO
C
	INTEGER*4 FIDS2(1), IDS2(1)
	REAL*4    TS2(1), PS2(1), OS2(1), XS2(1), YS2(1), MAGS2(1)
	REAL*4    PFT_CCS2(1), PRJ_CCS2(1), CMP_CCS2(1)
	POINTER   (TS2_P,TS2), (PS2_P,PS2), (OS2_P,OS2), (MAGS2_P,MAGS2)
	POINTER   (FIDS2_P,FIDS2), (IDS2_P,IDS2), (XS2_P,XS2), (YS2_P,YS2)
	POINTER   (PFT_CCS2_P,PFT_CCS2), (PRJ_CCS2_P,PRJ_CCS2)
	POINTER   (CMP_CCS2_P,CMP_CCS2)

	integer*4 i
C
C  ----------------------------------------------------------------------------
C
	write(*,*)'Orientation Refinement in Fourier space'
	write(*,*)'Based on original PFT [TSB]'
	write(*,*)'11-Apr-2003 [Xing-Misha]'
	call banner('OOR\0')
c	PRINT *, 'PFT: Model-based Polar Fourier Transform analysis'
c	PRINT *, '     10-Nov-2000 [TSB]'
	PRINT *, '   '
C	PRINT *, '     Delete data close to CTF nodes'
C	PRINT *, '     Spline fitting PFTS'
C	PRINT *, ' 3. Use CCOR of the Imaginary parts of PFTS to decide '
C	PRINT *, '    the Sign of Theta and to determine Omega or Omega+180.0 '
	PRINT *, ' 1. use real and imaginary parts of 2D-FFT',
	1    ' to refine orientations '
	PRINT *, ' 2. correct CTF of raw images before refine  '
	PRINT *, ' 3. use real and imaginary parts of 2D-FFT to refine center '
	PRINT *, '    right after refine THETA, PHI'
	PRINT *, ' 4. add CTF threshold 0.2 to select SF for comparing'
	PRINT *, ' 5. refine center using SF correlation instead of get_XY'
C	PRINT *, ' 7. Use IMG_LOP to delete noise around the virus with a soft filter'
C	PRINT *, ' 8. Use two filters to treat SFs before and after the first CTF peak, '		
C	PRINT *, '    treat both the raw images and the projections'
	PRINT *, ''
c	PRINT *, ' derived from "trypft_SF2DFFT66.for"'
	PRINT *, ' Compiled from "pft_532_5.f"'
	PRINT *, ' April-11-2003'
	PRINT *, ''
C
C	CALL ctime()
C
	CALL PIRADDEG (PI, RADIAN, DEGREE)
C  ************************
C  *    GET INPUT USER    *
C  ************************
	call key_info (ncol, nrow, nsec, ncolbin, nrowbin, nsecbin, nrot, 
	1 nannuli, npix, nfiles, nids, nids_tot, ctfmode, ic,sc,io,so, 
	2 radius2, bfactor, idim1, idim2,maphead,imghead,prjhead,pfthead)
C  *********************************************************
C  *    Allocate virtual memory for particle parameters    *
C  *********************************************************
	CALL GETMEM0 (NIDS_TOT, FIDS_P, IDS_P, TS_P, PS_P, OS_P, XS_P, 
	1 YS_P, MAGS_P, PFT_CCS_P, PRJ_CCS_P, CMP_CCS_P, FIDS2_P, 
	2 IDS2_P, TS2_P, PS2_P, OS2_P, XS2_P, YS2_P, MAGS2_P, 
	3 PFT_CCS2_P, PRJ_CCS2_P, CMP_CCS2_P)
C
	CALL FILL_PARAMS (NFILES, NIDS_TOT, NIDS, FIDS, 
	1 IDS, TS, PS, OS, XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS)

C  ****************************************************************************
C  *    GLOBAL OR REFINE MODE    *
C  *******************************
	IF(MODE.NE.3) THEN
	   stop "Global search is not allowed"
	ELSE
	   SORT_KEY = 3		! Sort on THETA values
	   CALL PARAM_SRT (SORT_KEY, NIDS_TOT, FIDS, IDS, TS, PS, OS, 
	1	XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS, 1)
	   CALL REFINE3_F4 (NCOL, NROW, NSEC, NCOLBIN, NROWBIN, NSECBIN,
	1	NROT, NANNULI, NPIX, NFILES, NIDS_TOT, FIDS, IDS, TS, 
	2	PS, OS, XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS, CTFMODE, PI,
	3	RADIAN, IC, SC, IO, SO, Radius2, BFACTOR, IDIM1, IDIM2, 
	4	MAPHEAD, IMGHEAD)
	   SORT_KEY = 12	! UNSORT BY FILE ID 1st, PARTICLE ID 2nd
	   CALL PARAM_SRT (SORT_KEY, NIDS_TOT, FIDS, IDS, TS, PS, OS, 
	1	XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS, 1)
	ENDIF
C
	CALL WRITE_PARAMS (NFILES, NIDS_TOT, FIDS, IDS, TS,PS,OS, 
	1    XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS)
C
	CLOSE (UNIT=2,  DISP='KEEP')		! LAST INPUT PARAMETER FILE
	CLOSE (UNIT=11, DISP='KEEP')		! LAST OUTPUT PARAMETER FILE
	CALL PIF_CLOSE (7)			! LAST INPUT IMAGE FILE
	IF(MODE.EQ.-1 .OR. MODE.EQ.-2) THEN
	   CLOSE (UNIT=3, DISP='KEEP')
	   CLOSE (UNIT=4, DISP='KEEP')
	ENDIF
CVMS	IF(TIMER.EQ.1) CALL STOP_WATCH_STOP
C  ****************************************************************************
C  *  Calculate average CCs and MAG factor if NIDS_TOT > 0    *
C  ************************************************************
	CALL CALC_AVGS (NIDS_TOT, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS)
	IF(VERBOSE.GE.1) THEN
	   PRINT *,'=========================================='
	   PRINT *,' Files EMPFT.RADS, EMPFT.RES1, EMPFT.RES2 '
	   PRINT *,'  written to default directory.'
	   PRINT *,'==========================================='
	   CLOSE (UNIT= 8, DISP='KEEP')
	   CLOSE (UNIT= 9, DISP='KEEP')
	   CLOSE (UNIT=10, DISP='KEEP')
	ENDIF
C
	CALL FREEMEM0 (FIDS_P, IDS_P, TS_P, PS_P, OS_P, XS_P, YS_P, 
	1    MAGS_P, PFT_CCS_P, PRJ_CCS_P, CMP_CCS_P, FIDS2_P, IDS2_P, 
	2    TS2_P, PS2_P, OS2_P, XS2_P, YS2_P, MAGS2_P, PFT_CCS2_P, 
	3    PRJ_CCS2_P, CMP_CCS2_P)
C
c	call ctime()
	CALL TSBEND
C
	END


C*****************************************************************************
C*                                                                           *
C*                 *****     SUBROUTINE LOADSTATS_R     *****                *
C*                                                                           *
C*    READ IN POLAR FFTS FOR THE SEARCH WINDOW TO SET UP PFT_AVGS AND SQ2    *
C*     ARRAYS.                                                               *
C*                                                                           *
C*    DECEMBER 1987  PURDUE UNIVERSITY VAX 11/750, T. BAKER                  *
C*    MAY 1992  PURDUE UNIVERSITY VAX 8550, J. TESMER.                       *
C*                                                                           *
C*    Moved THETA,PHI calculations to CALC_MOD_TPS, 30 Jan 1997, D. Belnap   *
C*    MODIFIED 23-DEC-1997 BY T. BAKER (Using memory for PFTS).              *
C*    MODIFIED 14-JAN-1998 BY TSB/RWA (*.PIF input of PFTs/PRJs).            *
C*    MODIFIED  6-MAR-1998 BY TSB (Make UNIX compatible).                    *
C*    MODIFIED 29-APR-1998 BY TSB (Adapted from PFTCC_FILL).                 *
C*    MODIFIED  4-MAY-1998 BY TSB (Implement sliding window).                *
C*    LAST MODIFIED 15-AUG-2003 BY MBS (Trap for div/zero (DPHIS = 0)).      *
C*****************************************************************************
C
	SUBROUTINE LOADSTATS_R (N_INCORE, PFTS, PFT_AVGS, NROT, NANNULI, 
	1 SQ2, VAR, RMS, AVER, TMP, MASK, THETA, PHI, MOD_TS, DPHIS, 
	3 NVIEW_THISIMG)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NROT, NANNULI, NVIEW_THISIMG, I, J, JN
	INTEGER*4 NROT_BY_4
	INTEGER*4 IR, R, NUM, VAR((NROT/4),NANNULI)
	INTEGER*4 N_INCORE, MASK(NPHI,NTHE,2), MODEL_ID
	INTEGER*4 ITH, IPH, ITHE, IPHI, ITHE1, IPHI1, ITHE2, IPHI2
	REAL*4    RMS((NROT/4),NANNULI), AVER((NROT/4),NANNULI)
	REAL*4    TMP((NROT/4),NANNULI), PFTS(1,1,1)
	REAL*4    PI, RADIAN, DEGREE, TMP2, FACTOR, TWOPIR
	REAL*4    PFT_AVGS(N_INCORE), SQ2(N_INCORE), PTS, AVE, AVE2
	REAL*4    PIXRES, THETA, PHI
	REAL*4    MOD_TS(NTHE), DPHIS(NTHE)
C  ----------------------------------------------------------------------------
C
	CALL PIRADDEG (PI, RADIAN, DEGREE)
	NROT_BY_4  = NROT/4
C  *****************************************************
C  *    COUNT UP THE VIEWS AVAILABLE FOR THIS IMAGE    *
C  *****************************************************
	NVIEW_THISIMG = 0
c	ITH           = NINT ( (THETA-MOD_TS(1))/DTHE ) + 1 !MBS 08/27/03
	ITH           = NINT ( (THETA-MOD_TS(1))/DTHE ) + 2 !Original
C	IPH           = NINT (ABS(PHI)/DPHIS(ITH)) + 1
	if(DPHIS(ITH) .ne. 0) then
	   IPH = NINT(PHI/DPHIS(ITH)) !MBS 08/15/03 trap for div by zero
	else
	   IPH = 0
	end if
	IPH = IPH + (NPHI+1)/2
	ITHE1         = ITH - 3
	ITHE2         = ITH + 3
	IPHI1         = IPH - 3
	IPHI2         = IPH + 3
	IF(ITHE1.LT.1)    ITHE1 =    1
	IF(ITHE2.GT.NTHE) ITHE2 = NTHE
	IF(IPHI1.LT.1)    IPHI1 =    1
	IF(IPHI2.GT.NPHI) IPHI2 = NPHI
C
	DO ITHE = ITHE1,ITHE2
	  DO IPHI = IPHI1,IPHI2
	    IF(MASK(IPHI,ITHE,1).GE.1) THEN	      ! IS THIS TEST NECESSARY?
	      NVIEW_THISIMG = NVIEW_THISIMG + 1
	    ENDIF
	  ENDDO
	ENDDO

CTSB	USE NVIEW_THISIMG = (ITHE2-ITHE1+1)*(IPHI2-IPHI1+1)	???

	FACTOR = 1.0/FLOAT(NVIEW_THISIMG)		! NORMALIZATION FACTOR
C
	DO ITHE = ITHE1,ITHE2
	  DO IPHI = IPHI1,IPHI2
	    IF(MASK(IPHI,ITHE,1).GE.1) THEN	     ! IS THIS TEST NECESSARY?
	      MODEL_ID = MASK(IPHI,ITHE,2)
	      PTS      = 0.0
	      DO IR = ANN_LO,ANN_HI
	        R      = MIN(NANNULI, IR+1)
	        TWOPIR = 2.0*PI*FLOAT(IR)
	        DO JN = JCUT,(NROT_BY_4 - 1)
	          IF(JN.GT.0) THEN				 ! TRAP FOR Jo
	             PIXRES = TWOPIR/FLOAT(JN)
	           ELSE 
	             PIXRES = 1000.0
	          ENDIF
	          IF(PIXRES.LT.PIXRES_MAX .OR. PIXRES.GT.PIXRES_MIN) 
	1           GO TO 200
	          J    = JN + 1
	          PTS  = PTS + 1.0
		  TMP2 = 0.0
                  IF(VAROPT) THEN
	            AVER(J,R) = AVER(J,R) + TMP2
                    VAR(J,R)  = 0
                  ENDIF
  200	        ENDDO
	      ENDDO
	    ENDIF
	  ENDDO					! DO IPHI = IPHI1,IPHI2
	ENDDO					! DO ITHE = ITHE1,ITHE2
C
        IF(.NOT.VAROPT) GO TO 500
C  ******************************
C  *    HERE IF SIGCUT > 0.0    *

C	***NOTE***  THIS NEEDS FIXING BECAUSE LOOP IS NOT CORRECT; ALSO ID NOT
C	RIGHT

C  ******************************
        AVE2 = 0.0
	DO IR = ANN_LO,ANN_HI
	  R      = MIN(NANNULI, IR+1)
	  TWOPIR = 2.0*PI*FLOAT(IR)
	  DO JN = JCUT,(NROT_BY_4 - 1)
	    IF(JN.GT.0) THEN					! TRAP FOR Jo
	       PIXRES = TWOPIR/FLOAT(JN)
	     ELSE
	       PIXRES = 1000.0
	    ENDIF
	    IF(PIXRES.LT.PIXRES_MAX .OR. PIXRES.GT.PIXRES_MIN) GO TO 300
	    J        = JN + 1
            RMS(J,R) = 0.0
	    AVE      = AVER(J,R)*FACTOR
            DO I = 1,NVIEW_THISIMG
              TMP2     = 0.0
              RMS(J,R) = RMS(J,R) + (TMP2-AVE)**2
	    ENDDO
            RMS(J,R) = SQRT(RMS(J,R)*FACTOR)
            AVE2     = AVE2 + RMS(J,R)
  300	  ENDDO
	ENDDO
        NUM  = 0
        AVE2 = AVE2/PTS
	DO IR = ANN_LO,ANN_HI
	  R      = MIN(NANNULI, IR + 1)
	  TWOPIR = 2.0*PI*FLOAT(IR)
	  DO JN = JCUT,(NROT_BY_4 - 1)
	    IF(JN.GT.0) THEN				! TRAP FOR Jo
	       PIXRES = TWOPIR/FLOAT(JN)
	     ELSE
	       PIXRES = 1000.0
	    ENDIF
	    IF(PIXRES.LT.PIXRES_MAX .OR. PIXRES.GT.PIXRES_MIN) GO TO 400
	    J = JN + 1
            IF (RMS(J,R).GT.SIGCUT*AVE2) THEN
              VAR(J,R) = 1
              NUM      = NUM + 1 
            ENDIF
  400	  ENDDO
	ENDDO
	PRINT *,'VARIANCE:'
	PRINT *,'USING POINTS > ',SIGCUT,'*',AVE2
	PRINT *,'POINTS FROM PFFT BEING USED: ',NUM
C
  500	NUM = 0
	DO ITHE = ITHE1,ITHE2
	   DO IPHI = IPHI1,IPHI2
	      IF(MASK(IPHI,ITHE,1).GE.1) THEN ! IS THIS TEST NECESSARY?
		 NUM           = NUM + 1
		 MODEL_ID      = MASK(IPHI,ITHE,2)
		 PTS           = 0.0
		 PFT_AVGS(NUM) = 0.0
		 SQ2(NUM)      = 0.0
	      ENDIF
	   ENDDO		! DO IPHI = IPHI1,IPHI2
	ENDDO			! DO ITHE = ITHE1,ITHE2
C
	RETURN
	END


C *****************************************************************************
C	EMPFT SUBROUTINES
C *****************************************************************************
C
C	REFINE
C	REFINE_CC_F4
C
C	CALC_AVGS
C       CALC_TP_MASK4
C	FILL_PARAMS
C	FREEMEM0
C	FREEMEM1
C	FREEMEM2
C	FREEMEM3
C	FREEMEM4
C	FREEMEM_REFINE_F4
C	GETMEM0
C	GETMEM1
C	GETMEM2
C	GETMEM3
C	GETMEM4_F4
C	GETMEM_REFINE
C	GETMEM_REFINE2
C	GET_BESTMAG_F4
C	GET_PRJSPFTS_R3
C	GET_THEPHI_PR3
C	GET_TPO_R3
C	INFO
C	LIST_CCS_F4
C	LOADSTATS_R
C	PFILE_INFO
C	PHIMAX
C	WRITE_PARAMS
C
C	PFTCC
C	CC_RADII
C	PFTCC_RES


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE REFINE_CC_F4     *****            *
C  *                                                                          *
C  *    CROSS-CORRELATE POLAR FFT DATA TO DETERMINE AND REFINE PARTICLE       *
C  *     ORIENTATIONS (REFINE MODE VERSION).                                  *
C  *                                                                          *
C  *	Note: PHI1 = PHIIN (the particle's PHI orientation)                   *
C  *                                                                          *
C  *    LOGICAL UNITS: 1 = FILE CONTAINING 3D MODEL                           *
C  *                   2 = INPUT PARAMETERS DATA FILE(s)                      *
C  *                   7 = INPUT RAW IMAGE DATA                               *
C  *                   8 = STATS OUTPUT FILE (EMPFT.RES1)                     *
C  *                   9 = STATS OUTPUT FILE (EMPFT.RADS)                     *
C  *                  10 = STATS OUTPUT FILE (EMPFT.RES2)                     *
C  *                                                                          *
C  *    Protocol:                                                             *
C  *     1. Compute polar map from raw image.                                 *
C  *     2. Compute polar FFT (PFT) of raw image.                             *
C  *     3. Cross-correlate raw PFT with data base of reference PFTs.         *
C  *     4. Cross-correlate raw image with "best" projection to resolve       *
C  *        PHI ambiquity and determine OMEGA.                                *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *                                                                          *
C  *    MODIFIED 19-JAN-1998 BY T. BAKER (Adapted from old PFTCC subroutine). *
C  *    MODIFIED  5-FEB-1998 BY T. BAKER (Clean up).                          *
C  *    MODIFIED  4-MAY-1998 BY T. BAKER (Implement new code).                *
C  *    MODIFIED 23-JUL-1999 BY TSB (Implement SBW's CTF code).               *
C  *    MODIFIED 25-AUG-1999 BY TSB (Implement TSB's CTF code).               *
C  *    MODIFIED 20-DEC-2002 BY MBS (IMG -> IMG_F4).                          *
C  *    MODIFIED 15-AUG-2003 BY MBS (Trap for div/zero (DPHIS = 0)).          *
C  *    MODIFIED 31-OCT-2003 BY MBS (added ctfmode to par listGET_TPO_R3)     *
C  *    LAST MODIFIED 7-DEC-2003 by MBS (Added corrct_B_factor for CTFMODE = 0*
C  ****************************************************************************
C
	SUBROUTINE REFINE_CC_F4 (IMG_NUM, NVIEW_THISIMG, NROT, 
	1  NANNULI, THE1, PHI1, IDIM1, IDIM2, NFILES, PRJS, 
	2  PFT_AVGS, MOD_TS, MOD_PS, IMG_BKG, NCOLBIN, NROWBIN, 
	3  NIDS_TOT, IMG_F4, IDS, TS, PS, OS, XS, YS, MAGS, PFT_CCS, 
	4  PRJ_CCS, CMP_CCS, N_INCORE, MASK, DPHIS, FFT1, FFT3, FFT4, 
	5  FFT5, MSKIMG, MSKPRJ, R4_IMG, BESTPRJ, PRJ_ROT, PFTIMG, PFFT, 
	6  R4_PIMG, FLTPMAP, POLAR_PRJ, TMP1, TMP2, TMP3, CCOR, CTFMODE, 
	7  FILE, IC, SC, IO, SO, IMG_CCOR, MAP3D, PRJS_SF, IMG_SF, 
	8  FFTA, FFTB, halfSize, Radius2, PRJS_SFB, BFACTOR)
C
	IMPLICIT NONE
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1 FOCUS_MIN, ANG_MAJ, CS_COEFF
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 LUN, I, J 
	INTEGER*4 CTFMODE,halfSize
	INTEGER*4 NCOLBIN, NROWBIN, FILE, INDEX, UNITS(999)
	INTEGER*4 NROT, NANNULI, IDIM1, IDIM2, NFILES
	INTEGER*4 IMG_NUM, NIMG, NIDS_TOT, NVIEW_THISIMG
	INTEGER*4 MASK(NPHI,NTHE,2)
	INTEGER*4 N1, N2_NUM, THISNSEC, FFT_IEND
	INTEGER*4 IDS(NIDS_TOT), N_INCORE, ITH, IPH
	INTEGER*2 MSKIMG(NCOLBIN,NROWBIN)                       !MBS 12.12.02
	REAL*4    IMG_F4(NCOLBIN,NROWBIN)                       !MBS 12.12.02
	real*4    IMG_BKG                                       !MBS 08.19.03
	REAL*4    COL_BY_2, ROW_BY_2, COL_BY_10, ROW_BY_10, DELX, DELY
	REAL*4    DPHIS(NTHE), WAVEL, TEMPFAC
	REAL*4    FFT1(2*IDIM1,IDIM2), FFT3(2*IDIM1,IDIM2)
	REAL*4    FFT4(2*IDIM1,IDIM2), FFT5(2*IDIM1,IDIM2) 
	REAL*4    R4_IMG(NCOLBIN,NROWBIN), R4_PIMG(NROT,NANNULI)
	REAL*4    MSKPRJ(NCOLBIN,NROWBIN), POLAR_PRJ(NROT,NANNULI)
	REAL*4    BESTPRJ(NCOLBIN,NROWBIN), PRJ_ROT(NCOLBIN,NROWBIN)
	REAL*4    PFFT(2*NROT,NANNULI), FLTPMAP(NROT,NANNULI)
	REAL*4    TMP1(NCOLBIN,NROWBIN), TMP2(2*IDIM1,IDIM2)
	REAL*4    TMP3(2*IDIM1,IDIM2), CCOR(N_INCORE)
	REAL*4    PFTIMG(IDIM1*2,IDIM2)
	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999), ANG_MAJ(999)
C
	REAL*4    TS(NIDS_TOT), PS(NIDS_TOT), OS(NIDS_TOT)
	REAL*4    XS(NIDS_TOT), YS(NIDS_TOT), MAGS(NIDS_TOT)
	REAL*4    PFT_CCS(NIDS_TOT), PRJ_CCS(NIDS_TOT)
	REAL*4    CMP_CCS(NIDS_TOT)
	REAL*4    PI, RADIAN, DEGREE
	REAL*4    DPHI, THE1, PHI1, DEL_PIXRES
	REAL*4    THE_BEST, PHI_BEST, OME_BEST
	REAL*4    CC_PFT, CC_PRJ, CC_CMP, CC_AVG
	REAL*4    XCEN, YCEN, XCEN_MOD, YCEN_MOD
	REAL*4    CC_RES_GLOBAL(2,20), CC_RAD_GLOBAL(2,200)
	REAL*4    BEST_MAG, MAG, MAG_LO
	REAL*4    MOD_TS(NTHE)
	REAL*4    MOD_PS(NPHI,NTHE), PFT_AVGS(N_INCORE)
	REAL*4    PRJS(1)
	REAL*4    IMG_SCALEF, FFT_FZERO, FFT_ORIGX, FFT_ORIGY
	LOGICAL   FLIP, FFT_FULL, MALLOC, SIGN
	DATA      CC_RES_GLOBAL/40*0.0/, CC_RAD_GLOBAL/400*0.0/
	REAL*4    MAP3D(NCOLBIN,NROWBIN, NROWBIN)
c$$$	REAL*4    IC, IO, SC, SO, IMG_CCOR, Radius2, BFACTOR
	REAL*4    IC, IO, IMG_CCOR, Radius2, BFACTOR
	INTEGER*4 SC, SO
C
	REAL*4    PRJS_SF(2*halfSize+1,2*halfSize+1, N_INCORE)
	REAL*4    PRJS_SFB(2*halfSize+1,2*halfSize+1, N_INCORE)
	REAL*4    IMG_SF(2*halfSize+1,2*halfSize+1)
	REAL*4    FFTA(IDIM1,IDIM2), FFTB(IDIM1,IDIM2)
	REAL*4    PROJN(NCOLBIN, NROWBIN), FLT_FAC

	real*4    rad_pol, delang, annlo, ANNHI
	real*4    resmin,resmax

C  ----------------------------------------------------------------------------
C
	CALL PIRADDEG (PI, RADIAN, DEGREE)
C  *******************************
C  *  Set up initial parameters  *
C  *******************************
	DEL_PIXRES = 0.025
	DEL_ANG    = 360.0/FLOAT(NROT)
	COL_BY_2   = FLOAT(NCOLBIN/2)
	ROW_BY_2   = FLOAT(NROWBIN/2)
	COL_BY_10  = FLOAT(NCOLBIN/10)
	ROW_BY_10  = FLOAT(NROWBIN/10)
	XCEN_MOD   = (NCOLBIN+1)/2  ! X center for PRJs (real space coordinate)
	YCEN_MOD   = (NROWBIN+1)/2  ! Y center for PRJs (real space coordinate)
	CC_AVG     = 0.0
	FFT_ORIGX = NCOLBIN/2
	FFT_ORIGY = NROWBIN/2
C  *****************************************************************
C  *   Convert raw IMG from INT*2 to REAL*4 format (IMG->R4_IMG)   *
C  *   Use MAGS(IMG_NUM) for initial MAG scale factor.             *
C  *   Use XS(IMG_NUM) and YS(IMG_NUM) for initial X,Y origins     *
C  *   If MAG does not equal 1.0, scale image by MAG	           *
C  *   Compute FFT of image					   *
C  *****************************************************************
C
c$$$c       Debug
c$$$	write(*,*)"A: img_num, TS, mod_ts:",img_num, TS(IMG_NUM), mod_ts(1)
C ****** Reset MAG to 1.0 ******
	MAGS(IMG_NUM) = 1.0
C ****** Reset MAG to 1.0 ******
C
	XCEN      = ((XS(IMG_NUM)+1.0)/BIN)-1.0      ! NEED IN BINNED FFT SPACE
	YCEN      = ((YS(IMG_NUM)+1.0)/BIN)-1.0      ! NEED IN BINNED FFT SPACE
	MAG       = MAGS(IMG_NUM)
	IF(MAG.NE.1.0) then
	   CALL INTERP2D (IMG_F4, NCOLBIN, NROWBIN, XCEN, YCEN, XCEN,
	1	YCEN, 0.0, MAG, 0.0)
	end if
c  	CALL IMG_R4_FFT (R4_IMG, NCOLBIN, NROWBIN, FFT1, IDIM1, IDIM2,   !MBS 3/01/03
c	1       0.0, FFT_ORIGX, FFT_ORIGY, FFT_FZERO, FFT_FULL, 0)
C
C  *****************************************************************
C  *    Low-pass Fourier filter the image to help TPO calculation  *
C  *    This replace the IMG_MAP call.                             *
C  *****************************************************************
C       MBS 12.12.02 img_fft -> img_r4_fft
C       MBS 3.01.03 IMG_F4 already mag-adjusted
	  CALL IMG_R4_FFT (IMG_F4, NCOLBIN, NROWBIN, FFT1, IDIM1, IDIM2, 
	1  IMG_BKG, FFT_ORIGX, FFT_ORIGY, FFT_FZERO, FFT_FULL, 0)
	  resmin = RES_MIN
	  resmax = RES_MAX
	  CALL FFT_FLT (FFT1, IDIM1, IDIM2, RESMIN, RESMAX,
	1 3.0, 3.0, FFT_FULL)
C       **************************************************************
C-------*  Apply CTF to FFT to downweight data according to Steve    *
C       *   Fuller scheme.  Implemented by SBW.                      *
C       **************************************************************
        IF((CTFMODE.NE.0) .or. (CTFMODE.NE.1)) THEN
	   IF(CTFMODE.EQ.2) THEN
	      TEMPFAC    = 0.0	! HARDWIRE THIS FOR NOW
C         ****************************************************************
C---------*  2-May-00 Hardwire input parameters to CTF_MULTIPLY such as  *
C         *  mode=1; Cc=0.0, BETA=0.0, dE=0.0; These will *not* impose   *
C         *  a decay factor on the data.                                 *
C         *  Put CTFMODE back MBS 1.6.03                                 *
C         ****************************************************************
	      CALL CTF_MULTIPLY (CTFMODE, FFT1, IDIM1, IDIM2, UNITS(FILE), 
	1	   BIN*PIXSIZ(FILE),PI,RADIAN,WAVEL(VOLTS(FILE)), 
	2	   FOCUS_MAJ(FILE), FOCUS_MIN(FILE), ANG_MAJ(FILE), 
	3	   AMP_FAC(FILE), CS_COEFF(FILE), 0.0, 0.0, 0.0,
	4	   VOLTS(FILE),TEMPFAC,FFT_FULL)      
	      WRITE(*,'(A,I4,A,2F8.4,A,F8.4)') 'Applied CTF to image ',
	1	   IDS(FILE), ' with defocus values', FOCUS_MAJ(FILE),
	2	   FOCUS_MIN(FILE), ' and ANG_MAJ =', ANG_MAJ(FILE)
	   else if (ctfmode .eq. 3) then
	      FLT_FAC = 0.1	!Hardwired!!!!
	      CALL CTF_DIVIDE_TRY3 (CTFMODE, FFT1,IDIM1,IDIM2,UNITS(FILE),
	1      PIXSIZ(FILE), pi, RADIAN, WAVEL(VOLTS(FILE)), FOCUS_MAJ(FILE),
	2	   FOCUS_MIN(FILE), ANG_MAJ(FILE), AMP_FAC(FILE), 
	3	   CS_COEFF(FILE), FLT_FAC, BFACTOR, FFT_FULL, 0.0, 0.0, 0.0, 
	4	   VOLTS(FILE), IDIM1)
	   end if
c       MBS Added corrct_B_factor for CTFMODE = 0 12.07.2003
	else if(CTFMODE.eq.0) then
	   CALL CORRCT_B_FACTOR (FFT1, IDIM1, IDIM2, UNITS(FILE), 
	1	PIXSIZ(FILE), BFACTOR, RESMAX, FFT_FULL)
	end if
	CALL FFT_2D_BACK (FFT1, IDIM1, IDIM2)
	CALL FFT_MAP_FILL (FFT1, IDIM1, IDIM2, R4_IMG, NCOLBIN, 
	1    NROWBIN, 1, 1, FFT_FULL)
C       MBS 12.12.02 fft_img_fill -> FFT_MAP_FILL
	CALL FFT_MAP_FILL (FFT1, IDIM1, IDIM2, IMG_F4, NCOLBIN, 
	1    NROWBIN, 1, 1, FFT_FULL)
	XCEN = XCEN + 1.0  		     ! Return to real space coordinates
	YCEN = YCEN + 1.0
C	*******************************************************************
C-------*  Sometimes EMPFT crashes with a division-by-0 error in          *
C	*  GET_THEPHI (called below). It turns out that GET_XY            *
C	*  (above) returns a bad origin due to poor matching between an   *
C	*  image and its reference, and then MAP_POLAR (below) fills the  *
C	*  polar array with zero since the origin can be too far off the  *
C	*  edge of the cartesian array. The div-by-0 occurs because the   *
C	*  sum-of-sqaures (SQ1) for the polar array is zero.  Check here  * 
C	*  for (XCEN,YCEN) being away from the center of the image array  *
C	*  (NCOLBIN/2,NROWBIN/2) and post warnings. If it is way off, the *
C	*  user can try to skip this particle and go on to the next,      *
C	*  rather than crash in GET_THEPHI.                               *
C	*******************************************************************
	DELX = ABS(XCEN-COL_BY_2)
	DELY = ABS(YCEN-ROW_BY_2)
	IF(DELX.GT.COL_BY_2 .OR. DELY.GT.ROW_BY_2)  THEN
	   WRITE (*,1000) IMG_NUM, XCEN*BIN, YCEN*BIN
	   PFT_CCS(IMG_NUM) = -9.999    !MBS Assign ridiculous values
	   PRJ_CCS(IMG_NUM) = -9.999    !MBS Assign ridiculous values
	   CMP_CCS(IMG_NUM) = -9.999    !MBS Assign ridiculous values
	   XS(IMG_NUM)      = COL_BY_2*BIN
	   YS(IMG_NUM)      = ROW_BY_2*BIN
	   return
	 ELSE IF (DELX.GT.COL_BY_10 .OR. DELY.GT.ROW_BY_10)  THEN
	    WRITE (*,1010) IMG_NUM, XCEN*BIN, YCEN*BIN
	    XCEN = FLOAT(NCOLBIN/2)
	    YCEN = FLOAT(NROWBIN/2)
	ENDIF

C  ******************************************************************
C  *    ITH,IPH ARE THE "MASK" COORDINATES OF THE CURRENT IMAGE     *
C  ******************************************************************
c	ITH = NINT ( (TS(IMG_NUM)-MOD_TS(1))/DTHE ) + 1 !MBS 08/27/03
	ITH = NINT ( (TS(IMG_NUM)-MOD_TS(1))/DTHE ) + 2 !original
	if(DPHIS(ITH) .ne. 0) then
	   IPH = NINT(PS(IMG_NUM)/DPHIS(ITH)) !MBS 08/15/03 trap for div by zero
	else
	   IPH = 0
	end if
	IPH = IPH + (NPHI+1)/2
c$$$c       Debug
c$$$	write(*,*)"In refine_cc_f4"
c$$$	write(*,'(a,i5,2f10.3,2i4)')"img_num, TS, mod_ts,ith,iph:",
c$$$	1    img_num, TS(IMG_NUM), mod_ts(1),ith,iph
c$$$c       Debug
C       MBS added ctfmode into parameter list in GET_TPO_R3 to eliminate
C       CTF calculations if ctfmode = 0
	CALL GET_TPO_R3 (NCOLBIN, NROWBIN, NROT, NANNULI, N_INCORE, 
	1    NVIEW_THISIMG, MASK, MOD_TS, MOD_PS, ITH, IPH, 
	2    R4_IMG, R4_PIMG, XCEN, YCEN, PI, THE_BEST, PHI_BEST, 
	5    OME_BEST, CC_PFT, CCOR, INDEX, OS(IMG_NUM), 
	7    IC, SC, IO, SO, IMG_CCOR, PRJS_SF, IMG_SF, IDIM1, IDIM2, 
	8    FFT_FULL, FFT_FZERO, FFT1, FFTA, FFTB, FILE, halfSize, 
	9    Radius2, PRJS_SFB, MAP3D,BFACTOR,CTFMODE)
C
	CALL MAP_PRJ (1, NCOLBIN, NROWBIN, NROWBIN, PROJN, BESTPRJ, 
	1    MAP3D, THE_BEST, PHI_BEST, 90.0, 1.0, 
	2    FLOAT(NROWBIN), 90., 90., 90.)
C
	CALL MAP_FFT_FILL (BESTPRJ, NCOLBIN, NROWBIN, 1, 1, FFT1, 
     1      IDIM1, IDIM2, FFT_FZERO, FFT_ORIGX, FFT_ORIGY, FFT_IEND, 
     2      FFT_FULL, 0)
	CALL FFT_2D (FFT1, IDIM1, IDIM2)
C
	CALL FFT_FLT (FFT1, IDIM1, IDIM2, RESMIN, RESMAX,
	1 3.0, 3.0, FFT_FULL)!MBS?????
C       
	  CALL CORRCT_B_FACTOR (FFT1, IDIM1, IDIM2, UNITS(FILE), 
	1  PIXSIZ(FILE), BFACTOR, RESMAX, FFT_FULL)
C
	CALL FFT_2D_BACK (FFT1, IDIM1, IDIM2)
	CALL FFT_MAP_FILL (FFT1, IDIM1, IDIM2, BESTPRJ, NCOLBIN,
	1      NROWBIN, 1, 1, fft_full)
	rad_pol = RADIUS
	delang  = DEL_ANG
	annlo   = ANN_LO
	ANNHI   = ANN_HI
	CALL MAP_POLAR (BESTPRJ, NCOLBIN, NROWBIN, POLAR_PRJ, NROT, 
	1 NANNULI, XCEN_MOD, YCEN_MOD, 1.0, RAD_pol, DELANG)
C
	CALL GET_CC_PRJ (R4_PIMG, POLAR_PRJ, NROT, NANNULI, 
	1    OME_BEST, FLIP, CC_PRJ)
C
	IF(FLIP) THEN
	   THE_BEST = 180.0 - THE_BEST
	   IF(THE_BEST.EQ.0.0 .OR. THE_BEST.EQ.180.0) THEN
	      PHI_BEST = 0.0
	   ELSE
	      PHI_BEST = PHI_BEST + 180.0
	   ENDIF
	   IF(THE_BEST.EQ.0.0 .and. THE_BEST.NE.180.0)
	1	OME_BEST = OME_BEST + 180.0
	   CALL MAP_XFLIP (BESTPRJ, NCOLBIN, NROWBIN, 0)
	ENDIF
C  *******************************************************************
C  *  Again test 1) for centers off the Cartesian array              *
C  *             2) for centers more than 10% from array center      *
C  *  Such origins may indicate poor data                            *
C  *******************************************************************
	DELX = ABS(XCEN-COL_BY_2)
	DELY = ABS(YCEN-ROW_BY_2)
	IF(DELX.GT.COL_BY_2 .OR. DELY.GT.ROW_BY_2)  THEN
	   WRITE(*,1000) IMG_NUM, XCEN*BIN, YCEN*BIN
	   PFT_CCS(IMG_NUM) = -9.999    !MBS Assign ridiculous values
	   PRJ_CCS(IMG_NUM) = -9.999    !MBS Assign ridiculous values
	   CMP_CCS(IMG_NUM) = -9.999    !MBS Assign ridiculous values
	   XS(IMG_NUM)      = COL_BY_2*BIN
	   YS(IMG_NUM)      = ROW_BY_2*BIN
	   return
	 ELSE IF (DELX.GT.COL_BY_10 .OR. DELY.GT.ROW_BY_10)  THEN
	   WRITE(*,1010) IMG_NUM, XCEN*BIN, YCEN*BIN
	   XCEN = COL_BY_2
	   YCEN = ROW_BY_2
	   CC_PFT = 0.0
	   CC_PRJ = 0.0
	   IMG_CCOR = 0.0
	ENDIF
C  **************************************************************************
C  *  Find/refine MAG scale factor and CMP coefficient, use a routine       *
C  *  adapted from CMP program, IMG and BESTPRJ are compared in real space  *
C  **************************************************************************
	IF(MAG_CEN.LE.0.0) THEN
	   MAG_LO = MAG     - (((MAG_NUM-1)/2)*MAG_STEP)
	  ELSE
	   MAG_LO = MAG_CEN - (((MAG_NUM-1)/2)*MAG_STEP)
	ENDIF
C       MBS 12.12.02 GET_BESTMAG -> GET_BESTMAG_F4
	CALL GET_BESTMAG_F4 (IMG_F4, BESTPRJ, TMP1, TMP2, TMP3, NCOLBIN, 
	1    NROWBIN, IDIM1, IDIM2, XCEN-1.0, YCEN-1.0, OME_BEST, MAG_LO, 
	3    BEST_MAG, CC_CMP, IMG_BKG, FFT_ORIGX, FFT_ORIGY)
C
C ****** Reset MAG to 1.0 ******
	BEST_MAG = 1.0                        !MBS Xing resets it to 1 because
c       he does not believe in the get_bestmag routine but wanted to get CC_CMP
C ****** Reset MAG to 1.0 ******
C
	MAG = BEST_MAG
C  *********************************************************************
C  *    IF VERBOSE > 0                                                   *
C  *    RECOMPUTE POLAR MAP OF RAW IMAGE WITH NEWEST X,Y ORIGIN        *
C  *    RECOMPUTE POLAR MAP OF ROTATED PROJECTION                      *
C  *    COMPUTE COEFFICIENTS AS A FUNCTION OF RESOLUTION AND RADIUS    *
C  *    WRITE TO EXTRA OUTPUT FILES (UNITS 8,9,10)                     *
C  *********************************************************************
	IF(VERBOSE.LE.0) GO TO 500
C       MBS 12.12.02 LIST_CCS -> LIST_CCS_F4
	CALL LIST_CCS_F4 (NCOLBIN, NROWBIN, IDIM1, IDIM2, NROT, 
	1 NANNULI, IMG_F4, MSKIMG, FFT1, FFT3, FFT4, FFT5, 
	2 PRJ_ROT, MSKPRJ, POLAR_PRJ, R4_PIMG, FLTPMAP, IMG_NUM, IMG_BKG, 
	3 XCEN, YCEN, XCEN_MOD, YCEN_MOD, FFT_ORIGX, FFT_ORIGY, 
	4 DEL_PIXRES, CC_AVG, CC_RES_GLOBAL, CC_RAD_GLOBAL)
 500	IF(THE_BEST.EQ.0.0 .OR. THE_BEST.EQ.180.0) then
	   DPHI = 0.0
	else
	   DPHI = DPHI_90 / SIND(THE_BEST)
	end if
C  *******************************************************************
C  *  If not asymmetric, check if orientation is in asymmetric unit  *
C  *  and return to asymmetric unit if orientation lies outside      *
C  *******************************************************************
	IF(SYM.nE.1)
	1    CALL TO_ASYM_UNIT (SYM, THE_BEST, PHI_BEST, OME_BEST)
c	WRITE(*,*)
  600	IF(OME_BEST.LT.0.0) OME_BEST = OME_BEST + 360.0      ! Keep positive
	XS(IMG_NUM)      = (XCEN*BIN) - 1.0		! FOURIER SPACE COORDS
	YS(IMG_NUM)      = (YCEN*BIN) - 1.0		! FOURIER SPACE COORDS
	TS(IMG_NUM)      = THE_BEST
	PS(IMG_NUM)      = PHI_BEST
	OS(IMG_NUM)      = OME_BEST
	MAGS(IMG_NUM)    = MAG
	PFT_CCS(IMG_NUM) = CC_PFT
	PRJ_CCS(IMG_NUM) = CC_PRJ
	CMP_CCS(IMG_NUM) = CC_CMP
C  ****************************************************************************
C  *  Write extra output to files if user wishes (VERBOSE > 0)  *
C  ************************************************************
ccc	nimgs not defined because only considering one image at a time here
ccc	IF (VERBOSE.GE.1) THEN
ccc	  J = NINT(1/(PIXRES_MAX*DEL_PIXRES))
ccc	  WRITE(8,1020)
ccc	  WRITE(8,'(A,10F7.3)') ' AVG', (CC_RES_GLOBAL(1,I)/FLOAT(NIMGS),
ccc	1				I=1,J)
ccc	  WRITE(10,1020)
ccc	  WRITE(10,'(A,10F7.3)') ' AVG', (CC_RES_GLOBAL(2,I)/FLOAT(NIMGS),
ccc	1				 I=1,J)
ccc	  WRITE(9,1030) NIMGS
ccc	  DO I = 1,ANN_HI+1
ccc	    WRITE(9,'(I4,2F7.3)') I-1, CC_RAD_GLOBAL(1,I)/FLOAT(NIMGS),
ccc	1			       CC_RAD_GLOBAL(2,I)/FLOAT(NIMGS)
ccc	  ENDDO
ccc	ENDIF
C
	RETURN
 1000	FORMAT (' WARNING - estimated center is off edge of array:',/,
	1       ' IMG_NUM, XCEN, YCEN = ', I, 2F10.3,/,
	2       ' Reset to the box center')
1010	FORMAT (' WARNING - estimated center is more than 10% from ',
	1       'array center:',/,' IMG_NUM, XCEN, YCEN = ', I, 2F10.3,/,
	2        'Reset to the box center')
 1020	FORMAT(74('-')) 
 1030	FORMAT (' AVERAGE correlation coefficients as function ',
	1	  'of radius for num images:',I4)
	END

C  ****************************************************************************
C  *									      *
C  *                 *****     SUBROUTINE CALC_AVGS     *****                 *
C  *									      *
C  *  Computes average and std values of MAG, CC_PFT, CC_PRJ, and CC_CMP.     *
C  *									      *
C  *    D. Belnap, July 1996						      *
C  *    MODIFIED 23-APR-1998 BY T. BAKER                                      *
C  *    LAST MODIFIED 01-DEC-2003 BY M. Sherman (Skip correlations that       *
C  *	were not calculated or set to -9.999; set up num counter)             *
C  ****************************************************************************
C
	SUBROUTINE CALC_AVGS (NIDS_TOT,MAGS,PFT_CCS,PRJ_CCS,CMP_CCS)
C
	IMPLICIT NONE
C
	INTEGER*4 ID, NIDS_TOT, num
	REAL*4    MAGS(NIDS_TOT), MAGS_AVG, MAGS_STD
	REAL*4    PFT_CCS(NIDS_TOT), PFTCC_AVG, PFTCC_STD
	REAL*4    PRJ_CCS(NIDS_TOT), PRJCC_AVG, PRJCC_STD
	REAL*4    CMP_CCS(NIDS_TOT), CMPCC_AVG, CMPCC_STD
C -----------------------------------------------------------------------------
C
	IF(NIDS_TOT.LE.0) RETURN
C
	WRITE(*,100) NIDS_TOT
  100	FORMAT (' ====================================================',
	1	'=====',
	2       /,' Compute average MAG factor and coefficients,',
	3	' N = ',I7,/,
	4       ' ====================================================',
	5	'=====')
C  ***************************
C  *  Initialze parameters   *
C  ***************************
	MAGS_AVG  = 0.0
	MAGS_STD  = 0.0
	PFTCC_AVG = 0.0
	PFTCC_STD = 0.0
	PRJCC_AVG = 0.0
	PRJCC_STD = 0.0
	CMPCC_AVG = 0.0
	CMPCC_STD = 0.0
C
C       Skip correlations that were not calculated or
C	set to -9.999; set up num counter MBS 12/1/03
	num = 0
	DO ID = 1,NIDS_TOT
	  if(PFT_CCS(ID) .ne.-9.999 .and.
	1    PRJ_CCS(ID) .ne.-9.999 .and.
	1    CMP_CCS(ID) .ne.-9.999) then
	     MAGS_AVG  = MAGS_AVG  + MAGS(ID)
	     PFTCC_AVG = PFTCC_AVG + PFT_CCS(ID)
	     PRJCC_AVG = PRJCC_AVG + PRJ_CCS(ID)
	     CMPCC_AVG = CMPCC_AVG + CMP_CCS(ID)
	     num = num + 1
	  end if
	ENDDO
c$$$	MAGS_AVG  = MAGS_AVG  / FLOAT(NIDS_TOT)
c$$$	PFTCC_AVG = PFTCC_AVG / FLOAT(NIDS_TOT)
c$$$	PRJCC_AVG = PRJCC_AVG / FLOAT(NIDS_TOT)
c$$$	CMPCC_AVG = CMPCC_AVG / FLOAT(NIDS_TOT)
	MAGS_AVG  = MAGS_AVG  / num
	PFTCC_AVG = PFTCC_AVG / num
	PRJCC_AVG = PRJCC_AVG / num
	CMPCC_AVG = CMPCC_AVG / num
C  ***************************************************************
C  *  Calculate and list coefficients and standard deviations    *
C  ***************************************************************
c	IF(NIDS_TOT.GT.1) THEN
	IF(num.GT.1) THEN
	  DO ID = 1,NIDS_TOT
	     if(PFT_CCS(ID) .ne.-9.999 .and.
	1	PRJ_CCS(ID) .ne.-9.999 .and.
	1	CMP_CCS(ID) .ne.-9.999) then
		MAGS_STD  =  MAGS_STD  + (MAGS(ID)    - MAGS_AVG)**2
		PFTCC_STD =  PFTCC_STD + (PFT_CCS(ID) - PFTCC_AVG)**2
		PRJCC_STD =  PRJCC_STD + (PRJ_CCS(ID) - PRJCC_AVG)**2
		CMPCC_STD =  CMPCC_STD + (CMP_CCS(ID) - CMPCC_AVG)**2
	     end if
	  ENDDO
c$$$	  MAGS_STD  = SQRT( MAGS_STD  / FLOAT(NIDS_TOT-1) )
c$$$	  PFTCC_STD = SQRT( PFTCC_STD / FLOAT(NIDS_TOT-1) )
c$$$	  PRJCC_STD = SQRT( PRJCC_STD / FLOAT(NIDS_TOT-1) )
c$$$	  CMPCC_STD = SQRT( CMPCC_STD / FLOAT(NIDS_TOT-1) )
	  MAGS_STD  = SQRT( MAGS_STD  / (num-1) )
	  PFTCC_STD = SQRT( PFTCC_STD / (num-1) )
	  PRJCC_STD = SQRT( PRJCC_STD / (num-1) )
	  CMPCC_STD = SQRT( CMPCC_STD / (num-1) )
	ENDIF
C
	WRITE(*,1000) '                     Average  Std. Dev.'
	WRITE(*,1000) '---------------------------------------'
	WRITE(*,1000) ' MAG factor      ', MAGS_AVG,  MAGS_STD
	WRITE(*,1000) ' PFT Coefficient ', PFTCC_AVG, PFTCC_STD
	WRITE(*,1000) ' PRJ Coefficient ', PRJCC_AVG, PRJCC_STD
	WRITE(*,1000) ' CMP Coefficient ', CMPCC_AVG, CMPCC_STD
 1000	FORMAT(A,2(F10.3))
C
	RETURN
	END



C  ****************************************************************************
C  *                                                                          *
C  *                 *****     SUBROUTINE FILL_PARAMS     *****               *
C  *                                                                          *
C  *    FILL UP THE PARAMETER ARRAYS BY READING DATA IN FROM ALL PARAMETER    *
C  *     FILES (UNIT = 2).                                                    *
C  *                                                                          *
C  *    FEBRUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.       *
C  *    LAST MODIFIED 26-MAY-1998 BY RWA/TSB (Bug in NN variable).            *
C  *    LAST MODIFIED 11-August-2003 BY MBS (switch to asymmetric triangle).  *
C  ****************************************************************************
C
	SUBROUTINE FILL_PARAMS (NFILES, NIDS_TOT, NIDS, FIDS, 
	1 IDS, TS, PS, OS, XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS)
C	
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 ID, NFILES, NIDS_TOT, FILE, NN, NIDS(999)
	INTEGER*4 FIDS(NIDS_TOT), IDS(NIDS_TOT), IDS_TOT
	REAL*4    TS(NIDS_TOT), PS(NIDS_TOT), OS(NIDS_TOT)
	REAL*4    XS(NIDS_TOT), YS(NIDS_TOT), MAGS(NIDS_TOT)
	REAL*4    PFT_CCS(NIDS_TOT), PRJ_CCS(NIDS_TOT)
	REAL*4    CMP_CCS(NIDS_TOT)
	CHARACTER IMG_FILENAME*256
	INTEGER*4 symm                  ! temp symmetry variable
	REAL*4    the, phi, ome         ! temp orientations
C  ----------------------------------------------------------------------------
C
	symm = sym                      !copy to symm for not use sym in parameters MBS
C
	NN = 1
C  *************************************************************
C  * Loop through PARAMETER files and retrieve particle info   *
C  *************************************************************
	DO 900 FILE = 1,NFILES
	  NN = NN -1		       ! Needed to handle EOF on 2nd PARAM file
	  IF(FILE.GT.1) CLOSE (UNIT=2, DISP='KEEP')
	  OPEN (UNIT=2, FILE=PARAM_FNAMES(FILE), STATUS='OLD', ERR=996,
	1      ACCESS='SEQUENTIAL', FORM='FORMATTED', READONLY)
C	1      ACCESS='SEQUENTIAL', FORM='FORMATTED', READONLY, SHARED)
	  READ (2,'(A)',ERR=997) 				! DUMMY READ #1
	  READ (2,'(A)',ERR=998)				! DUMMY READ #2
  100	  NN = NN + 1
	  IF(NN.GT.NIDS_TOT) GO TO 900
	  READ (2,'(I,9F)',END=900,ERR=999) IDS(NN), THE, PHI, OME,
	1  XS(NN), YS(NN), MAGS(NN), PFT_CCS(NN), PRJ_CCS(NN),
	2  CMP_CCS(NN)
	  FIDS(NN) = FILE
c       Always start from asymmetric triangle MBS 08/11/03
c       Check for negative thetas MBS 02/04/04
	IF(THE.lt.0.0) THEN
	   THE = -THE
	   PHI = PHI + 180.0
	ENDIF
	  if(symm .ne. 1)
	1      CALL TO_ASYM_UNIT (SYMm, THE, PHI, OME)
	  TS(NN) = the
	  PS(NN) = phi
	  OS(NN) = ome
c$$$c       Debug
c$$$	write(*,*)'nn,ts',nn,TS(NN)
	  GO TO 100
  900	ENDDO
	NN = NN -1
 	RETURN
C
  996	PRINT *, '***********************************'
	PRINT *, '* ERROR trying to OPEN PARAM file *'
	CALL TSBEND
  997	PRINT *, '****************************************************'
	PRINT *, '* FILL_PARAMS ERROR reading 1st line of PARAM file *'
	CALL TSBEND
  998	PRINT *, '****************************************************'
	PRINT *, '* FILL_PARAMS ERROR reading 2nd line of PARAM file *'
	CALL TSBEND
  999	PRINT *, '**************************************************'
	PRINT *, '* FILL_PARAMS ERROR reading data from PARAM file *'
	CALL TSBEND
	END


C  ****************************************************************************
C  *                                                                          *
C  *                  *****     SUBROUTINE FREEMEM0     *****                 *
C  *                                                                          *
C  *    RELEASE MEMORY TAKEN BY DYNAMIC MEMORY ARRAYS                         *
C  *                                                                          *
C  *    FEBRUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.       *
C  *    LAST MODIFIED 26-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE FREEMEM0 (FIDS_P, IDS_P, TS_P, PS_P, OS_P, XS_P, 
	1 YS_P, MAGS_P, CC1S_P, CC2S_P, CC3S_P, FIDS2_P, IDS2_P, 
	2 TS2_P, PS2_P, OS2_P, XS2_P, YS2_P, MAGS2_P, PFT_CCS2_P, 
	3 PRJ_CCS2_P, CMP_CCS2_P)
C
	IMPLICIT NONE
C
	INTEGER*4 FIDS(1), IDS(1)
	REAL*4    TS(1), PS(1), OS(1), XS(1), YS(1), MAGS(1), CC1S(1)
	REAL*4    CC2S(1), CC3S(1)
	POINTER   (TS_P,TS), (PS_P,PS), (OS_P,OS), (MAGS_P,MAGS)
	POINTER   (FIDS_P,FIDS), (IDS_P,IDS), (XS_P,XS), (YS_P,YS)
	POINTER   (CC1S_P,CC1S), (CC2S_P,CC2S), (CC3S_P,CC3S)
C
	INTEGER*4 FIDS2(1), IDS2(1)
	REAL*4    TS2(1), PS2(1), OS2(1), XS2(1), YS2(1), MAGS2(1)
	REAL*4    PFT_CCS2(1), PRJ_CCS2(1), CMP_CCS2(1)
	POINTER   (TS2_P,TS2), (PS2_P,PS2), (OS2_P,OS2), (MAGS2_P,MAGS2)
	POINTER   (FIDS2_P,FIDS2), (IDS2_P,IDS2),(XS2_P,XS2),(YS2_P,YS2)
	POINTER   (PFT_CCS2_P,PFT_CCS2), (PRJ_CCS2_P,PRJ_CCS2)
	POINTER   (CMP_CCS2_P,CMP_CCS2)
C
C  ----------------------------------------------------------------------------
C
	CALL FREE (FIDS_P)
	CALL FREE (IDS_P)
	CALL FREE (TS_P)
	CALL FREE (PS_P)
	CALL FREE (OS_P)
	CALL FREE (XS_P)
	CALL FREE (YS_P)
	CALL FREE (MAGS_P)
	CALL FREE (CC1S_P)
	CALL FREE (CC2S_P)
	CALL FREE (CC3S_P)
C
	CALL FREE (FIDS2_P)
	CALL FREE (IDS2_P)
	CALL FREE (TS2_P)
	CALL FREE (PS2_P)
	CALL FREE (OS2_P)
	CALL FREE (XS2_P)
	CALL FREE (YS2_P)
	CALL FREE (MAGS2_P)
	CALL FREE (PFT_CCS2_P)
	CALL FREE (PRJ_CCS2_P)
	CALL FREE (CMP_CCS2_P)
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE FREEMEM1     *****                *
C  *                                                                          *
C  *    RELEASE MEMORY TAKEN BY DYNAMIC MEMORY ARRAYS                         *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    LAST MODIFIED 27-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE FREEMEM1 (PRJ_AVG_P, PFTS_P, PRJS_P, THETAS_P,PHIS_P, 
	1 PFTS_AMP_SPFIT_P, PFTS_AMP_P, IMG_AMP_SPFIT_P, IMG_AMP_P, 
	2 PFTS_IMG_P, MAP3D_P)
C
	IMPLICIT NONE
C
	REAL*4    THETAS(1), PHIS(1), PRJ_AVG(1), PFTS(1), PRJS(1)
	POINTER   (THETAS_P,THETAS), (PHIS_P,PHIS)
	POINTER   (PRJ_AVG_P,PRJ_AVG), (PFTS_P,PFTS), (PRJS_P,PRJS)
	REAL*4    PFTS_AMP_SPFIT(1), PFTS_AMP(1), PFTS_IMG(1)
	REAL*4    IMG_AMP(1), IMG_AMP_SPFIT(1), MAP3D(1)
	POINTER   (PFTS_AMP_SPFIT_P,PFTS_AMP_SPFIT)
	POINTER   (PFTS_AMP_P,PFTS_AMP)
	POINTER   (IMG_AMP_P, IMG_AMP), (IMG_AMP_SPFIT_P, IMG_AMP_SPFIT)
	POINTER   (PFTS_IMG_P, PFTS_IMG), (MAP3D_P,MAP3D)
C  ----------------------------------------------------------------------------
C
	CALL FREE (PRJ_AVG_P)
	CALL FREE (PFTS_P)
	CALL FREE (PRJS_P)
	CALL FREE (THETAS_P)
	CALL FREE (PHIS_P)
	CALL FREE (PFTS_AMP_SPFIT_P)
	CALL FREE (PFTS_AMP_P)
	CALL FREE (IMG_AMP_P)
	CALL FREE (IMG_AMP_SPFIT_P)
	CALL FREE (MAP3D_P)
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE FREEMEM2     *****                *
C  *                                                                          *
C  *    RELEASE MEMORY TAKEN BY DYNAMIC MEMORY ARRAYS                         *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    LAST MODIFIED 27-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE FREEMEM2 (PRJ_P, PROJN_P, TRP_P, PFFT_P, 
	2 PMAP_P, A_P, ITMP_P, FFT6_P)
C
	IMPLICIT NONE
C
	INTEGER*4 ITMP(1)
	REAL*4    PRJ(1), PROJN(1), TRP(1), PFFT(1), A(1)
	REAL*4    PMAP(1), FFT6(1)
	POINTER   (PRJ_P,PRJ), (PROJN_P,PROJN)
	POINTER   (TRP_P,TRP), (PFFT_P,PFFT), (PMAP_P,PMAP)
	POINTER   (A_P,A), (ITMP_P,ITMP), (FFT6_P,FFT6)
C  ----------------------------------------------------------------------------
C
	CALL FREE (PRJ_P)
	CALL FREE (PROJN_P)
	CALL FREE (TRP_P)
	CALL FREE (PFFT_P)
	CALL FREE (PMAP_P)
	CALL FREE (A_P)
	CALL FREE (ITMP_P)
	CALL FREE (FFT6_P)
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE FREEMEM3     *****                *
C  *                                                                          *
C  *    RELEASE MEMORY TAKEN BY DYNAMIC MEMORY ARRAYS                         *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    LAST MODIFIED 27-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE FREEMEM3 (PFT_AVGS_P,SQ2_P,VAR_P,RMS_P,AVER_P,TMP_P)
C
	IMPLICIT NONE
C
	REAL*4  PFT_AVGS(1), TMP(1), SQ2(1), VAR(1), RMS(1), AVER(1)
	POINTER (PFT_AVGS_P,PFT_AVGS), (TMP_P,TMP), (SQ2_P,SQ2)
	POINTER (VAR_P,VAR), (RMS_P,RMS), (AVER_P,AVER)
C  ----------------------------------------------------------------------------
C
	CALL FREE (PFT_AVGS_P)
	CALL FREE (SQ2_P)
	CALL FREE (VAR_P)
	CALL FREE (RMS_P)
	CALL FREE (AVER_P)
	CALL FREE (TMP_P)
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                *****     SUBROUTINE FREEMEM4_F4     *****                *
C  *                                                                          *
C  *    FREE UP MEMORY.                                                       *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    MODIFIED 26-FEB-1998 BY TSB (Make UNIX compatible).                   *
C  *    LAST MODIFIED 20-DEC-2002 BY MBS (Added IMG_F4_P).                    *
C  ****************************************************************************
C
	SUBROUTINE FREEMEM4_F4 (FFT1_P, FFT3_P, IMG_P, R4_IMG_P, PRJ_P, 
	1 PRJ_ROT_P, PFTIMG_P, PFFT_P, PMAP_P, FLTPMAP_P, PPRJ_P, A_P, 
	2 CCOR_P, ICOR_P, ITMP_P)
C
	IMPLICIT NONE
C
	INTEGER*4 NROT, NANNULI, NVIEW_MOD
	INTEGER*2 IMG(1), ICOR(1), ITMP(1)
	REAL*4    IMG_F4(1)                                                   !MBS 12.12.02
	REAL*4    FFT1(1), FFT3(1), R4_IMG(1), PRJ(1), PRJ_ROT(1)
	REAL*4    PFTIMG(1), PFFT(1), PMAP(1), FLTPMAP(1)
	REAL*4    PPRJ(1), A(1), CCOR(1)
	POINTER   (FFT1_P,FFT1), (FFT3_P,FFT3), (IMG_P,IMG)
	POINTER   (IMG_F4_P,IMG_F4)                                           !MBS 12.12.02
	POINTER   (R4_IMG_P,R4_IMG), (PRJ_P,PRJ), (PRJ_ROT_P,PRJ_ROT)
	POINTER   (PFTIMG_P,PFTIMG), (PFFT_P,PFFT), (PMAP_P,PMAP)
	POINTER   (FLTPMAP_P,FLTPMAP), (PPRJ_P,PPRJ), (A_P,A)
	POINTER   (CCOR_P,CCOR), (ICOR_P,ICOR), (ITMP_P,ITMP)
C  ----------------------------------------------------------------------------
C
	CALL FREE (FFT1_P)
	CALL FREE (FFT3_P)
	CALL FREE (IMG_P)
	CALL FREE (IMG_F4_P)
	CALL FREE (R4_IMG_P)
	CALL FREE (PRJ_P)
	CALL FREE (PRJ_ROT_P)
	CALL FREE (PFTIMG_P)
	CALL FREE (PFFT_P)
	CALL FREE (PMAP_P)
	CALL FREE (FLTPMAP_P)
	CALL FREE (PPRJ_P)
	CALL FREE (A_P)
	CALL FREE (CCOR_P)
	CALL FREE (ICOR_P)
	CALL FREE (ITMP_P)
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *             *****     SUBROUTINE FREEMEM_REFINE_F4     *****             *
C  *                                                                          *
C  *    RELEASE MEMORY TAKEN BY DYNAMIC MEMORY ARRAYS IN REFINE MODE.         *
C  *                                                                          *
C  *    APRIL 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.          *
C  *    MODIFIED  4-DEC-98 BY SBW (Add FFT6 for CTF stuff).                   *
C  *    LAST MODIFIED 20-DEC-02 BY MBS (Add IMG_F4_P).                        *
C  ****************************************************************************
C
	SUBROUTINE FREEMEM_REFINE_F4 (MAP3D_P, MOD_TS_P, MOD_PS_P,PRJ_P, 
	1 PROJN_P, PMAP_P, PFFT_P, ITMP_P, VAR_P, RMS_P, AVER_P, TMP_P, 
	2 IMG_P, IMG_F4_P, FFT1_P, FFT3_P, FFT4_P, FFT5_P, FFT6_P, 
	3 MSKIMG_P, MSKPRJ_P, R4_IMG_P, BESTPRJ_P, PRJ_ROT_P, PFTIMG_P, 
	4 R4_PIMG_P, FLTPMAP_P, POLAR_PRJ_P, A_ARRAY_P, 
	5 TMP1_P, TMP2_P, TMP3_P, FFTA_P, FFTB_P)
C
	IMPLICIT NONE
C
	INTEGER*2 IMG(1)
	INTEGER*4 ITMP(1)
	REAL*4    IMG_F4(1)
	REAL*4    MAP3D(1), MOD_TS(1), MOD_PS(1), PMAP(1), TMP(1)
        REAL*4    PRJ(1), PROJN(1), PFFT(1), VAR(1), RMS(1), AVER(1)
	REAL*4    FFT1(1), FFT3(1), FFT4(1), FFT5(1), FFT6(1)
	REAL*4    A_ARRAY(1)
	REAL*4    R4_IMG(1), R4_PIMG(1), BESTPRJ(1), PRJ_ROT(1)
	REAL*4    PFTIMG(1), POLAR_PRJ(1), MSKIMG(1), MSKPRJ(1)
	REAL*4    FLTPMAP(1), TMP1(1), TMP2(1), TMP3(1), CCOR(1)
	REAL*4    PRJS(1), PFTS(1), PFT_AVGS(1), SQ2(1)
	POINTER   (ITMP_P,ITMP), (MOD_TS_P,MOD_TS), (MOD_PS_P,MOD_PS)
	POINTER   (PFFT_P,PFFT), (PMAP_P,PMAP), (TMP_P,TMP)
	POINTER   (MAP3D_P,MAP3D), (PRJ_P,PRJ), (PROJN_P,PROJN)
	POINTER   (IMG_P,IMG), (VAR_P,VAR), (RMS_P,RMS), (AVER_P,AVER)
	POINTER   (IMG_F4_P,IMG_F4)
	POINTER   (FFT1_P,FFT1), (FFT3_P,FFT3), (FFT4_P,FFT4)
	POINTER   (FFT5_P,FFT5), (FFT6_P,FFT6)
	pointer   (R4_IMG_P,R4_IMG), (R4_PIMG_P,R4_PIMG)
	POINTER   (MSKIMG_P,MSKIMG), (MSKPRJ_P,MSKPRJ), (PFTS_P,PFTS)
	POINTER   (BESTPRJ_P,BESTPRJ), (PRJ_ROT_P,PRJ_ROT)
	POINTER   (PFTIMG_P,PFTIMG), (PFT_AVGS_P,PFT_AVGS), (SQ2_P,SQ2)
	POINTER   (FLTPMAP_P,FLTPMAP), (POLAR_PRJ_P,POLAR_PRJ)
	POINTER   (CCOR_P,CCOR), (A_ARRAY_P,A_ARRAY), (PRJS_P,PRJS)
	POINTER   (TMP1_P,TMP1), (TMP2_P,TMP2), (TMP3_P,TMP3)
C
	REAL*4    PRJS_SF(1), IMG_SF(1), PRJS_SFB(1)
	POINTER   (PRJS_SF_P, PRJS_SF), (IMG_SF_P, IMG_SF)
	POINTER   (PRJS_SFB_P, PRJS_SFB)
	real*4    FFTA(1), FFTB(1)
	pointer   (FFTA_P, FFTA), (FFTB_P, FFTB)
C
C  ----------------------------------------------------------------------------
C
	CALL FREE (IMG_P)
	CALL FREE (IMG_F4_P)
	CALL FREE (MAP3D_P)
	CALL FREE (MOD_TS_P)
	CALL FREE (MOD_PS_P)
	CALL FREE (PRJ_P)
	CALL FREE (PROJN_P)
	CALL FREE (PMAP_P)
	CALL FREE (PFTS_P)
	CALL FREE (PFFT_P)
	CALL FREE (ITMP_P)
	CALL FREE (VAR_P)
	CALL FREE (RMS_P)
	CALL FREE (AVER_P)
	CALL FREE (TMP_P)
	CALL FREE (R4_IMG_P)
	CALL FREE (R4_PIMG_P)
	CALL FREE (FFT1_P)
	CALL FREE (FFT3_P)
	CALL FREE (FFT4_P)
	CALL FREE (FFT5_P)
	CALL FREE (FFT6_P)
	CALL FREE (MSKIMG_P)
	CALL FREE (MSKPRJ_P)
	CALL FREE (BESTPRJ_P)
	CALL FREE (PRJ_ROT_P)
	CALL FREE (PFTIMG_P)
	CALL FREE (PFT_AVGS_P)
	CALL FREE (SQ2_P)
	CALL FREE (FLTPMAP_P)
	CALL FREE (POLAR_PRJ_P)
	CALL FREE (A_ARRAY_P)
	CALL FREE (CCOR_P)
	CALL FREE (TMP1_P)
	CALL FREE (TMP2_P)
	CALL FREE (TMP3_P)
	CALL FREE (PRJS_SF_P)
	CALL FREE (IMG_SF_P)
	CALL FREE (FFTA_P)
	CALL FREE (FFTB_P)
	CALL FREE (PRJS_SFB_P)
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE GETMEM0     *****                 *
C  *                                                                          *
C  *    SET UP DYNAMIC MEMORY ARRAYS                                          *
C  *                                                                          *
C  *    FEBRUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.       *
C  *    LAST MODIFIED 26-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE GETMEM0 (NIDS_TOT, FIDS_P, IDS_P, TS_P, PS_P, OS_P, 
	1 XS_P, YS_P, MAGS_P, CC1S_P, CC2S_P, CC3S_P, FIDS2_P, IDS2_P, 
	2 TS2_P, PS2_P, OS2_P, XS2_P, YS2_P, MAGS2_P, PFT_CCS2_P, 
	3 PRJ_CCS2_P, CMP_CCS2_P)
C
	IMPLICIT NONE
C
	INTEGER*4 NIDS_TOT, FIDS(1), IDS(1)
	REAL*4    TS(1), PS(1), OS(1), XS(1), YS(1), MAGS(1), CC1S(1)
	REAL*4    CC2S(1), CC3S(1)
	POINTER   (TS_P,TS), (PS_P,PS), (OS_P,OS), (MAGS_P,MAGS)
	POINTER   (FIDS_P,FIDS), (IDS_P,IDS), (XS_P,XS), (YS_P,YS)
	POINTER   (CC1S_P,CC1S), (CC2S_P,CC2S), (CC3S_P,CC3S)
C
	INTEGER*4 FIDS2(1), IDS2(1)
	REAL*4    TS2(1), PS2(1), OS2(1), XS2(1), YS2(1), MAGS2(1)
	REAL*4    PFT_CCS2(1), PRJ_CCS2(1), CMP_CCS2(1)
	POINTER   (TS2_P,TS2), (PS2_P,PS2), (OS2_P,OS2), (MAGS2_P,MAGS2)
	POINTER   (FIDS2_P,FIDS2), (IDS2_P,IDS2), (XS2_P,XS2)
	POINTER   (YS2_P,YS2)
	POINTER   (PFT_CCS2_P,PFT_CCS2), (PRJ_CCS2_P,PRJ_CCS2)
	POINTER   (CMP_CCS2_P,CMP_CCS2)
	LOGICAL   MALLOC
C  ----------------------------------------------------------------------------
C
	FIDS_P = MALLOC(4*NIDS_TOT)
	IDS_P  = MALLOC(4*NIDS_TOT)
	TS_P   = MALLOC(4*NIDS_TOT)
	PS_P   = MALLOC(4*NIDS_TOT)
	OS_P   = MALLOC(4*NIDS_TOT)
	XS_P   = MALLOC(4*NIDS_TOT)
	YS_P   = MALLOC(4*NIDS_TOT)
	MAGS_P = MALLOC(4*NIDS_TOT)
	CC1S_P = MALLOC(4*NIDS_TOT)
	CC2S_P = MALLOC(4*NIDS_TOT)
	CC3S_P = MALLOC(4*NIDS_TOT)
C
	FIDS2_P = MALLOC(4*NIDS_TOT)
	IDS2_P  = MALLOC(4*NIDS_TOT)
	TS2_P   = MALLOC(4*NIDS_TOT)
	PS2_P   = MALLOC(4*NIDS_TOT)
	OS2_P   = MALLOC(4*NIDS_TOT)
	XS2_P   = MALLOC(4*NIDS_TOT)
	YS2_P   = MALLOC(4*NIDS_TOT)
	MAGS2_P = MALLOC(4*NIDS_TOT)
	PFT_CCS2_P = MALLOC(4*NIDS_TOT)
	PRJ_CCS2_P = MALLOC(4*NIDS_TOT)
	CMP_CCS2_P = MALLOC(4*NIDS_TOT)
C
	IF(FIDS_P.EQ.0 .OR. IDS_P.EQ.0 .OR. TS_P.EQ.0 .OR. PS_P.EQ.0 
	1 .OR. OS_P.EQ.0 .OR. MAGS_P.EQ.0 .OR. CC1S_P.EQ.0 
	2 .OR. CC2S_P.EQ.0 .OR. CC3S_P.EQ.0) THEN
c	IF(FIDS_P.EQ.0 .OR. IDS_P.EQ.0 .OR. TS_P.EQ.0 .OR. PS_P.EQ.0 .OR. 
c	1 OS_P.EQ.0 .OR. MAGS_P.EQ.0 .OR. CC1S_P.EQ.0 .OR. CC2S_P.EQ.0
c	2 .OR. CC3S_P.EQ.0) THEN
	  PRINT *, '*************************************************'
	  PRINT *, '* GETMEM0 memory allocation error: EXIT program *'
	  CALL TSBEND
	ENDIF
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE GETMEM1     *****                 *
C  *                                                                          *
C  *    SET UP DYNAMIC MEMORY ARRAYS                                          *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    LAST MODIFIED 27-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE GETMEM1 (NCOLBIN, NROWBIN, NPIX, 
	1 PRJ_AVG_P, PFTS_P, PRJS_P, THETAS_P, PHIS_P, NFILES, 
	2 PFTS_AMP_SPFIT_P, PFTS_AMP_P, IMG_AMP_SPFIT_P, IMG_AMP_P,
	3 PFTS_IMG_P)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NCOLBIN, NROWBIN, NPIX, NFILES
	REAL*4    THETAS(1), PHIS(1), PRJ_AVG(1), PFTS(1), PRJS(1)
	REAL*4    PFTS_AMP_SPFIT(1), PFTS_AMP(1), PFTS_IMG(1)
	REAL*4    IMG_AMP(1), IMG_AMP_SPFIT(1)
	POINTER   (THETAS_P,THETAS), (PHIS_P,PHIS)
	POINTER   (PRJ_AVG_P,PRJ_AVG), (PFTS_P,PFTS), (PRJS_P,PRJS)
	POINTER   (PFTS_AMP_SPFIT_P,PFTS_AMP_SPFIT)
	POINTER   (PFTS_AMP_P,PFTS_AMP)
	POINTER   (IMG_AMP_P, IMG_AMP), (IMG_AMP_SPFIT_P, IMG_AMP_SPFIT)
	POINTER   (PFTS_IMG_P, PFTS_IMG)
	LOGICAL   MALLOC
C  ----------------------------------------------------------------------------
C
	IMG_AMP_P        = MALLOC (4*NPIX)
	IMG_AMP_SPFIT_P  = MALLOC (4*NPIX)
	THETAS_P  = MALLOC (4*NVIEW_MOD)
	PHIS_P    = MALLOC (4*NVIEW_MOD)
	PRJ_AVG_P = MALLOC (4*NCOLBIN*NROWBIN)
	PFTS_P    = MALLOC (4*NPIX*NVIEW_MOD*NFILES)
	PRJS_P    = MALLOC (4*NCOLBIN*NROWBIN*NVIEW_MOD*NFILES)
	PFTS_AMP_P       = MALLOC (4*NPIX*NVIEW_MOD*NFILES)
	PFTS_AMP_SPFIT_P = MALLOC (4*NPIX*NVIEW_MOD*NFILES)
	PFTS_IMG_P       = MALLOC (4*NPIX*NVIEW_MOD*NFILES)
	IF(PRJ_AVG_P.EQ.0 .OR. PFTS_P.EQ.0 .OR. PRJS_P.EQ.0 .OR. 
	1 THETAS_P.EQ.0 .OR. PHIS_P.EQ.0 .OR. PFTS_IMG_P.EQ.0) THEN
	  PRINT *, '*************************************************'
	  PRINT *, '* GETMEM1 memory allocation error: EXIT program *'
	  CALL TSBEND
	ENDIF
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE GETMEM2     *****                 *
C  *                                                                          *
C  *    SET UP DYNAMIC MEMORY ARRAYS                                          *
C  *                                                                          *
C  *    DECEMBER 1997 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.       *
C  *    MODIFIED 27-FEB-1998 BY TSB (Make UNIX compatible).                   *
C  *    LAST MODIFIED  4-DEC-98 BY SBW (Add FFT6 for CTF stuff).              *
C  ****************************************************************************
C
	SUBROUTINE GETMEM2 (NCOL, NROW, NSEC, NROT, NANNULI, idim1, 
	1 idim2, PRJ_P, PROJN_P, TRP_P, MAP3D_P, PFFT_P, PMAP_P, 
	2 A_P, ITMP_P, FFT6_P)
C
	IMPLICIT NONE
C
	INTEGER*4 NCOL, NROW, NSEC, NROT, NANNULI, IDIM1, IDIM2, ITMP(1)
	REAL*4    MAP3D(1), PRJ(1), PROJN(1), TRP(1), PFFT(1), A(1)
	REAL*4    PMAP(1), FFT6(1)
	POINTER   (MAP3D_P,MAP3D), (PRJ_P,PRJ), (PROJN_P,PROJN)
	POINTER   (TRP_P,TRP), (PFFT_P,PFFT), (PMAP_P,PMAP)
	POINTER   (A_P,A), (ITMP_P,ITMP), (FFT6_P,FFT6)
	LOGICAL   MALLOC
C  ----------------------------------------------------------------------------
C
	MAP3D_P = MALLOC (4*NCOL*NROW*NSEC)
	PRJ_P   = MALLOC (4*NCOL*NROW)
	PROJN_P = MALLOC (4*NCOL*NROW)
	TRP_P   = MALLOC (4*NCOL*NROW)
	PFFT_P  = MALLOC (4*2*NROT*NANNULI)
	PMAP_P  = MALLOC (4*NROT*NANNULI)
	A_P     = MALLOC (4*2*NROT)
	ITMP_P  = MALLOC (4*NCOL)
	FFT6_P  = MALLOC (8*IDIM1*IDIM2)
	IF(MAP3D_P.EQ.0 .OR. PRJ_P.EQ.0 .OR. PROJN_P.EQ.0 .OR. 
	1 TRP_P.EQ.0 .OR. PFFT_P.EQ.0 .OR. PMAP_P.EQ.0 .OR.
	2 A_P.EQ.0 .OR. ITMP_P.EQ.0 .or. FFT6_P.EQ.0) THEN
	  PRINT *, '*************************************************'
	  PRINT *, '* GETMEM2 memory allocation error: EXIT program *'
	  CALL TSBEND
	ENDIF
C
	RETURN
	END	


C  ****************************************************************************
C  *                                                                          *
C  *                   *****     SUBROUTINE GETMEM3     *****                 *
C  *                                                                          *
C  *    SET UP DYNAMIC MEMORY ARRAYS                                          *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    LAST MODIFIED 27-FEB-1998 BY TSB (Make UNIX compatible).              *
C  ****************************************************************************
C
	SUBROUTINE GETMEM3 (NPIX, NROT, NANNULI, PFT_AVGS_P, 
	1 SQ2_P, VAR_P, RMS_P, AVER_P, TMP_P)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NPIX, NROT, NANNULI
	REAL*4    PFT_AVGS(1), TMP(1), SQ2(1), VAR(1), RMS(1), AVER(1)
	POINTER   (PFT_AVGS_P,PFT_AVGS), (TMP_P,TMP), (SQ2_P,SQ2)
	POINTER   (VAR_P,VAR), (RMS_P,RMS), (AVER_P,AVER)
	LOGICAL   MALLOC
C  ----------------------------------------------------------------------------
C
	PFT_AVGS_P = MALLOC (4*NVIEW_MOD)
	SQ2_P      = MALLOC (4*NVIEW_MOD)
	VAR_P      = MALLOC (4*NPIX)
	RMS_P      = MALLOC (4*NPIX)
	AVER_P     = MALLOC (4*NPIX)
	TMP_P      = MALLOC (4*NROT*NANNULI)
	IF(PFT_AVGS_P.EQ.0 .OR. SQ2_P.EQ.0 .OR. VAR_P.EQ.0 .OR. 
	1 RMS_P.EQ.0 .OR. AVER_P.EQ.0 .OR. TMP_P.EQ.0) THEN
	  PRINT *, '*************************************************'
	  PRINT *, '* GETMEM3 memory allocation error: EXIT program *'
	  CALL TSBEND
	ENDIF
C
	RETURN
	END	



C  ****************************************************************************
C  *                                                                          *
C  *                  *****     SUBROUTINE GET_TPO_R3     *****               *
C  *                                                                          *
C  *    DETERMINE THETA, PHI, AND OMEGA FOR CURRENT IMAGE.                    *
C  *                                                                          *
C  *    Protocol:                                                             *
C  *     1. Compute polar image from raw (Cartesian) image (R4_IMG->R4_PIMG). *
C  *     2. Compute polar FFT (PFFT) from R4_PIMG.                            *
C  *     3. Convert image PFT to amplitudes & extract desired portion.        *
C  *     4. Find average value of image PFT (also termed particle PFT).       *
C  *     5. Cross-correlate particle PFT with all reference PFTS to get THETA *
C  *         and PHI (|PHI| for n22 and 532 symmetry, both THETA and PHI may  *
C  *         change if the object is asymmetric or n-fold.                    *
C  *     6. Load the best model projection (BESTPRJ) corresponding to THETA   *
C  *         and |PHI|.                                                       *
C  *     7. Convert BESTPRJ to polar projection (POLAR_PRJ).                  *
C  *     8. Cross-correlate POLAR_PRJ with R4_PIMG to determine sign of       *
C  *         PHI and value of OMEGA (also THETA if SYM=1-20).                 *
C  *                                                                          *
C  *    APRIL 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.          *
C  *                                                                          *
C  *    MODIFIED 29-APR-1998 BY T. BAKER (REFINE specific).                   *
C  *    MODIFIED 19-MAR-2003 BY M. Sherman (rewamped code).                   *
C  *    LAST MODIFIED 31-OCT-2003 BY MBS (added ctfmode to par list,          *
C  *    same for GET_THEPHI_PR3, GET_THEPHI_PR3_mult)                         *
C  ****************************************************************************
C
	SUBROUTINE GET_TPO_R3 (NCOLBIN, NROWBIN, NROT, NANNULI, NUNIQUE, 
	1    NVIEWS, MASK, MOD_TS, MOD_PS, ITH, IPH, R4_IMG, R4_PIMG, 
	2    XCEN, YCEN, PI, THE_BEST, PHI_BEST, OME_BEST, CC_PFT, 
	5    CCOR, INDEX, OME_RAWIMG, IC, SC, IO, SO, IMG_CCOR, 
	7    PRJS_SF, IMG_SF, IDIM1, IDIM2, FFT_FULL, FFT_FZERO,FFT1,FFTA,
	8    FFTB,FILE,halfSize,Radius2,PRJS_SFB,map3d,BFACTOR,CTFMODE)
C
	IMPLICIT NONE
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1 FOCUS_MIN, ANG_MAJ, CS_COEFF

	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999),FOCUS_MIN(999),ANG_MAJ(999),UNITS(999)


	INTEGER*4 I, J, NCOLBIN, NROWBIN, NROT, NANNULI, halfSize
	INTEGER*4 NUNIQUE, INDEX, CTFMODE
	INTEGER*4 NVIEWS, MASK(NPHI,NTHE,2), ITH, IPH
	REAL*4    XCEN, YCEN
	REAL*4    CC_PFT, BFACTOR, resmin,resmax
	REAL*4    PI, THE_BEST, PHI_BEST, OME_BEST
	REAL*4    MOD_TS(NTHE), MOD_PS(NPHI,NTHE)
	REAL*4    R4_IMG(NCOLBIN*NROWBIN)
	REAL*4    R4_PIMG(NROT,NANNULI)
	REAL*4    CCOR(NUNIQUE)
	REAL*4    IC, IO, IMG_CCOR, OME_RAWIMG, OMEGA
	REAL*4    X_TMP, Y_TMP
	INTEGER*4 SC, SO
C
	INTEGER*4 IDIM1, IDIM2
	REAL*4    IMG_SF  ((2*halfSize+1)*(2*halfSize+1))
	REAL*4    IMG_SFB ((2*halfSize+1)*(2*halfSize+1))
	REAL*4    PRJS_SF (2*halfSize+1,2*halfSize+1, NUNIQUE)
	REAL*4    PRJS_SFB(2*halfSize+1,2*halfSize+1, NUNIQUE)
	REAL*4    PRJS_AVG(halfSize+1)
	REAL*4    IMG_AVG (halfSize+1)
	REAL*4    AMAX,INDEX_BEST, XCEN_BEST, YCEN_BEST
	REAL*4    R4_IMG_BAK(NCOLBIN*NROWBIN)
	INTEGER*4 ITH_TMP, IPH_TMP, ID, ITH_BEST, IPH_BEST
C
	REAL*4    Radius2
	REAL*4    FFT_FZERO
	REAL*4    best_omega, best_the, best_phi
	real*4    best_ccpft, best_x, best_Y
	REAL*4    FFT_ORIGX, FFT_ORIGY
	REAL*4    FFT1(2*IDIM1, IDIM2)
	REAL*4    FFT_cent(2*(2*halfSize+1),(2*halfSize+1))
	REAL*4    FFTA(IDIM1, IDIM2), FFTB(IDIM1, IDIM2)
	INTEGER*4 best_index, index2, tmp1, FILE
	integer*4 row, col, row_min, col_min, row_max, col_max
	integer*4 nrowcolbin,cent_dim
	REAL*4    rad_pol, delang
	real*4    omega_img
	REAL*4    MAP3D(NCOLBIN,NROWBIN,NROWBIN)
	REAL*4    projn(NCOLBIN*NROWBIN),bestprj(NCOLBIN*NROWBIN)
	REAL*4    prj_SF ((2*halfSize+1)*(2*halfSize+1))
	REAL*4    prj_SFB((2*halfSize+1)*(2*halfSize+1))
C
	LOGICAL   MASK_JUDGE, FFT_FULL

C -----------------------------------------------------------------------------
C       

	resmin = res_min
	resmax = res_max
	rad_pol = RADIUS
	delang  = DEL_ANG

	FFT_ORIGX = FLOAT(NCOLBIN/2)
	FFT_ORIGY = FLOAT(NROWBIN/2)
C
	row_min = IDIM2/2 - halfSize
	row_max = IDIM2/2 + halfSize
	col_min = IDIM1/2 + 1 - halfSize
	col_max = IDIM1/2 + 1 + halfSize

	cent_dim = 2*halfSize + 1

	MASK_JUDGE = .FALSE.
	AMAX = -1.0
	BEST_CCPFT = -99.0
	Y_TMP=YCEN
	X_TMP=XCEN
	OMEGA = OME_RAWIMG
	nrowcolbin = NROWBIN*NCOLBIN
C
	CALL GET_PRJS_AMPAVG (PRJS_SF, PRJS_SFB, MASK, 
	1    NUNIQUE, ITH, IPH, halfSize, PRJS_AVG)
C
C **************************************
C		Refine centers using SF
C **************************************
C
	tmp1 = 0
	FFT_FULL = .TRUE.
C
c       MBS rotate image to -omega to match projections and calculate 
c       average amplitudes
	do i = 1, nrowcolbin
	   R4_IMG_BAK(i) = R4_IMG(i)
	end do
	CALL interp2d (R4_IMG_BAK, NCOLBIN, NROWBIN, XCEN-1.0, YCEN-1.0, 
	1    XCEN-1.0, YCEN-1.0, -OMEGA, 1.0, 0.0)
	CALL IMG_R4_FFT (R4_IMG_BAK,NCOLBIN, NROWBIN, FFT1,IDIM1,IDIM2, 
	1    0.0, FFT_ORIGX, FFT_ORIGY, FFT_FZERO, FFT_FULL, 0)
	call Fill_FFT_cent(IDIM1, IDIM2, FFT1, FFT_Cent, halfSize, 
	1    halfSize)
	CALL FFT_REARRANGE_SF(FFT_Cent,IMG_SF,IMG_SFB,
	1    cent_dim,cent_dim,XCEN,YCEN, IDIM1, IDIM2)
	CALL GET_IMG_AMPAVG (IMG_SF, IMG_SFB, halfSize, IMG_AVG)
c       Now do rotational search in omega:
	DO OMEGA = OME_RAWIMG-SO*IO, OME_RAWIMG+SO*IO,IO
	   do i = 1, nrowcolbin
	      R4_IMG_BAK(i) = R4_IMG(i)
	   end do
	   CALL interp2d (R4_IMG_BAK, NCOLBIN, NROWBIN, XCEN-1.0, YCEN-1.0, 
	1	XCEN-1.0, YCEN-1.0, -OMEGA, 1.0, 0.0)
	   CALL IMG_R4_FFT (R4_IMG_BAK,NCOLBIN, NROWBIN, FFT1,IDIM1,IDIM2, 
	1	0.0, FFT_ORIGX, FFT_ORIGY, FFT_FZERO, FFT_FULL, 0)
	   call Fill_FFT_cent(IDIM1, IDIM2, FFT1, FFT_Cent, halfSize, 
	1    halfSize)
	   CALL FFT_REARRANGE_SF(FFT_Cent,IMG_SF,IMG_SFB,
	1    cent_dim,cent_dim,XCEN,YCEN, IDIM1, IDIM2)
	   CALL GET_IMG_AMPAVG (IMG_SF, IMG_SFB, halfSize, IMG_AVG)
	   CALL GET_THEPHI_PR3 (PRJS_SF, IMG_SF, NUNIQUE, 
	1	CC_PFT, CCOR, PI, INDEX, MASK, ITH, IPH, 
	2	NVIEWS, MASK_JUDGE, ITH_TMP, IPH_TMP, 
	3	IDIM1,IDIM2, FILE, halfSize, 
	4	PRJS_SFB, IMG_SFB, PRJS_AVG, IMG_AVG, CTFMODE)
C
	   IF (CC_PFT.GT.BEST_CCPFT) THEN
	      BEST_CCPFT = CC_PFT
	      BEST_INDEX  = INDEX
	      BEST_X = XCEN
	      BEST_Y = YCEN
	      BEST_OMEGA = OMEGA
	      best_the = MOD_TS(ITH_TMP)
	      best_phi = MOD_PS(IPH_TMP,ITH_TMP)
	      ITH_BEST = ITH_TMP
	      IPH_BEST = IPH_TMP
	   ENDIF
ca	   print *, 'get_tpo_r:', CC_PFT, BEST_CCPFT, BEST_OMEGA, OMEGA
	ENDDO  !Omega = ....

	INDEX    = BEST_INDEX
	THE_BEST = best_the
	PHI_BEST = best_phi
	OME_BEST = BEST_OMEGA
	CC_PFT   = BEST_CCPFT
	OMEGA  = OME_BEST
	ITH = ITH_BEST
	IPH = IPH_BEST

c$$$c       Rotate the image copy to the best omega found etc.
c$$$	do i = 1, nrowcolbin
c$$$	   R4_IMG_BAK(i) = R4_IMG(i)
c$$$	end do
c$$$	CALL interp2d (R4_IMG_BAK, NCOLBIN, NROWBIN, XCEN-1.0, YCEN-1.0, 
c$$$	1    XCEN-1.0, YCEN-1.0, -OMEGA, 1.0, 0.0)
c$$$	CALL IMG_R4_FFT (R4_IMG_BAK,NCOLBIN, NROWBIN, FFT1,IDIM1,IDIM2, 
c$$$	1    0.0, FFT_ORIGX, FFT_ORIGY, FFT_FZERO, FFT_FULL, 0)
c$$$	call Fill_FFT_cent(IDIM1, IDIM2, FFT1, FFT_Cent, halfSize, 
c$$$	1    halfSize)
c$$$c       From now on all is needed, is to apply phase shift (equivalent to
c$$$c       shifting of origin). Copy original transform into an intermediate array
	BEST_CCPFT = -999.
	DO YCEN=Y_TMP-SC*IC, Y_TMP+SC*IC, IC
	   DO XCEN=X_TMP-SC*IC, X_TMP+SC*IC, IC
	      do i = 1, nrowcolbin
		 R4_IMG_BAK(i) = R4_IMG(i)
	      end do
	      CALL interp2d(R4_IMG_BAK,NCOLBIN,NROWBIN,XCEN-1.0,YCEN-1.0, 
	1	   XCEN-1.0, YCEN-1.0, -OMEGA, 1.0, 0.0)
	      CALL IMG_R4_FFT (R4_IMG_BAK,NCOLBIN,NROWBIN,FFT1,IDIM1,
	1	   IDIM2,0.0,FFT_ORIGX,FFT_ORIGY,FFT_FZERO,FFT_FULL,0)
	      call Fill_FFT_cent(IDIM1, IDIM2, FFT1, FFT_Cent, halfSize, 
	1	   halfSize)
c	      Shift the center to Xcen, Ycen:
	      CALL FFT_REARRANGE_SF(FFT_Cent,IMG_SF,IMG_SFB,
	1	   cent_dim,cent_dim,XCEN,YCEN, IDIM1, IDIM2)
c       Get correlation between the image and map projection in
c       best_the, best_phi direction
	      CALL GET_THEPHI_PR3_mult(PRJS_SF,PRJS_SFB,IMG_SF,IMG_SFB,
	1	   PRJS_AVG,IMG_AVG,CC_PFT,PI,IDIM1,IDIM2,FILE,halfSize,
	2	   ith,iph,NUNIQUE,MASK,MASK_JUDGE, CTFMODE)
	      IF (CC_PFT.GT.BEST_CCPFT) THEN
		 BEST_CCPFT = CC_PFT
		 BEST_X = XCEN
		 BEST_Y = YCEN
	      ENDIF      
 	   ENDDO
	   tmp1 = tmp1 + 1
ca	   WRITE(*,900) 'Centers: ',tmp1,BEST_X,BEST_Y,XCEN,YCEN,BEST_CCPFT
	ENDDO

	XCEN     = BEST_X
	yCEN     = BEST_y
	INDEX    = BEST_INDEX
	THE_BEST = best_the
	PHI_BEST = best_phi
	OME_BEST = BEST_OMEGA
	CC_PFT   = BEST_CCPFT
	CALL MAP_POLAR (R4_IMG, NCOLBIN, NROWBIN, R4_PIMG, NROT, 
	1    NANNULI, xcen, ycen, 1.0, RAD_pol, DELANG)
 900	FORMAT(A,I3, 5F10.4)
	RETURN
	END

C*****************************************************************************
C*                                                                           *
C*                     *****     SUBROUTINE KEY_INFO     *****               *
C*                                                                           *
C*    ROUTINE FOR OBTAINING INFORMATION FOR RUNNING EMPFT PROGRAM.           *
C*    Modified from rtn INFO().  This rtn uses keyword input now.            *
C*                                                                           *
C*    Originated  14-JUL-2004 BY RWA.                                        *
C*****************************************************************************
C
	SUBROUTINE key_info (NCOL, NROW, NSEC, NCOLBIN, NROWBIN, NSECBIN,
	1 NROT, NANNULI, NPIX, NFILES, NIDS, NIDS_TOT, CTFMODE, IC,SC,IO,SO,
	2 Radius2, BFACTOR,IDIM1,IDIM2,MAPHEAD,IMGHEAD,PRJHEAD,PFTHEAD)
C
	IMPLICIT NONE
C
	include 'PIFHEAD.INC'
	RECORD /PIFHEAD_T/ MAPHEAD, IMGHEAD, PRJHEAD, PFTHEAD
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES
C
	INCLUDE '../libEMF/INCfiles/INFOHEAD.INC'
	RECORD /INFOHEAD_T/ INFOHEAD
C
	INTEGER*4 NIMG, BIN, TIMER, CTFMODE
	INTEGER*4 NIDS_TOT, NIDS(999), NCOLBIN, NROWBIN, NSECBIN
	INTEGER*4 NCOL, NROW, NSEC, NVIEW_MOD, NTHE, NPHI, NANNULI, NV
	INTEGER*4 I, N_POWER2, NROT, MODE, VERBOSE, NPIX, T_MATRIX(2,3)
	INTEGER*4 IDIM1, IDIM2, FFT_NFAC, JCUT, ANN_LO, ANN_HI, SYM
	INTEGER*4 MAG_NUM, MAG_NORM,Radius2
	INTEGER*4 NROTD, NANNULID, NVIEW_MODD, SYMD, NFILES
	REAL*4    PIXSIZE, RECIP_SF
	REAL*4    DEL_ANG, RADIUS , PFTRAD_LO, PFTRAD_HI, PFTRAD_STEP
	REAL*4    PI, RADIAN, DEGREE, DEL_PIXRES, RES_MIN, RES_MAX
	REAL*4    PIXRES_MIN, PIXRES_MAX, RES_TST, SIGCUT, MAG_LO
	REAL*4    MAG_HI
	REAL*4    DANG, DTHE, DPHI_90, TEST, MAG_CEN, MAG_STEP
	REAL*4    RES_LO, RES_HI, DTHED, DPHI_90D, XCEN_MODEL
	REAL*4    YCEN_MODEL
	CHARACTER PFT_FILENAME*256, PRJ_FILENAME*256, MAP_FILENAME*256
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256, FNAME*256
	character pftrads_fname*512
	character pftres1_fname*512
	character pftres2_fname*512
	LOGICAL   VAROPT
	BYTE      READWRITE(4)
	REAL*4    IC, IO, BFACTOR, FILT_FAC
	INTEGER*4 SC, SO
C  ----------------------------------------------------------------------------
C
	call parse_key_input(infoHead)
Ca
C  ****************************************************************************
C  *  Input LINE #1: MODE, BIN, SYM, DANG, CTFMODE, VERBOSE, TIMER              *
C  *                  (3I,F,3I)                                               *
C  *                                                                          *
C  *  MODE  =  3 Refine mode (Basically it's always 3 here in OOR)            *
C  *  BIN   = 1 Binfactor of 1 (no data compression)                          *
C  *        = 2 Binfactor of 2 (compresses 3D data by 8X; 2D data by 4X)      *
C  *  VERBOSE = 0 Minimal output (refined orientations and origins)             *
C  *        = 1 Correlation coefficients written out as functions of radius   *
C  *             and resolution in output files EMPFT.* (* = RADS,RES1,RES2)  *
C  *        = 2 gives VERBOSE=1 output and correlation coefficients for each    *
C  *             possible view                                                *
C  *  TIMER = 0 gives no information on program segment times.                *
C  *        = 1 reports program segment times.                                *
C  ****************************************************************************
	MODE         = infoHead.input_mode
	BIN          = infoHead.bin_factor
	SYM          = infoHead.symmetry
        DANG         = infoHead.delta_theta
	CTFMODE      = infoHead.ctf_mode
	VERBOSE      = infoHead.verbose
	FILT_FAC     = infoHead.filter_factor_1
	RADIUS2      = infoHead.radius
	BFACTOR      = infoHead.temperature_factor
	IF(TIMER.NE.1) TIMER = 0		       ! A PHANTOM SWITCH [TSB]
	IF(MODE.LT.-2 .OR. MODE.GT.4 .OR.MODE.EQ.0) MODE = 1
	IF(SYM.LE.0) SYM = 532
	IF(BIN.NE.2) BIN =   1         	       ! Current limit for this set = 2
	mode = 3
	IF(MODE.EQ.3 .AND. VERBOSE.GT.0) THEN
	  VERBOSE = 0
	  PRINT *,'VERBOSE set to 0 because you are in REFINE mode'
	ENDIF
	IF(CTFMODE.LE.0 .OR. CTFMODE.GT.3) CTFMODE = 0
	IF(DANG.LE.0.0) DANG = 1.0
	IF(DANG.LT.0.1) DANG = 0.1
	DTHE    = DANG
	DPHI_90 = DANG
	IF(VERBOSE.NE.1 .AND. VERBOSE.NE.2) VERBOSE = 0
C       ******************************************************************
C-------*    Determine NVIEW_MOD, the number of model projection views   *
C       *     used in GLOBAL mode and check for sampling intervals.      *
C       ******************************************************************
	IF(MODE.EQ.3) THEN					! REFINE MODE
	   NVIEW_MOD = 49			  ! Hardwired for now
	 ELSE   						! GLOBAL MODE
	    print *, 'Warning - resetting mode == 3'
	    mode = 3
	ENDIF
C
	PFTRAD_LO    = infoHead.pftrad_lo
	PFTRAD_HI    = infoHead.pftrad_hi
	PFTRAD_STEP  = infoHead.pftrad_step
	ANN_LO       = infoHead.annulus_low
	ANN_HI       = infoHead.annulus_high
	IF(PFTRAD_LO.LE.1.0)     PFTRAD_LO = 1.0
	IF(PFTRAD_STEP.LE.0.0) PFTRAD_STEP = 1.0
C
	RES_LO       = infoHead.resolution_low
	RES_HI       = infoHead.resolution_high
	JCUT         = infoHead.jcut
	SIGCUT       = infoHead.sigcut
	IC           = infoHead.delta_xy
	SC           = infoHead.num_delta_xy
	IO           = infoHead.delta_omega
	SO           = infoHead.num_delta_omega
	IF(JCUT.EQ.0) JCUT = 1
	IF(JCUT.LT.0) JCUT = 0
C  **************************************************************
C  *    VAROPT IS CURRENTLY NOT RECCOMMENDED BECAUSE OF BUGS    *
C  *    *****     NEED TO TEST THIS !                           *
C  **************************************************************
	VAROPT = .FALSE.
	IF(SIGCUT.GT.0.0) VAROPT = .TRUE.
C
	MAG_CEN      = infoHead.mag_cen
	MAG_STEP     = infoHead.mag_step
	MAG_NUM      = infoHead.mag_num
	MAG_NORM     = infoHead.mag_norm
C
	MAP_FILENAME = infoHead.model_filename
	IF(MODE.GT.0) THEN
	  MAPHEAD.LUNUM = 1
	  READWRITE(1)  = 'r'
	  CALL PIF_OPEN (MAP_FILENAME, MAPHEAD.LUNUM, READWRITE) 
	  CALL PIF_READ_GH (MAPHEAD)
	  IF(MAPHEAD.MODE.EQ.20) THEN
	     MAPHEAD.IMGNUM = 1                              ! LOAD UP "IMG" #1
	     CALL PIF_READ_DH (MAPHEAD)
	     NCOL = MAPHEAD.NUMCOLS
	     NROW = MAPHEAD.NUMROWS
	     NSEC = MAPHEAD.NUMSECS
	   ELSE
	     PRINT *, 'Invalid PIF mode for 3D MAP'
	     CALL TSBEND
	  ENDIF
	ENDIF
C
	PRJ_FILENAME   = infoHead.prj_filename
	IF(PRJ_FILENAME.EQ.' ')  PRJ_FILENAME = 'PFT.PRJS'
C
	PFT_FILENAME   = infoHead.pft_filename
	IF(PFT_FILENAME.EQ.' ')  PFT_FILENAME = 'PFT.PFTS'
C
	pftrads_fname   = infoHead.pftrads_filename
	IF(pftrads_fname.EQ.' ')  pftrads_fname = 'EMPFT.RADS'
C
	pftres1_fname   = infoHead.pftres1_filename
	IF(pftres1_fname.EQ.' ')  pftres1_fname = 'EMPFT.RES1'
C
	pftres2_fname   = infoHead.pftres2_filename
	IF(pftres2_fname.EQ.' ')  pftres2_fname = 'EMPFT.RES2'
	IF(MODE.GT.0) GO TO 100
C       ***************************************************************
C-------*  Need to get NCOL & NROW if reading in stored PRJs/PFTs)    *
C       ***************************************************************
	PRJHEAD.LUNUM = 3
	READWRITE(1)  = 'r'
	CALL PIF_OPEN (PRJ_FILENAME, PRJHEAD.LUNUM, READWRITE)
	CALL PIF_READ_GH (PRJHEAD)
	IF(PRJHEAD.MODE.NE.21) THEN
	  PRINT *, 'Invalid PIF mode for PRJ file'
	  CALL TSBEND
	ENDIF
C       ****************************************************
C-------* NOTE: NSEC must be reduced by one because the    *
C       *       last entry in the PRJ file is PRJ_AVG.     *
C       ****************************************************
	PRJHEAD.IMGNUM = 1
	CALL PIF_READ_DH (PRJHEAD)
	NCOL = PRJHEAD.NUMCOLS
	NROW = PRJHEAD.NUMROWS
	NSEC = PRJHEAD.NUMSECS - 1	     ! Last one is the circular average
	PFTHEAD.LUNUM = 4
	READWRITE(1)  = 'r'
	CALL PIF_OPEN (PFT_FILENAME, PFTHEAD.LUNUM, READWRITE)
	CALL PIF_READ_GH (PFTHEAD)
	IF(PFTHEAD.MODE.NE.22) THEN
	  PRINT *, 'Invalid PIF mode for PFT file'
	  CALL TSBEND
	ENDIF
	PFTHEAD.IMGNUM = 1
	CALL PIF_READ_DH (PFTHEAD)
	NROTD      = PFTHEAD.NUMCOLS
	NANNULID   = PFTHEAD.NUMROWS
	NVIEW_MODD = PFTHEAD.NUMSECS
C  ****************************************************************************
C  *  Input line #11: Parameter data FILENAME  *
C  ********************************************
  100	CALL PFILE_INFO (NCOL, NROW, NSEC, NFILES, NIDS, 
	1 NIDS_TOT, PIXSIZE, IMGHEAD)
C  ************************************************************
C  *    CHECK DEFAULTS FROM INPUT LINES #2-4                  *
C  *                                                          *
C  *    CHECK VALUE OF PFTRAD_HI AGAINST SIZE OF NCOL,NROW    *
C  *     AND USE TO DETERMINE VALUE OF NROT                   *
C  ************************************************************
	RADIUS = (FLOAT(MIN(NCOL-1,NROW-1))/2.0)
	IF(PFTRAD_HI.GT.RADIUS)  PFTRAD_HI = RADIUS
	IF(PFTRAD_HI.LE.0.0)     PFTRAD_HI = RADIUS
	NANNULI = (PFTRAD_HI - PFTRAD_LO + 1.0)/PFTRAD_STEP
	CALL PIRADDEG (PI, RADIAN, DEGREE)
	TEST     = 2*PFTRAD_HI*PI      ! NEED SUFFICIENT SAMPLING AT PFTRAD_HI
	N_POWER2 = 0
	DO WHILE(TEST.GT.1)
	  N_POWER2 = N_POWER2 + 1
	  TEST     = TEST/2.0
	ENDDO
	NROT     = 2**N_POWER2
	DEL_ANG  = 360.0/FLOAT(NROT)
	FFT_NFAC = 1		     ! DOUBLES DEFAULT FFT SIZE FOR CC FUNCTION
	CALL FFT_SETDIM_DEF_SAME (NCOL, NROW, IDIM1, IDIM2, FFT_NFAC)
C
	IF(MODE.GT.0) GO TO 200
C  ********************************************************
C  *    HERE IF READING IN PFT/PRJ DATA FROM DISK.        *
C  *    CHECK TO SEE THAT PFT/PRJ DATA FILES AGREE WITH   *
C  *     INPUTS FROM COMMAND BATCH JOB.                   *
C  ********************************************************
C	better check out use of nrot
c
	IF(NROTD*4 .NE. NROT .OR. NANNULID .NE. NANNULI 
	1  .OR. NVIEW_MODD .NE. NVIEW_MOD) THEN
	   PRINT *, 'NROT,NANNULI,NVIEW_MOD - ', NROT,NANNULI,NVIEW_MOD
	   PRINT *, 'NROTD,NANNULID,NVIEW_MODD - ', NROTD,NANNULID,
	1	NVIEW_MODD
	   PRINT *, 'Mismatch between PFT file and batch job params.'
	  CALL TSBEND
	ENDIF
  200	PFTRAD_STEP = 1.0
	DEL_PIXRES  = 0.025
	IF(JCUT.GE.NROT/4)   JCUT  = (NROT/4) - 1
	IF(PIXSIZE.LE.0.0) PIXSIZE = 1.0		  ! PIXEL UNITS ASSUMED
	RES_TST = FLOAT(IDIM1/2)*PIXSIZE
	IF(RES_LO.LE.0.0 .OR. RES_LO.GT.RES_TST) RES_LO = RES_TST
	RES_TST = 2.0*PIXSIZE
	IF(RES_HI.LT.RES_TST) RES_HI = RES_TST
	PIXRES_MIN = RES_LO/PIXSIZE
	PIXRES_MAX = RES_HI/PIXSIZE
	RES_MIN    = FLOAT(IDIM1)/PIXRES_MIN
	RES_MAX    = FLOAT(IDIM1)/PIXRES_MAX
	IF(VERBOSE.LT.1) GO TO 300
C  ******************************************
C  *  Extra output files if VERBOSE = 1 or 2  *
C  ******************************************
	OPEN (UNIT=8, FILE=pftres1_fname, STATUS='NEW',
	1     ACCESS='SEQUENTIAL', FORM='FORMATTED')
	WRITE(8,'(2A,/,A)') 'Correlation coefficients as function of',
	1 ' resolution (1/pixel) for unmasked data',
	2 ' (computed after X,Y origin refinement)'
	WRITE(8,'(A,10F7.3)')  ' IMG', (I*DEL_PIXRES, I=1,10)
	WRITE(8,1000)
	OPEN (UNIT=9, FILE=pftrads_fname, STATUS='NEW',
	1     ACCESS='SEQUENTIAL', FORM='FORMATTED')
	WRITE(9,'(2A,/,A)')' Correlation coefficients as a function of',
	1 ' radius', '  (computed after X,Y origin refinement)'
	OPEN (UNIT=10, FILE=pftres2_fname, STATUS='NEW',
	1     ACCESS='SEQUENTIAL', FORM='FORMATTED')
	WRITE(10,'(2A,/,A)') ' Correlation coefficients as function',
	1  ' of resolution (1/pixel) for ANN_LO to ANN_HI masked data',
	2  ' (computed after X,Y origin refinement)'
	WRITE(10,'(A,10F7.3)')  ' IMG', (I*DEL_PIXRES, I=1,10)
	WRITE(10,1000)
  300	IF(MAG_NUM.LE.0) MAG_NUM = 1
	IF(MOD(MAG_NUM,2).EQ.0) MAG_NUM = MAG_NUM + 1
	IF(IABS(MODE).EQ.1 .AND. MAG_CEN.LE.0.0) THEN
	  MAG_CEN = 1.0
	  Write(*,*)
	1  'GLOBAL SEARCH mode - you have no MAG values to read in'
	ENDIF
C  ******************************************************************
C  *    BEFORE EXITING THIS ROUTINE, THIS IS WHERE WE MAKE SURE     *
C  *    EVERTHING IS TREATED CORRECTLY TO ACCOMODATE BINNING.       *
C  ******************************************************************
	NCOLBIN = (NCOL/BIN) + 1
	NROWBIN = (NROW/BIN) + 1
	NSECBIN = (NSEC/BIN) + 1
	IF(MOD(NCOLBIN,2).EQ.0) NCOLBIN = NCOLBIN - 1
	IF(MOD(NROWBIN,2).EQ.0) NROWBIN = NROWBIN - 1
	IF(MOD(NSECBIN,2).EQ.0) NSECBIN = NSECBIN - 1
	NROT       = NROT/BIN
	IDIM1      = IDIM1/BIN
	IDIM2      = IDIM2/BIN
	NANNULI    = NANNULI/BIN
	NPIX       = NROT*NANNULI/4
	DEL_ANG    = DEL_ANG*FLOAT(BIN)
	ANN_LO     = ANN_LO/BIN
	ANN_HI     = ANN_HI/BIN
	RADIUS     = FLOAT(NANNULI)
	PFTRAD_LO  = PFTRAD_LO/FLOAT(BIN)
	IF(PFTRAD_LO.LT.1.0) PFTRAD_LO = 1.0
	PFTRAD_HI  = PFTRAD_HI/FLOAT(BIN)
	XCEN_MODEL = (NCOLBIN+1)/2
	YCEN_MODEL = (NROWBIN+1)/2
	IF(ANN_LO.LT.1)       ANN_LO = 0
	IF(ANN_LO.GE.NANNULI) ANN_LO = NANNULI - 1
	IF(ANN_HI.LE.0)       ANN_HI = NANNULI - 1
	IF(ANN_HI.GE.NANNULI) ANN_HI = NANNULI - 1
	WRITE(*,1001) ' EMPFT parameters'
	WRITE(*,1001) ' ------------------------------------------------'
	WRITE(*,1001) ' BIN factor          ', BIN
	WRITE(*,1001) ' Symmetry            ', SYM
	IF(MODE.EQ.3) THEN
	   WRITE(*,1001) ' # model views/window', NVIEW_MOD
	 ELSE
	   WRITE(*,1001) ' # model views/asu   ', NVIEW_MOD
	ENDIF
	WRITE(*,1001) ' # particle images   ', NIDS_TOT
	WRITE(*,1001) ' FFT_DIM1,FFT_DIM2   ', IDIM1, IDIM2
	WRITE(*,1002) ' RES_LO,RES_HI       ', RES_LO, RES_HI
	WRITE(*,1002) ' RES_MIN,RES_MAX     ', RES_MIN, RES_MAX
	WRITE(*,1002) ' PIXRES_MIN/MAX      ', PIXRES_MIN, PIXRES_MAX
	WRITE(*,1002) ' Pixel size          ', PIXSIZE
	WRITE(*,1001) ' NANNULI             ', NANNULI
	WRITE(*,1002) ' PFTRAD_LO,PFTRAD_HI ', PFTRAD_LO, PFTRAD_HI
	WRITE(*,1001) ' ANN_LO,ANN_HI       ', ANN_LO, ANN_HI
	WRITE(*,1001) ' NROT                ', NROT
	WRITE(*,1002) ' Delta omega         ', DEL_ANG
	WRITE(*,1002) ' Model XCEN,YCEN     ', XCEN_MODEL, YCEN_MODEL
	WRITE(*,1003) ' STEP & INCREMENT  ', IC, SC, IO, SO
	WRITE(*,1002) ' MAG(_CEN,_STEP,_NUM)', MAG_CEN, MAG_STEP,
	1    MAG_NUM
	IF(MAG_CEN.GT.0.0) THEN
	   MAG_LO = MAG_CEN - (((MAG_NUM-1)/2)*MAG_STEP)
	   MAG_HI = MAG_CEN + (((MAG_NUM-1)/2)*MAG_STEP)
	   WRITE(*,1002) ' MAG range(LO,HI)    ', MAG_LO, MAG_HI
	 ELSE
	   WRITE(*,'(2A)') ' MAG factor centered on input values from',
	1 ' data file'
	ENDIF
	WRITE(*,1001) '------------------------------------------------'
	RETURN
C
  999	PRINT *, '************************************************'
	PRINT *, '* BATCH COMMAND JOB input ERROR - STOP program *'
	CALL TSBEND
 1000	FORMAT(74('-'))
 1001	FORMAT(A,2I9)
 1002	FORMAT(A,2F9.3,I9)
 1003	FORMAT(A,F9.2,i3,f9.2,i3)
	END


C  ****************************************************************************
C  *                                                                          *
C  *                *****     SUBROUTINE LIST_CCS_F4     *****                *
C  *                                                                          *
C  *    COMPUTES AND THEN LISTS CORRELATION COEFFICIENTS AS A FUNCTION OF     *
C  *     PARTICLE RADIUS AND RESOLUTION.                                      *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    MODIFIED  4-JUN-1998 BY TSB (Use BIN to get correct radii).           *
C  *    LAST MODIFIED  20-DEC-2002 BY MBS (IMG -> IMG_F4, rid of R4_IMG).     *
C  ****************************************************************************
C
	SUBROUTINE LIST_CCS_F4 (NCOL, NROW, IDIM1, IDIM2, NROT, 
	1 NANNULI, IMG_F4, MSKIMG,  FFT1, FFT3, FFT4, FFT5, 
	2 PRJ_ROT, MSKPRJ, PPRJ, PMAP, FLTPMAP, IMG_NUM, IMG_BKG, 
	3 XCEN, YCEN, XCEN_MOD, YCEN_MOD, FFT_ORIGX, FFT_ORIGY, 
	4 DEL_PIXRES, CC_AVG, CC_RES_GLOBAL, CC_RAD_GLOBAL)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NCOL, NROW, IDIM1, IDIM2, NROT, NANNULI, IMG_NUM
	INTEGER*4 I, J
	real*4    IMG_BKG                                       !MBS 08.19.03
	REAL*4    MSKIMG(NCOL,NROW)                                   !MBS 12.12.02
	REAL*4    IMG_F4(NCOL,NROW)                                   !MBS 12.12.02
	REAL*4    XCEN, YCEN, XCEN_MOD, YCEN_MOD, FFT_ORIGX, FFT_ORIGY
	REAL*4    CCRES(2,20), CCRAD(2,200), DEL_PIXRES
	REAL*4    CC_RES_GLOBAL(2,20), CC_AVG, CC_RAD_GLOBAL(2,200)
	REAL*4    SUMRSQ, FFT_FZERO
	REAL*4    FFT1(2*IDIM1,IDIM2), FFT3(2*IDIM1,IDIM2)
	REAL*4    FFT4(2*IDIM1,IDIM2), FFT5(2*IDIM1,IDIM2)
	REAL*4    MSKPRJ(NCOL,NROW)
	REAL*4    PRJ_ROT(NCOL,NROW), FLTPMAP(NROT,NANNULI)
	REAL*4    PMAP(NROT,NANNULI), PPRJ(NROT,NANNULI)
	REAL*4    ANNLO, ANNHI, RAD_pol, DELANG, RESMIN, RESMAX
	LOGICAL   FFT_FULL
C  ----------------------------------------------------------------------------
C

	ANNLO   = ANN_LO
	ANNHI   = ANN_HI
	RAD_pol = RADIUS
	DELANG  = DEL_ANG
	RESMIN  = RES_MIN
	RESMAX  = RES_MAX
	CALL COPY_R4 (IMG_F4, MSKIMG, NCOL*NROW)                  !MBS 12.12.02
	CALL MAP_2DMASK (MSKIMG, NCOL, NROW, ANNLO, 
	1 ANNHI, XCEN, YCEN, IMG_BKG)                             !MBS 12.12.02
	CALL MAP_FFT_FILL (MSKIMG, NCOL, NROW, 1, 1, FFT4, IDIM1, IDIM2, 
	1 FFT_FZERO, FFT_ORIGX, FFT_ORIGY, FFT_FULL, 0)           !MBS 12.12.02
	CALL FFT_2D (FFT4, IDIM1, IDIM2)
	CALL COPY_R4 (PRJ_ROT, MSKPRJ, NCOL*NROW)
	CALL MAP_2DMASK (MSKPRJ, NCOL, NROW,ANNLO,ANNHI,
	1 XCEN_MOD, YCEN_MOD, 0.0)
	CALL MAP_FFT_FILL (MSKPRJ, NCOL, NROW, 1, 1, FFT5, IDIM1, IDIM2, 
	2 FFT_FZERO, FFT_ORIGX, FFT_ORIGY, FFT_FULL, 0)
	CALL FFT_2D (FFT5, IDIM1, IDIM2)
	CALL PFTCC_RES (FFT1, FFT3, FFT4, FFT5, IDIM1, IDIM2, CCRES, 
	2 DEL_PIXRES)
c	CALL FFT_FLT (FFT1, IDIM1, IDIM2, RESMIN, RESMAX, 0.0, 0.0, .TRUE.)
	CALL FFT_2DBT_R4 (FFT1, IDIM1, IDIM2, IMG_F4, NCOL, NROW,
	1 FFT_FULL, 0)                                            !MBS 12.12.02
	CALL MAP_POLAR (IMG_F4, NCOL, NROW, FLTPMAP, NROT, NANNULI, 
	1 XCEN,YCEN, 1.0, RAD_pol, DELANG)                        !MBS 12.12.02
	CALL MAP_POLAR (PRJ_ROT, NCOL, NROW, PPRJ, NROT, NANNULI, 
	1 XCEN_MOD, YCEN_MOD, 1.0, RAD_pol, DELANG)
	CALL CC_RADII (PMAP, FLTPMAP, PPRJ, NROT, NANNULI, CCRAD)
C
	SUMRSQ = 0.0
	DO I = 1,ANN_HI+1
	  CC_RAD_GLOBAL(1,I) = CC_RAD_GLOBAL(1,I) + CCRAD(1,I)
	  CC_RAD_GLOBAL(2,I) = CC_RAD_GLOBAL(2,I) + CCRAD(2,I)
	  IF(I.GE.ANN_LO) THEN
	    SUMRSQ = SUMRSQ + (BIN*(I-1))**2
	    CC_AVG = CC_AVG + CCRAD(2,I)*((BIN*(I-1))**2)
	  ENDIF
	ENDDO
	IF(SUMRSQ.GT.0.0) THEN
	   CC_AVG = CC_AVG/SUMRSQ
	 ELSE
	   CC_AVG = 0.0
	ENDIF
	J = NINT(1/(PIXRES_MAX*DEL_PIXRES))
	DO I = 1,J
	  CC_RES_GLOBAL(1,I) = CC_RES_GLOBAL(1,I) + CCRES(1,I)
	  CC_RES_GLOBAL(2,I) = CC_RES_GLOBAL(2,I) + CCRES(2,I)
	ENDDO
	WRITE (8,1000) IMG_NUM, (CCRES(1,I),I=1,J)
	WRITE(10,1000) IMG_NUM, (CCRES(2,I),I=1,J)
	WRITE (9,1010) ' CCs vs radius for particle', IMG_NUM
	DO I = 1,ANN_HI+1
	  J = (I-1)*BIN
	  WRITE (9,1020) J, CCRAD(1,I), CCRAD(2,I)
	ENDDO
	WRITE (9,1030) CC_AVG, ' = Average CC for ANN_LO,ANN_HI=', 
	1 ANN_LO, ANN_HI
C
 1000	FORMAT(I6,20F7.3)
 1010	FORMAT(A,I6)
 1020	FORMAT(I5,2F10.3)
 1030	FORMAT(F7.3,A,2I4)
	RETURN
	END
C  ************************************************************************
C  *								          *
C  *              *****     SUBROUTINE PFILE_INFO     *****               *
C  *								          *
C  *    READS IN  ALL PARAMETER FILENAMES AND CHECKS THE CORRESPONDING    *
C  *     IMAGE FILES TO BE SURE THE IMAGE AND MODEL DATA HAVE THE         *
C  *     SAME DIMENSIONS.                                                 *
C  *								          *
C  *	Prompts user for a parameter data file, opens file, counts        *
C  *	number of particles, finds highest particle number, gets          *
C  *	image filename from first line of file.			          *
C  *								          *
C  *	Adapted from PFT_GETPFILE written by DMB, May 1995.               *
C  *								          *
C  *	MODIFIED 19-JAN-1998 BY T. BAKER (Multiple inout files).          *
C  *	MODIFIED 22-JAN-1998 BY T. BAKER (CTF parameters).                *
C  *	MODIFIED  3-FEB-1998 BY T. BAKER (Save IMG_FILENAMES).            *
C  *    MODIFIED  6-MAR-1998 BY TSB (Make UNIX compatible).               *
C  *    MODIFIED  5-MAY-1998 BY TSB (Pass min pixel size to INFO).        *
C  *    MODIFIED 25-AUG-1999 BY TSB (Add UNITS variable).                 *
C  *    LAST MODIFIED 7-MAR-2003 BY MBS (rid of pifhead, add imghead      *
C  *    to params).                                                       *
C  ************************************************************************
C
	SUBROUTINE PFILE_INFO (NCOL, NROW, NSEC, NFILES, NIDS, 
	1 NIDS_TOT, PIXSIZE, IMGHEAD)
C
	IMPLICIT NONE
C
	include 'PIFHEAD.INC'
	RECORD /PIFHEAD_T/ IMGHEAD
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1 FOCUS_MIN, ANG_MAJ, CS_COEFF
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NIMG, LUN, MAX_ID, ID, NIDS_TOT
	INTEGER*4 NCOL, NROW, NSEC, NCIMG, NRIMG, NFILES
	INTEGER*4 NIDS(999), UNITS(999)
	REAL*4    IMG_SCALEF, PIXSIZ(999), VOLTS(999), AMP_FAC(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999), ANG_MAJ(999)
	REAL*4    CS_COEFF(999), PIXSIZE
	CHARACTER FNAME*256
	BYTE      READWRITE(4)
C------------------------------------------------------------------------------
C
	NFILES   = 0			 ! COUNT THE NUMBER OF PARAMETER FILES
	NIDS_TOT = 0			 ! COUNT THE NUMBER OF IMAGES TO SEARCH
	PIXSIZE  = 1000000.0		 ! Minimum pixel size 
C
  	WRITE(*,'(A)') '$PARAMETER DATA FILES: '
  100	READ (*,'(A)',ERR=950,END=999) FNAME
C  *************************************************************
C  *  Open file and read first line, make sure it is not data  *
C  *************************************************************
	IF(NFILES.NE.0) THEN
	  CLOSE (UNIT=2, DISP='KEEP')
	  CALL PIF_CLOSE (7)
	ENDIF
	OPEN (UNIT=2, FILE=FNAME, STATUS='OLD', ERR=960,
	1     ACCESS='SEQUENTIAL', FORM='FORMATTED', READONLY)
	NFILES = NFILES + 1
	IF(NFILES.GT.999) GO TO 940
	READ (2,'(A)',ERR=970) IMG_FNAMES(NFILES)
Ca	READ (2,'(F,I,6F)',ERR=975) PIXSIZ(NFILES), UNITS(NFILES), 
	READ (2,*,ERR=975) PIXSIZ(NFILES), UNITS(NFILES), 
	1 VOLTS(NFILES), AMP_FAC(NFILES), FOCUS_MAJ(NFILES), 
	2 FOCUS_MIN(NFILES), ANG_MAJ(NFILES), CS_COEFF(NFILES)
	PARAM_FNAMES(NFILES) = FNAME
	IF(PIXSIZ(NFILES).LT.PIXSIZE) PIXSIZE = PIXSIZ(NFILES)
	PRINT *
	WRITE(*,'(A,I4,2A)') ' File #', NFILES, ' is: ',
	1    IMG_FNAMES(NFILES)
C  ************************************************************
C  *    OPEN UP PIF FORMAT IMAGE FILE AND CHECK TO MAKE SURE  *
C  *     MODEL MAP DATA SIZE AND IMAGE SIZE ARE EQUIVALENT    *
C  ************************************************************
	IMGHEAD.LUNUM = 7
	READWRITE(1)  = 'r'
	CALL PIF_OPEN (IMG_FNAMES(NFILES), IMGHEAD.LUNUM, READWRITE)
	CALL PIF_READ_GH (IMGHEAD)
	NIMG  = IMGHEAD.NIMG
	NCIMG = IMGHEAD.NUMCOLS
        NRIMG = IMGHEAD.NUMROWS
	IF(NCIMG.NE.NCOL .OR. NRIMG.NE.NROW) THEN
	  PRINT *, '=================================================='
	  PRINT *, '|  ERROR!!  Sizes of MAP and IMAGE are unequal.  |'
	  PRINT *, '|  Make them equivalent then rerun program.      |'
	  PRINT *, '=================================================='
	  PRINT *, 'NCOL (map)   = ', NCOL
	  PRINT *, 'NROW (map)   = ', NROW
	  PRINT *, 'NSEC (map)   = ', NSEC
	  PRINT *, 'NCOL (image) = ', NCIMG
	  PRINT *, 'NROW (image) = ', NRIMG
	  CALL TSBEND
	ENDIF
	IF(IABS(MODE).EQ.1) THEN
	  NIDS(NFILES) = NIMG
	  MAX_ID       = NIMG
	  NIDS_TOT     = NIDS_TOT + NIMG
	  GO TO 300
	ENDIF
C  ********************************************************************
C  *  Count particles in parameter file, find one with highest value  *
C  ********************************************************************
	NIDS(NFILES) = 0
	MAX_ID       = 0
  200	READ (2,'(I)',ERR=980,END=300) ID
	NIDS_TOT     = NIDS_TOT + 1
	NIDS(NFILES) = NIDS(NFILES) + 1
	MAX_ID       = MAX(MAX_ID, ID)
	GO TO 200
C  ****************************************************
C  *    HERE WHEN AT END OF CURRENT PARAMETER FILE    *
C  ****************************************************
  300	IF(NIDS(NFILES).EQ.0)  GO TO 990
	GO TO 100			! CHECK FOR ADDITIONAL PARAMETER FILES
C  ***************************************************************************
C  *    ERROR MESSAGES    *
C  ************************
  940	PRINT *, '***********************************************'
	PRINT *, '* More than 999 parameter files: EXIT program *'
	CALL TSBEND
  950	PRINT *, '**************************************************'
	PRINT *, '* Error reading parameter filename: EXIT program *'
	CALL TSBEND
  960	PRINT *, '**********************************************'
	PRINT *, '* Error opening parameter file: EXIT program *'
	CALL TSBEND
  970	PRINT *, '**********************************************'
	PRINT *, '* Error reading IMAGE filename: EXIT program *'
	CALL TSBEND
  975	PRINT *, '**********************************************'
	PRINT *, '* READ ERROR: 2nd record of PARAMETER file   *'
	PRINT *, '* HINT: Did you specify UNITS?               *'
	CALL TSBEND
  980	PRINT *, '*******************************************'
	PRINT *, '* Error reading particle ID: EXIT program *'
	CALL TSBEND
  990	PRINT *, '*********************************************'
	PRINT *, '* PARAM file contains no data: EXIT program *'
	CALL TSBEND
  999	WRITE(*,'(I5,A,I6,A)') NFILES, ' PARAMETER files and total of', 
	1 NIDS_TOT, ' images for PFT search.'
	CLOSE (UNIT=2, DISP='KEEP')
	CALL PIF_CLOSE (7)
	RETURN
C
	END


C  ****************************************************************************
C  *                                                                          *
C  *                      *****   FUNCTION PHIMAX   *****                     *
C  *                                                                          *
C  *  Determines PHIMAX based on SYM chosen			              *
C  *                                                                          *
C  *  PURDUE UNIVERSITY DEC 3000, Jan,Jun 1994, T. Baker and D. Belnap        *
C  *  Added 222, 522, 622, & 722 symmetries, Dec 1996, Jan 1997, D. Belnap    *
C  *  Added 2-20 n-fold symmetries, 30 Jan 1997, D. Belnap                    *
C  *    LAST MODIFIED 24-APR-1998 BY T. BAKER (Renamed and added more SYM.)   *
C  ****************************************************************************
C
	REAL*4 FUNCTION PHIMAX (SYM, THETA)
C
	IMPLICIT NONE
C
	INTEGER*4 SYM
	REAL*4    THETA
C -----------------------------------------------------------------------------
C
	PHIMAX = 180.0				      ! FOR ASYMMETRIC OBJECTS
	IF (SYM.EQ.532) THEN
	   CALL ICO_CALCPHIMAX (THETA, PHIMAX)
	 ELSE IF (SYM.GE.2 .AND. SYM.LE.20) THEN
	   PHIMAX = 360.0 / FLOAT(SYM*2)
	 ELSE IF (SYM.EQ.222) THEN
	   PHIMAX = 360.0/4.0				    ! = 90.0 degrees
	 ELSE IF (SYM.EQ.32) THEN
	   PHIMAX = 360.0/6.0				    ! = 60.0 degrees
	 ELSE IF (SYM.EQ.422) THEN
	   PHIMAX = 360.0/8.0				    ! = 45.0 degrees
	 ELSE IF (SYM.EQ.52) THEN
	   PHIMAX = 360.0/10.0				    ! = 36.0 degrees
	 ELSE IF (SYM.EQ.622) THEN
	   PHIMAX = 360.0/12.0				    ! = 30.0 degrees
	 ELSE IF (SYM.EQ.72) THEN
	   PHIMAX = 360.0/14.0				    ! = 25.7 degrees
	 ELSE IF (SYM.EQ.822) THEN
	   PHIMAX = 360.0/16.0				    ! = 22.5 degrees
	 ELSE IF (SYM.EQ.92) THEN
	   PHIMAX = 360.0/18.0				    ! = 20.0 degrees
	ENDIF
C
	RETURN
	END


C  ****************************************************************************
C  *                                                                          *
C  *                *****     SUBROUTINE WRITE_PARAMS     *****               *
C  *                                                                          *
C  *    WRITES OUT THE CURRENT PARAMETER ARRAYS TO NEW PARAMETER FILES        *
C  *     FILES (UNIT = 11).                                                   *
C  *    ALSO NORMALIZES THE MAG SCALE FACTORS IF MAG_NORM SET = TO 1.         *
C  *                                                                          *
C  *    FEBRUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.       *
C  *    LAST MODIFIED  6-FEB-1998 BY T. BAKER (Add MAG normalization).        *
C  ****************************************************************************
C
	SUBROUTINE WRITE_PARAMS (NFILES, NIDS_TOT, FIDS, IDS, TS,PS,OS, 
	1 XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS)
C	
	IMPLICIT NONE
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1 FOCUS_MIN, ANG_MAJ, CS_COEFF
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*2 STRING_LEN
	INTEGER*4 NFILES, NIDS_TOT, FILE, FILE_LAST
	INTEGER*4 ID, FIDS(NIDS_TOT), IDS(NIDS_TOT), UNITS(999)
	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999)
	REAL*4    ANG_MAJ(999), MAG_FACTOR
	REAL*4    TS(NIDS_TOT), PS(NIDS_TOT), OS(NIDS_TOT)
	REAL*4    XS(NIDS_TOT), YS(NIDS_TOT), MAGS(NIDS_TOT)
	REAL*4    PFT_CCS(NIDS_TOT), PRJ_CCS(NIDS_TOT)
	REAL*4    CMP_CCS(NIDS_TOT)
	CHARACTER IMG_FILENAME*256, NEW_FNAME*256
C  ----------------------------------------------------------------------------
C
C  *****************************************************
C  *    NORMALIZE MAG SCALE FACTORS IF MAG_NORM = 1    *
C  *****************************************************
	IF(MAG_NORM.EQ.1) THEN
	  MAG_FACTOR = 0.0
	  DO ID = 1, NIDS_TOT
	    MAG_FACTOR = MAG_FACTOR + MAGS(ID)
	  ENDDO
	  MAG_FACTOR = MAG_FACTOR/FLOAT(NIDS_TOT)
	  WRITE(*,'(A,F8.4,A)') ' Mean of MAG scale factors (=', 
	1  MAG_FACTOR,') normalized to 1.0'
	  MAG_FACTOR = 1.0/MAG_FACTOR
	  DO ID = 1, NIDS_TOT
	    MAGS(ID) = MAGS(ID)*MAG_FACTOR
	  ENDDO
	ENDIF
C
	CALL INCREMENT_FNAME (PARAM_FNAMES(1), NEW_FNAME)
	OPEN (UNIT=11, FILE=NEW_FNAME, STATUS='UNKNOWN',
	1 ACCESS='SEQUENTIAL', FORM='FORMATTED')
	IMG_FILENAME = IMG_FNAMES(1)
	IF(IMG_FILENAME(1:1).EQ.' ') THEN
	   WRITE(11,'(A)') IMG_FILENAME(1:STRING_LEN(IMG_FILENAME))
	 ELSE
	   WRITE(11,'(A)') ' '//IMG_FILENAME(1:STRING_LEN(IMG_FILENAME))
	ENDIF
	WRITE(11,1010) PIXSIZ(1), UNITS(1), VOLTS(1), AMP_FAC(1), 
	1 FOCUS_MAJ(1), FOCUS_MIN(1), ANG_MAJ(1), CS_COEFF(1)
	FILE_LAST = FIDS(1)
C
	DO ID = 1,NIDS_TOT
	  FILE = FIDS(ID)
	  IF(FILE_LAST.NE.FILE) THEN
	    CLOSE (UNIT=11, DISP='KEEP')
	    CALL INCREMENT_FNAME (PARAM_FNAMES(FILE), NEW_FNAME)
c       MBS Beware of "unknown" status of dat files!
	    OPEN (UNIT=11, FILE=NEW_FNAME, STATUS='unknown',
	1         ACCESS='SEQUENTIAL', FORM='FORMATTED')
	    IMG_FILENAME = IMG_FNAMES(FILE)
	    IF(IMG_FILENAME(1:1).EQ.' ') THEN
	       WRITE(11,'(A)') IMG_FILENAME(1:STRING_LEN(IMG_FILENAME))
	     ELSE
	       WRITE(11,'(A)') ' '//
	1	     IMG_FILENAME(1:STRING_LEN(IMG_FILENAME))
	    ENDIF
	    WRITE(11,1010) PIXSIZ(FILE), UNITS(FILE), VOLTS(FILE), 
	1    AMP_FAC(FILE), FOCUS_MAJ(FILE), FOCUS_MIN(FILE), 
	2    ANG_MAJ(FILE), CS_COEFF(FILE)
	    FILE_LAST = FIDS(ID)
	  ENDIF
	  WRITE(11,1020) IDS(ID), TS(ID), PS(ID), OS(ID), XS(ID), YS(ID),
	1  MAGS(ID), PFT_CCS(ID), PRJ_CCS(ID), CMP_CCS(ID)
	ENDDO
	CLOSE (UNIT=11, DISP='KEEP')
C
 	RETURN
 1000	FORMAT (1X,I6,',',4(F8.3,','),F7.3,',',3(F6.3,','),F6.3)
 1010	FORMAT (1X,F8.4,',',I2,',',F10.1,',',3(F10.4,','),
	1    F10.4,',',F8.2)
 1020	FORMAT (1X,I6,',',4(F8.3,','),F7.3,',',3(F6.3,','),F6.3)
C
	END


C*****************************************************************************
C*                                                                           *
C*                  *****     SUBROUTINE CC_RADII     *****                  *
C*                                                                           *
C*    FIND CORRELATION COEFICIENTS AS FUNCTION OF ANNULUS(RADIUS).           *
C*                                                                           *
C*    NOTE: REAL SPACE CORRELATION COEFFICIENTS ARE USED.                    *
C*                                                                           *
C*    CCRAD(1,N) = Correlation coefficient: PMAP vs. POLAR_PRJ.              *
C*    CCRAD(2,N) = Correlation coefficient: filtered PMAP vs. POLAR_PRJ.     *
C*                                                                           *
C*    FEBRUARY 1993  PURDUE UNIVERSITY VAX 8550, T. BAKER.                   *
C*    LAST MODIFIED  4-JUN-1998 BY TSB (Renamed CC_RADII).                   *
C*****************************************************************************
C
	SUBROUTINE CC_RADII (PMAP, FLTPMAP, POLAR_PRJ, NROT, NANNULI, 
	1 CCRAD)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NROT, NANNULI, I, II, J
	REAL*8    SUM12(2,200), SQ1(200), SQ2(200), SQ3(200)
	REAL*4    PMAP(NROT,NANNULI), POLAR_PRJ(NROT,NANNULI)
	REAL*4    FLTPMAP(NROT,NANNULI), DIV1, DIV2, DIV3
	REAL*4    CCRAD(2,200), AVG1(200), AVG2(200), AVG3(200)
C  ----------------------------------------------------------------------------
C
	DO I = 1,MIN(NANNULI, ANN_HI+1)
	  AVG1(I) = 0.0
	  AVG2(I) = 0.0
	  AVG3(I) = 0.0
	  SQ1(I)  = 0.0
	  SQ2(I)  = 0.0
	  SQ3(I)  = 0.0
	  DO II = 1,2
	    CCRAD(II,I) = 0.0
	    SUM12(II,I) = 0.0
	  ENDDO
	  DO J = 1,NROT
	    SUM12(1,I) = SUM12(1,I) + PMAP(J,I)*POLAR_PRJ(J,I)
	    SUM12(2,I) = SUM12(2,I) + FLTPMAP(J,I)*POLAR_PRJ(J,I)
	    AVG1(I)    = AVG1(I) + PMAP(J,I)
	    AVG2(I)    = AVG2(I) + POLAR_PRJ(J,I)
	    AVG3(I)    = AVG3(I) + FLTPMAP(J,I)
	    SQ1(I)     = SQ1(I) + PMAP(J,I)**2
	    SQ2(I)     = SQ2(I) + POLAR_PRJ(J,I)**2
	    SQ3(I)     = SQ3(I) + FLTPMAP(J,I)**2
	  ENDDO
	  AVG1(I) = AVG1(I)/FLOAT(NROT)
	  AVG2(I) = AVG2(I)/FLOAT(NROT)
	  AVG3(I) = AVG3(I)/FLOAT(NROT)
	  DIV1    = SQ1(I) - FLOAT(NROT)*(AVG1(I)**2)
	  DIV2    = SQ2(I) - FLOAT(NROT)*(AVG2(I)**2)
	  DIV3    = SQ3(I) - FLOAT(NROT)*(AVG3(I)**2)
	  DIV1    = sign(SQRT(abs(DIV1)), DIV1)
	  DIV2    = sign(SQRT(abs(DIV2)), DIV2)
	  DIV3    = sign(SQRT(abs(DIV3)), DIV3)
	  IF(DIV1*DIV2.LE.0.0) THEN
	     CCRAD(1,I) = 0.0
	   ELSE
	     CCRAD(1,I) = (SUM12(1,I)-AVG1(I)*AVG2(I)*FLOAT(NROT)) /
	1                           (DIV1*DIV2)
	  ENDIF
	  IF(DIV3*DIV2.LE.0.0) THEN
	     CCRAD(2,I) = 0.0
	   ELSE
	     CCRAD(2,I) = (SUM12(2,I)-AVG3(I)*AVG2(I)*FLOAT(NROT)) /
	1                           (DIV3*DIV2)
	  ENDIF
	ENDDO
C
	RETURN
	END


	SUBROUTINE PFTCC_RES (FFT1, FFT3, FFT4, FFT5, IDIM1, IDIM2, CC,
	1 DEL_PIXRES)
C
	IMPLICIT NONE
C
	INTEGER*4 IDIM1, IDIM2, IX, IY, XX, BIN, IXTEST
	INTEGER*4 I, J, IX2, IY2
	REAL*4    CC(2,20), RES, YSQ
	REAL*4    FFT1(2*IDIM1,IDIM2), FFT3(2*IDIM1,IDIM2)
	REAL*4    FFT4(2*IDIM1,IDIM2), FFT5(2*IDIM1,IDIM2)
	REAL*4    DEL_PIXRES, SUM13(20), SUM45(20), PTS(20)
	REAL*4    AVG1(20), AVG3(20), AVG4(20), AVG5(20)
	REAL*4    SQ1(20), SQ3(20), SQ4(20), SQ5(20)
	REAL*4    DIV1, DIV3, DIV4, DIV5, AMP1, AMP3, AMP4, AMP5
C  ----------------------------------------------------------------------------
C
	IXTEST  = IDIM1/2
C
C  ***************************
C  *    INITIALIZE ARRAYS    *
C  ***************************
	DO I = 1,20
	  DO J = 1,2
	    CC(J,I)    = 0.0
	  ENDDO
	  PTS(I)   = 0.0
	  SUM13(I) = 0.0
	  SUM45(I) = 0.0
	  AVG1(I)  = 0.0
	  AVG3(I)  = 0.0
	  AVG4(I)  = 0.0
	  AVG5(I)  = 0.0
	  SQ1(I)   = 0.0
	  SQ3(I)   = 0.0
	  SQ4(I)   = 0.0
	  SQ5(I)   = 0.0
	ENDDO
C
	IY2 = (IDIM2/2) + 1
	DO IY = 1,IY2
	  YSQ = FLOAT(IY-1)**2
	  IX2 = IDIM1
	  IF(IY.EQ.1 .OR. IY.EQ.IY2) THEN
	    IX2 = IDIM1/2
	  ENDIF
	  DO IX = 1,IX2
	    IF(IY.EQ.1 .AND. IX.EQ.1) GO TO 100
	    XX = (IX-1)*2 + 1
	    IF(IX.GT.IXTEST) THEN
	       RES = SQRT ( YSQ + FLOAT(IDIM1-IX+1)**2 ) 
	     ELSE
	       RES = SQRT ( YSQ + FLOAT(IX-1)**2 )
	    ENDIF
	    AMP1 = SQRT(FFT1(XX,IY)**2 + FFT1(XX+1,IY)**2)
	    AMP3 = SQRT(FFT3(XX,IY)**2 + FFT3(XX+1,IY)**2)
	    AMP4 = SQRT(FFT4(XX,IY)**2 + FFT4(XX+1,IY)**2)
	    AMP5 = SQRT(FFT5(XX,IY)**2 + FFT5(XX+1,IY)**2)
	    BIN        = INT( RES/(IDIM1*DEL_PIXRES) ) + 1
	    IF(BIN.GT.20) GO TO 100
	    PTS(BIN)   = PTS(BIN) + 1.0
	    SUM13(BIN) = SUM13(BIN) + AMP1*AMP3
	    SUM45(BIN) = SUM45(BIN) + AMP4*AMP5
	    AVG1(BIN)  = AVG1(BIN) + AMP1
	    AVG3(BIN)  = AVG3(BIN) + AMP3
	    AVG4(BIN)  = AVG4(BIN) + AMP4
	    AVG5(BIN)  = AVG5(BIN) + AMP5
	    SQ1(BIN)   = SQ1(BIN) + AMP1**2
	    SQ3(BIN)   = SQ3(BIN) + AMP3**2
	    SQ4(BIN)   = SQ4(BIN) + AMP4**2
	    SQ5(BIN)   = SQ5(BIN) + AMP5**2
  100	  ENDDO
	ENDDO
C
	DO I = 1,20
	  IF(PTS(I).LE.0.0) THEN
	     AVG1(I) = 0.0
	     AVG3(I) = 0.0
	     AVG4(I) = 0.0
	     AVG5(I) = 0.0
	   ELSE
	     AVG1(I) = AVG1(I)/PTS(I)
	     AVG3(I) = AVG3(I)/PTS(I)
	     AVG4(I) = AVG4(I)/PTS(I)
	     AVG5(I) = AVG5(I)/PTS(I)
	  ENDIF
	  DIV1 = SQ1(I) - PTS(I)*(AVG1(I)**2)
	  DIV3 = SQ3(I) - PTS(I)*(AVG3(I)**2)
	  DIV4 = SQ4(I) - PTS(I)*(AVG4(I)**2)
	  DIV5 = SQ5(I) - PTS(I)*(AVG5(I)**2)
	  DIV1 = sign(SQRT(abs(DIV1)), DIV1)
	  DIV3 = sign(SQRT(abs(DIV3)), DIV3)
	  DIV4 = sign(SQRT(abs(DIV4)), DIV4)
	  DIV5 = sign(SQRT(abs(DIV5)), DIV5)
	  IF(DIV1*DIV3.LE.0.0) THEN
	     CC(1,I) = 0.0
	   ELSE
	     CC(1,I) = (SUM13(I)-AVG1(I)*AVG3(I)*PTS(I))/(DIV1*DIV3)
	  ENDIF
	  IF(DIV4*DIV5.LE.0.0) THEN
	     CC(2,I) = 0.0
	   ELSE
	     CC(2,I) = (SUM45(I)-AVG4(I)*AVG5(I)*PTS(I))/(DIV4*DIV5)
	  ENDIF
	ENDDO
C
	RETURN
	END

C  ****************************************************************************
C  *                                                                          *
C  *               *****     SUBROUTINE REFINE3_F4     *****                  *
C  *                                                                          *
C  *    REFINE THETA,PHI,OMEGA,X,Y PARAMETERS (MODE.EQ.3).                    *
C  *                                                                          *
C  *    APRIL 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.          *
C  *                                                                          *
C  *    MODIFIED  4-MAY-1998 BY T. BAKER (Sliding window code).               *
C  *    MODIFIED  4-DEC-98 BY SBW (Add FFT6 for CTF stuff).                   *
C  *    MODIFIED  20-DEC-02 BY MBS (made R4 compatible,added idim1,idim2      *
C  *    to params).                                                           *
C  *    LAST MODIFIED 7-MAR-2003 BY MBS (rid of pifhead, add heads to params).*
C  ****************************************************************************
C
	SUBROUTINE REFINE3_F4 (NCOL, NROW, NSEC, NCOLBIN, NROWBIN, 
     1 NSECBIN, NROT, NANNULI, NPIX, NFILES, NIDS_TOT, FIDS, IDS, TS, 
     2 PS, OS, XS, YS, MAGS, PFT_CCS, PRJ_CCS, CMP_CCS, CTFMODE, PI, 
     3 RADIAN, IC, SC, IO, SO, Radius2, BFACTOR, IDIM1, IDIM2,
     4 MAPHEAD, IMGHEAD)
C
	IMPLICIT NONE
C
	include 'VAX_MINMAX.INC'
	include 'PIFHEAD.INC'
	RECORD /PIFHEAD_T/ MAPHEAD, IMGHEAD
C
	COMMON /PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
     1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
     2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
     3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
     4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	VOLATILE /PFT_COM/
	INTEGER*4 I, NCOL, NROW, NSEC, NCOLBIN, NROWBIN, NSECBIN
C river This line is added for the GETMEM_REFINE2
	INTEGER*4 J,NUM_THIS_THE,ITEST, Radius2
	INTEGER*4 FILE, FILE_LAST, NUNIQUE, NVIEW_THISIMG, ITHE1, ITHE2
	INTEGER*4 NROT, NANNULI, NPIX
	INTEGER*4 IDIM1, IDIM2
	INTEGER*4 NFILES, IMG_MIN, IMG_MAX
	INTEGER*4 N_INCORE, NIDS_TOT, FIDS(NIDS_TOT), IDS(NIDS_TOT)
C river	INTEGER*4 WINDOW, OFFSET, IMG_NUM, IMG_NUM1, IMG_LAST, CTFMODE
C river changed for the new refine mode
	INTEGER*4 OFFSET, IMG_NUM,IMG_NUM1,IMG_NUM2,IMG_LAST,CTFMODE
	real*4    IMG_BKG                                       !MBS 08.19.03
	REAL*4    TS(NIDS_TOT), PS(NIDS_TOT), OS(NIDS_TOT)
	REAL*4    XS(NIDS_TOT), YS(NIDS_TOT), MAGS(NIDS_TOT)
	REAL*4    PFT_CCS(NIDS_TOT), PRJ_CCS(NIDS_TOT)
	REAL*4    CMP_CCS(NIDS_TOT)
	REAL*4    IMG_AVG, IMG_STD
	REAL*4    THE1, THE2, PHI1
	REAL*4    PHIMAX
	REAL*4    THE_MAX, PI, RADIAN
	CHARACTER IMG_FILENAME*256
	BYTE      READWRITE(4)
	LOGICAL   MALLOC, SLIDE, N_INCORE_ALLOCATION
C
C	ARRAY POINTERS
C
	INTEGER*2 IMG(1), MSKIMG(1)
	INTEGER*4 ITMP(1), MASK(1)
	REAL*4    IMG_F4(1)    !MBS 12.12.2002
	REAL*4    MOD_TS(1), MOD_PS(1), PFTS(1), PMAP(1), DPHIS(1)
	REAL*4    MAP3D(1), PRJ(1), PRJS(1), PROJN(1), PFFT(1)
	REAL*4    PFT_AVGS(1), TMP(1), SQ2(1), VAR(1), RMS(1), AVER(1)
	REAL*4    FFT1(1), FFT3(1), FFT4(1), FFT5(1), FFT6(1), R4_IMG(1)
	REAL*4    BESTPRJ(1), PRJ_ROT(1), CCOR(1), MSKPRJ(1)
	REAL*4    R4_PIMG(1), FLTPMAP(1), POLAR_PRJ(1)
	REAL*4    TMP1(1), TMP2(1), TMP3(1), A_ARRAY(1), PFTIMG(1)
 
	INTEGER*2 ALLOCATE_STAT

	POINTER   (MOD_TS_P,MOD_TS), (MOD_PS_P,MOD_PS), (ITMP_P,ITMP)
	POINTER   (PFTS_P,PFTS), (PFFT_P,PFFT), (PMAP_P,PMAP)
	POINTER   (MAP3D_P,MAP3D), (PRJS_P,PRJS), (PROJN_P,PROJN)
	POINTER   (PFT_AVGS_P,PFT_AVGS), (TMP_P,TMP), (SQ2_P,SQ2)
	POINTER   (PRJ_P,PRJ), (VAR_P,VAR), (RMS_P,RMS), (AVER_P,AVER)
	POINTER   (IMG_P,IMG), (DPHIS_P,DPHIS), (MASK_P,MASK)
	POINTER   (IMG_F4_P,IMG_F4)                                  !MBS 12.12.2002
	POINTER   (FFT1_P,FFT1), (FFT3_P,FFT3), (A_ARRAY_P,A_ARRAY)
	POINTER   (FFT4_P,FFT4), (FFT5_P,FFT5), (FFT6_P,FFT6)
	POINTER   (FLTPMAP_P,FLTPMAP)
	POINTER   (R4_IMG_P,R4_IMG), (BESTPRJ_P,BESTPRJ)
	POINTER   (PRJ_ROT_P,PRJ_ROT), (POLAR_PRJ_P,POLAR_PRJ)
	POINTER   (PFTIMG_P,PFTIMG), (R4_PIMG_P,R4_PIMG)
	POINTER   (TMP1_P,TMP1), (TMP2_P,TMP2), (TMP3_P,TMP3)
	POINTER   (CCOR_P,CCOR), (MSKIMG_P,MSKIMG), (MSKPRJ_P,MSKPRJ)
	REAL*4    IC, IO, IMG_CCOR
	INTEGER*4 SC, SO
C
	REAL*4    PFTS_AMP_SPFIT(1), PFTS_AMP(1), PRJS_SF(1)
	REAL*4    IMG_AMP(1), IMG_AMP_SPFIT(1), IMG_SF(1)
	real*4    FFTA(1), FFTB(1), PRJS_SFB(1), BFACTOR
	POINTER   (PFTS_AMP_SPFIT_P,PFTS_AMP_SPFIT)
	POINTER   (PFTS_AMP_P,PFTS_AMP)
	POINTER   (IMG_AMP_P, IMG_AMP), (IMG_AMP_SPFIT_P, IMG_AMP_SPFIT)
	POINTER   (PRJS_SF_P, PRJS_SF), (IMG_SF_P, IMG_SF)
	pointer   (FFTA_P, FFTA), (FFTB_P, FFTB), (PRJS_SFB_P, PRJS_SFB)
	INTEGER*4 halfSize
C
C -----------------------------------------------------------------------------
C
	OFFSET = 0
	WRITE(*,1001)
 1001	FORMAT('  View # Theta    Phi',/,21('-'),/)
C
C  ****************************************************
C  *    GET THETA BOUNDS ON GRID VALUES               *
C  *    DETERMINE SIZE OF MASK ARRAY (2,NPHI,NTHE)    *
C  ****************************************************
	THE1 = (NINT( (TS(1) - INT(TS(1))) /DTHE) * DTHE) + INT(TS(1)) 
     1 - 3*DTHE
	THE2 = (NINT( (TS(NIDS_TOT) - INT(TS(NIDS_TOT))) /DTHE) * DTHE) 
     1 + NINT(TS(NIDS_TOT)) + 3*DTHE
c	IF(THE1.LT.0.0)   THE1 =   0.0  !MBS 08/15/03
c	IF(THE2.GT.180.0) THE2 = 180.0  !MBS 08/15/03
	NTHE = NINT((THE2-THE1)/DTHE) + 2
c	NTHE = NINT((THE2-THE1)/DTHE) + 4   !MBS to calc outside of asymm unit
	NPHI = 2*NINT((PHIMAX(SYM,90.)/DPHI_90)) + 1 + 3 + 3	! "+3" to allow extra
	NVIEW_MOD = NTHE*NPHI		       ! Mask built for NVIEW_MOD views
C  **********************************************************************
C  *  Allocate virtual memory and read entire 3D MAP into core memory.  *
C  **********************************************************************
C       MBS 12.12.02 GETMEM_REFINE -> GETMEM_REFINE_F4
	CALL GETMEM_REFINE_F4 (NPIX, NROT, NANNULI, NCOLBIN, NROWBIN, 
	1 NSECBIN, IDIM1, IDIM2, MAP3D_P, MOD_TS_P,MOD_PS_P, 
	2 IMG_P, IMG_F4_P, R4_IMG_P, R4_PIMG_P, PRJ_P, PROJN_P, 
	3 BESTPRJ_P, PMAP_P, PFFT_P, VAR_P, RMS_P,AVER_P,MASK_P,DPHIS_P, 
	4 FFT1_P, FFT3_P, FFT4_P, FFT5_P, FFT6_P, MSKIMG_P, MSKPRJ_P, 
	5 PRJ_ROT_P, PFTIMG_P, FLTPMAP_P, POLAR_PRJ_P, A_ARRAY_P,ITMP_P, 
	6 TMP_P, TMP1_P, TMP2_P, TMP3_P, FFTA_P, FFTB_P)

	CALL PIF_READ_WMAPI2 (MAPHEAD, MAP3D, NCOLBIN, NROWBIN, NSECBIN) 
	CALL CALC_TP_MASK4 (NIDS_TOT, TS, PS, MOD_TS, MOD_PS, MASK, 
	1    DPHIS, THE1, THE2, NUNIQUE)
C  **************************************************************************
c  *    Allocate memory for PRJS and PFTS arrays                            *
C  *    If not enough, for now STOP program. Later add SLIDING WINDOW code  *
C  **************************************************************************
	halfSize   = INT(RES_MAX) + 1
	CALL GETMEM_REFINE3 (NUNIQUE, N_INCORE, NCOLBIN, NROWBIN, 
	1 NPIX, CCOR_P, PFT_AVGS_P, SQ2_P, PRJS_P, MASK, 
	2 ITHE2, SLIDE, PRJS_SF_P, IMG_SF_P, IDIM1, IDIM2, halfSize, 
	3 PRJS_SFB_P)
	THE_MAX       = MOD_TS(ITHE2)-3*DTHE  ! MAX THETA ALLOWED THIS WINDOW
	FILE_LAST     = FIDS(1)
	IMGHEAD.LUNUM = 7                       ! UNIT FOR INPUT IMAGE FILE
	READWRITE(1)  = 'r'

C  **********************************************************
C  *    OPEN UP THE VERY FIRST IMAGE FILE AND ITS HEADER    *
C  **********************************************************
	IMG_NUM1 = 1
	IMG_NUM2 = NIDS_TOT
	FILE         = FIDS(1)
	IMG_FILENAME = IMG_FNAMES(FILE)
	CALL PIF_OPEN (IMG_FILENAME, IMGHEAD.LUNUM, READWRITE)
	CALL PIF_READ_GH (IMGHEAD)
C         ******************************************************
C---------*  Calculate PRJS and PFTS in current search window  *
C         ******************************************************
100	CALL GET_PRJSPFTS_R3 (N_INCORE, NCOLBIN, NROWBIN, 
     1  NSECBIN, NROT, NANNULI, PROJN, PRJ, PRJS, MAP3D, MOD_TS, 
     2  MOD_PS, ITMP, PMAP, PFFT, A_ARRAY, MASK, 
     3  ITHE1, ITHE2, OFFSET, CTFMODE, IDIM1, IDIM2, FFT6, PI, RADIAN,
     4  IMG_NUM1,IMG_NUM2,the1,dphis,ps,ts,nids_tot, PRJS_SF, 
     6  FFTA, FFTB, halfSize, Radius2, BFACTOR, PRJS_SFB)
C         *******************************************
C---------*    MAIN LOOP THROUGH NIDS_TOT IMAGES    *
C         *******************************************
C river	  DO IMG_NUM = IMG_NUM1,NIDS_TOT
          DO IMG_NUM = IMG_NUM1, IMG_NUM2
	    IMG_LAST = IMG_NUM
	    IF(TS(IMG_NUM).GT.THE_MAX) GO TO 800   ! TRAP FOR TOO LARGE THETA 
	    FILE         = FIDS(IMG_NUM)
	    IMG_FILENAME = IMG_FNAMES(FILE)
C	    *********************************************
C-----------*    OPEN THE CORRECT IMAGE DATA FILE       *
C	    *********************************************
	IF(FILE_LAST.NE.FILE) THEN               ! CLOSE CURRENT IMAGE FILE
	   CALL PIF_CLOSE (IMGHEAD.LUNUM)
	   CALL PIF_OPEN (IMG_FILENAME, IMGHEAD.LUNUM, READWRITE)
	   CALL PIF_READ_GH (IMGHEAD)
	   FILE_LAST = FILE
	ENDIF
C	    ***********************
C-----------*  Read in raw IMG    *
C	    ***********************
	    IMGHEAD.IMGNUM = IDS(IMG_NUM)        ! SPECIFY WHICH "IMG" YOU WANT
	    CALL PIF_READ_DH (IMGHEAD)
	    IMG_BKG = IMGHEAD.IMG_BKG
C         MBS 12.07.02
	    if(IMGHEAD.MODE .EQ.0) then
	       CALL PIF_READ_IMGI1 (IMGHEAD, IMG,
	1	    NCOLBIN, NROWBIN, IDS(IMG_NUM))
C       CONVERT IMG TO FLOAT FORMAT
	       CALL IMG_MAP (IMG,IMG_F4,NCOLBIN,NROWBIN,IMG_BKG) !MBS 12.07.02
	       IMG_BKG = 0.
	    else if(IMGHEAD.MODE .EQ.1) then
	       call pif_read_imgi2 (IMGHEAD, IMG,NCOLBIN, NROWBIN, 
	1	    IDS(IMG_NUM))                                 !MBS 12.07.02
	       CALL IMG_MAP (IMG,IMG_F4,NCOLBIN,NROWBIN, IMG_BKG) !MBS 12.07.02
	       IMG_BKG = 0.
	    else if(IMGHEAD.MODE .EQ.2) then
	       call pif_read_imgi4 (IMGHEAD, IMG_F4, NCOLBIN, NROWBIN, 
	1	    IDS(IMG_NUM))                                 !MBS 12.07.02
	  else if(IMGHEAD.MODE .EQ.6) then
	     CALL PIF_READ_IMGI1 (IMGHEAD, IMG, NCOLBIN, NROWBIN, 
	1	  IDS(IMG_NUM))                                   !MBS 12.07.02
	       CALL IMG_MAP (IMG, IMG_F4,NCOLBIN,NROWBIN,IMG_BKG) !MBS 12.07.02
	       IMG_BKG = 0.
	  else if(IMGHEAD.MODE .EQ.7) then
	       call pif_read_imgi2 (IMGHEAD, IMG, NCOLBIN, NROWBIN, 
	1	    IDS(IMG_NUM))                                 !MBS 12.07.02
	       CALL IMG_MAP (IMG, IMG_F4,NCOLBIN,NROWBIN,IMG_BKG) !MBS 12.07.02
	       IMG_BKG = 0.
	  else if(IMGHEAD.MODE .EQ.46) then
	     CALL pif_read_imgi4(IMGHEAD, IMG_F4,
	1	    NCOLBIN, NROWBIN, IDS(IMG_NUM))               !MBS 08.18.03
C	    ********************************************  !MBS 12.10.02
	  else
	     write(*,*)"Image format not implemented yet"
	     stop
	  end if
c       MBS: Get img backgrnd as (1,1) pixel for now
	  IMG_BKG = IMG_F4(1)

	  do i = 1, ncolbin*nrowbin
	     IMG_F4(i) = IMG_F4(i) - IMG_BKG
	  end do
	  IMG_BKG = 0.
C	    ********************************************
C-----------*  Load PFT_AVGS and SQ2 for this IMAGE    *
C           ********************************************
	  CALL LOADSTATS_R (N_INCORE, PFTS, PFT_AVGS, NROT, NANNULI, 
	1      SQ2, VAR, RMS, AVER, TMP, MASK, 
	2      TS(IMG_NUM), PS(IMG_NUM), MOD_TS, DPHIS, NVIEW_THISIMG)
C           ***************************************************************
C-----------*  Determine best THETA,PHI,OMEGA, and origin for this image  *
C           ***************************************************************
C           !MBS 12.12.02 (img_f4)

	    CALL REFINE_CC_F4 (IMG_NUM, NVIEW_THISIMG, NROT, 
	1  NANNULI, THE1, PHI1, IDIM1, IDIM2, NFILES, PRJS, 
	2  PFT_AVGS, MOD_TS, MOD_PS, IMG_BKG, NCOLBIN, NROWBIN, 
	3  NIDS_TOT, IMG_F4, IDS, TS, PS, OS, XS, YS, MAGS, PFT_CCS, 
	4  PRJ_CCS, CMP_CCS, N_INCORE, MASK, DPHIS, FFT1, FFT3, FFT4, 
	5  FFT5, MSKIMG, MSKPRJ, R4_IMG, BESTPRJ, PRJ_ROT, PFTIMG, PFFT, 
	6  R4_PIMG, FLTPMAP, POLAR_PRJ, TMP1, TMP2, TMP3, CCOR, CTFMODE, 
	7  FILE, IC, SC, IO, SO, IMG_CCOR, MAP3D, PRJS_SF, IMG_SF, 
	8  FFTA, FFTB, halfSize, Radius2, PRJS_SFB, BFACTOR)

	    WRITE(*,1000)  IMG_NUM, TS(IMG_NUM), PS(IMG_NUM), OS(IMG_NUM),
	1	 XS(IMG_NUM), YS(IMG_NUM), MAGS(IMG_NUM), PFT_CCS(IMG_NUM), 
	2	 PRJ_CCS(IMG_NUM), CMP_CCS(IMG_NUM)

	 ENDDO			! DO IMG_NUM = IMG_NUM1,NIDS_TOT
C river   go to 900
c       Debug
c$$$	 write(*,*)"IMG_NUM,NIDS_TOT,img_num1,img_num2",
c$$$	1     IMG_NUM,NIDS_TOT,img_num1,img_num2
c       Debug
 800	 if(IMG_NUM .GT. NIDS_TOT) GOTO 900
	 img_num1 = img_num2 + 1
	 img_num2 = nids_tot

	 GO TO 100
C  ****************************************************************************
C  *    DEALLOCATE UNNEEDED MEMORY BEFORE EXITING SUBROUTINE    *
C  **************************************************************
 900	 CALL FREEMEM_REFINE_F4 (MAP3D_P, MOD_TS_P, MOD_PS_P,PRJ_P, 
	1     PROJN_P, PMAP_P, PFFT_P, ITMP_P, VAR_P, RMS_P, AVER_P, TMP_P, 
	2     IMG_P, IMG_F4_P, FFT1_P, FFT3_P, FFT4_P, FFT5_P, FFT6_P, 
	3     MSKIMG_P, MSKPRJ_P, R4_IMG_P, BESTPRJ_P, PRJ_ROT_P, PFTIMG_P, 
	4     R4_PIMG_P, FLTPMAP_P, POLAR_PRJ_P, A_ARRAY_P, 
	5     TMP1_P, TMP2_P, TMP3_P, FFTA_P, FFTB_P)
	 CALL PIF_CLOSE (1)

	 RETURN
 1000	 FORMAT (1X,I6,',',4(F8.3,','),F7.3,',',3(F6.3,','),F6.3)
	 END

C  ****************************************************************************
C  *                                                                          *
C  *                 *****     SUBROUTINE CALC_TP_MASK     *****              *
C  *                                                                          *
C  *    "MASK" ARRAY SERVES TWO PURPOSES:                                     *
C  *      1. TALLY # MODEL PROJECTIONS NEEDED AT EACH THE,PHI GRID POINT.     *
C  *      2. SIGNIFY AN ID FOR EACH UNIQUE PROJECTION THAT THEN IDENTIFIES    *
C  *         THE LOCATION OF APPROPRIATE DATA IN THE PFTS AND PRJS ARRAYS.    *
C  *                                                                          *
C  *    APRIL 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.          *
C  *                                                                          *
C  *    MODIFIED 28-APR-1998 BY T. BAKER (Implement new code).                *
C  *    LAST MODIFIED 15-AUG-2003 BY MBS (Trap for div/zero (DPHIS = 0)).     *
C  ****************************************************************************
C
	SUBROUTINE CALC_TP_MASK4 (NIDS_TOT, TS, PS, MOD_TS, 
     1 MOD_PS, MASK, DPHIS, THE1, THE2, NUNIQUE)
C
	IMPLICIT NONE
C
	COMMON /PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
     1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
     2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
     3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
     4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NIDS_TOT, ITH, IPH, ITHE, IPHI, ID
	INTEGER*4 I, J, ITHETA, NUM, NUNIQUE, NNPHI, MASK(NPHI,NTHE,2)
	REAL*4    THETA, PHI, MOD_TS(NTHE), MOD_PS(NPHI,NTHE)
	REAL*4    PHT_PHIMAX, TS(NIDS_TOT), PS(NIDS_TOT), DPHIS(NTHE)
	REAL*4    DELTA, THE1, THE2, DPHI, SPEED, ITMP
	LOGICAL   SIGN
C  ----------------------------------------------------------------------------
C
	NUM     = 0
	NUNIQUE = 0
C  ***************************
C  *    INITIALIZE "MASK"    *
C  ***************************
	DO J = 1,NTHE
	  DO I = 1,NPHI
	    MASK (I,J,1) = -1
	    MASK (I,J,2) = -1
	  ENDDO
	ENDDO
C  **********************************************
C  *    FILL UP DPHIS, MOD_TS, MOD_PS ARRAYS    *
C  **********************************************
	DO J = 1,NTHE
	  THETA = THE1 + (J-1)*DTHE
	  MOD_TS(J) = THETA
	  IF(THETA.EQ.0.0 .OR. THETA.EQ.180.0) THEN
	     DPHIS(J) = 0.0
	   ELSE
	     DPHIS(J) = DPHI_90/SIND(THETA)
	  ENDIF
	  DO I = 1,NPHI
	    ITMP = I - (NPHI+1)/2
	    MOD_PS(I,J) = ITMP*DPHIS(J)
	  ENDDO
	ENDDO
C  *************************************************************
C  *    RESET TS,PS ARRAYS SO T,P VALUES LIE AT GRID POINTS    *
C  *************************************************************
	DO ID = 1, NIDS_TOT
	  SIGN   = .TRUE.
	  ITHETA = NINT(TS(ID))
	  DELTA  = TS(ID) - ITHETA
	  TS(ID) = ITHETA + NINT(DELTA/DTHE)*DTHE
	  if(sind(ts(id)) .ne. 0.0) then !MBS 08/15/03 trap for div by zero
	     DPHI   = DPHI_90/SIND(TS(ID))
	     PS(ID) = NINT(PS(ID)/DPHI) * DPHI
	  else
	     PS(ID) = 0.0
	  end if
c	  IF(PS(ID).LT.0.0) SIGN = .FALSE.
c	  IF(.NOT.SIGN) PS(ID) = -PS(ID)
	ENDDO
C  ********************************************************
C  *    MAP THE "WINDOW" FOR EACH PARTICLE INTO "MASK"    *
C  ********************************************************
	DO ID = 1, NIDS_TOT
c	   ITH = NINT((TS(ID)-THE1)/DTHE) + 1 !MBS 08/27/03
	   ITH = NINT((TS(ID)-THE1)/DTHE) + 2 !Original
	   DO 300 ITHE = ITH-3,ITH+3
	      IF(ITHE.GE.1 .AND. ITHE.LE.NTHE) THEN
C       IPH = NINT( ABS(PS(ID))/DPHIS(ITHE) ) + 1
		 if(DPHIS(ITHE) .ne. 0) then
		    IPH = NINT(PS(ID)/DPHIS(ITHE)) !MBS 08/15/03 trap for div by zero
		 else
		    IPH = 0
		 end if
		 IPH = IPH + (NPHI+1)/2
		 DO 200 IPHI = IPH-3,IPH+3
		    IF(IPHI.GE.1 .AND. IPHI.LE.NPHI) THEN
		       MASK(IPHI,ITHE,1) = MASK(IPHI,ITHE,1) + 1
		       NUM               = NUM + 1
		       IF(MASK(IPHI,ITHE,1).EQ.0) then
			  NUNIQUE = NUNIQUE + 1
			  mask(iphi,ithe,1)=mask(iphi,ithe,1)+1
		       endif
		    ENDIF
 200		 ENDDO		! LOOP OVER PHI GRID POINTS
	      ENDIF
 300	   ENDDO		! LOOP OVER THETA GRID POINTS
	ENDDO			! LOOP OVER ALL PARTICLES
C
	PRINT '(I8,A)', NUM, ' total number of projections needed'
	PRINT '(I8,A)', NUNIQUE, ' of these are unique'
	SPEED = FLOAT(NUM)/FLOAT(NUNIQUE)
	PRINT '(A,F10.2,A)', ' Potential compute savings:', SPEED,
     1  ' times'
C
	RETURN
 1000	FORMAT(1X,66I2)
	END
C  ****************************************************************************
C  *                                                                          *
C  *                *****     SUBROUTINE GETMEM_REFINE3     *****             *
C  *                                                                          *
C  *    DETERMINE SIZE OF PRJ/PFT ARRAYS THAT CAN BE PROCESSED.               *
C  *                                                                          *
C  *    IF CAN'T CALCULATE AND STORE ALL NEEDED MODEL VIEWS, THEN RESORT TO   *
C  *     SLIDING WINDOW APPROACH, IN WHICH THE ASYMMETRIC UNIT IS DIVIDED     *
C  *     INTO SMALL ENOUGH WINDOWS TO PROCESS A SUB-SET OF IMAGES.            *
C  *                                                                          *
C  *    MAY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.            *
C  ****************************************************************************
C
	SUBROUTINE GETMEM_REFINE3 (NUNIQUE, N_INCORE, NCOLBIN, NROWBIN, 
	1 NPIX, CCOR_P, PFT_AVGS_P, SQ2_P, PRJS_P, 
	2 MASK, ITHE2, SLIDE, PRJS_SF_P, IMG_SF_P, IDIM1, IDIM2, halfSize, 
	3 PRJS_SFB_P)
C
	IMPLICIT NONE
C
	COMMON /PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
     1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
     2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
     3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
     4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NUNIQUE, N_INCORE, NCOLBIN, NROWBIN,I, J
	INTEGER*4 NUM_THIS_THE, ITEST, NPIX, MASK(NPHI,NTHE,2),ITHE2
	INTEGER*4 ROB_TEST, TMP2, N_INCORE_MAX
	INTEGER*8 TOT_MEM, TMP
	REAL*4    CCOR(1), PFT_AVGS(1), SQ2(1), PRJS(1), PRJS_SFB(1)
	POINTER   (CCOR_P,CCOR), (PFT_AVGS_P,PFT_AVGS)
	POINTER   (SQ2_P,SQ2), (PRJS_P,PRJS), (PRJS_SFB_P, PRJS_SFB)

	REAL*4    PRJS_SF(1)
	REAL*4    IMG_SF(1)
	POINTER   (PRJS_SF_P, PRJS_SF), (IMG_SF_P, IMG_SF)
	INTEGER   MALLOC
	LOGICAL   SLIDE, N_INCORE_ALLOCATION, TEST
	INTEGER*4 IDIM1, IDIM2, halfSize, N_INCORE_TMP
C
C  ----------------------------------------------------------------------------
C
	N_INCORE_MAX = ( NPHI + 6 ) * 7
	TMP2       = 0
	SLIDE      = .FALSE.
	ITHE2      = NTHE
	N_INCORE   = NUNIQUE
	N_INCORE_ALLOCATION = .FALSE.
	NPIX = (2*halfSize + 1)*(2*halfSize + 1)

	TMP = 2400000000

	N_INCORE_TMP = TMP / (8*NPIX)

	if (N_INCORE.GT.N_INCORE_TMP .and. N_INCORE_TMP.GT.N_INCORE_MAX) 
	1 N_INCORE = N_INCORE_MAX
C
	if (N_INCORE.GT.N_INCORE_MAX) N_INCORE = N_INCORE_MAX
C
	if (N_INCORE_TMP.LE.N_INCORE_MAX .and. N_INCORE.GT.N_INCORE_TMP)
	1 N_INCORE = N_INCORE_TMP

	   TMP = 8*NPIX*N_INCORE + 3*4*N_INCORE + 3*4*NPIX

	DO 5 WHILE (TMP.GE.2400000000)
	   N_INCORE = N_INCORE - 10
	   TMP = 8*NPIX*N_INCORE + 3*4*N_INCORE + 3*4*NPIX
5	CONTINUE

	DO 200 WHILE (N_INCORE_ALLOCATION .EQ. .FALSE.)

 30	   IF(PRJS_SF_P.NE.0) THEN 
	      CALL FREE (PRJS_SF_P)
	      PRJS_SF_P = MALLOC (4*NPIX*N_INCORE)
	   ELSE
	      PRJS_SF_P = MALLOC (4*NPIX*N_INCORE)
	   ENDIF
C       
	   IF(PRJS_SFB_P.NE.0) THEN 
	      CALL FREE (PRJS_SFB_P)
	      PRJS_SFB_P = MALLOC (4*NPIX*N_INCORE)
	   ELSE
	      PRJS_SFB_P = MALLOC (4*NPIX*N_INCORE)
	   ENDIF
C       
	   IF(PRJS_P.NE.0) THEN 
	      CALL FREE (PRJS_P)
	      PRJS_P = MALLOC (1)
	   ELSE
	      PRJS_P = MALLOC (1)
	   ENDIF
C       
	   IF(IMG_SF_P.NE.0) THEN 
	      CALL FREE (IMG_SF_P)
	      IMG_SF_P = MALLOC (4*NPIX)
	   ELSE
	      IMG_SF_P = MALLOC (4*NPIX)
	   ENDIF
	   
	   IF(CCOR_P.NE.0) THEN 
	      CALL FREE (CCOR_P)
	      CCOR_P = MALLOC (4*N_INCORE)
	   ELSE
	      CCOR_P = MALLOC (4*N_INCORE)
	   ENDIF
	   
	   IF(PFT_AVGS_P.NE.0) THEN 
	      CALL FREE (PFT_AVGS_P)
	      PFT_AVGS_P = MALLOC (4*N_INCORE)
	   ELSE
	      PFT_AVGS_P = MALLOC (4*N_INCORE)
	   ENDIF
	   
	   IF(SQ2_P.NE.0) THEN 
	      CALL FREE (SQ2_P)
	      SQ2_P = MALLOC (4*N_INCORE)
	   ELSE
	      SQ2_P = MALLOC (4*N_INCORE)
	   ENDIF
C       
C       ****** Finish ******
C       
	   IF(CCOR_P.EQ.0) THEN
C       print *, 'CCOR_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ELSE IF(IMG_SF_P.EQ.0) THEN
C       print *, 'PFT_AVG_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ELSE IF(PRJS_SF_P.EQ.0) THEN
C       print *, 'PFT_AVG_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ELSE IF(PRJS_SFB_P.EQ.0) THEN
C       print *, 'PFT_AVG_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ELSE IF(PFT_AVGS_P.EQ.0) THEN
C       print *, 'PFT_AVG_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ELSE IF(SQ2_P.EQ.0) THEN
C       print *, 'SQ2_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ELSE IF(PRJS_P.EQ.0) THEN
C       print *, 'PRJS_P is error!!!'
	      N_INCORE_ALLOCATION= .FALSE.
	      N_INCORE = N_INCORE - 10
	      GOTO 30
	   ENDIF
C       
	   IF (TMP2 .EQ. 1) GOTO 300
C       
	   TOT_MEM = 8*NPIX*N_INCORE + 3*4*N_INCORE + 4*NPIX
	   
	   PRINT *, 'MAX ### OF PRJS AND TOTAL MEMORY:', N_INCORE, TOT_MEM
	   
	   TMP = TOT_MEM
	   
	   DO 10 WHILE (TMP.GE.2400000000 .and. TOT_MEM.GT.(TMP-100000000))
	      N_INCORE = N_INCORE - 10
	      TOT_MEM = 8*NPIX*N_INCORE + 3*4*N_INCORE + 4*NPIX
 10	   CONTINUE
	   
	   TMP2 = TMP2 + 1
	   
 200	CONTINUE

C         **********************************************************
C---------*    HERE IF NEED TO INCREASE NUMBER OF SEARCH WINDOWS   *
C         **********************************************************

 300	IF (N_INCORE .LT. 49) THEN
          PRINT *,
     1  '*******************************************************'
          PRINT *,
     1  '* GETMEM_REFINE less than 49   error: EXIT program    *'
          PRINT *,
     1  '*******************************************************'
	  CALL TSBEND
	ENDIF
	
	SLIDE    = .TRUE.
	
	print *, ''
	print *, 'N_INCORE, NCOLBIN, NPIX: ', N_INCORE, NCOLBIN, NPIX
	print *, '4*NCOLBIN*NROWBIN*N_INCORE: ',
	1    4*NCOLBIN*NROWBIN*N_INCORE
	print *, '8*NPIX*N_INCORE: ', 8*NPIX*N_INCORE
	print *, '3*4*N_INCORE: ', 3*4*N_INCORE
	print *, '4*NPIX: ', 4*NPIX
	print *, ''
	
	PRINT *, 'THE ### OF PRJS AND TOTAL MEMORY:', N_INCORE, TOT_MEM
	
	RETURN
 1000	format(1x, 3(I4,','),I12,',', I6,',', I9, I12)	
	END	

C  ****************************************************************************
C  *                                                                          *
C  *             *****     SUBROUTINE GET_PRJSPFTS_R3     *****               *
C  *                                                                          *
C  *    STEPS CARRIED OUT BY THIS ROUTINE (REFINE MODE ONLY):                 *
C  *     1. PROJECT 3D MAP ALONG NVIEW_MOD THETA,PHI,OMEGA(=90) ORIENTATIONS. *
C  *     2. 2D PROJECTION SAVED IN PRJ ARRAY.                                 *
C  *     3. IMPOSE CTF ON PROJECTION [OPTIONAL]                               *
C  *     4. CONVERT 2D CARTESIAN PROJECTION MAP INTO 2D POLAR MAP.            *
C  *     5. FOURIER TRANSFORM THE POLAR MAP.                                  *
C  *     6. SAVE POLAR FFT IN PFTS ARRAY.                                     *
C  *                                                                          *
C  *    APRIL 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.          *
C  *                                                                          *
C  *    MODIFIED  5-MAY-1998 BY T. BAKER (Implement sliding window).          *
C  *    MODIFIED 23-JUL-1999 BY TSB (Implement SBW's CTF code).               *
C  *    MODIFIED 25-AUG-1999 BY TSB (Implement TSB's CTF code).               *
C  *    MODIFIED 15-AUG-2003 BY MBS (Trap for div/zero (DPHIS = 0)).          *
C  *    LAST MODIFIED 31-OCT-2003 BY MBS (bug in ctfmode = 0,                 *
C  *    see note in the code)                                                 *
C  ****************************************************************************
C
	SUBROUTINE GET_PRJSPFTS_R3 (N_INCORE, NCOLBIN, NROWBIN,
     1 NSECBIN, NROT, NANNULI, PROJN, PRJ, PRJS, MAP3D, MOD_TS, 
     2 MOD_PS, ITMP, PMAP, PFFT, A, MASK,
     3 ITHE1,ITHE2, OFFSET,CTFMODE, IDIM1, IDIM2, FFT6,PI, RADIAN,
     4 IMG_NUM1,IMG_NUM2,the1,dphis,ps,ts,nids_tot, PRJS_SF, 
     5 FFTA, FFTB, halfSize,Radius2, BFACTOR, PRJS_SFB)
c
	IMPLICIT NONE
C
	COMMON /CTF_COM/ PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
     1 FOCUS_MIN, ANG_MAJ, CS_COEFF
C
        VOLATILE /CTF_COM/

	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NCOLBIN, NROWBIN, NSECBIN, NROT, NANNULI, ID
	INTEGER*4 I, J, ITMP(NCOLBIN), N_INCORE, ITH, IPH
	INTEGER*4 MASK(NPHI,NTHE,2), ITHE1, ITHE2, ID_LAST, OFFSET 
	INTEGER*4 ID_PRINT, CTFMODE, IDIM1, IDIM2, FFT_IEND, FILE
	INTEGER*4 UNITS(999), halfSize, Radius2
	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999), ANG_MAJ(999)
	REAL*4    WAVELENGTH, WAVEL, TEMPFAC, PI, RADIAN
	REAL*4    MOD_TS(NTHE), MOD_PS(NPHI,NTHE)
	REAL*4    PFFT(NROT*2,NANNULI), A(NROT*2), PROJN(NCOLBIN,NROWBIN)
	REAL*4    MAP3D(NCOLBIN,NROWBIN,NSECBIN), PRJ(NCOLBIN,NROWBIN)
	REAL*4    PRJS(1), PMAP(NROT,NANNULI)
	REAL*4    PRJS_SF(2*halfSize+1,2*halfSize+1, N_INCORE)
	REAL*4    PRJS_SFB(2*halfSize+1,2*halfSize+1, N_INCORE)
	REAL*4    FFT_ORIGX, FFT_ORIGY, FFT_FZERO, FFT6(2*IDIM1,IDIM2)
	DATA	  ID_LAST/0/
	SAVE      ID_LAST
	REAL*4    FFT_cent(2*(2*halfSize+1),(2*halfSize+1))
	LOGICAL   FFT_FULL
C river
	INTEGER*4 IMG_NUM1,IMG_NUM2,IMG_NUM,PRJ_NUM,NIDS_TOT
	integer*4 ithe_search,iphi_search,ithe,iphi,last_the
	REAL*4    THE1,dphis(nthe),ps(NIDS_TOT),ts(NIDS_TOT)
	logical   no_free_pointer
C
	REAL*4    BFACTOR, FLT_FAC
	REAL*4    FFTA(IDIM1,IDIM2), FFTB(IDIM1,IDIM2)
	INTEGER*4 row, col,cent_dim
	REAL*4    resmin, resmax

C
C  ----------------------------------------------------------------------------
	cent_dim = 2*halfSize + 1

	FFT_ORIGX = FLOAT(NCOLBIN/2)
	FFT_ORIGY = FLOAT(NROWBIN/2)
	resmin = res_min
	resmax = res_max
cMBS	WRITE(*,1000)
C river **********************************************************************
C river *  Get as many as possible projections in memory                     *
C river **********************************************************************
        PRJ_NUM=1
        IMG_NUM=IMG_NUM1
        last_the=1
        no_free_pointer=.false.
        do while (IMG_NUM.le.IMG_NUM2)
c	   ITH = NINT( ((TS(IMG_NUM)-THE1)/DTHE) ) + 1 !MBS 08/27/03
	   ITH = NINT( ((TS(IMG_NUM)-THE1)/DTHE) ) + 2 !original
	   ITHE = ITH -3
c$$$c       Debug
c$$$	write(*,*)"In get_prjspfts"
c$$$	write(*,'(a,i5,2f10.3,2i4)')"img_num, TS, the1,ithe,nthe:",
c$$$	1    img_num, TS(IMG_NUM), the1,ithe,nthe
c$$$c       Debug
	   DO while (ithe.le.ITH+3)
	      IF(ITHE.GE.1 .AND. ITHE.LE.NTHE) THEN
C       IPH = NINT( PS(IMG_NUM)/DPHIS(ITHE) ) + 1
		 if(DPHIS(ITHE) .ne. 0) then
		    IPH = NINT(PS(IMG_NUM)/DPHIS(ITHE)) !MBS 08/15/03 trap for div by zero
		 else
		    IPH = 0
		 end if
		 IPH = IPH + (NPHI+1)/2
		 IPHI = IPH-3
		 
c	print *,'inside GET_PRJSPFTS_R3 3',ITH,IPH,PS(IMG_NUM),DPHIS(ITHE) !Debug
		 DO while (iphi.le.iph+3)
		    IF(IPHI.GE.1 .AND. IPHI.LE.NPHI
	1		 .and.mask(iphi,ithe,2).eq.-1 ) THEN
		       id = 0
		       if(img_num1.eq.1)then 
			  id=prj_num ! first cycle fill the array
		       else 
			  ithe_search=last_the
C       river first search for free pointer
			  do while(ithe_search.le.nthe
     1                     .and.(.not.no_free_pointer)) 
			     iphi_search=1
			     do while(iphi_search.le.nphi)
				if(mask(iphi_search,ithe_search,1).eq.0
	1			     .and.mask(iphi_search,ithe_search,2)
	2			     .ne.-1)then
				   ID=mask(iphi_search,ithe_search,2)
				   mask(iphi_search,ithe_search,1)=-1
				   mask(iphi_search,ithe_search,2)=-1
				   last_the = ithe_search
				   iphi_search=nphi ! jump out of the iphi search
				   ithe_search=nthe ! jump out of the ithe search
				endif
				iphi_search=iphi_search+1
			     end do
 			     ithe_search=ithe_search+1
			  end do
C river **************************************************************
C river if can not move one particle, it will force to get the first
C river image's projection 
c       river **************************************************************
			  if(id.eq.0.and.img_num.le.img_num1
	1		       .and.(.not.no_free_pointer))then
c$$$	print *, 'Not enough memory, projections can not be reused.',
c$$$	1 ' Projections needed by the first particle will be discarded.'
			     last_the = 1
			     ithe_search = last_the 
			     no_free_pointer = .true.
			  endif
C river search first available pointer which is not used by the current
C river image
			  if(id.eq.0.and.img_num.le.img_num1
     1                   .and.no_free_pointer)then
c			     offset = offset - 1 !one useful projection has been deleted
			     do while(ithe_search.le.nthe) 
				iphi_search=1
				do while(iphi_search.le.nphi)
				   if(mask(iphi_search,ithe_search,2).ne.-1
     1                     .and.(.not.((iphi_search.ge.(iph-3)).and.
     2                     (iphi_search.le.(iph+3)).and.(ithe_search
     3                     .ge.(ith-3)).and.(ithe_search.le.(ith+3)))))  
     4                     then
				      id=mask(iphi_search,ithe_search,2)
				      mask(iphi_search,ithe_search,2)=-1 
				      last_the = ithe_search
				      ithe_search=nthe ! jump out ithe search
				      iphi_search=nphi ! jump out iphi search
				   endif
				   iphi_search=iphi_search+1
				end do
				ithe_search=ithe_search+1
			     end do
			  endif                                
		       endif
C river ******************************************************************
C river get the projection
c river ******************************************************************
		       if(id.ne.0)then
			  ID_PRINT = prj_num + OFFSET 
			  if(verbose)WRITE(*,1010) id_print, MOD_TS(ITHe),
	1		       MOD_PS(IPHi,ITHe)
			  PRJ_NUM=PRJ_NUM+1
			  if(prj_num.gt.n_incore)then
			     img_num2=img_num-1
			     iphi=iph+3
			     ithe=ith+3
			  endif
C           **************************************************************
C-----------*    NOTE USE OF OMEGA = 90.0 TO GET CORRECT ORIENTATIONS    *
C           **************************************************************
C
			  CALL MAP_PRJ (1, NCOLBIN, NROWBIN, NSECBIN, PROJN,
	1		       PRJ, MAP3D, MOD_TS(ITHe), MOD_PS(IPHi,ITHe), 
	2		       90.0, 1.0, FLOAT(NSECBIN), 90., 90., 90.)
C
C           **************************************************
C-----------*    Impose CTF to projections if CTFMODE != 0   *
C           *    Code implemented by SBW                     *
C           **************************************************

C       Removed ctfmode != 0 condition statement MBS 10/31/03
c			  if(CTFMODE .nE. 0) then
			  CALL IMG_R4_FFT (PRJ, NCOLBIN, NROWBIN, FFT6,
	1		       IDIM1, IDIM2, 0.0, FFT_ORIGX, FFT_ORIGY, 
	2		       FFT_FZERO, FFT_FULL, 0)
			  CALL FFT_FLT (FFT6, IDIM1, IDIM2, 
	1		       RESMIN, RESMAX, 3.0, 3.0, FFT_FULL)
C
			  FILE = 1
c       FLT_FAC = 0.1 !Hardwired
C
C       Assume that the 3D map has already been B-factor corrected. MBS March 2003
C       It is much cheaper than correct reprojections for every particle!
c       Put it back to allow individual B-factors for separate micrographs
			  CALL CORRCT_B_FACTOR (FFT6, IDIM1, IDIM2, 
	1		       units(FILE), PIXSIZ(FILE), BFACTOR, 
	2		       RESMAX, FFT_FULL)
			  call Fill_FFT_cent(IDIM1, IDIM2, FFT6, FFT_cent,
	1		       halfsize,halfsize)
			  CALL FFT_REARRANGE_SF(FFT_cent,PRJS_SF(1,1,ID),
	1		       PRJS_SFB(1,1,ID), cent_dim,cent_dim, 
	2		       FFT_ORIGX+1.0, FFT_ORIGY+1.0, IDIM1, IDIM2)
c			  ENDIF
C
C ****** PREPARE DATA OF THE PFTS ******
c river ***********************************************************************
c river finish get the projection
c river ***********************************************************************
			  mask(iphi,ithe,2)=id
		       else
C river  can not find more free memory for projection
			  img_num2=img_num-1
			  iphi=iph+3
			  ithe=ith+3
		       endif
		    ENDIF
		    iphi=iphi+1
		 end do         ! LOOP OVER PHI GRID POINTS
	      ENDIF
	      ithe=ithe+1	
	   end do               ! LOOP OVER THETA GRID POINTS
	   img_num=img_num+1
	end do                  !Outer loop over img_num
C       
	OFFSET = OFFSET + PRJ_NUM - 1
	if(verbose)write(*,1020) prj_num - 1 
 1010	FORMAT ('+',I5,F8.2,F7.2)
 1020   format ( I5,' new projections have been done.' )
        RETURN
	END
C  **************************************************************************
C  *                                                                        *
C  *                 *****     SUBROUTINE GET_THEPHI_PR3     *****          *
C  *                                                                        *
C  *    DETERMINES THE BEST THETA,PHI ORIENTATION (REFINE MODE).            *
C  *                                                                        *
C  *    APRIL 1998 PURDUE UNIVERSITY VAX 8550, T. BAKER.                    *
C  *                                                                        *
C  *	MODIFIED 29-APR-1998 BY T. BAKER (Adapted from PFTCC_GETTP).        *
C  *    MODIFIED 15-AUG-2003 BY MBS (Trap for id < 0).                      *
C  *    LAST MODIFIED 31-OCT-2003 BY MBS (added ctfmode to par list)        *
C  **************************************************************************
C
	SUBROUTINE GET_THEPHI_PR3 (PRJS_SF, IMG_SF, NUNIQUE, CCPFT, 
	1 CCOR, PI, INDEX, MASK, ITH, IPH, NVIEWS, MASK_JUDGE, 
	2 ITH_BEST, IPH_BEST, IDIM1,IDIM2, FILE, halfSize, PRJS_SFB, 
	3 IMG_SFB, PRJS_AVG, IMG_AVG, ctfmode)
C
	IMPLICIT NONE
C
	include 'VAX_MINMAX.INC'
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1    FOCUS_MIN, ANG_MAJ, CS_COEFF
c
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT
C
	INTEGER*4 INDEX, NUNIQUE, I, J, ID, R
	INTEGER*4 ITHE, IPHI, ITH, IPH,CTFMODE
	INTEGER*4 NVIEWS, NUM, ITHE1, ITHE2, IPHI1, IPHI2
	INTEGER*4 IDIM1,IDIM2, NCOLBIN,NROWBIN, halfSize
	REAL*4    CCOR(NVIEWS), TOT, AMAX, CCPFT, DIV1, DIV2, PI
	REAL*4    A, B, A2, B2
	REAL*8    SUMA, SUMB, SUMAB, SUMASQ, SUMBSQ
	REAL*8    SUMA2, SUMB2, SUMAB2, SUMASQ2, SUMBSQ2
	REAL*8    t1, t2, t3, t01, t02
C
	INTEGER*4 MASK(NPHI,NTHE,2)
	REAL*4    PRJS_SF  (2*halfSize+1,2*halfSize+1,NUNIQUE)
	REAL*4    PRJS_SFB (2*halfSize+1,2*halfSize+1,NUNIQUE)
	REAL*4    IMG_SF  (2*halfSize+1,2*halfSize+1)
	REAL*4    IMG_SFB (2*halfSize+1,2*halfSize+1)
	REAL*4    PRJS_AVG(halfSize+1)
	REAL*4    IMG_AVG (halfSize+1)
	real*4    RES_MIN_Q2, RES_MAX_Q2
	real*4    XTMP, YTMP, RAD, D_STAR, CCA, CCB
	INTEGER*4 ITH_BEST, IPH_BEST, UNITS(999), FILE, RAD2
	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999), ANG_MAJ(999)
	real*4    TMP, WAVELENGTH, wavel, C, CTF, TMP1, TMP2
	real*4    FOCUS, CS, VOLT
	LOGICAL   MASK_JUDGE
C
C  ----------------------------------------------------------------------------
C
	index = -999
	ITH_BEST = -999
	IpH_BEST = -999
	AMAX = -999.0
	RES_MIN_Q2 = RES_MIN**2
	RES_MAX_Q2 = RES_MAX**2
	WAVELENGTH = WAVEL (VOLTS(FILE))
	FOCUS      = FOCUS_MAJ(FILE)
	CS         = CS_COEFF(FILE)
	VOLT       = VOLTS(FILE)
C
	if(units(FILE) .eq. 1) then
	   TMP = 10.0/(IDIM1*PIXSIZ(FILE))  ! CONVERT TO NM UNITS
	else if (units(FILE) .eq. 2) then
	   TMP =  1.0/(IDIM1*PIXSIZ(FILE))  ! ALREADY IN NM UNITS
	end if
C
C  *********************************
C  *    INITIALIZE "CCOR" ARRAY    *
C  *********************************
	DO I = 1,NVIEWS
	  CCOR(I) = R4_MIN
	ENDDO
C
	NUM   = 0
	ITHE1 = ITH - 3                
	ITHE2 = ITH + 3
	IPHI1 = IPH - 3
	IPHI2 = IPH + 3
	IF(ITHE1.LT.1)    ITHE1 =    1
	IF(ITHE2.GT.NTHE) ITHE2 = NTHE
	IF(IPHI1.LT.1)    IPHI1 =    1
	IF(IPHI2.GT.NPHI) IPHI2 = NPHI
C
	DO ITHE = ITHE1,ITHE2
	   DO IPHI = IPHI1,IPHI2
	      IF (MASK_JUDGE.EQ..FALSE.) THEN
		 MASK(IPHI,ITHE,1) = MASK(IPHI,ITHE,1) - 1 ! UPDATE MASK
		 MASK_JUDGE = .TRUE.
	      ENDIF
	      ID   = MASK(IPHI,ITHE,2)
	      NUM = NUM + 1
	      if(id .le. 0) go to 300 !MBS There is no corresponding projection
	      
	      SUMA    = 0.0d0
	      SUMB    = 0.0d0
	      SUMAB   = 0.0d0
	      SUMASQ  = 0.0d0
	      SUMBSQ  = 0.0d0
	      SUMA2   = 0.0d0
	      SUMB2   = 0.0d0
	      SUMAB2  = 0.0d0
	      SUMASQ2 = 0.0d0
	      SUMBSQ2 = 0.0d0
	      TOT     = 0.0
C       
	      DO I = 1, halfSize*2 + 1
		 YTMP = FLOAT(I - halfSize - 1)
		 DO J = 1, halfSize*2 + 1
		    XTMP   = FLOAT(J - halfSize - 1)
		    RAD    = XTMP**2 + YTMP**2
		    D_STAR = SQRT(RAD)
		    RAD2   = INT(SQRT(RAD)+0.5)+1
c		    RAD2   = nINT(D_STAR)+1
		    D_STAR = D_STAR * TMP
		    if(ctfmode .ne. 0) then
		       C = CTF (D_STAR, pi, WAVELENGTH,FOCUS, 0.0, CS, 
	1		    0.0, 0.0, 0.0, VOLT)
		    else
		       C = 1.0
		    end if

		    TMP1 = SQRT(PRJS_SF(J,I,ID)**2 + PRJS_SFB(J,I,ID)**2)
		    TMP2 = SQRT(IMG_SF(J,I)**2 + IMG_SFB(J,I)**2)

		    IF( (RAD.GT.RES_MIN_Q2).AND.(RAD.LE.RES_MAX_Q2).AND.
	1		 (ABS(C).GT.0.2).AND.(RAD2.Lt.halfSize).AND.
	2		 (TMP1.GE.PRJS_AVG(RAD2)*0.5).AND.
	3		 (TMP2.GE.IMG_AVG(RAD2)*0.6) ) THEN
C
		       A       = IMG_SF(J,I)
		       B       = PRJS_SF(J,I,ID)
		       A2      = IMG_SFB(J,I)
		       B2      = PRJS_SFB(J,I,ID)
		       SUMA    = SUMA + A
		       SUMB    = SUMB + B
		       SUMAB   = SUMAB  + A*B
		       SUMASQ  = SUMASQ + A*A
		       SUMBSQ  = SUMBSQ + B*B
		       SUMA2   = SUMA2 + A2
		       SUMB2   = SUMB2 + B2
		       SUMAB2  = SUMAB2  + A2*B2
		       SUMASQ2 = SUMASQ2 + A2*A2
		       SUMBSQ2 = SUMBSQ2 + B2*B2
		       TOT     = TOT + 1.0
		    ENDIF
 		 ENDDO
	      ENDDO 
C
	      CCA = 0.0
	      CCB = 0.0
	      if(tot .ne. 0) then
		 t01 = SUMASQ*TOT - (SUMA*SUMA)
		 t02 = SUMBSQ*TOT - (SUMB*SUMB)
		 t1 = SQRT(t01)
		 t2 = SQRT(t02)
		 t3 = SUMAB*TOT - (SUMA*SUMB)
C       
		 if ((T1*T2) .ne. 0.0) then
		    CCA = T3 / (T1*T2)
		 endif
C       
		 t01 = SUMASQ2*TOT - (SUMA2*SUMA2)
		 t02 = SUMBSQ2*TOT - (SUMB2*SUMB2)
		 t1 = SQRT(t01)
		 t2 = SQRT(t02)
		 t3 = SUMAB2*TOT - (SUMA2*SUMB2)
C       
		 if ((T1*T2) .ne. 0.0) then
		    CCB = T3 / (T1*T2)
		 endif
	      endif
C       
	      CCOR(NUM) = (CCA+CCB) / 2.0
C       
	      IF(CCOR(NUM).GT.AMAX) THEN
		 AMAX     = CCOR(NUM)
		 INDEX    = ID
		 ITH_BEST = ITHE
		 IPH_BEST = IPHI
	      ENDIF
300	   ENDDO		! DO IPHI = IPHI1,IPHI2
	ENDDO			! DO ITHE = ITHE1,ITHE2
C
	CCPFT = AMAX
C       
	RETURN
	END

C  ***************************************************************************
C  *                                                                         *
C  *            *****     SUBROUTINE GET_THEPHI_PR3_mult     *****           *
C  *                                                                         *
C  *    DETERMINES THE BEST THETA,PHI ORIENTATION (REFINE MODE).             *
C  *                                                                         *
C  *    APRIL 1998 PURDUE UNIVERSITY VAX 8550, T. BAKER.                     *
C  *                                                                         *
C  *	MODIFIED 29-APR-1998 BY T. BAKER (Adapted from PFTCC_GETTP).         *
C  *	MODIFIED 18-MAR-2003 BY MBS (Adapted from GET_THEPHI_PR3).           *
C  *    LAST MODIFIED 31-OCT-2003 BY MBS (added ctfmode to par list)         *
C  ***************************************************************************
C
	SUBROUTINE GET_THEPHI_PR3_mult (PRJs_SF, PRJs_SFB, IMG_SF, IMG_SFB, 
	1    PRJS_AVG, IMG_AVG, CC_PFT, PI, IDIM1,IDIM2, FILE, halfSize,
	2    ith, iph,NUNIQUE,MASK,MASK_JUDGE, ctfmode)

C
	IMPLICIT NONE
C
	include 'VAX_MINMAX.INC'
C
	COMMON/CTF_COM/PIXSIZ, UNITS, VOLTS, AMP_FAC, FOCUS_MAJ, 
	1    FOCUS_MIN, ANG_MAJ, CS_COEFF
c
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT
C
	INTEGER*4 I, J, ID, R
	INTEGER*4 IDIM1,IDIM2, halfSize
	INTEGER*4 NUNIQUE,ith,iph, CTFMODE
	REAL*4    TOT, CC_PFT, PI
	REAL*4    A, B, A2, B2
	REAL*8    SUMA, SUMB, SUMAB, SUMASQ, SUMBSQ
	REAL*8    SUMA2, SUMB2, SUMAB2, SUMASQ2, SUMBSQ2
	REAL*8    t1, t2, t3, t01, t02
C
	INTEGER*4 MASK(NPHI,NTHE,2)
	REAL*4    PRJS_SF  (2*halfSize+1,2*halfSize+1,NUNIQUE)
	REAL*4    PRJS_SFB (2*halfSize+1,2*halfSize+1,NUNIQUE)
	REAL*4    IMG_SF  (2*halfSize+1,2*halfSize+1)
	REAL*4    IMG_SFB (2*halfSize+1,2*halfSize+1)
	REAL*4    PRJS_AVG(halfSize+1)
	REAL*4    IMG_AVG (halfSize+1)
	real*4    RES_MIN_Q2, RES_MAX_Q2
	real*4    XTMP, YTMP, RAD, D_STAR, CCA, CCB
	INTEGER*4 UNITS(999), FILE, RAD2
	REAL*4    PIXSIZ(999), VOLTS(999), AMP_FAC(999), CS_COEFF(999)
	REAL*4    FOCUS_MAJ(999), FOCUS_MIN(999), ANG_MAJ(999)
	real*4    TMP, WAVELENGTH, wavel, C, CTF, TMP1, TMP2
	real*4    FOCUS, CS, VOLT
	LOGICAL   MASK_JUDGE
C
C  ----------------------------------------------------------------------------
C
	RES_MIN_Q2 = RES_MIN**2
	RES_MAX_Q2 = RES_MAX**2
	WAVELENGTH = WAVEL (VOLTS(FILE))
	FOCUS      = FOCUS_MAJ(FILE)
	CS         = CS_COEFF(FILE)
	VOLT       = VOLTS(FILE)

	if(units(FILE) .eq. 1) then
	   TMP = 10.0/(IDIM1*PIXSIZ(FILE))  ! CONVERT TO NM UNITS
	else if (units(FILE) .eq. 2) then
	   TMP =  1.0/(IDIM1*PIXSIZ(FILE))  ! ALREADY IN NM UNITS
	end if
C
	IF (MASK_JUDGE.EQ..FALSE.) THEN
	   MASK(IPH,ITH,1) = MASK(IPH,ITH,1) - 1 ! UPDATE MASK
	   MASK_JUDGE = .TRUE.
	ENDIF
	ID      = MASK(IPH,ITH,2)
	if(id .le. 0) go to 300	!MBS There is no corresponding projection

	SUMA    = 0.0d0
	SUMB    = 0.0d0
	SUMAB   = 0.0d0
	SUMASQ  = 0.0d0
	SUMBSQ  = 0.0d0
	SUMA2   = 0.0d0
	SUMB2   = 0.0d0
	SUMAB2  = 0.0d0
	SUMASQ2 = 0.0d0
	SUMBSQ2 = 0.0d0
	TOT     = 0.0
C       
	DO I = 1, halfSize*2 + 1
	   YTMP = FLOAT(I - halfSize - 1)
	   DO J = 1, halfSize*2 + 1
	      XTMP = FLOAT(J - halfSize - 1)
	      RAD = XTMP**2 + YTMP**2
	      D_STAR = SQRT(RAD)
c	      RAD2   = INT(SQRT(RAD)+0.5)
	      RAD2   = INT(SQRT(RAD)+0.5)+1
	      D_STAR = D_STAR * TMP
	      if(ctfmode .ne. 0) then
		 C = CTF (D_STAR, pi, WAVELENGTH,FOCUS, 0.0, CS, 
	1	      0.0, 0.0, 0.0, VOLT)
	      else
		 C = 1.0
	      end if
	      
	      TMP1 = SQRT(PRJS_SF(J,I,ID)**2 + PRJS_SFB(J,I,ID)**2)
	      TMP2 = SQRT(IMG_SF(J,I)**2 + IMG_SFB(J,I)**2)
	      IF( (RAD.GT.RES_MIN_Q2).AND.(RAD.LE.RES_MAX_Q2).AND.
	1	   (ABS(C).GT.0.2).AND.(RAD2.Lt.halfSize).AND.
	2	   (TMP1.GE.PRJS_AVG(RAD2)*0.5).AND.
	3	   (TMP2.GE.IMG_AVG(RAD2)*0.6) ) THEN
C       
		 A       = IMG_SF (J,I)
		 B       = PRJs_SF (J,I,id)
		 A2      = IMG_SFB(J,I)
		 B2      = PRJs_SFB(J,I,id)
		 SUMA    = SUMA    + A
		 SUMB    = SUMB    + B
		 SUMAB   = SUMAB   + A*B
		 SUMASQ  = SUMASQ  + A*A
		 SUMBSQ  = SUMBSQ  + B*B
		 SUMA2   = SUMA2   + A2
		 SUMB2   = SUMB2   + B2
		 SUMAB2  = SUMAB2  + A2*B2
		 SUMASQ2 = SUMASQ2 + A2*A2
		 SUMBSQ2 = SUMBSQ2 + B2*B2
		 TOT     = TOT     + 1.0
	      ENDIF
	   ENDDO
	ENDDO
C       
 300	CCA = 0.0
	CCB = 0.0
	if(tot .ne. 0) then
	   t01 = SUMASQ*TOT - (SUMA*SUMA)
	   t02 = SUMBSQ*TOT - (SUMB*SUMB)
	   t1 = SQRT(t01)
	   t2 = SQRT(t02)
	   t3 = (SUMAB*TOT - (SUMA*SUMB))
C       
	   if ((T1*T2) .ne. 0.0) then
	      CCA = T3 / (T1*T2)
	   endif
C       
	   t01 = SUMASQ2*TOT - (SUMA2*SUMA2)
	   t02 = SUMBSQ2*TOT - (SUMB2*SUMB2)
	   t1 = SQRT(t01)
	   t2 = SQRT(t02)
	   t3 = (SUMAB2*TOT - (SUMA2*SUMB2))
C       
	   if ((T1*T2) .ne. 0.0) then
	      CCB = T3 / (T1*T2)
	   endif
	end if
C       
	CC_PFT = (CCA+CCB) / 2.0
C       
	RETURN
	END
C
C  ****************************************************************************
C  *                                                                          *
C  *                  *****     SUBROUTINE CTF_DIVIDE_TRY3    *****           *
C  *                                                                          *
C  *    DIVIDE IMAGE FFT BY CTF.                                              *
C  *                                                                          *
C  *    THE FORMULA USED IS:                                                  *
C  *                                                                          *
C  *        FFTnewvalue = (FFToldvalue*TEMPFAC)/(ABS(CTF) + WIENER)           *
C  *                                                                          *
C  *    WIENER  = FACTOR (USUALLY 0.2) TO MAKE SURE NOISE IS NOT MAGNIFIED    *
C  *               TOO MUCH NEAR THE NODES OF THE CTF.                        *
C  *    TEMPFAC = TEMPERATURE FACTOR, SPECIFIED IN ANGSTROMS SQUARED, IS      *
C  *               USED TO SHARPEN UP HIGHER RESOUTION FEATURES.  VALUES      *
C  *               FOR CRYOTEM DATA TYPICALLY RANGE BETWEEN 100-500 A**2.     *
C  *               THIS TEMPERATURE FACTOR IS APPLIED AS AN INVERSE T_FAC.    *
C  *    MODE    = 0 DO NOTHING!                                               *
C  *            = 1 CORRECT PHASES AND AMPLITUDES                             *
C  *            = 2 FLIP PHASES WHERE NECESSARY; LEAVE AMPLITUDES UNCHANGED   *
C  *            = 3 ADJUST AMPLITUDES; LEAVE PHASES UNCHANGED                 *
C  *                                                                          *
C  *    AUGUST 1999 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.         *
C  *    MODIFIED 12-AUG-1999 BY TSB (Fixed XTEST,YTEST problems).             *
C  *    LAST MODIFIED 17-AUG-1999 BY TSB (Resorted to old faithful formula).  *
C  ****************************************************************************
C
	SUBROUTINE CTF_DIVIDE_TRY3 (MODE, FFT, IDIM1, IDIM2, UNITS, 
	1 PIXSIZE, PI, RADIAN, WAVELENGTH, FOCUS_MAJ, FOCUS_MIN, 
	2 ANG_MAJ, AMP_FAC,  Cs, WIENER, TEMPFAC, FFT_FULL, Cc, BETA,
	3 dE, VOLTS, FFT_DIM)
C
	IMPLICIT NONE
C
	INTEGER*4 mode,CTFMODE, IDIM1, IDIM2, ICOUNT, IX, IY, IY2
	INTEGER*4 UNITS, XTEST, YTEST, I, NUM_ZERO, FILE, FILE_TMP 
	REAL*4    FFT(1), YSQ, CTF, D_STAR, PIXSIZE, PSI, PHI, C, T
	REAL*4    FOCUS, FOCUS_MAJ, FOCUS_MIN, ANG_MAJ, WIENER, TEMPFAC
	REAL*4    PI, RADIAN, WAVELENGTH, AMP_FAC, Cs, X, Y, TMP, FACTOR
	REAL*4    Cc, BETA, dE, VOLTS, RAD_FIRSTPEAK, RAD
	REAL*4    START, END_PT, ZERO(200), ZONE(200), THRESHOLD
	INTEGER*4 FFT_DIM
	LOGICAL   FFT_FULL, ASTIG, CTF_LOG
	PARAMETER (THRESHOLD=0.2)
C  ----------------------------------------------------------------------------
C
	IF(MODE.LE.0 .OR. MODE.GT.3) RETURN
c	For future changes to input actual CTF MODE and filter #, 
c       for now valid only for mode = 3 MBS 4/20/2003
	if(mode .eq. 3)then
	   CTFMODE = 1
	else if(mode .eq. 2)then
	   CTFMODE = 3
	else if(mode .eq. 1)then
	   CTFMODE = 3
	end if
C
	PHI = ANG_MAJ * RADIAN		! ANGLE OF FOCUS_MAJ WRT X-DIRECTION
C
	IF(UNITS.EQ.2) THEN
	   TMP = 1.0/(IDIM1*PIXSIZE)			! ALREADY IN NM UNITS
	 ELSE if(UNITS.EQ.1) THEN
	   TMP = 10.0/(IDIM1*PIXSIZE)			! CONVERT TO NM UNITS
	ENDIF
C
	IF(FOCUS_MAJ.EQ.FOCUS_MIN) THEN
	   ASTIG = .FALSE.
	   FOCUS = FOCUS_MAJ
	 ELSE
	   ASTIG = .TRUE.
	ENDIF
C
	XTEST  = IDIM1/2
	YTEST  = IDIM2/2
	ICOUNT = -1
C
	IF(FFT_FULL) THEN
	   IY2 = IDIM2
	ELSE
	   IY2 = YTEST + 1
	ENDIF
C
C *************************************************
C * Determine the positions of CTF nodes and the  *
C * area in which the data are not used by EM3DR. *
C *************************************************
C
 66	DO IY = 1,IY2		! PROCESS FFT A ROW AT A TIME
	   IF(IY.GT.YTEST) THEN
	      Y = FLOAT(IY-IDIM2-1)
	   ELSE
	      Y = FLOAT(IY-1)
	   ENDIF
	   YSQ = Y**2
	   DO IX = 1,IDIM1	! PROCESS CURRENT ROW
	      ICOUNT = ICOUNT + 2
	      IF(IX.LE.XTEST) THEN
		 X = FLOAT(IX-1) ! RIGHT HALF OF FFT
	      ELSE
		 X = FLOAT(IX-IDIM1-1) ! LEFT HALF OF FFT
	      ENDIF
	      D_STAR = SQRT(YSQ + X**2) * TMP
	      IF(ABS(X).GT.0.9) THEN ! TRAP FOR FFT Y-AXIS DATA
		 PSI = ATAN2 (Y,X) ! ATAN2 GIVES ANGLE IN RADIANS
	      ELSE
		 IF(Y.GE.0.0) THEN
		    PSI =  PI/2.0
		 ELSE
		    PSI = -PI/2.0
		 ENDIF
	      ENDIF
C
	      IF(ASTIG) CALL CTF_ASTIG (FOCUS,FOCUS_MAJ,FOCUS_MIN,PSI,PHI)
C
	      C = CTF (D_STAR, PI, WAVELENGTH, FOCUS, AMP_FAC, Cs,Cc,BETA,
	1	   dE, VOLTS)
 100	      T = EXP ((TEMPFAC/400.0) * D_STAR**2) ! ADJUSTED FOR NM

	      FACTOR = (T*ABS(C))/(C*C+WIENER*(1.0-abs(C)))
C
 666	      IF(CTFMODE.EQ.1) THEN ! ADJUST PHASES AND AMPLITUDES
		 IF(C.GT.0.0) FACTOR = -FACTOR
		 FFT(ICOUNT)   = FACTOR * FFT(ICOUNT)
		 FFT(ICOUNT+1) = FACTOR * FFT(ICOUNT+1)
	      ELSE IF(CTFMODE.EQ.2) THEN ! ADJUST PHASES ONLY
		 IF(C.GT.0.0) THEN
		    FFT(ICOUNT)   = -FFT(ICOUNT)
		    FFT(ICOUNT+1) = -FFT(ICOUNT+1)
		 ENDIF
	      ELSE IF(CTFMODE.EQ.3) THEN ! ADJUST AMPLITUDES ONLY
		 IF(C.GT.0.0) FACTOR = -FACTOR
		 FFT(ICOUNT)   = FACTOR * FFT(ICOUNT)
		 FFT(ICOUNT+1) = FACTOR * FFT(ICOUNT+1)
	      ENDIF
	   ENDDO
	ENDDO
C
	RETURN
	END



	SUBROUTINE CORRCT_B_FACTOR (FFT, IDIM1, IDIM2,UNITS, 
	1PIXSIZE, TEMPFAC, resmax, FFT_FULL)
C
	IMPLICIT NONE
C
	INTEGER*4 IDIM1, IDIM2, ICOUNT, IX, IY, IY2
	INTEGER*4 UNITS, XTEST, YTEST, I
	REAL*4    FFT(1), YSQ, D_STAR_sq, PIXSIZE, T, TEMPFAC
	REAL*4    X, Y, TMP
	REAL*4    resmax
	LOGICAL   FFT_FULL
C  ----------------------------------------------------------------------------
C
	resmax = resmax**2			      ! Square resolution limit
	IF(UNITS.EQ.2) THEN
	   TMP = 1.0/(IDIM1*PIXSIZE)			! ALREADY IN NM UNITS
	ELSE if(UNITS.EQ.1) THEN
	   TMP = 10.0/(IDIM1*PIXSIZE)			! CONVERT TO NM UNITS
	ENDIF
	XTEST  = IDIM1/2
	YTEST  = IDIM2/2
	ICOUNT = -1
	IF(FFT_FULL) THEN
	   IY2 = IDIM2
	ELSE
	   IY2 = YTEST + 1
	ENDIF
 66	DO IY = 1,IY2		                ! PROCESS FFT A ROW AT A TIME
	   IF(IY.GT.YTEST) THEN
	      Y = FLOAT(IY-IDIM2-1)
	   ELSE
	      Y = FLOAT(IY-1)
	   ENDIF
	   YSQ = Y**2
	   DO IX = 1,IDIM1	                ! PROCESS CURRENT ROW
	      ICOUNT = ICOUNT + 2
	      IF(IX.LE.XTEST) THEN
		 X = FLOAT(IX-1)                ! RIGHT HALF OF FFT
	      ELSE
		 X = FLOAT(IX-IDIM1-1)          ! LEFT HALF OF FFT
	      ENDIF
	      D_STAR_sq = (YSQ + X**2) * TMP**2
	      if(D_STAR_sq.le.resmax)then
 100		 T = EXP ((TEMPFAC/400.0) * D_STAR_sq) ! ADJUSTED FOR NM
		 FFT(ICOUNT)   = T * FFT(ICOUNT)
		 FFT(ICOUNT+1) = T * FFT(ICOUNT+1)
	      end if
	   ENDDO
	ENDDO
	RETURN
	END


c       separate fft into real and imaginary parts and apply phase shift
c       to get the right phase origin. Idim1 and idim2 are the dimensions
c       of the part selected for separation. fft_dim1; fft_dim2 are the 
c       dimensions of the full fft transform. Fft_origx; fft_origy, are the
c       new origin coordinates
	SUBROUTINE FFT_REARRANGE_SF(FFT1, FFTA, FFTB, IDIM1, IDIM2, 
     1         FFT_ORIGX, FFT_ORIGY,fft_dim1,fft_dim2)
C
	IMPLICIT NONE
C
	INTEGER*4 I,J,IDIM1,IDIM2
	INTEGER*4 fft_dim1,fft_dim2
	REAL*4    FFT1(IDIM1*2, IDIM2)
	REAL*4    FFTA(IDIM1,IDIM2)
	REAL*4    FFTB(IDIM1,IDIM2)
	REAL*4    X, Y, XSHIFT, YSHIFT, XSHIFT_STEP, YSHIFT_STEP, SHIFT
	REAL*4    COS_SHIFT, SIN_SHIFT, FFT_ORIGX, FFT_ORIGY
	REAL*4    PI, A, B, AMP
	PARAMETER (PI=3.1415926536)
C ------------------------------------------------------------------------
C
	DO J = 1, IDIM2
	   DO I =1, IDIM1
	     FFTA(I,J) = FFT1(2*I-1,J)
	     FFTB(I,J) = FFT1(2*I,J)
	   ENDDO
	ENDDO

	XSHIFT_STEP = FFT_ORIGX*2.0*PI/FLOAT(fft_DIM1)
	YSHIFT_STEP = FFT_ORIGY*2.0*PI/FLOAT(fft_DIM2)
C
	DO J = 1, IDIM2
	   YSHIFT = YSHIFT_STEP * FLOAT(J-IDIM2/2-1)
	   DO I = 1, IDIM1
	      X = FLOAT(I-IDIM1/2-1)
	      SHIFT     = XSHIFT_STEP*X + YSHIFT
	      COS_SHIFT = COS(SHIFT)
	      SIN_SHIFT = SIN(SHIFT)
	      A =  FFTA(I,J)*COS_SHIFT + FFTB(I,J)*SIN_SHIFT
	      B = -FFTA(I,J)*SIN_SHIFT + FFTB(I,J)*COS_SHIFT
	      FFTA(I,J) = A
	      FFTB(I,J) = B
	   ENDDO
	ENDDO

	RETURN
	END
	SUBROUTINE GET_PRJS_AMPAVG (PRJS_SF, PRJS_SFB, MASK, 
	1 NUNIQUE, ITH, IPH, halfSize, PRJS_AVG)
C
	IMPLICIT NONE
C
	include 'VAX_MINMAX.INC'
C
	COMMON /PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
     1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
     2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
     3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
     4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NUNIQUE, I, J, ID, halfSize, ITHE1, ITHE2,IPHI1,IPHI2
	INTEGER*4 MASK(NPHI,NTHE,2), ITHE, IPHI, ITH, IPH
	INTEGER*4 RAD, MAX, MIN
	REAL*4    TMP, X, Y
C
	REAL*4    PRJS_SF(2*halfSize+1,2*halfSize+1, NUNIQUE)
	REAL*4    PRJS_SFB(2*halfSize+1,2*halfSize+1, NUNIQUE)
	REAL*4    PRJS_AVG(halfSize+1), PRJS_NUM(halfSize+1)
C
C  ----------------------------------------------------------------------------
C
	MIN = 999
	MAX = -1
	ITHE1 = ITH - 3
	ITHE2 = ITH + 3
	IPHI1 = IPH - 3
	IPHI2 = IPH + 3
	IF(ITHE1.LT.1)    ITHE1 =    1
	IF(ITHE2.GT.NTHE) ITHE2 = NTHE
	IF(IPHI1.LT.1)    IPHI1 =    1
	IF(IPHI2.GT.NPHI) IPHI2 = NPHI
C
	DO I = 1, halfSize + 1
	   PRJS_AVG(I) = 0.0
	   PRJS_NUM(I) = 0.0
	ENDDO
C
	DO ITHE = ITHE1,ITHE2
	   DO IPHI = IPHI1,IPHI2
	      ID   = MASK(IPHI,ITHE,2)
C
	      DO I = 1, halfSize*2 + 1
		 X = (I - halfSize - 1)**2
		 DO J = 1, halfSize*2 + 1
		    Y = (J - halfSize - 1)**2
		    RAD = INT(SQRT(X + Y)+0.5)+1
		    IF (RAD .Le. halfSize) THEN
		       PRJS_NUM(RAD) = PRJS_NUM(RAD) + 1.0
		       TMP = SQRT(PRJS_SF(J,I,ID)**2 + PRJS_SFB(J,I,ID)**2)
		       PRJS_AVG(RAD) = PRJS_AVG(RAD) + TMP
		    ENDIF
		 ENDDO
	      ENDDO
	   ENDDO
	ENDDO
C
	DO I = 1, halfSize + 1
	   if(PRJS_NUM(I).ne.0) then
	      PRJS_AVG(I) = PRJS_AVG(I) / PRJS_NUM(I)
	   end if
	ENDDO
C
	RETURN
	END
	
	SUBROUTINE GET_IMG_AMPAVG (IMG_SF, IMG_SFB, halfSize, IMG_AVG)
C
	IMPLICIT NONE
C
	include 'VAX_MINMAX.INC'
C
	INTEGER*4 I, J, halfSize, RAD
	REAL*4    TMP, X, Y
C
	REAL*4    IMG_SF(2*halfSize+1,2*halfSize+1)
	REAL*4    IMG_SFB(2*halfSize+1,2*halfSize+1)
	REAL*4    IMG_AVG(halfSize+1), IMG_NUM(halfSize+1)
C
C  ----------------------------------------------------------------------------
C
	DO I = 1, halfSize + 1
	   IMG_AVG(I) = 0.0
	   IMG_NUM(I) = 0.0
	ENDDO
C       
	DO J = 1, halfSize*2 + 1
	   Y = (J - halfSize - 1)**2
	   DO I = 1, halfSize*2 + 1
	      X = (I - halfSize - 1)**2
	      RAD = INT(SQRT(X + Y)+0.5)
	      IF (RAD .LT. halfSize) THEN
		 IMG_NUM(RAD+1) = IMG_NUM(RAD+1) + 1.0
		 TMP = SQRT(IMG_SF(i,j)**2 + IMG_SFB(i,j)**2)
		 IMG_AVG(RAD+1) = IMG_AVG(RAD+1) + TMP
	      ENDIF
	   ENDDO
	ENDDO
C
	DO I = 1, halfSize + 1
	   if(IMG_NUM(I).ne.0.) then
	      IMG_AVG(I) = IMG_AVG(I) / IMG_NUM(I)
	   else
	      IMG_AVG(I) = 0.
	   end if
	ENDDO
C
	RETURN
	END
C  ***************************************************************************
C  *                                                                         *
C  *                 *****     SUBROUTINE GET_BESTMAG_F4     *****           *
C  *                                                                         *
C  *	Compares an image against a model 2D projection over a range of      *
C  *    magnifications.  Returns the MAG factor with the best PRJ/IMG        *
C  *    correlation.  Correlation done in real space.	                     *
C  *								             *
C  *	Modified from IMG_CMP_PRJ, July 1996, by D. Belnap                   *
C  *    MODIFIED  8-JAN-1998 BY T. BAKER (Changed from SCALE to MAG)         *
C  *    MODIFIED 22-APR-1998 BY T. BAKER (Renamed from PFTCC_GETMAG).        *
C  *    LAST MODIFIED 20-DEC-2002 BY MBS (IMG -> IMG_F4).                    *
C  ***************************************************************************
C
	SUBROUTINE GET_BESTMAG_F4 (IMG_F4, PRJ, TMP, FFTI, FFTM, NCOL, 
	1 NROW, IDIM1, IDIM2,X_CEN, Y_CEN, OME_BEST, MAG_LO, BEST_MAG, 
	2 BEST_CCOEF, IMG_BKG, FFT_ORIGX, FFT_ORIGY)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 I, J, K, NCOL, NROW, IDIM1, IDIM2
	real*4    IMG_BKG                                       !MBS 08.19.03
	REAL*4    IMG_F4(NCOL,NROW)                             !MBS 12.12.02
	REAL*4    TOT, PRJ(NCOL,NROW), TMP(NCOL,NROW), A, B
	REAL*4    FFTI(2*IDIM1,IDIM2), FFTM(2*IDIM1,IDIM2)
	REAL*4    MAG, MAG_LO, X_CEN, Y_CEN
	REAL*4    CCOEF, BEST_CCOEF, BEST_MAG, OME_BEST
	REAL*4    RADLO, RADHI, RADDSQ, RADLOSQ, RADHISQ, YSQ
	REAL*4    FFT_ORIGX, FFT_ORIGY, FFT_FZERO
	REAL*8    SUMA, SUMB, SUMASQ, SUMBSQ, SUMAB
c	REAL*4    resmin, resmax
	LOGICAL   FFT_FULL
C  ---------------------------------------------------------------------------
C
C  *******************************
C  *    Initialize parameters    *
C  *******************************
	radlo      = FLOAT(ANN_LO)
	radhi      = FLOAT(ANN_HI)
c	resmin     = RES_Min
c	resmax     = RES_MAX
	RADLOSQ    = RADLO*RADLO
	RADHISQ    = RADHI*RADHI
	BEST_CCOEF = 0.0
	BEST_MAG   = 1.0
C  **************************************************************
C  *  Filter in reciprocal space if RESMIN or RESMAX are > 0.0  *
C  **************************************************************
c       MBS not needed, already filtered in calling routine

C  ***************************************************************
C  *    LOOP THROUGH SPECIFIED MAGNIFICATION SCALE FACTOR RANGE  *
C  ***************************************************************
	MAG = MAG_LO
	DO I = 1,MAG_NUM		! INITIALIZE PARAMETERS FOR EACH LOOP
	  SUMA    = 0.0
	  SUMB    = 0.0
	  SUMAB   = 0.0
	  SUMASQ  = 0.0
	  SUMBSQ  = 0.0
	  TOT     = 0.0
	  DO K = 1,NROW		         ! Load PRJ ("Gold standard") into TMP
	    DO J = 1,NCOL
	      TMP(J,K) = PRJ(J,K)
	    ENDDO
	  ENDDO
C	  *******************************************************
C	  *  Scale projection and move it to X,Y origin of IMG  *
C	  *******************************************************
	  CALL INTERP2D (TMP, NCOL, NROW, X_CEN, Y_CEN,
	1  FLOAT(NCOL-1)/2.0, FLOAT(NROW-1)/2.0, OME_BEST, MAG, 0.0)
C	  **********************************************************
C	  *  Loop through IMG and PRJ, get values for coefficient  *
C	  *  calculation, J and K must be within RADLO and RADHI   *
C	  **********************************************************
	  DO K = 1,NROW
	    YSQ = (FLOAT(K)-Y_CEN+1.0)**2
	    DO J = 1,NCOL
	      RADDSQ = (FLOAT(J)-X_CEN+1.0)**2 + YSQ
	      IF(RADDSQ.GE.RADLOSQ .AND. RADDSQ.LE.RADHISQ) THEN
c       MBS background was subtracted already in the calling routine
c		A      = FLOAT(IMG(J,K)) - IMG_BKG
		A      = IMG_F4(J,K) 
		B      = TMP(J,K)
		SUMA   = SUMA + A
		SUMB   = SUMB + B
		SUMAB  = SUMAB + A*B
		SUMASQ = SUMASQ + A*A
		SUMBSQ = SUMBSQ + B*B
		TOT    = TOT + 1.0
	      ENDIF
	    ENDDO
	  ENDDO
C	  *************************************
C	  *  Compute correlation coefficient  *
C	  *************************************
	  IF(TOT.EQ.0. .OR. (SUMASQ-SUMA*SUMA).EQ.0. .OR.
	1  (SUMBSQ-SUMB*SUMB).EQ.0.) THEN
	     CCOEF = 1.0E-20
	   ELSE
	     CCOEF = (SUMAB - (SUMA*SUMB)/TOT) /
	1            (SQRT(SUMASQ - (SUMA*SUMA)/TOT)*
	2             SQRT(SUMBSQ - (SUMB*SUMB)/TOT))
	  ENDIF
	  IF(ABS(CCOEF).GT.ABS(BEST_CCOEF)) THEN	! Test for best CCOEF
	    BEST_CCOEF = CCOEF
	    BEST_MAG   = MAG
	  ENDIF
D	  WRITE(6,'(A,2F10.4)') ' CCOEF, MAG =', CCOEF, MAG
	  MAG = MAG + MAG_STEP			     ! Increment MAG i DO loop
	ENDDO
C
 999	RETURN
	END

C  ****************************************************************************
C  *                                                                          *
C  *             *****     SUBROUTINE GETMEM_REFINE_F4     *****              *
C  *                                                                          *
C  *    SET UP DYNAMIC MEMORY ARRAYS FOR REFINE MODE                          *
C  *                                                                          *
C  *    APRIL 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.          *
C  *    MODIFIED  4-DEC-98 BY SBW (Add FFT6 for CTF stuff).                   *
C  *    LAST MODIFIED 20-DEC-02 BY MBS (Changed img_p->img_f4_p).             *
C  ****************************************************************************
C
	SUBROUTINE GETMEM_REFINE_F4 (NPIX, NROT, NANNULI, NCOLBIN, 
	1 NROWBIN,NSECBIN, IDIM1, IDIM2, MAP3D_P, MOD_TS_P, 
	2 MOD_PS_P, IMG_P, IMG_F4_P, R4_IMG_P, R4_PIMG_P, PRJ_P, 
	3 PROJN_P, BESTPRJ_P, PMAP_P, PFFT_P, VAR_P, RMS_P, AVER_P, 
	4 MASK_P, DPHIS_P, FFT1_P, FFT3_P, FFT4_P, FFT5_P, FFT6_P, 
	5 MSKIMG_P, MSKPRJ_P, PRJ_ROT_P, PFTIMG_P, FLTPMAP_P, 
	6 POLAR_PRJ_P, A_ARRAY_P, ITMP_P, TMP_P, TMP1_P, TMP2_P,TMP3_P,
	7 FFTA_P, FFTB_P)
C
	IMPLICIT NONE
C
	COMMON /PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
     1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
     2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
     3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
     4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NPIX, NROT, NANNULI, NCOLBIN, NROWBIN, NSECBIN
	INTEGER*4 ITMP(1), MASK(1), IDIM1, IDIM2
	INTEGER*2 IMG(1)
	REAL*4    IMG_F4(1)
	REAL*4    MAP3D(1), MOD_TS(1), MOD_PS(1), PMAP(1)
        REAL*4    PRJ(1), PROJN(1), PFFT(1), A_ARRAY(1), DPHIS(1)
	REAL*4    VAR(1), RMS(1), AVER(1), TMP(1)
	REAL*4    FFT1(1), FFT3(1), FFT4(1), FFT5(1), FFT6(1)
	REAL*4    R4_IMG(1), R4_PIMG(1), BESTPRJ(1), PRJ_ROT(1)
	REAL*4    PFTIMG(1), FLTPMAP(1), POLAR_PRJ(1)
	REAL*4    CCOR(1), TMP1(1), TMP2(1), TMP3(1)
	REAL*4    MSKIMG(1), MSKPRJ(1)
	POINTER   (ITMP_P,ITMP), (MOD_TS_P,MOD_TS), (MOD_PS_P,MOD_PS)
	POINTER   (PFFT_P,PFFT), (PMAP_P,PMAP), (A_ARRAY_P,A_ARRAY)
        POINTER   (MAP3D_P,MAP3D), (PROJN_P,PROJN), (PRJ_P,PRJ)
	POINTER   (VAR_P,VAR), (RMS_P,RMS), (AVER_P,AVER), (TMP_P,TMP)
	POINTER   (IMG_P,IMG), (DPHIS_P,DPHIS), (MASK_P,MASK)
	POINTER   (IMG_F4_P,IMG_F4)
	POINTER   (FFT1_P,FFT1), (FFT3_P,FFT3), (FFT4_P,FFT4)
	POINTER   (FFT5_P,FFT5), (FFT6_P,FFT6)
	pointer   (R4_IMG_P,R4_IMG), (R4_PIMG_P,R4_PIMG)
	POINTER   (MSKIMG_P,MSKIMG), (MSKPRJ_P,MSKPRJ)
	POINTER   (BESTPRJ_P,BESTPRJ), (PRJ_ROT_P,PRJ_ROT)
	POINTER   (PFTIMG_P,PFTIMG)
	POINTER   (FLTPMAP_P,FLTPMAP), (POLAR_PRJ_P,POLAR_PRJ)
	POINTER   (TMP1_P,TMP1), (TMP2_P,TMP2), (TMP3_P,TMP3)
	real*4    FFTA(1), FFTB(1)
	pointer   (FFTA_P, FFTA), (FFTB_P, FFTB)
	LOGICAL   MALLOC
C  ----------------------------------------------------------------------------
C
	MAP3D_P     = MALLOC (4*NCOLBIN*NROWBIN*NSECBIN)
	MOD_TS_P    = MALLOC (4*NTHE)
	MOD_PS_P    = MALLOC (4*NPHI*NTHE)
	IMG_F4_P    = MALLOC (4*NCOLBIN*NROWBIN)              !MBS 12.12.02
	IMG_P       = MALLOC (2*NCOLBIN*NROWBIN) 
	R4_IMG_P    = MALLOC (4*NCOLBIN*NROWBIN)	      ! FLOAT RAW IMAGE
	R4_PIMG_P   = MALLOC (4*NROT*NANNULI)		! FLOAT POLAR RAW IMAGE
	PRJ_P       = MALLOC (4*NCOLBIN*NROWBIN)
	PROJN_P     = MALLOC (4*NCOLBIN*NROWBIN)
	PMAP_P      = MALLOC (4*NROT*NANNULI)
	PFFT_P      = MALLOC (4*2*NROT*NANNULI)
	VAR_P       = MALLOC (2*NPIX)
	RMS_P       = MALLOC (4*NPIX)
	AVER_P      = MALLOC (4*NPIX)
	MASK_P      = MALLOC (4*2*NPHI*NTHE)
	DPHIS_P     = MALLOC (4*NTHE)
	FFT1_P      = MALLOC (8*IDIM1*IDIM2)      	     ! FFT OF RAW IMAGE
	FFT3_P      = MALLOC (8*IDIM1*IDIM2) 		      ! MOD PRJ @ T,P,O
	FFT4_P      = MALLOC (8*IDIM1*IDIM2)		       ! MASKED RAW IMG
	FFT5_P      = MALLOC (8*IDIM1*IDIM2)		       ! MASKED MOD PRJ
	FFT6_P      = MALLOC (8*IDIM1*IDIM2)                 ! FOR APPLYING CTF
	FFTA_P      = MALLOC (4*IDIM1*IDIM2)
	FFTB_P      = MALLOC (4*IDIM1*IDIM2)
	BESTPRJ_P   = MALLOC (4*NCOLBIN*NROWBIN)	! BEST MODEL PROJECTION
	PRJ_ROT_P   = MALLOC (4*NCOLBIN*NROWBIN)
	PFTIMG_P    = MALLOC (NROT*NANNULI)
	FLTPMAP_P   = MALLOC (4*NROT*NANNULI)
	POLAR_PRJ_P = MALLOC (4*NROT*NANNULI)
	MSKIMG_P    = MALLOC (2*NCOLBIN*NROWBIN)
	MSKPRJ_P    = MALLOC (4*NCOLBIN*NROWBIN)
	A_ARRAY_P   = MALLOC (4*2*NROT)
	ITMP_P      = MALLOC (4*NCOLBIN)
	TMP_P       = MALLOC (NROT*NANNULI)
	TMP1_P      = MALLOC (4*NCOLBIN*NROWBIN)
	TMP2_P      = MALLOC (8*IDIM1*IDIM2)
	TMP3_P      = MALLOC (8*IDIM1*IDIM2)
C
	IF(  ITMP_P.EQ.0 .OR.    MOD_TS_P.EQ.0 .OR.  MOD_PS_P.EQ.0 .OR.
	1   MAP3D_P.EQ.0 .OR.     PROJN_P.EQ.0 .OR.    PMAP_P.EQ.0 .OR. 
	2    PFFT_P.EQ.0 .OR.       VAR_P.EQ.0 .OR.     RMS_P.EQ.0 .OR.
	3    AVER_P.EQ.0 .OR.       TMP_P.EQ.0 .OR.     PRJ_P.EQ.0 .OR.
	5     IMG_P.EQ.0 .OR.      MASK_P.EQ.0 .OR.   DPHIS_P.EQ.0 .OR.
	6    FFT1_P.EQ.0 .OR.      FFT3_P.EQ.0 .OR.    FFT4_P.EQ.0 .OR. 
	7    FFT5_P.EQ.0 .OR.    MSKIMG_P.EQ.0 .OR.  MSKPRJ_P.EQ.0 .OR.
	8  R4_IMG_P.EQ.0 .OR.   BESTPRJ_P.EQ.0 .OR. PRJ_ROT_P.EQ.0 .OR. 
	9  PFTIMG_P.EQ.0 .OR.   R4_PIMG_P.EQ.0 .OR. FLTPMAP_P.EQ.0 .OR. 
	1 POLAR_PRJ_P.EQ.0 .OR. A_ARRAY_P.EQ.0 .OR.    FFT6_P.EQ.0 .OR. 
	2    TMP1_P.EQ.0 .OR.      TMP2_P.EQ.0 .OR.    TMP3_P.EQ.0 .OR. 
	3  IMG_F4_P.EQ.0 .OR.      FFTA_P.EQ.0 .OR.    FFTB_P.EQ.0) THEN
	  PRINT *,'*******************************************************'
	  PRINT *,'* GETMEM_REFINE memory allocation error: EXIT program *'
	  CALL TSBEND
	ENDIF
C
	RETURN
	END	
C  ****************************************************************************
C  *                                                                          *
C  *                *****     SUBROUTINE GETMEM4_F4     *****                 *
C  *                                                                          *
C  *    SET UP DYNAMIC MEMORY ARRAYS                                          *
C  *                                                                          *
C  *    JANUARY 1998 PURDUE UNIVERSITY AlphaStation 500/400, T. BAKER.        *
C  *    MODIFIED 26-FEB-1998 BY TSB (Make UNIX compatible).                   *
C  *    LAST MODIFIED 20-DEC-2002 BY MBS (Reserve additional IMG_F4 memory)   *
C  ****************************************************************************
C
	SUBROUTINE GETMEM4_F4 (NCOL, NROW, IDIM1, IDIM2, NROT, NANNULI, 
	1 FFT1_P, FFT3_P, IMG_P, IMG_F4_P, R4_IMG_P, PRJ_P, 
	2 PRJ_ROT_P, PFTIMG_P, PFFT_P, PMAP_P, FLTPMAP_P, PPRJ_P, A_P, 
	3 CCOR_P, ICOR_P, ITMP_P)
C
	IMPLICIT NONE
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NCOL, NROW, IDIM1, IDIM2, NROT, NANNULI
	INTEGER*2 IMG(1), ICOR(1), ITMP(1)
	REAL*4    IMG_F4(1)
	REAL*4    FFT1(1), FFT3(1), R4_IMG(1), PRJ(1), PRJ_ROT(1), A(1)
	REAL*4    PFTIMG(1), PFFT(1), PMAP(1), FLTPMAP(1), PPRJ(1)
	REAL*4    CCOR(1)
	POINTER   (FFT1_P,FFT1), (FFT3_P,FFT3), (IMG_P,IMG)
	POINTER   (IMG_F4_P,IMG_F4)
	POINTER   (R4_IMG_P,R4_IMG), (PRJ_P,PRJ), (PRJ_ROT_P,PRJ_ROT)
	POINTER   (PFTIMG_P,PFTIMG), (PFFT_P,PFFT), (PMAP_P,PMAP)
	POINTER   (FLTPMAP_P,FLTPMAP), (PPRJ_P,PPRJ), (A_P,A)
	POINTER   (CCOR_P,CCOR), (ICOR_P,ICOR), (ITMP_P,ITMP)
	LOGICAL   MALLOC
C  ----------------------------------------------------------------------------
C
	FFT1_P    = MALLOC (8*IDIM1*IDIM2)      		    ! RAW IMAGE
	FFT3_P    = MALLOC (8*IDIM1*IDIM2) 		      ! MOD PRJ @ T,P,O
	IMG_P     = MALLOC (2*NCOL*NROW)
	IMG_F4_P  = MALLOC (4*NCOL*NROW)                            !MBS 12.12.02
	R4_IMG_P  = MALLOC (4*NCOL*NROW)
	PRJ_P     = MALLOC (4*NCOL*NROW)
	PRJ_ROT_P = MALLOC (4*NCOL*NROW)
	PFTIMG_P  = MALLOC (NROT*NANNULI)
	PFFT_P    = MALLOC (4*2*NROT*NANNULI)
	PMAP_P    = MALLOC (4*NROT*NANNULI)
	FLTPMAP_P = MALLOC (4*NROT*NANNULI)
	PPRJ_P    = MALLOC (4*NROT*NANNULI)
	A_P       = MALLOC (4*2*NROT)
	CCOR_P    = MALLOC (4*NVIEW_MOD)
	ICOR_P    = MALLOC (2*NVIEW_MOD)
	ITMP_P    = MALLOC (4*NCOL)
	IF(FFT1_P  .EQ.0 .OR. FFT3_P.EQ.0 .OR. IMG_P    .EQ.0 .OR. 
	1 R4_IMG_P .EQ.0 .OR. PRJ_P .EQ.0 .OR. PRJ_ROT_P.EQ.0 .OR. 
	2 PFTIMG_P .EQ.0 .OR. PFFT_P.EQ.0 .OR.PMAP_P    .EQ.0 .OR. 
	3 FLTPMAP_P.EQ.0 .OR. PPRJ_P.EQ.0 .OR.A_P       .EQ.0 .OR. 
	4 CCOR_P   .EQ.0 .OR. ICOR_P.EQ.0 .OR. ITMP_P   .EQ.0 .OR. 
	5 IMG_F4_P .EQ.0) THEN
	  PRINT *, '*************************************************'
	  PRINT *, '* GETMEM4 memory allocation error: EXIT program *'
	  CALL TSBEND
	ENDIF
C
	RETURN
	END	

C  ****************************************************************************
C  *                                                                          *
C  *                 *****     SUBROUTINE GET_CC_PRJ     *****                *
C  *                                                                          *
C  *    FIND BEST OMEGA VALUE AND TEST FOR BEST PHI (POSITIVE OR NEGATIVE).   *
C  *                                                                          *
C  *    NOTE: REAL SPACE CORRELATION COEFFICIENTS ARE USED.                   *
C  *                                                                          *
C  *    MAY 1992  PURDUE UNIVERSITY VAX 8550, T. BAKER.                       *
C  *                                                                          *
C  *    MODIFIED 29-APR-1998 BY T. BAKER (Renamed from PFTCC_GETPO).          *
C  *    LAST MODIFIED 07-MAY-2004 BY M. Sherman (put fixed difference         *
C  *    threshold for best2 vs. best1 and simplified formulas).               *
C  ****************************************************************************
C
	SUBROUTINE GET_CC_PRJ (R4_PIMG, POLAR_PRJ, NROT, NANNULI, 
	1    OME_BEST, FLIP, CCPRJ_BEST)
C
	IMPLICIT NONE
C
	include 'VAX_MINMAX.INC'
C
	COMMON/PFT_COM/ MODE, BIN, VERBOSE, TIMER, MAG_CEN, MAG_STEP, 
	1 MAG_NUM, MAG_NORM, XCEN_MODEL, YCEN_MODEL, SYM, DTHE, DPHI_90, 
	2 PFTRAD_LO, PFTRAD_HI, JCUT, PIXRES_MIN, PIXRES_MAX, ANN_LO, 
	3 ANN_HI, VAROPT, SIGCUT, RES_MIN, RES_MAX, RADIUS, 
	4 DEL_ANG, NVIEW_MOD, NTHE, NPHI, PARAM_FNAMES, IMG_FNAMES

C       Common definitions
	INTEGER*4 BIN, VERBOSE, TIMER, MODE, SYM
	INTEGER*4 NTHE, NPHI, NVIEW_MOD, JCUT, ANN_LO, ANN_HI
	INTEGER*4 MAG_NUM, MAG_NORM
	REAL*4    PFTRAD_LO, PFTRAD_HI, PIXRES_MIN, PIXRES_MAX
	REAL*4    XCEN_MODEL, YCEN_MODEL, RES_MIN, RES_MAX
	REAL*4    DTHE, DPHI_90, DEL_ANG, MAG_CEN, MAG_STEP
	REAL*4    RADIUS, SIGCUT
	CHARACTER PARAM_FNAMES(999)*256, IMG_FNAMES(999)*256
	LOGICAL   VAROPT

	INTEGER*4 NROT, NANNULI, I, II, J, IXOFF
	INTEGER*4 XBEST1, XBEST2
	REAL*4    R4_PIMG(NROT,NANNULI), POLAR_PRJ(NROT,NANNULI)
	REAL*4    A, B, SUMA, SUMB, SUMAB, SUMASQ, SUMBSQ
	real*4    suma2, sumb2, sumab2
	REAL*4    CCPRJ_BEST, TOT, OME_BEST
	LOGICAL   FLIP
	real*4    CCPRJ_BEST1, CCPRJ_BEST2
C  -----------------------------------------------------------------------------
C
	FLIP = .FALSE.
	SUMA = 0.0
	SUMB = 0.0
	TOT  = 0.0
C
	DO J = ANN_LO,ANN_HI
	  DO I = 1,NROT
	    SUMA = SUMA + R4_PIMG(I,J+1)*J
	    SUMB = SUMB + POLAR_PRJ(I,J+1)*J
	    TOT  = TOT + 1.0
	  ENDDO
	ENDDO
	suma2  = suma*suma/tot
	sumb2  = sumb*sumb/tot
	sumab2 = suma*sumb/tot
C
C  ********************************************
C  *    CHECK FOR POSITIVE PHI ANGLE FIRST    *
C  ********************************************
C
	IF (OME_BEST .LT. 0.0) OME_BEST = OME_BEST + 360.0
	  IXOFF = (360.0-OME_BEST)/DEL_ANG
	  SUMAB  = 0.0
	  SUMASQ = 0.0
	  SUMBSQ = 0.0
	  DO J = ANN_LO,ANN_HI
	    DO I = 1,NROT
	      II     = I + IXOFF
	      IF(II.GT.NROT) II = II - NROT
	      A      = R4_PIMG(I,J+1)*J
	      B      = POLAR_PRJ(II,J+1)*J
	      SUMAB  = SUMAB  + A*B
	      SUMASQ = SUMASQ + A*A
	      SUMBSQ = SUMBSQ + B*B
	    ENDDO
	  ENDDO
	  CCPRJ_BEST1 = (SUMAB - SUMAB2) /
     1         (SQRT((SUMASQ - SUMA2)*(SUMBSQ - SUMB2)))
	  CCPRJ_BEST = CCPRJ_BEST1
C
c       MBS disabled, there is no reason to flip hand in refinement May 11 2004
c$$$C  ******************************************
c$$$C  *    NOW CHECK FOR NEGATIVE PHI ANGLE    *
c$$$C  ******************************************
c$$$C
c$$$	  SUMAB  = 0.0
c$$$	  SUMASQ = 0.0
c$$$	  SUMBSQ = 0.0
c$$$	  DO J = ANN_LO,ANN_HI
c$$$	    DO I = 1,NROT
c$$$	      II     = NROT + 1 - I - IXOFF
c$$$	      IF(II.LE.0) II = II + NROT
c$$$	      A      = R4_PIMG(I,J+1)*J
c$$$	      B      = POLAR_PRJ(II,J+1)*J
c$$$	      SUMAB  = SUMAB  + A*B
c$$$	      SUMASQ = SUMASQ + A*A
c$$$	      SUMBSQ = SUMBSQ + B*B
c$$$	    ENDDO
c$$$	  ENDDO
c$$$	  CCPRJ_BEST2 = (SUMAB - SUMAB2) /
c$$$     1         (SQRT((SUMASQ - SUMA2)*(SUMBSQ - SUMB2)))
c$$$C
c$$$Ca      Calculate the difference between max2 and max1.  If it's 
c$$$Ca      significant at all then see if we have to flip it or not.
c$$$Ca      We put this in because there are subtle precision differences
c$$$Ca      between out SGI Indigo (IRIX 5.3) and our Alpha VAX (VMS 6.1).
c$$$C
c$$$	     IF(CCPRJ_BEST2.GT.(CCPRJ_BEST1+0.0001)) THEN
c$$$		FLIP       = .TRUE.
c$$$		CCPRJ_BEST = CCPRJ_BEST2
c$$$	     ENDIF

	RETURN
	END

C  **************************************************************************
C  *                                                                        *
C  *            *****     SUBROUTINE Fill_FFT_cent     *****                *
C  *                                                                        *
C  *    Fill real and imaginary parts of a complex array with data          *
C  *    from central region in FFT space                                    *
C  *                                                                        *
C  *    March 2003 MBS  PURDUE UNIVERSITY                                   *
C  *                                                                        *
C  *    LAST MODIFIED 19-Mar-2003 BY M. Sherman                             *
C  **************************************************************************
	subroutine Fill_FFT_cent(IDIM1, IDIM2, FFTA, FFTB,xsize,ysize) 

	implicit none

	integer*4 idim1, idim2
	integer*4 xsize, ysize, xsize2
	integer*4 row, col
	integer*4 i, j, idim1p1, idim2p1, idim2d2

	real*4    ffta(2*idim1, idim2)
	real*4    fftb(-2*xsize:2*xsize+1,-ysize:ysize)

	xsize2  = 2*xsize
	idim1p1 = idim1*2 + 1
	idim2p1 = idim2   + 1
	idim2d2 = idim2/2

	do j = 1, idim2
	   row = j
	   if(row.gt.idim2d2) row = j-idim2
	   if (iabs(row) .le. ysize) then
	      do i = 1, 2*idim1,2
		 col = i - 1
		 if(col .gt.idim1) col = i-idim1p1
		 if (iabs(col).le. xsize2) then
		    fftb(col  ,row) = ffta(i  ,j)
		    fftb(col+1,row) = ffta(i+1,j)
		 end if
	      enddo
	   endif
	enddo
	return
	end


C  ***************************************************************************
C  *                                                                         *
C  *            *****     SUBROUTINE img_map           *****                 *
C  *                                                                         *
C  *    Copy data from I*2 array into a real*4 array                         *
C  *    and subtract background from it                                      *
C  *    August 2003 MBS  PURDUE UNIVERSITY                                   *
C  *                                                                         *
C  *    LAST MODIFIED 19-Aug-2003 BY M. Sherman                              *
C  ***************************************************************************
	subroutine IMG_MAP (IMG, IMG_F4, NCOL, NROW, IMG_BKG)

	implicit none

	integer*4 i, NCOL, NROW
	integer*2 IMG(NCOL*NROW)
	real*4    IMG_BKG
	real*4    IMG_F4(NCOL*NROW)

	do i = 1, NCOL*NROW
	   IMG_F4(i) = IMG(i) - IMG_BKG
	end do
	return
	end
